/* 
 * (c) 2011 Saarland University, Telecommunications Lab 
 * (prrt@nt.uni-saarland.de)
 *
 * The contents of this file are subject to the Mozilla Public License
 + Version 1.1 (the "License"); you may not use this file except in
 + compliance with the License. You may obtain a copy of the License at
 + http://www.mozilla.org/MPL/
 + 
 + Software distributed under the License is distributed on an "AS IS"
 + basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 + License for the specific language governing rights and limitations
 + under the License.
 + 
 + The Initial Developer of the Original Code is Saarland University.
 + All Rights Reserved.
 + 
 + Contributor(s): Yongtao Shuai (Saarland University).
 + 
 + Disclaimer: This Alpha version of PRRT is not supported and stable yet.
 + Do not use it in a production environment: it may crash your device, 
 + so please make sure that you have a backup of your data before
 + installing PRRT.
 */

#include "AdaptByLookup.hpp"

#include <cerrno>       // for errnos

#include <stdio.h>
#include <time.h>
#include <sys/time.h> // gettimeofday

using namespace std;

namespace Adaptation {


  AdaptByLookup::AdaptByLookup()
  : m_gplr   ( 0 )
  , m_delay200_cell (200)
  {
    m_sock = 0;
  }

  AdaptByLookup::~AdaptByLookup()
  {
    this->stop();
  }
  
  void AdaptByLookup::run()
  {
    pthread_mutex_lock(&m_mutex);
    try {

    while ( m_is_running ) {

      updateStats();
      isStatsChanged( 0, 0, 0, 0, 0, 0, 0, &m_grtt, 0, &m_gplr );

      // Look up the optimized parameter set.
      const OptimizedParameterSet *opt = searchOptParamSet();
      
      if (opt == 0) {
        std::cerr << "AdaptByLookup run() no opt found!" << std::endl;
      }
      // Check if the parameter set is changed.
      if ( opt && isOptParamSetChanged( opt ) ) {
        // Apply the optimized parameter set.
        applyOptParamSet( opt );

        time_t rawtime;
        struct tm * timeinfo;

        time ( &rawtime );
        timeinfo = localtime ( &rawtime );

        std::cout << "AdaptByLookup run() grtt= " << m_stats.lastGRTT << " srtt= " << m_rtt << " plr= " << (double (m_stats.lastGPLR)) / STATS_SCALE << " splr= " << m_plr << " opt : " << *opt << " time=" << asctime (timeinfo);
      }
      else {
        struct timeval current_time;
        ::gettimeofday(&current_time, 0);

        // Set up 'timespec' structure. Check for overrun on the 'nsec' field.
        long nsec = m_update_interval % 1000 * 1000000;
        long current_time_nsec = current_time.tv_usec * 1000;
        long nsec_from_max = 1000000000L - nsec;

        struct timespec timeout;
        timeout.tv_sec = current_time.tv_sec + m_update_interval / 1000;

        if(nsec_from_max > current_time_nsec) {
          timeout.tv_nsec = current_time_nsec + nsec;
        }
        else {
          // Handle nsec overrun that would cause pthread_cond_timedwait() to return EINVAL.
          ++timeout.tv_sec;
          timeout.tv_nsec = current_time_nsec - nsec_from_max;
        }

        int result = pthread_cond_timedwait(&m_condition, &m_mutex, &timeout);

        if (result == ETIMEDOUT) {
          
        }
        else if (result == 0) {
          
        }
        else {
          // This never happens.
          // As with most strange things that never happen, we would like to know when it happens.
          // In software, never can be any time, and we never want to miss it.
          std::cerr << "> AdaptByLookup::run() : Error in relativeTimedWait" << std::endl;
        }
    
      }

    }

    } catch (...) {
      std::cerr << "AdaptByLookup::run() something wrong!" << endl;
    }

    pthread_mutex_unlock(&m_mutex);
  }

  double AdaptByLookup::smoothPLR(double plr, double fall_acceleration, double max_deviation, double min_deviation)
  {
    static int fall_cnt = 0;
    static double last_plr = plr;
    static double last_smoothed_plr = plr;
    double diff = plr - last_plr;
    double smoothed_plr = last_smoothed_plr;
    // Use deviation to avoid smoothed PLR far from original one if PLR oscillating.
    double deviation = last_smoothed_plr - plr;
    
    if (min_deviation < deviation && deviation < max_deviation) {
      fall_cnt = 0;
    }
    else if (0 < diff && deviation < max_deviation) {
      if (diff < max_deviation) {
        smoothed_plr = last_smoothed_plr + diff;
      }
      else {
        // Avoid steep increasing when last smoothed plr is already greater than orginal one.
        smoothed_plr = plr + max_deviation;
      }
      fall_cnt = 0;
    }
    else if (diff < 0) {
      // Drop like free fall
      fall_cnt++;
      smoothed_plr = last_smoothed_plr - fall_acceleration * fall_cnt * fall_cnt;
      // Smoothed value should be not less than the orignal plr.
      if (smoothed_plr < plr) {
        smoothed_plr = plr;
        fall_cnt = 0;
      }
    }
    else {
      // two cases: (deviation >= max_deviation && diff >= 0) or (deviation <= min_deviation && diff = 0)
//       if (max_deviation <= deviation) {
//         smoothed_plr = last_smoothed_plr - fall_acceleration * fall_cnt * fall_cnt / 10;
//       }
//       fall_cnt = 0;
    }
    
    last_smoothed_plr = smoothed_plr;
    last_plr = plr;
    m_plr = smoothed_plr;
    return smoothed_plr;
  }
  
  uint16_t AdaptByLookup::smoothRTT(uint16_t rtt, double fall_acceleration, double max_deviation, double min_deviation)
  {
    // RTT is already smoothed in PRRT module by exponential smoothing. alpha=7/8
    // We do not smooth it.
    static int fall_cnt = 0;
    static double last_rtt = rtt;
    static double last_smoothed_rtt = rtt;
    double diff = rtt - last_rtt;
    double smoothed_rtt = last_smoothed_rtt;
    // Use deviation to avoid smoothed PLR far from original one if PLR oscillating.
    double deviation = last_smoothed_rtt - rtt;
    
    if (min_deviation < deviation && deviation < max_deviation) {
      fall_cnt = 0;
    }
    else if (0 < diff && deviation < max_deviation) {
      if (diff < max_deviation) {
        smoothed_rtt = last_smoothed_rtt + diff;
      }
      else {
        // Avoid steep increasing when last smoothed rtt is already greater than orginal one.
        smoothed_rtt = rtt + max_deviation;
      }
      fall_cnt = 0;
    }
    else if (diff < 0) {
      // Drop like free fall
      fall_cnt++;
      smoothed_rtt = last_smoothed_rtt - fall_acceleration * fall_cnt * fall_cnt;
      // Smoothed value should be not less than the orignal rtt.
      if (smoothed_rtt < rtt) {
        smoothed_rtt = rtt;
        fall_cnt = 0;
      }
    }
    else {
      // two cases: (deviation >= max_deviation && diff >= 0) or (deviation <= min_deviation && diff = 0)
//       if (max_deviation <= deviation) {
//         smoothed_rtt = last_smoothed_rtt - fall_acceleration * fall_cnt * fall_cnt / 10;
//       }
//       fall_cnt = 0;
    }
    
    last_smoothed_rtt = smoothed_rtt;
    last_rtt = rtt;
    m_rtt = smoothed_rtt;
    return static_cast<uint16_t>(smoothed_rtt);
  }

  const OptimizedParameterSet* AdaptByLookup::searchOptParamSet()
  {
    double cur_plr = static_cast<double>( m_gplr );
    cur_plr = cur_plr / STATS_SCALE;

    double plr = smoothPLR(cur_plr, 0.000001, 0.05, 0.02);  // the corresponding matlab code: g=0.0001/max=5/min=2 (drops roughly 2% in 100s)
    uint16_t rtt = smoothRTT(m_grtt, 0.001, 20, 10);  // drops roughly 0.2ms in 10s. NOTE dependent on the update time!!!
    rtt = m_grtt;

    return m_delay200_cell.getOptSet(1, rtt, plr);
  }


  DelayCell::DelayCell(int delay)
  : m_delay ( delay )
  , m_num_recv_map ()
  {
  }

  DelayCell::~DelayCell()
  {
    // Release all stored parameters.
    removeMap();
  }

  OptimizedParameterSet* DelayCell::getOptSet(int num_recv, int rtt, double plr)
  {
    num_recv_map::iterator rcv_it = m_num_recv_map.find( num_recv );

    if ( rcv_it == m_num_recv_map.end() ) {
      if (num_recv == 1) {
	add20MbpsOneReceiverMap();
      }
      else if (num_recv == 2) {
        //addTwoReceiversMap();
        add12MbpsTwoReceiversMap();
      }
      else if (num_recv == 3) {
        addThreeReceiversMap();
      }
      else {
        std::cerr << "num_recv not supported!" << std::endl;
        return 0;
      }
      rcv_it = m_num_recv_map.find( num_recv );
    }

    rtt_map::iterator rtt_it = rcv_it->second->lower_bound( rtt );
    if (rtt_it == rcv_it->second->end()) {
      std::cerr << "DelayCell::getOptSet() ERROR : rtt=" << rtt << " bigger than all rtt parameters" << std::endl;
      return 0;
    }
    plr_map::iterator plr_it = rtt_it->second->lower_bound( plr );
    if (plr_it == rtt_it->second->end()) {
      std::cerr << "DelayCell::getOptSet() ERROR : plr=" << plr << " bigger than all plr parameters" << std::endl;
      return 0;
    }
    OptimizedParameterSet *opt = plr_it->second;
    return opt;
  }

  void DelayCell::addOneReceiverMap()
  {
    // Targeted Delay = 0.1 & #receivers = 1
    num_recv_map::iterator rcv_it = m_num_recv_map.find(1);

    if ( rcv_it != m_num_recv_map.end() ) {
      // The parameters for one receiver already exist. No need to create.
      return;
    }

    rtt_map *rtt_opts_map = new rtt_map;
    plr_map *plr_opts_map = new plr_map;

//    /*********** Adding Optimum of RTT = 5 ms **************/
//    // Simplify Module from Manuel.
//    // PLR = 0.005
//    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.005, new OptimizedParameterSet(1, 2, 0, "1 1", 3)) );
//    // PLR = 0.01
//    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.01, new OptimizedParameterSet(10, 3, 0, "1 1 1", 27)) );
//    // PLR = 0.03
//    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.03, new OptimizedParameterSet(1, 3, 0, "1 1 1", 3)) );
//    // PLR = 0.06
//    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.06, new OptimizedParameterSet(1, 4, 0, "1 1 1 1", 3)) );
//    // PLR = 0.09
//    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.09, new OptimizedParameterSet(1, 5, 0, "1 1 1 1 1", 3)) );
//    // PLR = 0.1
//    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.1, new OptimizedParameterSet(1, 6, 0, "1 1 1 1 2", 3)) );
//    // PLR = 0.15
//    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.15, new OptimizedParameterSet(1, 7, 0, "1 1 1 1 3", 3)) );
//    // PLR = 0.2, use key=2 means using this opt if plr>0.2
//    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(2.0, new OptimizedParameterSet(1, 8, 0, "1 1 1 1 4", 3)) );
//
//    
//    rtt_opts_map->insert( pair<int, plr_map*>(5, plr_opts_map) );
//    /*********** Added Optimum of RTT = 5 ms **************/

//    plr_opts_map = new plr_map;

    /*********** Adding Optimum of RTT = 10 ms **************/
    // Simplify Module from Manuel.
    // PLR = 0.18
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.18, new OptimizedParameterSet(1, 0, 0, "1 1 1 1 5", 3)) );
    // PLR = 0.2
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.2, new OptimizedParameterSet(64, 64, 41, "", 174)) );
    // PLR = 0.25
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.25, new OptimizedParameterSet(64, 64, 51, "", 174)) );
    // PLR = 0.3, use key=3 means using this opt if plr>0.3
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(3.0, new OptimizedParameterSet(64, 64, 63, "", 174)) );
    
    // RTT=10ms
    rtt_opts_map->insert( pair<int, plr_map*>(10, plr_opts_map) );
    /*********** Added Optimum of RTT = 10 ms **************/

    plr_opts_map = new plr_map;

    /*********** Adding Optimum of RTT = 20 ms **************/
    // Simplify Module from Manuel.
    // PLR = 0.15
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.15, new OptimizedParameterSet(1, 0, 0, "1 1 1 5", 3)) );
    // PLR = 0.17
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.17, new OptimizedParameterSet(62, 62, 35, "", 168)) );
    // PLR = 0.20
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.2, new OptimizedParameterSet(62, 62, 40, "", 168)) );
    // PLR = 0.25
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.25, new OptimizedParameterSet(62, 62, 50, "", 168)) );
    // PLR = 0.3, use key=3 means using this opt if plr>0.3
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(3.0, new OptimizedParameterSet(62, 62, 61, "", 168)) );
    
    // RTT=20ms
    rtt_opts_map->insert( pair<int, plr_map*>(20, plr_opts_map) );
    /*********** Added Optimum of RTT = 20 ms **************/

    plr_opts_map = new plr_map;

    /*********** Adding Optimum of RTT = 30 ms **************/
    // Simplify Module from Manuel.
    // PLR = 0.12
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.12, new OptimizedParameterSet(1, 0, 0, "1 1 6", 3)) );
    // PLR = 0.14
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.14, new OptimizedParameterSet(60, 60, 29, "", 163)) );
    // PLR = 0.16
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.16, new OptimizedParameterSet(60, 60, 33, "", 163)) );
    // PLR = 0.18
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.18, new OptimizedParameterSet(60, 60, 36, "", 163)) );
    // PLR = 0.20
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.2, new OptimizedParameterSet(60, 60, 40, "", 163)) );
    // PLR = 0.25, use key=2.5 means using this opt if plr>0.25
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(2.5, new OptimizedParameterSet(60, 60, 49, "", 163)) );
    
    // RTT=30ms
    rtt_opts_map->insert( pair<int, plr_map*>(30, plr_opts_map) );
    /*********** Added Optimum of RTT = 30 ms **************/

//    plr_opts_map = new plr_map;
//
//    /*********** Adding Optimum of RTT = 40 ms **************/
//    // Simplify Module from Manuel.
//    // PLR = 0.12
//    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.12, new OptimizedParameterSet(1, 0, 0, "1 1 6", 3)) );
//    // PLR = 0.25, use key=2.5 means using this opt if plr>0.25
//    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(2.5, new OptimizedParameterSet(60, 60, 49, "", 163)) );
//    
//    // RTT=40ms
//    rtt_opts_map->insert( pair<int, plr_map*>(40, plr_opts_map) );
//    /*********** Added Optimum of RTT = 40 ms **************/

    plr_opts_map = new plr_map;

    /*********** Adding Optimum of RTT = 50 ms **************/
    // Simplify Module from Manuel.
    // PLR = 0.08
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.08, new OptimizedParameterSet(1, 0, 0, "1 4", 3)) );
    // PLR = 0.10
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.1, new OptimizedParameterSet(57, 57, 22, "", 155)) );
    // PLR = 0.12
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.12, new OptimizedParameterSet(57, 57, 25, "", 155)) );
    // PLR = 0.14
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.14, new OptimizedParameterSet(57, 57, 28, "", 155)) );
    // PLR = 0.16
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.16, new OptimizedParameterSet(57, 57, 32, "", 155)) );
    // PLR = 0.18
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.18, new OptimizedParameterSet(57, 57, 35, "", 155)) );
    // PLR = 0.20
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.20, new OptimizedParameterSet(57, 57, 38, "", 155)) );
    // PLR = 0.25, use key=2.5 means using this opt if plr>0.25
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(2.5, new OptimizedParameterSet(57, 57, 48, "", 155)) );
    
    // RTT=50ms
    rtt_opts_map->insert( pair<int, plr_map*>(50, plr_opts_map) );
    /*********** Added Optimum of RTT = 50 ms **************/

//    plr_opts_map = new plr_map;
//
//    /*********** Adding Optimum of RTT = 60 ms **************/
//    // Simplify Module from Manuel.
//    // PLR = 0.06
//    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.06, new OptimizedParameterSet(25, 25, 3, "9", 208)) );
//    // PLR = 0.25, use key=2.5 means using this opt if plr>0.25
//    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(2.5, new OptimizedParameterSet(55, 55, 47, "", 149)) );
//    
//    // RTT=60ms
//    rtt_opts_map->insert( pair<int, plr_map*>(60, plr_opts_map) );
//    /*********** Added Optimum of RTT = 60 ms **************/

    plr_opts_map = new plr_map;

    /*********** Adding Optimum of RTT = 80 ms **************/
    // Simplify Module from Manuel.
    // PLR = 0.03
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.03, new OptimizedParameterSet(1, 0, 0, "3", 3)) );
    // PLR = 0.06
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.06, new OptimizedParameterSet(14, 14, 2, "7", 38)) );
    // PLR = 0.08
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.08, new OptimizedParameterSet(51, 51, 18, "", 138)) );
    // PLR = 0.10
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.10, new OptimizedParameterSet(51, 51, 21, "", 138)) );
    // PLR = 0.12
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.12, new OptimizedParameterSet(51, 51, 24, "", 138)) );
    // PLR = 0.14
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.14, new OptimizedParameterSet(51, 51, 27, "", 138)) );
    // PLR = 0.16
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.16, new OptimizedParameterSet(51, 51, 30, "", 138)) );
    // PLR = 0.18
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.18, new OptimizedParameterSet(51, 51, 33, "", 138)) );
    // PLR = 0.2
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.2, new OptimizedParameterSet(51, 51, 36, "", 138)) );
    // PLR = 0.25, use key=2.5 means using this opt if plr>0.25
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(2.5, new OptimizedParameterSet(51, 51, 44, "", 138)) );
    
    // RTT=80ms
    rtt_opts_map->insert( pair<int, plr_map*>(80, plr_opts_map) );
    /*********** Added Optimum of RTT = 80 ms **************/

    plr_opts_map = new plr_map;

    /*********** Adding Optimum of RTT = 100 ms **************/
    // Simplify Module from Manuel.
    // PLR = 0.05
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.05, new OptimizedParameterSet(1, 0, 0, "4", 3)) );
    // PLR = 0.06
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.06, new OptimizedParameterSet(47, 47, 14, "", 127)) );
    // PLR = 0.08
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.08, new OptimizedParameterSet(47, 47, 17, "", 127)) );
    // PLR = 0.10
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.10, new OptimizedParameterSet(47, 47, 20, "", 127)) );
    // PLR = 0.12
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.12, new OptimizedParameterSet(47, 47, 23, "", 127)) );
    // PLR = 0.14
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.14, new OptimizedParameterSet(47, 47, 25, "", 127)) );
    // PLR = 0.16
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.16, new OptimizedParameterSet(47, 47, 28, "", 127)) );
    // PLR = 0.18
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.18, new OptimizedParameterSet(47, 47, 31, "", 127)) );
    // PLR = 0.20
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.2, new OptimizedParameterSet(47, 47, 34, "", 127)) );
    // PLR = 0.25, use key=2.5 means using this opt if plr>0.25
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(2.5, new OptimizedParameterSet(47, 47, 42, "", 127)) );
    
    // RTT=100ms, use key=1000 means using this opt if rtt>100ms
    rtt_opts_map->insert( pair<int, plr_map*>(1000, plr_opts_map) );
    /*********** Added Optimum of RTT = 80 ms **************/

    m_num_recv_map.insert( pair<int, rtt_map*>(1, rtt_opts_map) );
 std::cout << "DelayCell addOneReceiverMap() #recv_map=" << m_num_recv_map.size() << " #rtt_map=" << rtt_opts_map->size() << std::endl;
  }

  void DelayCell::add12MbpsOneReceiverMap()
  {
    // Targeted Delay = 0.1 & #receivers = 1
    num_recv_map::iterator rcv_it = m_num_recv_map.find(1);

    if ( rcv_it != m_num_recv_map.end() ) {
      // The parameters for one receiver already exist. No need to create.
      return;
    }

    rtt_map *rtt_opts_map = new rtt_map;
    plr_map *plr_opts_map = new plr_map;

plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.001, new OptimizedParameterSet(1, 1, 0, "1 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.010, new OptimizedParameterSet(10, 4, 0, "1 1 1 1 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.020, new OptimizedParameterSet(1, 3, 0, "1 1 1 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.030, new OptimizedParameterSet(1, 3, 0, "1 1 1 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.040, new OptimizedParameterSet(1, 4, 0, "1 1 1 1 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.050, new OptimizedParameterSet(1, 4, 0, "1 1 1 1 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.060, new OptimizedParameterSet(1, 4, 0, "1 1 1 1 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.070, new OptimizedParameterSet(1, 5, 0, "1 1 1 1 1 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.080, new OptimizedParameterSet(1, 5, 0, "1 1 1 1 1 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.090, new OptimizedParameterSet(1, 5, 0, "1 1 1 1 1 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.100, new OptimizedParameterSet(1, 6, 0, "1 1 1 1 2 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.110, new OptimizedParameterSet(1, 6, 0, "1 1 1 1 2 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.120, new OptimizedParameterSet(1, 6, 0, "1 1 1 1 2 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.130, new OptimizedParameterSet(1, 6, 0, "1 1 1 1 2 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.140, new OptimizedParameterSet(1, 7, 0, "1 1 1 1 3 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.150, new OptimizedParameterSet(1, 7, 0, "1 1 1 1 3 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.160, new OptimizedParameterSet(1, 7, 0, "1 1 1 1 3 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.170, new OptimizedParameterSet(1, 7, 0, "1 1 1 1 3 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.180, new OptimizedParameterSet(1, 8, 0, "1 1 1 1 4 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.190, new OptimizedParameterSet(1, 8, 0, "1 1 1 1 4 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.200, new OptimizedParameterSet(1, 8, 0, "1 1 1 1 4 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.250, new OptimizedParameterSet(1, 9, 0, "1 1 1 1 5 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(3.000, new OptimizedParameterSet(1, 11, 0, "1 1 1 2 6 ", 1)) );
rtt_opts_map->insert( pair<int, plr_map*>(10, plr_opts_map) );
plr_opts_map = new plr_map;
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.001, new OptimizedParameterSet(1, 1, 0, "1 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.010, new OptimizedParameterSet(10, 4, 0, "1 1 1 1 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.020, new OptimizedParameterSet(1, 3, 0, "1 1 1 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.030, new OptimizedParameterSet(1, 3, 0, "1 1 1 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.040, new OptimizedParameterSet(1, 4, 0, "1 1 1 1 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.050, new OptimizedParameterSet(1, 4, 0, "1 1 1 1 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.060, new OptimizedParameterSet(1, 4, 0, "1 1 1 1 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.070, new OptimizedParameterSet(1, 5, 0, "1 1 1 2 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.080, new OptimizedParameterSet(1, 5, 0, "1 1 1 2 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.090, new OptimizedParameterSet(1, 5, 0, "1 1 1 2 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.100, new OptimizedParameterSet(1, 6, 0, "1 1 1 3 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.110, new OptimizedParameterSet(1, 6, 0, "1 1 1 3 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.120, new OptimizedParameterSet(1, 6, 0, "1 1 1 3 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.130, new OptimizedParameterSet(1, 6, 0, "1 1 1 3 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.140, new OptimizedParameterSet(1, 7, 0, "1 1 1 4 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.150, new OptimizedParameterSet(1, 7, 0, "1 1 1 4 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.160, new OptimizedParameterSet(1, 7, 0, "1 1 1 4 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.170, new OptimizedParameterSet(1, 7, 0, "1 1 1 4 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.180, new OptimizedParameterSet(1, 8, 0, "1 1 1 5 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.190, new OptimizedParameterSet(1, 8, 0, "1 1 1 5 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.200, new OptimizedParameterSet(1, 8, 0, "1 1 1 5 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.250, new OptimizedParameterSet(1, 10, 0, "1 1 2 6 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(3.000, new OptimizedParameterSet(1, 12, 0, "1 1 2 8 ", 1)) );
rtt_opts_map->insert( pair<int, plr_map*>(20, plr_opts_map) );
plr_opts_map = new plr_map;
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.001, new OptimizedParameterSet(1, 1, 0, "1 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.010, new OptimizedParameterSet(1, 3, 0, "1 1 1 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.020, new OptimizedParameterSet(1, 3, 0, "1 1 1 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.030, new OptimizedParameterSet(1, 3, 0, "1 1 1 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.040, new OptimizedParameterSet(1, 4, 0, "1 1 2 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.050, new OptimizedParameterSet(1, 4, 0, "1 1 2 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.060, new OptimizedParameterSet(1, 4, 0, "1 1 2 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.070, new OptimizedParameterSet(1, 5, 0, "1 1 3 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.080, new OptimizedParameterSet(1, 5, 0, "1 1 3 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.090, new OptimizedParameterSet(1, 5, 0, "1 1 3 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.100, new OptimizedParameterSet(1, 6, 0, "1 1 4 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.110, new OptimizedParameterSet(1, 6, 0, "1 1 4 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.120, new OptimizedParameterSet(1, 6, 0, "1 1 4 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.130, new OptimizedParameterSet(1, 6, 0, "1 1 4 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.140, new OptimizedParameterSet(1, 7, 0, "1 1 5 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.150, new OptimizedParameterSet(1, 7, 0, "1 1 5 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.160, new OptimizedParameterSet(1, 7, 0, "1 1 5 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.170, new OptimizedParameterSet(1, 7, 0, "1 1 5 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.180, new OptimizedParameterSet(1, 8, 0, "1 1 6 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.190, new OptimizedParameterSet(1, 8, 0, "1 1 6 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.200, new OptimizedParameterSet(1, 8, 0, "1 1 6 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(2.500, new OptimizedParameterSet(1, 10, 0, "1 2 7 ", 1)) );
rtt_opts_map->insert( pair<int, plr_map*>(30, plr_opts_map) );
plr_opts_map = new plr_map;
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.001, new OptimizedParameterSet(1, 1, 0, "1 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.010, new OptimizedParameterSet(1, 3, 0, "1 2 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.020, new OptimizedParameterSet(1, 3, 0, "1 2 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.030, new OptimizedParameterSet(1, 3, 0, "1 2 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.040, new OptimizedParameterSet(1, 4, 0, "1 3 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.050, new OptimizedParameterSet(1, 4, 0, "1 3 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.060, new OptimizedParameterSet(1, 4, 0, "1 3 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.070, new OptimizedParameterSet(1, 5, 0, "1 4 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.080, new OptimizedParameterSet(1, 5, 0, "1 4 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.090, new OptimizedParameterSet(1, 5, 0, "1 4 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.100, new OptimizedParameterSet(1, 6, 0, "1 5 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.110, new OptimizedParameterSet(1, 6, 0, "1 5 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.120, new OptimizedParameterSet(1, 6, 0, "1 5 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.130, new OptimizedParameterSet(1, 6, 0, "1 5 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.140, new OptimizedParameterSet(1, 7, 0, "1 6 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.150, new OptimizedParameterSet(1, 7, 0, "1 6 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.160, new OptimizedParameterSet(1, 7, 0, "1 6 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.170, new OptimizedParameterSet(1, 7, 0, "1 6 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.180, new OptimizedParameterSet(10, 15, 3, "4 8 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.190, new OptimizedParameterSet(10, 15, 3, "4 8 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(2.000, new OptimizedParameterSet(10, 16, 3, "5 8 ", 10)) );
rtt_opts_map->insert( pair<int, plr_map*>(40, plr_opts_map) );
plr_opts_map = new plr_map;
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.001, new OptimizedParameterSet(1, 1, 0, "1 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.010, new OptimizedParameterSet(1, 3, 0, "1 2 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.020, new OptimizedParameterSet(1, 3, 0, "1 2 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.030, new OptimizedParameterSet(1, 3, 0, "1 2 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.040, new OptimizedParameterSet(1, 4, 0, "1 3 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.050, new OptimizedParameterSet(1, 4, 0, "1 3 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.060, new OptimizedParameterSet(1, 4, 0, "1 3 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.070, new OptimizedParameterSet(1, 5, 0, "1 4 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.080, new OptimizedParameterSet(1, 5, 0, "1 4 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.090, new OptimizedParameterSet(1, 5, 0, "1 4 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.100, new OptimizedParameterSet(1, 6, 0, "1 5 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.110, new OptimizedParameterSet(1, 6, 0, "1 5 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.120, new OptimizedParameterSet(1, 6, 0, "1 5 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.130, new OptimizedParameterSet(1, 6, 0, "1 5 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.140, new OptimizedParameterSet(1, 7, 0, "1 6 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.150, new OptimizedParameterSet(1, 7, 0, "1 6 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.160, new OptimizedParameterSet(1, 7, 0, "1 6 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.170, new OptimizedParameterSet(1, 7, 0, "1 6 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.180, new OptimizedParameterSet(10, 15, 3, "4 8 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.190, new OptimizedParameterSet(10, 15, 3, "4 8 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(2.000, new OptimizedParameterSet(10, 16, 3, "5 8 ", 10)) );
rtt_opts_map->insert( pair<int, plr_map*>(50, plr_opts_map) );
plr_opts_map = new plr_map;
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.001, new OptimizedParameterSet(1, 1, 0, "1 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.010, new OptimizedParameterSet(1, 3, 0, "3 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.020, new OptimizedParameterSet(1, 3, 0, "3 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.030, new OptimizedParameterSet(1, 3, 0, "3 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.040, new OptimizedParameterSet(10, 6, 1, "5 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.050, new OptimizedParameterSet(10, 7, 1, "6 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.060, new OptimizedParameterSet(10, 8, 1, "7 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.070, new OptimizedParameterSet(10, 8, 1, "7 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.080, new OptimizedParameterSet(10, 9, 2, "7 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.090, new OptimizedParameterSet(10, 9, 2, "7 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.100, new OptimizedParameterSet(10, 10, 2, "8 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.110, new OptimizedParameterSet(10, 11, 2, "9 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.120, new OptimizedParameterSet(10, 11, 3, "8 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.130, new OptimizedParameterSet(10, 12, 3, "9 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.140, new OptimizedParameterSet(10, 12, 3, "9 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.150, new OptimizedParameterSet(10, 13, 3, "10 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.160, new OptimizedParameterSet(10, 13, 3, "10 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.170, new OptimizedParameterSet(10, 14, 4, "10 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.180, new OptimizedParameterSet(10, 15, 4, "11 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.190, new OptimizedParameterSet(10, 15, 4, "11 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(2.000, new OptimizedParameterSet(10, 16, 4, "12 ", 10)) );
rtt_opts_map->insert( pair<int, plr_map*>(60, plr_opts_map) );
plr_opts_map = new plr_map;
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.001, new OptimizedParameterSet(1, 1, 0, "1 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.010, new OptimizedParameterSet(1, 3, 0, "3 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.020, new OptimizedParameterSet(1, 3, 0, "3 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.030, new OptimizedParameterSet(1, 3, 0, "3 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.040, new OptimizedParameterSet(10, 6, 1, "5 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.050, new OptimizedParameterSet(10, 7, 1, "6 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.060, new OptimizedParameterSet(10, 8, 1, "7 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.070, new OptimizedParameterSet(10, 8, 1, "7 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.080, new OptimizedParameterSet(10, 9, 2, "7 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.090, new OptimizedParameterSet(10, 9, 2, "7 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.100, new OptimizedParameterSet(10, 10, 2, "8 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.110, new OptimizedParameterSet(10, 11, 2, "9 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.120, new OptimizedParameterSet(10, 11, 3, "8 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.130, new OptimizedParameterSet(10, 12, 3, "9 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.140, new OptimizedParameterSet(10, 12, 3, "9 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.150, new OptimizedParameterSet(10, 13, 3, "10 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.160, new OptimizedParameterSet(10, 13, 3, "10 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.170, new OptimizedParameterSet(10, 14, 4, "10 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.180, new OptimizedParameterSet(10, 15, 4, "11 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.190, new OptimizedParameterSet(10, 15, 4, "11 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(2.000, new OptimizedParameterSet(10, 16, 4, "12 ", 10)) );
rtt_opts_map->insert( pair<int, plr_map*>(70, plr_opts_map) );
plr_opts_map = new plr_map;
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.001, new OptimizedParameterSet(1, 1, 0, "1 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.010, new OptimizedParameterSet(1, 3, 0, "3 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.020, new OptimizedParameterSet(1, 3, 0, "3 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.030, new OptimizedParameterSet(1, 3, 0, "3 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.040, new OptimizedParameterSet(10, 6, 1, "5 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.050, new OptimizedParameterSet(10, 7, 1, "6 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.060, new OptimizedParameterSet(10, 8, 1, "7 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.070, new OptimizedParameterSet(10, 8, 1, "7 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.080, new OptimizedParameterSet(10, 9, 2, "7 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.090, new OptimizedParameterSet(10, 9, 2, "7 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.100, new OptimizedParameterSet(10, 10, 2, "8 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.110, new OptimizedParameterSet(10, 11, 2, "9 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.120, new OptimizedParameterSet(10, 11, 3, "8 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.130, new OptimizedParameterSet(10, 12, 3, "9 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.140, new OptimizedParameterSet(10, 12, 3, "9 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.150, new OptimizedParameterSet(10, 13, 3, "10 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.160, new OptimizedParameterSet(10, 13, 3, "10 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.170, new OptimizedParameterSet(10, 14, 4, "10 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.180, new OptimizedParameterSet(10, 15, 4, "11 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.190, new OptimizedParameterSet(10, 15, 4, "11 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(2.000, new OptimizedParameterSet(10, 16, 4, "12 ", 10)) );
rtt_opts_map->insert( pair<int, plr_map*>(80, plr_opts_map) );
plr_opts_map = new plr_map;
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.001, new OptimizedParameterSet(1, 1, 0, "1 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.010, new OptimizedParameterSet(1, 3, 0, "3 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.020, new OptimizedParameterSet(1, 3, 0, "3 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.030, new OptimizedParameterSet(1, 3, 0, "3 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.040, new OptimizedParameterSet(10, 6, 1, "5 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.050, new OptimizedParameterSet(10, 7, 1, "6 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.060, new OptimizedParameterSet(10, 8, 1, "7 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.070, new OptimizedParameterSet(10, 8, 1, "7 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.080, new OptimizedParameterSet(10, 9, 2, "7 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.090, new OptimizedParameterSet(10, 9, 2, "7 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.100, new OptimizedParameterSet(10, 10, 2, "8 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.110, new OptimizedParameterSet(10, 11, 2, "9 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.120, new OptimizedParameterSet(10, 11, 3, "8 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.130, new OptimizedParameterSet(10, 12, 3, "9 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.140, new OptimizedParameterSet(10, 12, 3, "9 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.150, new OptimizedParameterSet(10, 13, 3, "10 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.160, new OptimizedParameterSet(10, 13, 3, "10 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.170, new OptimizedParameterSet(10, 14, 4, "10 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.180, new OptimizedParameterSet(10, 15, 4, "11 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.190, new OptimizedParameterSet(10, 15, 4, "11 ", 10)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(2.000, new OptimizedParameterSet(10, 16, 4, "12 ", 10)) );
rtt_opts_map->insert( pair<int, plr_map*>(1000, plr_opts_map) );

    m_num_recv_map.insert( pair<int, rtt_map*>(1, rtt_opts_map) );
 std::cout << "DelayCell add12MbpsOneReceiverMap() #recv_map=" << m_num_recv_map.size() << " #rtt_map=" << rtt_opts_map->size() << std::endl;
  }

  void DelayCell::addTwoReceiversMap()
  {
    // Targeted Delay = 0.1 & #receivers = 2
    num_recv_map::iterator rcv_it = m_num_recv_map.find(2);

    if ( rcv_it != m_num_recv_map.end() ) {
      // The parameters for one receiver already exist. No need to create.
      return;
    }
    rtt_map *rtt_opts_map = new rtt_map;
    plr_map *plr_opts_map = new plr_map;
//plr_opts_map = new plr_map;
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.001, new OptimizedParameterSet(10, 2, 0, "1 1 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.010, new OptimizedParameterSet(10, 4, 0, "1 1 1 1 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.020, new OptimizedParameterSet(10, 5, 0, "1 1 1 2 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.030, new OptimizedParameterSet(10, 6, 0, "1 1 1 3 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.040, new OptimizedParameterSet(10, 6, 0, "1 1 1 3 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.050, new OptimizedParameterSet(9, 7, 0, "1 1 1 1 3 ", 25)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.060, new OptimizedParameterSet(9, 7, 0, "1 1 1 1 3 ", 25)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.070, new OptimizedParameterSet(9, 8, 0, "1 1 1 1 4 ", 25)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.080, new OptimizedParameterSet(9, 8, 0, "1 1 1 1 4 ", 25)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.090, new OptimizedParameterSet(9, 9, 0, "1 1 1 1 5 ", 25)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.100, new OptimizedParameterSet(9, 10, 0, "1 1 1 2 5 ", 25)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.110, new OptimizedParameterSet(9, 10, 0, "1 1 1 2 5 ", 25)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.120, new OptimizedParameterSet(9, 11, 0, "1 1 1 2 6 ", 25)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.130, new OptimizedParameterSet(9, 11, 0, "1 1 1 2 6 ", 25)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.140, new OptimizedParameterSet(9, 12, 0, "1 1 1 2 7 ", 25)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.150, new OptimizedParameterSet(9, 12, 0, "1 1 1 2 7 ", 25)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.160, new OptimizedParameterSet(9, 13, 0, "1 1 1 2 8 ", 25)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.170, new OptimizedParameterSet(9, 13, 0, "1 1 1 2 8 ", 25)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.180, new OptimizedParameterSet(9, 14, 0, "1 1 1 3 8 ", 25)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.190, new OptimizedParameterSet(9, 15, 0, "1 2 2 4 6 ", 25)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.200, new OptimizedParameterSet(9, 15, 0, "1 2 2 4 6 ", 25)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.250, new OptimizedParameterSet(1, 9, 0, "1 1 1 2 4 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(3.000, new OptimizedParameterSet(1, 11, 0, "1 1 1 2 6 ", 3)) );
rtt_opts_map->insert( pair<int, plr_map*>(10, plr_opts_map) );
plr_opts_map = new plr_map;
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.001, new OptimizedParameterSet(10, 2, 0, "1 1 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.010, new OptimizedParameterSet(10, 4, 0, "1 1 2 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.020, new OptimizedParameterSet(10, 5, 0, "1 1 3 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.030, new OptimizedParameterSet(10, 6, 0, "1 1 4 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.040, new OptimizedParameterSet(10, 6, 0, "1 1 4 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.050, new OptimizedParameterSet(3, 5, 0, "1 1 1 2 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.060, new OptimizedParameterSet(3, 5, 0, "1 1 1 2 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.070, new OptimizedParameterSet(3, 6, 0, "1 1 1 3 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.080, new OptimizedParameterSet(3, 6, 0, "1 1 1 3 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.090, new OptimizedParameterSet(3, 7, 0, "1 1 1 4 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.100, new OptimizedParameterSet(3, 7, 0, "1 1 1 4 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.110, new OptimizedParameterSet(3, 7, 0, "1 1 1 4 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.120, new OptimizedParameterSet(3, 8, 0, "1 1 1 5 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.130, new OptimizedParameterSet(3, 8, 0, "1 1 1 5 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.140, new OptimizedParameterSet(3, 8, 0, "1 1 1 5 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.150, new OptimizedParameterSet(3, 9, 0, "1 1 2 5 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.160, new OptimizedParameterSet(3, 9, 0, "1 1 2 5 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.170, new OptimizedParameterSet(3, 9, 0, "1 1 2 5 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.180, new OptimizedParameterSet(10, 15, 2, "2 4 7 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.190, new OptimizedParameterSet(10, 15, 2, "2 4 7 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.200, new OptimizedParameterSet(10, 16, 2, "2 4 8 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.250, new OptimizedParameterSet(1, 10, 0, "1 1 2 6 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(3.000, new OptimizedParameterSet(1, 12, 0, "1 1 2 8 ", 3)) );
rtt_opts_map->insert( pair<int, plr_map*>(20, plr_opts_map) );
plr_opts_map = new plr_map;
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.001, new OptimizedParameterSet(10, 2, 0, "1 1 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.010, new OptimizedParameterSet(5, 3, 0, "1 1 1 ", 14)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.020, new OptimizedParameterSet(5, 4, 0, "1 1 2 ", 14)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.030, new OptimizedParameterSet(5, 5, 0, "1 1 3 ", 14)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.040, new OptimizedParameterSet(5, 5, 0, "1 1 3 ", 14)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.050, new OptimizedParameterSet(5, 6, 0, "1 1 4 ", 14)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.060, new OptimizedParameterSet(5, 6, 0, "1 1 4 ", 14)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.070, new OptimizedParameterSet(5, 7, 0, "1 1 5 ", 14)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.080, new OptimizedParameterSet(1, 5, 0, "1 1 3 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.090, new OptimizedParameterSet(1, 5, 0, "1 1 3 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.100, new OptimizedParameterSet(1, 6, 0, "1 1 4 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.110, new OptimizedParameterSet(1, 6, 0, "1 1 4 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.120, new OptimizedParameterSet(1, 6, 0, "1 1 4 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.130, new OptimizedParameterSet(1, 6, 0, "1 1 4 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.140, new OptimizedParameterSet(1, 7, 0, "1 1 5 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.150, new OptimizedParameterSet(1, 7, 0, "1 1 5 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.160, new OptimizedParameterSet(1, 7, 0, "1 1 5 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.170, new OptimizedParameterSet(1, 7, 0, "1 1 5 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.180, new OptimizedParameterSet(10, 15, 3, "4 8 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.190, new OptimizedParameterSet(10, 15, 3, "4 8 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.200, new OptimizedParameterSet(5, 12, 1, "2 3 6 ", 14)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(2.500, new OptimizedParameterSet(1, 10, 0, "1 2 7 ", 3)) );
rtt_opts_map->insert( pair<int, plr_map*>(30, plr_opts_map) );
plr_opts_map = new plr_map;
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.001, new OptimizedParameterSet(10, 2, 0, "1 1 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.010, new OptimizedParameterSet(1, 3, 0, "1 2 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.020, new OptimizedParameterSet(1, 3, 0, "1 2 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.030, new OptimizedParameterSet(1, 3, 0, "1 2 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.040, new OptimizedParameterSet(1, 4, 0, "1 3 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.050, new OptimizedParameterSet(1, 4, 0, "1 3 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.060, new OptimizedParameterSet(1, 4, 0, "1 3 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.070, new OptimizedParameterSet(1, 5, 0, "1 4 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.080, new OptimizedParameterSet(1, 5, 0, "1 4 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.090, new OptimizedParameterSet(1, 5, 0, "1 4 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.100, new OptimizedParameterSet(10, 10, 2, "3 5 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.110, new OptimizedParameterSet(10, 11, 2, "3 6 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.120, new OptimizedParameterSet(10, 11, 2, "3 6 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.130, new OptimizedParameterSet(10, 12, 2, "4 6 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.140, new OptimizedParameterSet(10, 12, 3, "3 6 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.150, new OptimizedParameterSet(10, 13, 3, "3 7 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.160, new OptimizedParameterSet(10, 13, 3, "3 7 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.170, new OptimizedParameterSet(10, 14, 3, "4 7 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.180, new OptimizedParameterSet(10, 15, 3, "4 8 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.190, new OptimizedParameterSet(10, 15, 3, "4 8 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(2.000, new OptimizedParameterSet(10, 16, 4, "4 8 ", 28)) );
rtt_opts_map->insert( pair<int, plr_map*>(40, plr_opts_map) );
plr_opts_map = new plr_map;
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.001, new OptimizedParameterSet(5, 2, 0, "1 1 ", 14)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.010, new OptimizedParameterSet(5, 3, 0, "1 2 ", 14)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.020, new OptimizedParameterSet(1, 3, 0, "1 2 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.030, new OptimizedParameterSet(1, 3, 0, "1 2 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.040, new OptimizedParameterSet(1, 4, 0, "1 3 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.050, new OptimizedParameterSet(1, 4, 0, "1 3 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.060, new OptimizedParameterSet(1, 4, 0, "1 3 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.070, new OptimizedParameterSet(1, 5, 0, "1 4 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.080, new OptimizedParameterSet(1, 5, 0, "1 4 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.090, new OptimizedParameterSet(1, 5, 0, "1 4 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.100, new OptimizedParameterSet(1, 6, 0, "1 5 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.110, new OptimizedParameterSet(1, 6, 0, "1 5 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.120, new OptimizedParameterSet(1, 6, 0, "1 5 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.130, new OptimizedParameterSet(5, 9, 1, "3 5 ", 14)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.140, new OptimizedParameterSet(10, 12, 4, "8 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.150, new OptimizedParameterSet(10, 13, 4, "9 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.160, new OptimizedParameterSet(5, 10, 2, "3 5 ", 14)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.170, new OptimizedParameterSet(5, 11, 2, "3 6 ", 14)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.180, new OptimizedParameterSet(5, 11, 2, "3 6 ", 14)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.190, new OptimizedParameterSet(10, 15, 5, "10 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(2.000, new OptimizedParameterSet(5, 12, 2, "4 6 ", 14)) );
rtt_opts_map->insert( pair<int, plr_map*>(50, plr_opts_map) );
plr_opts_map = new plr_map;
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.001, new OptimizedParameterSet(1, 1, 0, "1 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.010, new OptimizedParameterSet(1, 3, 0, "3 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.020, new OptimizedParameterSet(10, 5, 1, "4 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.030, new OptimizedParameterSet(10, 6, 1, "5 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.040, new OptimizedParameterSet(10, 6, 1, "5 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.050, new OptimizedParameterSet(10, 7, 1, "6 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.060, new OptimizedParameterSet(10, 8, 2, "6 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.070, new OptimizedParameterSet(10, 8, 2, "6 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.080, new OptimizedParameterSet(10, 9, 2, "7 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.090, new OptimizedParameterSet(10, 9, 2, "7 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.100, new OptimizedParameterSet(10, 10, 3, "7 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.110, new OptimizedParameterSet(10, 11, 3, "8 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.120, new OptimizedParameterSet(10, 11, 3, "8 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.130, new OptimizedParameterSet(10, 12, 3, "9 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.140, new OptimizedParameterSet(10, 12, 4, "8 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.150, new OptimizedParameterSet(10, 13, 4, "9 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.160, new OptimizedParameterSet(10, 13, 4, "9 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.170, new OptimizedParameterSet(10, 14, 4, "10 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.180, new OptimizedParameterSet(10, 15, 5, "10 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.190, new OptimizedParameterSet(10, 15, 5, "10 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(2.000, new OptimizedParameterSet(10, 16, 5, "11 ", 28)) );
rtt_opts_map->insert( pair<int, plr_map*>(60, plr_opts_map) );
plr_opts_map = new plr_map;
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.001, new OptimizedParameterSet(1, 1, 0, "1 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.010, new OptimizedParameterSet(1, 3, 0, "3 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.020, new OptimizedParameterSet(10, 5, 1, "4 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.030, new OptimizedParameterSet(10, 6, 1, "5 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.040, new OptimizedParameterSet(10, 6, 1, "5 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.050, new OptimizedParameterSet(10, 7, 1, "6 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.060, new OptimizedParameterSet(10, 8, 2, "6 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.070, new OptimizedParameterSet(10, 8, 2, "6 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.080, new OptimizedParameterSet(10, 9, 2, "7 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.090, new OptimizedParameterSet(10, 9, 2, "7 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.100, new OptimizedParameterSet(10, 10, 3, "7 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.110, new OptimizedParameterSet(10, 11, 3, "8 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.120, new OptimizedParameterSet(10, 11, 3, "8 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.130, new OptimizedParameterSet(10, 12, 3, "9 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.140, new OptimizedParameterSet(10, 12, 4, "8 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.150, new OptimizedParameterSet(10, 13, 4, "9 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.160, new OptimizedParameterSet(10, 13, 4, "9 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.170, new OptimizedParameterSet(10, 14, 4, "10 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.180, new OptimizedParameterSet(10, 15, 5, "10 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.190, new OptimizedParameterSet(10, 15, 5, "10 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(2.000, new OptimizedParameterSet(10, 16, 5, "11 ", 28)) );
rtt_opts_map->insert( pair<int, plr_map*>(70, plr_opts_map) );
plr_opts_map = new plr_map;
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.001, new OptimizedParameterSet(1, 1, 0, "1 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.010, new OptimizedParameterSet(1, 3, 0, "3 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.020, new OptimizedParameterSet(10, 5, 1, "4 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.030, new OptimizedParameterSet(10, 6, 1, "5 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.040, new OptimizedParameterSet(10, 6, 1, "5 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.050, new OptimizedParameterSet(10, 7, 1, "6 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.060, new OptimizedParameterSet(10, 8, 2, "6 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.070, new OptimizedParameterSet(10, 8, 2, "6 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.080, new OptimizedParameterSet(10, 9, 2, "7 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.090, new OptimizedParameterSet(10, 9, 2, "7 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.100, new OptimizedParameterSet(10, 10, 3, "7 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.110, new OptimizedParameterSet(10, 11, 3, "8 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.120, new OptimizedParameterSet(10, 11, 3, "8 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.130, new OptimizedParameterSet(10, 12, 3, "9 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.140, new OptimizedParameterSet(10, 12, 4, "8 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.150, new OptimizedParameterSet(10, 13, 4, "9 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.160, new OptimizedParameterSet(10, 13, 4, "9 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.170, new OptimizedParameterSet(10, 14, 4, "10 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.180, new OptimizedParameterSet(10, 15, 5, "10 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.190, new OptimizedParameterSet(10, 15, 5, "10 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(2.000, new OptimizedParameterSet(10, 16, 5, "11 ", 28)) );
rtt_opts_map->insert( pair<int, plr_map*>(80, plr_opts_map) );
plr_opts_map = new plr_map;
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.001, new OptimizedParameterSet(1, 1, 0, "1 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.010, new OptimizedParameterSet(3, 3, 0, "3 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.020, new OptimizedParameterSet(1, 3, 0, "3 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.030, new OptimizedParameterSet(1, 3, 0, "3 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.040, new OptimizedParameterSet(1, 4, 0, "4 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.050, new OptimizedParameterSet(3, 5, 1, "4 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.060, new OptimizedParameterSet(3, 5, 1, "4 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.070, new OptimizedParameterSet(3, 6, 1, "5 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.080, new OptimizedParameterSet(3, 6, 1, "5 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.090, new OptimizedParameterSet(3, 7, 1, "6 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.100, new OptimizedParameterSet(3, 7, 1, "6 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.110, new OptimizedParameterSet(3, 7, 1, "6 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.120, new OptimizedParameterSet(3, 8, 1, "7 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.130, new OptimizedParameterSet(3, 8, 1, "7 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.140, new OptimizedParameterSet(3, 8, 2, "6 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.150, new OptimizedParameterSet(3, 9, 2, "7 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.160, new OptimizedParameterSet(3, 9, 2, "7 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.170, new OptimizedParameterSet(3, 9, 2, "7 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.180, new OptimizedParameterSet(3, 10, 2, "8 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(1.900, new OptimizedParameterSet(3, 10, 2, "8 ", 9)) );
rtt_opts_map->insert( pair<int, plr_map*>(1000, plr_opts_map) );
    // RTT=100ms, use key=1000 means using this opt if rtt>100ms
    m_num_recv_map.insert( pair<double, rtt_map*>(2, rtt_opts_map) );

 std::cout << "DelayCell addTwoReceiversMap() #recv_map=" << m_num_recv_map.size() << " #rtt_map=" << rtt_opts_map->size() << std::endl;
return;
    //rtt_map *rtt_opts_map = new rtt_map;
    //plr_map *plr_opts_map = new plr_map;


    /*********** Adding Optimum of RTT = 20 ms **************/
    // Simplify Module from Manuel.
    // PLR = 0.001
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.001, new OptimizedParameterSet(33, 33, 0, "1 1", 114)) );
    // PLR = 0.02
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.02, new OptimizedParameterSet(18, 18, 0, "1 1 4", 49)) );
    // PLR = 0.06
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.06, new OptimizedParameterSet(3, 5, 0, "1 1 2", 8)) );
    // PLR = 0.11
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.11, new OptimizedParameterSet(3, 7, 0, "1 1 1 4", 8)) );
    // PLR = 0.14
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.14, new OptimizedParameterSet(33, 33, 5, "11", 89)) );
    // PLR = 0.17
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.17, new OptimizedParameterSet(33, 33, 6, "6 12", 89)) );
    // PLR = 0.2
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.2, new OptimizedParameterSet(33, 33, 7, "7 15", 89)) );
    // PLR = 0.25
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.25, new OptimizedParameterSet(1, 0, 0, "1 1 2 6", 89)) );
    // PLR = 0.3, use key=3.0 means using this opt if plr>0.30
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(3.0, new OptimizedParameterSet(1, 0, 0, "1 1 2 8", 89)) );
 
    rtt_opts_map->insert( pair<double, plr_map*>(20, plr_opts_map) );

    /*********** Adding Optimum of RTT = 40 ms **************/
    plr_opts_map = new plr_map;
    // Simplify Module from Manuel.
    // PLR = 0.001
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.001, new OptimizedParameterSet(14, 14, 0, "1 1", 38)) );
    // PLR = 0.03
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.03, new OptimizedParameterSet(1, 0, 0, "1 2", 3)) );
    // PLR = 0.06
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.06, new OptimizedParameterSet(1, 0, 0, "1 3", 3)) );
    // PLR = 0.08
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.08, new OptimizedParameterSet(36, 36, 6, "9", 98)) );
    // PLR = 0.10
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.10, new OptimizedParameterSet(36, 36, 7, "10", 98)) );
    // PLR = 0.12
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.12, new OptimizedParameterSet(36, 36, 8, "12", 98)) );
    // PLR = 0.14
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.14, new OptimizedParameterSet(36, 36, 10, "12", 98)) );
    // PLR = 0.16
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.16, new OptimizedParameterSet(36, 36, 11, "13", 98)) );
    // PLR = 0.18
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.18, new OptimizedParameterSet(36, 36, 12, "15", 98)) );
    // PLR = 0.20
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.20, new OptimizedParameterSet(36, 36, 14, "16", 98)) );
    // PLR = 0.25, use key=2.5 means using this opt if plr>0.25
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(2.5, new OptimizedParameterSet(58, 58, 48, "", 98)) );
 
    rtt_opts_map->insert( pair<double, plr_map*>(40, plr_opts_map) );

    /*********** Adding Optimum of RTT = 70 ms **************/
    plr_opts_map = new plr_map;
    // Simplify Module from Manuel.
    // PLR = 0.001
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.001, new OptimizedParameterSet(1, 0, 0, "1", 3)) );
    // PLR = 0.01
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.01, new OptimizedParameterSet(20, 20, 1, "3", 54)) );
    // PLR = 0.02
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.02, new OptimizedParameterSet(20, 20, 1, "5", 54)) );
    // PLR = 0.04
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.04, new OptimizedParameterSet(20, 20, 2, "6", 54)) );
    // PLR = 0.06
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.06, new OptimizedParameterSet(20, 20, 3, "7", 54)) );
    // PLR = 0.08
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.08, new OptimizedParameterSet(20, 20, 4, "7", 54)) );
    // PLR = 0.10
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.10, new OptimizedParameterSet(20, 20, 4, "9", 54)) );
    // PLR = 0.12
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.12, new OptimizedParameterSet(20, 20, 5, "10", 54)) );
    // PLR = 0.14
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.14, new OptimizedParameterSet(20, 20, 6, "10", 54)) );
    // PLR = 0.16
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.16, new OptimizedParameterSet(20, 20, 7, "11", 54)) );
    // PLR = 0.18
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.18, new OptimizedParameterSet(20, 20, 8, "12", 54)) );
    // PLR = 0.20
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.2, new OptimizedParameterSet(20, 20, 9, "13", 54)) );
    // PLR = 0.25, use key=2.5 means using this opt if plr>0.25
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(2.5, new OptimizedParameterSet(53, 53, 46, "", 144)) );
 
    rtt_opts_map->insert( pair<double, plr_map*>(70, plr_opts_map) );

    /*********** Adding Optimum of RTT = 80 ms **************/
    plr_opts_map = new plr_map;
    // Simplify Module from Manuel.
    // PLR = 0.001
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.001, new OptimizedParameterSet(1, 0, 0, "1", 3)) );
    // PLR = 0.01
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.01, new OptimizedParameterSet(1, 0, 0, "3", 38)) );
    // PLR = 0.02
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.02, new OptimizedParameterSet(14, 14, 1, "4", 38)) );
    // PLR = 0.04
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.04, new OptimizedParameterSet(14, 14, 2, "5", 38)) );
    // PLR = 0.07
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.07, new OptimizedParameterSet(14, 14, 2, "7", 38)) );
    // PLR = 0.10
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.10, new OptimizedParameterSet(14, 14, 3, "8", 38)) );
    // PLR = 0.13
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.13, new OptimizedParameterSet(14, 14, 4, "9", 38)) );
    // PLR = 0.16
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.16, new OptimizedParameterSet(14, 14, 5, "10", 38)) );
    // PLR = 0.18
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.18, new OptimizedParameterSet(14, 14, 6, "11", 38)) );
    // PLR = 0.20
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.20, new OptimizedParameterSet(14, 14, 7, "12", 38)) );
    // PLR = 0.25, use key=2.5 means using this opt if plr>0.25
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(2.5, new OptimizedParameterSet(51, 51, 44, "", 138)) );
 
    rtt_opts_map->insert( pair<double, plr_map*>(80, plr_opts_map) );

    /*********** Adding Optimum of RTT = 100 ms **************/
    plr_opts_map = new plr_map;
    // Simplify Module from Manuel.
    // PLR = 0.001
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.001, new OptimizedParameterSet(1, 0, 0, "1", 3)) );
    // PLR = 0.01
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.01, new OptimizedParameterSet(3, 3, 0, "3", 8)) );
    // PLR = 0.03
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.02, new OptimizedParameterSet(1, 0, 0, "3", 3)) );
    // PLR = 0.04
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.04, new OptimizedParameterSet(47, 47, 11, "", 127)) );
    // PLR = 0.06
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.06, new OptimizedParameterSet(47, 47, 14, "", 127)) );
    // PLR = 0.08
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.08, new OptimizedParameterSet(47, 47, 17, "", 127)) );
    // PLR = 0.10
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.10, new OptimizedParameterSet(47, 47, 20, "", 127)) );
    // PLR = 0.12
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.12, new OptimizedParameterSet(47, 47, 23, "", 127)) );
    // PLR = 0.14
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.14, new OptimizedParameterSet(47, 47, 25, "", 127)) );
    // PLR = 0.16
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.16, new OptimizedParameterSet(47, 47, 28, "", 127)) );
    // PLR = 0.18
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.18, new OptimizedParameterSet(47, 47, 31, "", 127)) );
    // PLR = 0.20
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.20, new OptimizedParameterSet(47, 47, 34, "", 127)) );
    // PLR = 0.25, use key=2.5 means using this opt if plr>0.25
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(2.5, new OptimizedParameterSet(47, 47, 42, "", 127)) );
 
    rtt_opts_map->insert( pair<double, plr_map*>(1000, plr_opts_map) );

    // RTT=100ms, use key=1000 means using this opt if rtt>100ms
    m_num_recv_map.insert( pair<double, rtt_map*>(2, rtt_opts_map) );

 std::cout << "DelayCell addTwoReceiversMap() #recv_map=" << m_num_recv_map.size() << " #rtt_map=" << rtt_opts_map->size() << std::endl;
  }

  void DelayCell::add12MbpsTwoReceiversMap()
  {
    // Targeted Delay = 0.1 & #receivers = 2
    num_recv_map::iterator rcv_it = m_num_recv_map.find(2);

    if ( rcv_it != m_num_recv_map.end() ) {
      // The parameters for one receiver already exist. No need to create.
      return;
    }
    rtt_map *rtt_opts_map = new rtt_map;
    plr_map *plr_opts_map = new plr_map;

plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.001, new OptimizedParameterSet(10, 2, 0, "1 1 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.010, new OptimizedParameterSet(10, 4, 0, "1 1 1 1 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.020, new OptimizedParameterSet(10, 5, 0, "1 1 1 2 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.030, new OptimizedParameterSet(10, 6, 0, "1 1 1 3 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.040, new OptimizedParameterSet(10, 6, 0, "1 1 1 3 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.050, new OptimizedParameterSet(9, 7, 0, "1 1 1 1 3 ", 25)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.060, new OptimizedParameterSet(9, 7, 0, "1 1 1 1 3 ", 25)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.070, new OptimizedParameterSet(9, 8, 0, "1 1 1 1 4 ", 25)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.080, new OptimizedParameterSet(9, 8, 0, "1 1 1 1 4 ", 25)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.090, new OptimizedParameterSet(9, 9, 0, "1 1 1 1 5 ", 25)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.100, new OptimizedParameterSet(9, 10, 0, "1 1 1 2 5 ", 25)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.110, new OptimizedParameterSet(9, 10, 0, "1 1 1 2 5 ", 25)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.120, new OptimizedParameterSet(9, 11, 0, "1 1 1 2 6 ", 25)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.130, new OptimizedParameterSet(9, 11, 0, "1 1 1 2 6 ", 25)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.140, new OptimizedParameterSet(9, 12, 0, "1 1 1 2 7 ", 25)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.150, new OptimizedParameterSet(9, 12, 0, "1 1 1 2 7 ", 25)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.160, new OptimizedParameterSet(9, 13, 0, "1 1 1 2 8 ", 25)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.170, new OptimizedParameterSet(9, 13, 0, "1 1 1 2 8 ", 25)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.180, new OptimizedParameterSet(9, 14, 0, "1 1 1 3 8 ", 25)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.190, new OptimizedParameterSet(9, 15, 0, "1 2 2 4 6 ", 25)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.200, new OptimizedParameterSet(9, 15, 0, "1 2 2 4 6 ", 25)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.250, new OptimizedParameterSet(1, 9, 0, "1 1 1 2 4 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(3.000, new OptimizedParameterSet(1, 11, 0, "1 1 1 2 6 ", 3)) );
rtt_opts_map->insert( pair<int, plr_map*>(10, plr_opts_map) );
plr_opts_map = new plr_map;
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.001, new OptimizedParameterSet(10, 2, 0, "1 1 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.010, new OptimizedParameterSet(10, 4, 0, "1 1 2 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.020, new OptimizedParameterSet(10, 5, 0, "1 1 3 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.030, new OptimizedParameterSet(10, 6, 0, "1 1 4 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.040, new OptimizedParameterSet(10, 6, 0, "1 1 4 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.050, new OptimizedParameterSet(3, 5, 0, "1 1 1 2 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.060, new OptimizedParameterSet(3, 5, 0, "1 1 1 2 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.070, new OptimizedParameterSet(3, 6, 0, "1 1 1 3 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.080, new OptimizedParameterSet(3, 6, 0, "1 1 1 3 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.090, new OptimizedParameterSet(3, 7, 0, "1 1 1 4 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.100, new OptimizedParameterSet(3, 7, 0, "1 1 1 4 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.110, new OptimizedParameterSet(3, 7, 0, "1 1 1 4 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.120, new OptimizedParameterSet(3, 8, 0, "1 1 1 5 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.130, new OptimizedParameterSet(3, 8, 0, "1 1 1 5 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.140, new OptimizedParameterSet(3, 8, 0, "1 1 1 5 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.150, new OptimizedParameterSet(3, 9, 0, "1 1 2 5 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.160, new OptimizedParameterSet(3, 9, 0, "1 1 2 5 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.170, new OptimizedParameterSet(3, 9, 0, "1 1 2 5 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.180, new OptimizedParameterSet(10, 15, 2, "2 4 7 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.190, new OptimizedParameterSet(10, 15, 2, "2 4 7 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.200, new OptimizedParameterSet(10, 16, 2, "2 4 8 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.250, new OptimizedParameterSet(1, 10, 0, "1 1 2 6 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(3.000, new OptimizedParameterSet(1, 12, 0, "1 1 2 8 ", 3)) );
rtt_opts_map->insert( pair<int, plr_map*>(20, plr_opts_map) );
plr_opts_map = new plr_map;
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.001, new OptimizedParameterSet(10, 2, 0, "1 1 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.010, new OptimizedParameterSet(5, 3, 0, "1 1 1 ", 14)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.020, new OptimizedParameterSet(5, 4, 0, "1 1 2 ", 14)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.030, new OptimizedParameterSet(5, 5, 0, "1 1 3 ", 14)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.040, new OptimizedParameterSet(5, 5, 0, "1 1 3 ", 14)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.050, new OptimizedParameterSet(5, 6, 0, "1 1 4 ", 14)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.060, new OptimizedParameterSet(5, 6, 0, "1 1 4 ", 14)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.070, new OptimizedParameterSet(5, 7, 0, "1 1 5 ", 14)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.080, new OptimizedParameterSet(1, 5, 0, "1 1 3 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.090, new OptimizedParameterSet(1, 5, 0, "1 1 3 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.100, new OptimizedParameterSet(1, 6, 0, "1 1 4 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.110, new OptimizedParameterSet(1, 6, 0, "1 1 4 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.120, new OptimizedParameterSet(1, 6, 0, "1 1 4 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.130, new OptimizedParameterSet(1, 6, 0, "1 1 4 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.140, new OptimizedParameterSet(1, 7, 0, "1 1 5 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.150, new OptimizedParameterSet(1, 7, 0, "1 1 5 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.160, new OptimizedParameterSet(1, 7, 0, "1 1 5 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.170, new OptimizedParameterSet(1, 7, 0, "1 1 5 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.180, new OptimizedParameterSet(10, 15, 3, "4 8 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.190, new OptimizedParameterSet(10, 15, 3, "4 8 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.200, new OptimizedParameterSet(5, 12, 1, "2 3 6 ", 14)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(2.500, new OptimizedParameterSet(1, 10, 0, "1 2 7 ", 3)) );
rtt_opts_map->insert( pair<int, plr_map*>(30, plr_opts_map) );
plr_opts_map = new plr_map;
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.001, new OptimizedParameterSet(10, 2, 0, "1 1 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.010, new OptimizedParameterSet(1, 3, 0, "1 2 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.020, new OptimizedParameterSet(1, 3, 0, "1 2 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.030, new OptimizedParameterSet(1, 3, 0, "1 2 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.040, new OptimizedParameterSet(1, 4, 0, "1 3 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.050, new OptimizedParameterSet(1, 4, 0, "1 3 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.060, new OptimizedParameterSet(1, 4, 0, "1 3 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.070, new OptimizedParameterSet(1, 5, 0, "1 4 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.080, new OptimizedParameterSet(1, 5, 0, "1 4 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.090, new OptimizedParameterSet(1, 5, 0, "1 4 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.100, new OptimizedParameterSet(10, 10, 2, "3 5 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.110, new OptimizedParameterSet(10, 11, 2, "3 6 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.120, new OptimizedParameterSet(10, 11, 2, "3 6 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.130, new OptimizedParameterSet(10, 12, 2, "4 6 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.140, new OptimizedParameterSet(10, 12, 3, "3 6 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.150, new OptimizedParameterSet(10, 13, 3, "3 7 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.160, new OptimizedParameterSet(10, 13, 3, "3 7 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.170, new OptimizedParameterSet(10, 14, 3, "4 7 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.180, new OptimizedParameterSet(10, 15, 3, "4 8 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.190, new OptimizedParameterSet(10, 15, 3, "4 8 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(2.000, new OptimizedParameterSet(10, 16, 4, "4 8 ", 28)) );
rtt_opts_map->insert( pair<int, plr_map*>(40, plr_opts_map) );
plr_opts_map = new plr_map;
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.001, new OptimizedParameterSet(5, 2, 0, "1 1 ", 14)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.010, new OptimizedParameterSet(5, 3, 0, "1 2 ", 14)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.020, new OptimizedParameterSet(1, 3, 0, "1 2 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.030, new OptimizedParameterSet(1, 3, 0, "1 2 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.040, new OptimizedParameterSet(1, 4, 0, "1 3 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.050, new OptimizedParameterSet(1, 4, 0, "1 3 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.060, new OptimizedParameterSet(1, 4, 0, "1 3 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.070, new OptimizedParameterSet(1, 5, 0, "1 4 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.080, new OptimizedParameterSet(1, 5, 0, "1 4 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.090, new OptimizedParameterSet(1, 5, 0, "1 4 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.100, new OptimizedParameterSet(1, 6, 0, "1 5 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.110, new OptimizedParameterSet(1, 6, 0, "1 5 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.120, new OptimizedParameterSet(1, 6, 0, "1 5 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.130, new OptimizedParameterSet(5, 9, 1, "3 5 ", 14)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.140, new OptimizedParameterSet(10, 12, 4, "8 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.150, new OptimizedParameterSet(10, 13, 4, "9 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.160, new OptimizedParameterSet(5, 10, 2, "3 5 ", 14)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.170, new OptimizedParameterSet(5, 11, 2, "3 6 ", 14)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.180, new OptimizedParameterSet(5, 11, 2, "3 6 ", 14)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.190, new OptimizedParameterSet(10, 15, 5, "10 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(2.000, new OptimizedParameterSet(5, 12, 2, "4 6 ", 14)) );
rtt_opts_map->insert( pair<int, plr_map*>(50, plr_opts_map) );
plr_opts_map = new plr_map;
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.001, new OptimizedParameterSet(1, 1, 0, "1 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.010, new OptimizedParameterSet(1, 3, 0, "3 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.020, new OptimizedParameterSet(10, 5, 1, "4 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.030, new OptimizedParameterSet(10, 6, 1, "5 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.040, new OptimizedParameterSet(10, 6, 1, "5 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.050, new OptimizedParameterSet(10, 7, 1, "6 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.060, new OptimizedParameterSet(10, 8, 2, "6 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.070, new OptimizedParameterSet(10, 8, 2, "6 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.080, new OptimizedParameterSet(10, 9, 2, "7 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.090, new OptimizedParameterSet(10, 9, 2, "7 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.100, new OptimizedParameterSet(10, 10, 3, "7 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.110, new OptimizedParameterSet(10, 11, 3, "8 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.120, new OptimizedParameterSet(10, 11, 3, "8 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.130, new OptimizedParameterSet(10, 12, 3, "9 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.140, new OptimizedParameterSet(10, 12, 4, "8 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.150, new OptimizedParameterSet(10, 13, 4, "9 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.160, new OptimizedParameterSet(10, 13, 4, "9 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.170, new OptimizedParameterSet(10, 14, 4, "10 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.180, new OptimizedParameterSet(10, 15, 5, "10 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.190, new OptimizedParameterSet(10, 15, 5, "10 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(2.000, new OptimizedParameterSet(10, 16, 5, "11 ", 28)) );
rtt_opts_map->insert( pair<int, plr_map*>(60, plr_opts_map) );
plr_opts_map = new plr_map;
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.001, new OptimizedParameterSet(1, 1, 0, "1 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.010, new OptimizedParameterSet(1, 3, 0, "3 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.020, new OptimizedParameterSet(10, 5, 1, "4 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.030, new OptimizedParameterSet(10, 6, 1, "5 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.040, new OptimizedParameterSet(10, 6, 1, "5 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.050, new OptimizedParameterSet(10, 7, 1, "6 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.060, new OptimizedParameterSet(10, 8, 2, "6 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.070, new OptimizedParameterSet(10, 8, 2, "6 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.080, new OptimizedParameterSet(10, 9, 2, "7 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.090, new OptimizedParameterSet(10, 9, 2, "7 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.100, new OptimizedParameterSet(10, 10, 3, "7 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.110, new OptimizedParameterSet(10, 11, 3, "8 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.120, new OptimizedParameterSet(10, 11, 3, "8 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.130, new OptimizedParameterSet(10, 12, 3, "9 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.140, new OptimizedParameterSet(10, 12, 4, "8 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.150, new OptimizedParameterSet(10, 13, 4, "9 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.160, new OptimizedParameterSet(10, 13, 4, "9 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.170, new OptimizedParameterSet(10, 14, 4, "10 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.180, new OptimizedParameterSet(10, 15, 5, "10 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.190, new OptimizedParameterSet(10, 15, 5, "10 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(2.000, new OptimizedParameterSet(10, 16, 5, "11 ", 28)) );
rtt_opts_map->insert( pair<int, plr_map*>(70, plr_opts_map) );
plr_opts_map = new plr_map;
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.001, new OptimizedParameterSet(1, 1, 0, "1 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.010, new OptimizedParameterSet(1, 3, 0, "3 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.020, new OptimizedParameterSet(10, 5, 1, "4 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.030, new OptimizedParameterSet(10, 6, 1, "5 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.040, new OptimizedParameterSet(10, 6, 1, "5 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.050, new OptimizedParameterSet(10, 7, 1, "6 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.060, new OptimizedParameterSet(10, 8, 2, "6 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.070, new OptimizedParameterSet(10, 8, 2, "6 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.080, new OptimizedParameterSet(10, 9, 2, "7 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.090, new OptimizedParameterSet(10, 9, 2, "7 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.100, new OptimizedParameterSet(10, 10, 3, "7 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.110, new OptimizedParameterSet(10, 11, 3, "8 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.120, new OptimizedParameterSet(10, 11, 3, "8 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.130, new OptimizedParameterSet(10, 12, 3, "9 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.140, new OptimizedParameterSet(10, 12, 4, "8 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.150, new OptimizedParameterSet(10, 13, 4, "9 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.160, new OptimizedParameterSet(10, 13, 4, "9 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.170, new OptimizedParameterSet(10, 14, 4, "10 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.180, new OptimizedParameterSet(10, 15, 5, "10 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.190, new OptimizedParameterSet(10, 15, 5, "10 ", 28)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(2.000, new OptimizedParameterSet(10, 16, 5, "11 ", 28)) );
rtt_opts_map->insert( pair<int, plr_map*>(80, plr_opts_map) );
plr_opts_map = new plr_map;
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.001, new OptimizedParameterSet(1, 1, 0, "1 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.010, new OptimizedParameterSet(3, 3, 0, "3 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.020, new OptimizedParameterSet(1, 3, 0, "3 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.030, new OptimizedParameterSet(1, 3, 0, "3 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.040, new OptimizedParameterSet(1, 4, 0, "4 ", 3)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.050, new OptimizedParameterSet(3, 5, 1, "4 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.060, new OptimizedParameterSet(3, 5, 1, "4 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.070, new OptimizedParameterSet(3, 6, 1, "5 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.080, new OptimizedParameterSet(3, 6, 1, "5 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.090, new OptimizedParameterSet(3, 7, 1, "6 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.100, new OptimizedParameterSet(3, 7, 1, "6 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.110, new OptimizedParameterSet(3, 7, 1, "6 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.120, new OptimizedParameterSet(3, 8, 1, "7 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.130, new OptimizedParameterSet(3, 8, 1, "7 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.140, new OptimizedParameterSet(3, 8, 2, "6 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.150, new OptimizedParameterSet(3, 9, 2, "7 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.160, new OptimizedParameterSet(3, 9, 2, "7 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.170, new OptimizedParameterSet(3, 9, 2, "7 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.180, new OptimizedParameterSet(3, 10, 2, "8 ", 9)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(1.900, new OptimizedParameterSet(3, 10, 2, "8 ", 9)) );
rtt_opts_map->insert( pair<int, plr_map*>(1000, plr_opts_map) );

    m_num_recv_map.insert( pair<double, rtt_map*>(2, rtt_opts_map) );

 std::cout << "DelayCell add12MbpsTwoReceiversMap() #recv_map=" << m_num_recv_map.size() << " #rtt_map=" << rtt_opts_map->size() << std::endl;
return;
  }

  void DelayCell::add12MbpsThreeReceiversMap()
  {
  }

  void DelayCell::addThreeReceiversMap()
  {
    // Targeted Delay = 0.1 & #receivers = 3
    num_recv_map::iterator rcv_it = m_num_recv_map.find(3);

    if ( rcv_it != m_num_recv_map.end() ) {
      // The parameters for one receiver already exist. No need to create.
      return;
    }

    rtt_map *rtt_opts_map = new rtt_map;
    plr_map *plr_opts_map = new plr_map;

    /*********** Adding Optimum of RTT = 10 ms **************/
    // Simplify Module from Manuel.
    // PLR = 0.001
    //~ plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.001, new OptimizedParameterSet(42, 42, 0, "1 1", 114)) );
    //~ // PLR = 0.01
    //~ plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.01, new OptimizedParameterSet(31, 31, 0, "1 1 3", 84)) );
    //~ // PLR = 0.04
    //~ plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.04, new OptimizedParameterSet(20, 20, 3, "1 1 1 5", 54)) );
    //~ // PLR = 0.06
    //~ plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.06, new OptimizedParameterSet(9, 9, 0, "1 1 1 1 4", 24)) );
    //~ // PLR = 0.09
    //~ plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.09, new OptimizedParameterSet(42, 42, 4, "4 9", 114)) );
    //~ // PLR = 0.1
    //~ plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.1, new OptimizedParameterSet(20, 20, 1, "1 2 3 6", 54)) );
    //~ // PLR = 0.11
    //~ plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.11, new OptimizedParameterSet(42, 42, 5, "5 10", 114)) );
    //~ // PLR = 0.13
    //~ plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.13, new OptimizedParameterSet(53, 53, 13, "13", 144)) );
    //~ // PLR = 0.14
    //~ plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.14, new OptimizedParameterSet(42, 42, 6, "6 12", 114)) );
    //~ // PLR = 0.19
    //~ plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.19, new OptimizedParameterSet(53, 53, 17, "18", 144)) );
    //~ // PLR = 0.20
    //~ plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.20, new OptimizedParameterSet(42, 42, 8, "8 16", 144)) );
    //~ // PLR = 0.25
    //~ plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.25, new OptimizedParameterSet(53, 53, 23, "23", 144)) );
    //~ // PLR = 0.30
    //~ plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.30, new OptimizedParameterSet(42, 42, 12, "12 24", 114)) );
    //~ // PLR = 0.35, use key=3.5 means using this opt if plr>0.35
    //~ plr_opts_map->insert( pair<double, OptimizedParameterSet*>(3.5, new OptimizedParameterSet(42, 42, 15, "15 30", 114)) );
 
    //~ rtt_opts_map->insert( pair<double, plr_map*>(10, plr_opts_map) );

    /*********** Adding Optimum of RTT = 30 ms **************/
    plr_opts_map = new plr_map;
    // Simplify Module from Manuel.
    // PLR = 0.001
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.001, new OptimizedParameterSet(23, 23, 0, "1 1", 62)) );
    // PLR = 0.01
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.01, new OptimizedParameterSet(5, 5, 0, "1 1 1", 14)) );
    // PLR = 0.03
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.03, new OptimizedParameterSet(5, 5, 0, "1 1 3", 14)) );
    // PLR = 0.05
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.05, new OptimizedParameterSet(23, 23, 2, "2 5", 62)) );
    // PLR = 0.07
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.07, new OptimizedParameterSet(23, 23, 2, "3 6", 62)) );
    // PLR = 0.09
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.09, new OptimizedParameterSet(23, 23, 3, "3 7", 62)) );
    // PLR = 0.11
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.11, new OptimizedParameterSet(42, 42, 9, "11", 114)) );
    // PLR = 0.13
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.13, new OptimizedParameterSet(42, 42, 11, "11", 114)) );
    // PLR = 0.15
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.15, new OptimizedParameterSet(42, 42, 12, "13", 114)) );
    // PLR = 0.17
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.17, new OptimizedParameterSet(42, 42, 14, "14", 114)) );
    // PLR = 0.19
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.19, new OptimizedParameterSet(42, 42, 15, "16", 114)) );
    // PLR = 0.25
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.25, new OptimizedParameterSet(42, 42, 19, "20", 114)) );
    // PLR = 0.30
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.30, new OptimizedParameterSet(42, 42, 24, "24", 114)) );
    // PLR = 0.35, use key=3.5 means using this opt if plr>0.35
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(3.5, new OptimizedParameterSet(23, 23, 10, "11 22", 114)) );
 
    rtt_opts_map->insert( pair<double, plr_map*>(30, plr_opts_map) );

    /*********** Adding Optimum of RTT = 60 ms **************/
    plr_opts_map = new plr_map;
    // Simplify Module from Manuel.
    // PLR = 0.001
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.001, new OptimizedParameterSet(1, 0, 0, "1", 3)) );
    // PLR = 0.01
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.01, new OptimizedParameterSet(25, 25, 1, "4", 68)) );
    // PLR = 0.03
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.03, new OptimizedParameterSet(25, 25, 2, "5", 68)) );
    // PLR = 0.05
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.05, new OptimizedParameterSet(25, 25, 3, "7", 68)) );
    // PLR = 0.07
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.07, new OptimizedParameterSet(25, 25, 4, "8", 68)) );
    // PLR = 0.09
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.09, new OptimizedParameterSet(25, 25, 5, "9", 68)) );
    // PLR = 0.11
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.11, new OptimizedParameterSet(25, 25, 6, "9", 68)) );
    // PLR = 0.13
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.13, new OptimizedParameterSet(25, 25, 7, "10", 68)) );
    // PLR = 0.15
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.15, new OptimizedParameterSet(25, 25, 8, "11", 68)) );
    // PLR = 0.17
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.17, new OptimizedParameterSet(25, 25, 9, "12", 68)) );
    // PLR = 0.19
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.19, new OptimizedParameterSet(25, 25, 10, "13", 68)) );
    // PLR = 0.20
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.20, new OptimizedParameterSet(25, 25, 11, "13", 68)) );
    // PLR = 0.25
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.25, new OptimizedParameterSet(25, 25, 14, "15", 68)) );
    // PLR = 0.30, use key=3.0 means using this opt if plr>0.30
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(3.0, new OptimizedParameterSet(25, 25, 18, "18", 68)) );
 
    rtt_opts_map->insert( pair<double, plr_map*>(60, plr_opts_map) );

    /*********** Adding Optimum of RTT = 80 ms **************/
    plr_opts_map = new plr_map;
    // Simplify Module from Manuel.
    // PLR = 0.001
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.001, new OptimizedParameterSet(1, 0, 0, "1", 3)) );
    // PLR = 0.01
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.01, new OptimizedParameterSet(14, 14, 1, "3", 38)) );
    // PLR = 0.03
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.03, new OptimizedParameterSet(14, 14, 1, "5", 38)) );
    // PLR = 0.05
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.05, new OptimizedParameterSet(14, 14, 2, "6", 38)) );
    // PLR = 0.08
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.08, new OptimizedParameterSet(14, 14, 3, "7", 38)) );
    // PLR = 0.10
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.10, new OptimizedParameterSet(14, 14, 4, "7", 38)) );
    // PLR = 0.12
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.12, new OptimizedParameterSet(14, 14, 4, "9", 38)) );
    // PLR = 0.14
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.14, new OptimizedParameterSet(14, 14, 5, "9", 38)) );
    // PLR = 0.16
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.16, new OptimizedParameterSet(14, 14, 6, "9", 38)) );
    // PLR = 0.18
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.18, new OptimizedParameterSet(14, 14, 6, "11", 38)) );
    // PLR = 0.20
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.20, new OptimizedParameterSet(14, 14, 7, "11", 38)) );
    // PLR = 0.25
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.25, new OptimizedParameterSet(14, 14, 9, "13", 38)) );
    // PLR = 0.30, use key=3.0 means using this opt if plr>0.30
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(3.0, new OptimizedParameterSet(14, 14, 11, "15", 38)) );
 
    rtt_opts_map->insert( pair<double, plr_map*>(80, plr_opts_map) );

    /*********** Adding Optimum of RTT = 100 ms **************/
    plr_opts_map = new plr_map;
    // Simplify Module from Manuel.
    // PLR = 0.001
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.001, new OptimizedParameterSet(1, 0, 0, "1", 3)) );
    // PLR = 0.01
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.01, new OptimizedParameterSet(3, 3, 0, "3", 8)) );
// no retransmission for following parameters
    // PLR = 0.02
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.02, new OptimizedParameterSet(47, 47, 8, "", 127)) );
    // PLR = 0.04
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.04, new OptimizedParameterSet(47, 47, 11, "", 127)) );
    // PLR = 0.06
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.06, new OptimizedParameterSet(47, 47, 14, "", 127)) );
    // PLR = 0.08
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.08, new OptimizedParameterSet(47, 47, 17, "", 127)) );
    // PLR = 0.10
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.10, new OptimizedParameterSet(47, 47, 20, "", 127)) );
    // PLR = 0.12
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.12, new OptimizedParameterSet(47, 47, 23, "", 127)) );
    // PLR = 0.14
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.14, new OptimizedParameterSet(47, 47, 25, "", 127)) );
    // PLR = 0.16
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.16, new OptimizedParameterSet(47, 47, 28, "", 127)) );
    // PLR = 0.18
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.18, new OptimizedParameterSet(47, 47, 31, "", 127)) );
    // PLR = 0.20
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.20, new OptimizedParameterSet(47, 47, 34, "", 127)) );
    // PLR = 0.25, use key=2.5 means using this opt if plr>0.25
    plr_opts_map->insert( pair<double, OptimizedParameterSet*>(2.5, new OptimizedParameterSet(47, 47, 42, "", 127)) );
 
    rtt_opts_map->insert( pair<double, plr_map*>(1000, plr_opts_map) );

    // RTT=100ms, use key=1000 means using this opt if rtt>100ms
    m_num_recv_map.insert( pair<double, rtt_map*>(3, rtt_opts_map) );

 std::cout << "DelayCell addThreeReceiverMap() #recv_map=" << m_num_recv_map.size() << " #rtt_map=" << rtt_opts_map->size() << std::endl;
  }

  void DelayCell::add20MbpsOneReceiverMap()
  {
    // Targeted Delay = 0.3 & #receivers = 1
    num_recv_map::iterator rcv_it = m_num_recv_map.find(1);

    if ( rcv_it != m_num_recv_map.end() ) {
      // The parameters for one receiver already exist. No need to create.
      return;
    }

    rtt_map *rtt_opts_map = new rtt_map;
    plr_map *plr_opts_map = new plr_map;

plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.010, new OptimizedParameterSet(1, 5, 0, "1 1 1 1 1 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.020, new OptimizedParameterSet(1, 5, 0, "1 1 1 1 1 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.030, new OptimizedParameterSet(1, 6, 0, "1 1 1 1 2 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.040, new OptimizedParameterSet(1, 7, 0, "1 1 1 1 3 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.050, new OptimizedParameterSet(1, 7, 0, "1 1 1 1 3 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.060, new OptimizedParameterSet(1, 8, 0, "1 1 1 1 4 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.070, new OptimizedParameterSet(1, 8, 0, "1 1 1 1 4 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.080, new OptimizedParameterSet(1, 9, 0, "1 1 1 1 5 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.090, new OptimizedParameterSet(1, 9, 0, "1 1 1 1 5 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(1.000, new OptimizedParameterSet(1, 10, 0, "1 1 1 1 6 ", 1)) );
rtt_opts_map->insert( pair<int, plr_map*>(25, plr_opts_map) );
plr_opts_map = new plr_map;
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.010, new OptimizedParameterSet(1, 5, 0, "1 1 3 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.020, new OptimizedParameterSet(1, 5, 0, "1 1 3 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.030, new OptimizedParameterSet(1, 6, 0, "1 1 4 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.040, new OptimizedParameterSet(1, 7, 0, "1 1 5 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.050, new OptimizedParameterSet(1, 7, 0, "1 1 5 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.060, new OptimizedParameterSet(1, 8, 0, "1 1 6 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.070, new OptimizedParameterSet(1, 8, 0, "1 1 6 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.080, new OptimizedParameterSet(1, 9, 0, "1 1 7 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.090, new OptimizedParameterSet(1, 9, 0, "1 1 7 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(1.000, new OptimizedParameterSet(1, 10, 0, "1 1 8 ", 1)) );
rtt_opts_map->insert( pair<int, plr_map*>(50, plr_opts_map) );
plr_opts_map = new plr_map;
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.010, new OptimizedParameterSet(1, 5, 0, "1 4 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.020, new OptimizedParameterSet(1, 5, 0, "1 4 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.030, new OptimizedParameterSet(1, 6, 0, "1 5 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.040, new OptimizedParameterSet(1, 7, 0, "1 6 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.050, new OptimizedParameterSet(1, 7, 0, "1 6 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.060, new OptimizedParameterSet(1, 8, 0, "1 7 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.070, new OptimizedParameterSet(1, 8, 0, "1 7 ", 1)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.080, new OptimizedParameterSet(96, 34, 8, "9 17 ", 53)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.090, new OptimizedParameterSet(96, 36, 9, "9 18 ", 53)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(1.000, new OptimizedParameterSet(96, 39, 9, "10 20 ", 53)) );
rtt_opts_map->insert( pair<int, plr_map*>(75, plr_opts_map) );
plr_opts_map = new plr_map;
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.010, new OptimizedParameterSet(133, 13, 3, "10 ", 73)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.020, new OptimizedParameterSet(133, 18, 5, "13 ", 73)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.030, new OptimizedParameterSet(133, 22, 7, "15 ", 73)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.040, new OptimizedParameterSet(133, 26, 9, "17 ", 73)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.050, new OptimizedParameterSet(133, 30, 10, "20 ", 73)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.060, new OptimizedParameterSet(133, 33, 12, "21 ", 73)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.070, new OptimizedParameterSet(133, 37, 14, "23 ", 73)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.080, new OptimizedParameterSet(133, 30, 16, "24 ", 73)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.090, new OptimizedParameterSet(133, 43, 18, "25 ", 73)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(1.000, new OptimizedParameterSet(133, 50, 22, "28 ", 73)) );
rtt_opts_map->insert( pair<int, plr_map*>(100, plr_opts_map) );
plr_opts_map = new plr_map;
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.010, new OptimizedParameterSet(133, 13, 3, "10 ", 73)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.020, new OptimizedParameterSet(133, 18, 5, "13 ", 73)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.030, new OptimizedParameterSet(133, 22, 7, "15 ", 73)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.040, new OptimizedParameterSet(133, 26, 9, "17 ", 73)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.050, new OptimizedParameterSet(133, 30, 10, "20 ", 73)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.060, new OptimizedParameterSet(133, 33, 12, "21 ", 73)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.070, new OptimizedParameterSet(133, 37, 14, "23 ", 73)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.080, new OptimizedParameterSet(133, 40, 16, "24 ", 73)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.090, new OptimizedParameterSet(133, 43, 18, "25 ", 73)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(1.000, new OptimizedParameterSet(133, 47, 20, "27 ", 73)) );
rtt_opts_map->insert( pair<int, plr_map*>(125, plr_opts_map) );
plr_opts_map = new plr_map;
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.010, new OptimizedParameterSet(64, 10, 2, "8 ", 35)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.020, new OptimizedParameterSet(64, 13, 3, "10 ", 35)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.030, new OptimizedParameterSet(64, 16, 4, "12 ", 35)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.040, new OptimizedParameterSet(64, 19, 5, "14 ", 35)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.050, new OptimizedParameterSet(64, 21, 6, "15 ", 35)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.060, new OptimizedParameterSet(64, 23, 7, "16 ", 35)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.070, new OptimizedParameterSet(64, 25, 8, "17 ", 35)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.080, new OptimizedParameterSet(64, 27, 9, "18 ", 35)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(0.090, new OptimizedParameterSet(64, 29, 10, "19 ", 35)) );
plr_opts_map->insert( pair<double, OptimizedParameterSet*>(1.000, new OptimizedParameterSet(64, 32, 11, "21 ", 35)) );
rtt_opts_map->insert( pair<int, plr_map*>(1000, plr_opts_map) );

    m_num_recv_map.insert( pair<int, rtt_map*>(1, rtt_opts_map) );
 std::cout << "DelayCell add12MbpsOneReceiverMap() #recv_map=" << m_num_recv_map.size() << " #rtt_map=" << rtt_opts_map->size() << std::endl;
  }


  void DelayCell::removeMapOfNumReceivers(int num_recv)
  {
    num_recv_map::iterator rcv_it = m_num_recv_map.find(num_recv);
    
    if ( rcv_it == m_num_recv_map.end() ) {
      // The parameters for the given number of receivers doesn't exist. No need to remove.
      return;
    }

    m_num_recv_map.erase( rcv_it );
    rtt_map *rtt_opts_map = rcv_it->second;

    // Remove all parameters (PLR map) in RTT table.
    rtt_map::iterator rtt_it;
    for (rtt_it=rtt_opts_map->begin(); rtt_it != rtt_opts_map->end(); rtt_it++) {
      // Remove all parameters (OptimizedParameterSet) in PLR table of one RTT.
      plr_map::iterator plr_it;
      plr_map *plr_opts_map = rtt_it->second;
      for (plr_it=plr_opts_map->begin(); plr_it != plr_opts_map->end(); plr_it++) {
        delete plr_it->second;
      }
    }

  }

  void DelayCell::removeMap()
  {
    num_recv_map::iterator rcv_it;
    // Remove all parameters.
    for (rcv_it=m_num_recv_map.begin(); rcv_it != m_num_recv_map.end(); rcv_it++) {
      removeMapOfNumReceivers( rcv_it->first );
    }
  }


}
