/* 
 * (c) 2011 Saarland University, Telecommunications Lab 
 * (prrt@nt.uni-saarland.de)
 *
 * The contents of this file are subject to the Mozilla Public License
 + Version 1.1 (the "License"); you may not use this file except in
 + compliance with the License. You may obtain a copy of the License at
 + http://www.mozilla.org/MPL/
 + 
 + Software distributed under the License is distributed on an "AS IS"
 + basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 + License for the specific language governing rights and limitations
 + under the License.
 + 
 + The Initial Developer of the Original Code is Saarland University.
 + All Rights Reserved.
 + 
 + Contributor(s): Yongtao Shuai (Saarland University).
 + 
 + Disclaimer: This Alpha version of PRRT is not supported and stable yet.
 + Do not use it in a production environment: it may crash your device, 
 + so please make sure that you have a backup of your data before
 + installing PRRT.
 */


#ifndef OPTIMIZEDPARAMETERSET_HPP
#define OPTIMIZEDPARAMETERSET_HPP OPTIMIZEDPARAMETERSET_HPP

#include <vector>

// For debug
#include <iostream>
#include <sstream>  // ostringstream
#include <string.h>


namespace Adaptation {
  
  /** Class for the Optimized Parameters Set in the Adaptation.
   *
   * This class represents a set containing the specific parameters which are optimized by the Adapability in a certain scenario.
   */
  class OptimizedParameterSet
  {
    public :
      /** @name Constructor & Destructor
       */
      //@{
      /**
       */
      OptimizedParameterSet(unsigned int _k, unsigned int _r, unsigned int _n_p, unsigned int _rr,
                            unsigned int _last_rr, unsigned int _d_fec);

      OptimizedParameterSet(unsigned int _k, unsigned int _r, unsigned int _n_p, std::string _rr, unsigned int _d_fec);
      
      OptimizedParameterSet();

      OptimizedParameterSet(const OptimizedParameterSet &opt);
      
      ~OptimizedParameterSet();
      //@}
      
      /** Check if the parameters set is identical.
       */
      bool operator==(const OptimizedParameterSet &opt) const;

      /** 
       */
      OptimizedParameterSet & operator=(const OptimizedParameterSet &opt);

      /** Convert parameters to human readable string of the form k; r; n_p; d_fec; v_rr
       */
      std::string toString() const;

      /** Output operator for OptimizedParameterSet
       *
       * @param ostr output stream that should be used
       * @param opt that should be written to ostr.
       */
      friend std::ostream& operator<<(std::ostream &ostr, const OptimizedParameterSet &opt) {
        return ostr << opt.toString();
      }

      unsigned int getNumberOfDataPacketsPerBlock() const { return m_k; }
      unsigned int getNumberOfRedundancyPacketsPerBlock() const { return m_r; }
      unsigned int getNumberOfPresentRedundancyPacketsPerBlock() const { return m_n_p; }
      unsigned int getNumberOfRetransmissionRounds() const { return m_rr; }
      unsigned int getNumberOfRetransmissionsInLastRound() const { return m_last_rr; }
      int getDelayFEC() const { return m_delay_fec; }

      /** Get the retransmission rounds vector.
       *
       * NOTE Only call this method if the parameter is applied to the PRRT, otherwise it might get a wrong vector.
       * Because all OptimizedParameterSet objects share the same vector and could access at the same time.
       * We assume that only one parameter set is applied each time.
       */
      const std::vector<unsigned int> & getRetransmissionRoundsVector() const { return m_v_rr; }
      
    private :
      /** The parameters to be optimized by the Adaptability.
       */
      unsigned int m_k;                   ///< The number of the data packets per block (k).
      unsigned int m_r;                   ///< The number of the redundancy packets per block (r).
      unsigned int m_n_p;                 ///< The number of the precautionary redundancy packets per block (n_p).
      
      // m_v_rr can be determined by m_rr and m_last_rr, because only one redundancy packet is sent in all retransmission rounds except the last one.
      unsigned int m_rr;                  ///< The number of the retransmission rounds.
      unsigned int m_last_rr;             ///< The number of retransmissions in the last retransmission round.
      std::vector<unsigned int> m_v_rr;        ///< The number of retransmission rounds and the retransmission multipliers for each round (N_rt vector).
      
      int m_delay_fec;                        ///< The delay of FEC in millisecond i.e. the time of collecting k data packets for encoding. packet interval, we need datarate and packet length.

      static int m_num_opts;                ///< Counter the number of optimized parameters having been created, used for debug if they are released.
  };
  
}

#endif
