/*
 * (c) 2010 Saarland University, Telecommunications Lab
 * (prrt@nt.uni-saarland.de)
 *
 * The contents of this file are subject to the Mozilla Public License
 + Version 1.1 (the "License"); you may not use this file except in
 + compliance with the License. You may obtain a copy of the License at
 + http://www.mozilla.org/MPL/
 +
 + Software distributed under the License is distributed on an "AS IS"
 + basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 + License for the specific language governing rights and limitations
 + under the License.
 +
 + The Original Code is "Predictably Reliable Real-time Transport (PRRT)".
 +
 + The Initial Developer of the Original Code is Saarland University.
 + All Rights Reserved.
 +
 + Contributor(s): Jochen Gruen (Saarland University), Manuel Gorius
 + (Saarland University), Yongtao Shuai (Saarland University).
 +
 + Disclaimer: This Alpha version of PRRT is not supported and stable yet.
 + Do not use it in a production environment: it may crash your device,
 + so please make sure that you have a backup of your data before
 + installing PRRT.
 */

#ifndef MAINCLASSREROUTABLERELAY_H
#define	MAINCLASSREROUTABLERELAY_H

#include <iostream>
#include <signal.h>
#include <cstdlib>
#include <string.h>
#include <prrt/prrt_relay.h>
#include <mainClassPrrtRelay.h>
#include <CommandParser.h>
#include <CommandExecutor.h>
#include <GetSocketCommand.h>
#include <RerouteCommand.h>
#include <AddSocketCommand.h>
#include <SetIntParameterCommand.h>
#include <SetFloatParameterCommand.h>
#include <SetAHECCodingParameterCommand.h>
#include <SetOptionCommand.h>
#include <ShutdownCommand.h>
#include <SocketList.h>
#include <UnixSocketControlCommunication.h>
#include <ConsoleControlCommunication.h>
#include <ReroutableRelaySettings.h>
#include <ShutdownException.h>

using namespace std;
using namespace media_transport_protocol;
using namespace CommandLineArgumentLib;

class mainClassReroutableRelay : protected mainClassPrrtRelay
{
public:

    mainClassReroutableRelay()
    {
        forever = true;
    }
    ~mainClassReroutableRelay();

    int main(int argc, char** argv)
    {
        prrt_init();
        try
        {
            InitReroutableRelaySettings();
            try {
                mainClassPrrtRelay::InitRelay(argc, argv);
            } catch(...){
                if(this->prrtsocketRecv){
                    delete this->prrtsocketRecv;
                }
                if(this->prrtsocketSend){
                    delete this->prrtsocketSend;
                }
                if(this->relay){
                    delete this->relay;
                }
                throw;
            }

            InitControlCommunication();

            InitSocketConnector();

            CommandParser commandParser;
            InitCommandParser(commandParser);

            commandParser.start();

            while (forever)
            {
                ;
            }
        } catch (prrt_error &me) {
            cerr << "Error in prrt lib (ReroutableRelay): " << me.what() << endl;
        } catch (CLAException &parseException) {
            char* program = argv[0];
            show_usage(program, this->parser.GetOptionHelpMessage());
            cerr << "Error while parsing arguments: " << parseException.what() << endl;
        } catch (ShutdownException &shutdownException) {
            cerr << shutdownException.what() << endl;
        } catch (exception &e) {
            cerr << "Unexpected error: " << e.what() << endl;
        } catch (const char* c) {
            cerr << "Unexpected error: " << c << endl;
        }

        shutdown();
    }

    void shutdown()
    {
        forever = false;
        socketConnector.Shutdown();
    }

protected:
    mainClassReroutableRelay(const mainClassReroutableRelay& orig);

    void InitReroutableRelaySettings()
    {
        stringstream helpMessage;

        helpMessage.str("");
        helpMessage << "\n   Reroutable relay options:\n";
        helpMessage << "\t -csp <control socket path> \t\tPath to unix socket for communication with relay\n";
        socketPathOption.SetIdentifier("-csp");
        socketPathOption.SetHelpMessage(helpMessage.str());
        socketPathOption.SetSetting(reroutableRelaySettings.GetSocketPathSetting());
        parser.AddArgumentType(&socketPathOption);
    }

    void InitControlCommunication()
    {
        if (reroutableRelaySettings.UseSocketControlConnection())
        {
            controlCommunication = new UnixSocketControlCommunication();
            dynamic_cast<UnixSocketControlCommunication*> (controlCommunication)->CreateUnixSocket(reroutableRelaySettings.GetSocketPath());
        }
        else
        {
            controlCommunication = new ConsoleControlCommunication();
        }
    }

    void InitSocketConnector()
    {
        socketConnector.SetRelay(this->relay);
        socketConnector.AddReceivingSocket(this->prrtsocketRecv);
        socketConnector.AddSendingSocket(this->prrtsocketSend);
        socketConnector.SetReceiverSettings(&recvSettings);
        socketConnector.SetSenderSettings(&sendSettings);
        socketConnector.SetRelaySettings(&relaySettings);
        socketConnector.SetControlCommunication(controlCommunication);
        commandExecutor.SetSocketConnector(&socketConnector);
        commandExecutor.SetControlCommunication(controlCommunication);
    }

    void InitCommandParser(CommandParser &commandParser)
    {
        ShutdownCommand *shutdownCommand = new ShutdownCommand();
        shutdownCommand->SetKey("shutdown");
        shutdownCommand->SetHelpMessage("Shutdown the relay");
        shutdownCommand->SetCommandExecutor(&commandExecutor);
        commandParser.addCommand(shutdownCommand);

        GetSocketCommand *getSocketCommand = new GetSocketCommand();
        getSocketCommand->SetKey("sockets");
        getSocketCommand->SetHelpMessage("Print all available sockets");
        getSocketCommand->SetCommandExecutor(&commandExecutor);
        getSocketCommand->ExecuteFunction = &CommandExecutor::PrintSockets;
        commandParser.addCommand(getSocketCommand);

        RerouteCommand *rerouteCommand = new RerouteCommand();
        rerouteCommand->SetKey("reroute");
        rerouteCommand->SetHelpMessage("Parameter: <receiverIndex> <senderIndex>, route the traffic from socket with index receiverIndex to socket with index senderIndex");
        rerouteCommand->SetCommandExecutor(&commandExecutor);
        rerouteCommand->ExecuteFunction = &CommandExecutor::Reroute;
        commandParser.addCommand(rerouteCommand);

        AddSocketCommand *addReceiverSocketCommand = new AddSocketCommand();
        addReceiverSocketCommand->SetKey("addreceiver");
        addReceiverSocketCommand->SetHelpMessage("Parameter: <senderIp> <receiverIp>, Add a receiving socket with the given Ips");
        addReceiverSocketCommand->SetCommandExecutor(&commandExecutor);
        addReceiverSocketCommand->ExecuteFunction = &CommandExecutor::AddReceiverSocket;
        commandParser.addCommand(addReceiverSocketCommand);

        AddSocketCommand *addSenderSocketCommand = new AddSocketCommand();
        addSenderSocketCommand->SetKey("addsender");
        addSenderSocketCommand->SetHelpMessage("Parameter: <senderIp> <receiverIp>, Add a sending socket with the given Ips");
        addSenderSocketCommand->SetCommandExecutor(&commandExecutor);
        addSenderSocketCommand->ExecuteFunction = &CommandExecutor::AddSenderSocket;
        commandParser.addCommand(addSenderSocketCommand);

        SetIntParameterCommand<param_BW_est> *setBWestCommand = new SetIntParameterCommand<param_BW_est > ();
        setBWestCommand->SetKey("setBWest");
        setBWestCommand->SetHelpMessage("Parameter: <socketIndex> <bw_est>, Set the parameter bandwidth estimate of the sending socket with index socketIndex");
        setBWestCommand->SetCommandExecutor(&commandExecutor);
        setBWestCommand->ExecuteFunction = &CommandExecutor::SetReceiverParameter;
        commandParser.addCommand(setBWestCommand);

        SetAHECCodingParameterCommand *setAHECCodingParamsCommand = new SetAHECCodingParameterCommand();
        setAHECCodingParamsCommand->SetKey("setAHECCodingParams");
        setAHECCodingParamsCommand->SetHelpMessage("Parameter: <socketIndex> <d_target> <d_fec> <k> <n_p>, Set the AHEC coding parameters: delay target, delay of FEC, block size, parity packets of each cycle (in string form e.g. 0,1,3) of the sending socket with index socketIndex");
        setAHECCodingParamsCommand->SetCommandExecutor(&commandExecutor);
        setAHECCodingParamsCommand->ExecuteFunction = &CommandExecutor::SetSenderParameter;
        commandParser.addCommand(setAHECCodingParamsCommand);

        SetFloatParameterCommand<param_PLR> *setPlrCommand = new SetFloatParameterCommand<param_PLR > ();
        setPlrCommand->SetKey("setPlr");
        setPlrCommand->SetHelpMessage("Parameter: <socketIndex> <plr>, Set the PLR of the sending socket with index socketIndex");
        setPlrCommand->SetCommandExecutor(&commandExecutor);
        setPlrCommand->ExecuteFunction = &CommandExecutor::SetSenderParameter;
        commandParser.addCommand(setPlrCommand);

        SetOptionCommand<option_REPLACE> *sendReplacementsCommand = new SetOptionCommand<option_REPLACE > ();
        sendReplacementsCommand->SetKey("stopReplacements");
        sendReplacementsCommand->SetHelpMessage("Parameter: <socketIndex> <0|1> ..., Should the sending socket with index socketIndex stop sending replacements (0 = false, 1 = true)");
        sendReplacementsCommand->SetCommandExecutor(&commandExecutor);
        sendReplacementsCommand->ExecuteFunction = &CommandExecutor::SetOption;
        commandParser.addCommand(sendReplacementsCommand);

        commandParser.SetControlCommunication(controlCommunication);
    }

    virtual void show_usage(char* program, std::string optionMessage)
    {
        std::stringstream ss;
        ss << "\033[31mUsage: " << program << " [options] <sender> <relayname>\033[0m\n\n";
        ss << "List of available options:\n" << optionMessage;

        cerr << ss.str();
    }

    bool forever;
    CommandExecutor commandExecutor;
    SocketConnector socketConnector;
    ReroutableRelaySettings reroutableRelaySettings;
    CLAOptionWithParameter socketPathOption;
    ControlCommunicationBase *controlCommunication;

};

#endif	/* MAINCLASSREROUTABLERELAY_H */

