/* 
 * (c) 2010 Saarland University, Telecommunications Lab 
 * (prrt@nt.uni-saarland.de)
 *
 * The contents of this file are subject to the Mozilla Public License
 + Version 1.1 (the "License"); you may not use this file except in
 + compliance with the License. You may obtain a copy of the License at
 + http://www.mozilla.org/MPL/
 + 
 + Software distributed under the License is distributed on an "AS IS"
 + basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 + License for the specific language governing rights and limitations
 + under the License.
 + 
 + The Original Code is "Predictably Reliable Real-time Transport (PRRT)".
 + 
 + The Initial Developer of the Original Code is Saarland University.
 + All Rights Reserved.
 + 
 + Contributor(s): Jochen Gruen (Saarland University), Manuel Gorius
 + (Saarland University), Yongtao Shuai (Saarland University).
 + 
 + Disclaimer: This Alpha version of PRRT is not supported and stable yet.
 + Do not use it in a production environment: it may crash your device, 
 + so please make sure that you have a backup of your data before
 + installing PRRT.
 */

#include "delay_cc.h"

#include <iostream>
#include <math.h>

#define STEP 3000 

using namespace media_transport_protocol;
using namespace std;

prrt_delay_cc::prrt_delay_cc() {
	baseRTT = 0;
	m_avg_rtt = 0;

	min_rtt = new value_pair[MIN_RTT_SAMPLES];
	max_rtt = new value_pair[MAX_RTT_SAMPLES];
	min_end = min_rtt + MIN_RTT_SAMPLES;
	min_last = min_rtt;
	min_pair = min_rtt;
	max_end = max_rtt + MAX_RTT_SAMPLES;
	max_last = max_rtt;
	max_pair = max_rtt;
	min_pair->value = 20000;
	min_pair->expiration = MIN_RTT_SAMPLES;
	max_pair->value = 0;
	max_pair->expiration = MAX_RTT_SAMPLES;
	rtt_index = 0;

	window = 1;

	m_gamma = 0.5;
	m_beta = 5.0;
	m_alpha = 3.0;

	m_should_update = 0;
	m_last_lost_cnt = 0;
	m_last_avg_rtt = 0;
}

prrt_delay_cc::prrt_delay_cc(double alpha, double beta, double gamma)
: m_alpha(alpha), m_beta(beta), m_gamma(gamma)
{
	baseRTT = 0;
	m_avg_rtt = 0;

	min_rtt = new value_pair[MIN_RTT_SAMPLES];
	max_rtt = new value_pair[MAX_RTT_SAMPLES];
	min_end = min_rtt + MIN_RTT_SAMPLES;
	min_last = min_rtt;
	min_pair = min_rtt;
	max_end = max_rtt + MAX_RTT_SAMPLES;
	max_last = max_rtt;
	max_pair = max_rtt;
	min_pair->value = 20000;
	min_pair->expiration = MIN_RTT_SAMPLES;
	max_pair->value = 0;
	max_pair->expiration = MAX_RTT_SAMPLES;
	rtt_index = 0;

	window = 1;

  m_should_update = 0;
  m_last_lost_cnt = 0;
  m_last_avg_rtt = 0;
}

prrt_delay_cc::~prrt_delay_cc() {
  delete[] min_rtt;
  delete[] max_rtt;
}

void prrt_delay_cc::update_base_RTT(double RTT)
{
	float newRTT = (float)RTT;

	// Ascending Minima Algorithm by Richard Harter: http://richardhartersworld.com/cri/2001/slidingmin.html
	if (min_pair->expiration <= rtt_index){
		min_pair++;
		if (min_pair >= min_end)
			min_pair = min_rtt;
	}
	if (max_pair->expiration <= rtt_index){
		max_pair++;
		if (max_pair >= max_end)
			max_pair = max_rtt;
	}

	if (newRTT <= min_pair->value){
		min_pair->value = newRTT;
		min_pair->expiration = rtt_index + MIN_RTT_SAMPLES;
		min_last = min_pair;
	} else {
		while (min_last->value >= newRTT){
			if (min_last == min_rtt)
				min_last = min_end;
				--min_last;
		}
		++min_last;
		if (min_last == min_end)
			min_last = min_rtt;
		min_last->value = newRTT;
		min_last->expiration = rtt_index + MIN_RTT_SAMPLES;
	}

	baseRTT = (double)min_pair->value;

	if (newRTT >= max_pair->value){
		max_pair->value = newRTT;
		max_pair->expiration = rtt_index + MAX_RTT_SAMPLES;
		max_last = max_pair;
	} else {
		while (max_last->value <= newRTT){
			if (max_last == max_rtt)
				max_last = max_end;
				--max_last;
		}
		++max_last;
		if (max_last == max_end)
			max_last = max_rtt;
		max_last->value = newRTT;
		max_last->expiration = rtt_index + MAX_RTT_SAMPLES;
	}

	maxRTT = (double)max_pair->value;
	
	//cerr << "newRTT " << newRTT << ", baseRTT " << baseRTT << ", maxRTT " << maxRTT << endl;

	rtt_index++;
}

unsigned int prrt_delay_cc::calc_W(double recv_rate, double RTT, unsigned long lost_cnt)
{	
	//equation-based congestion control derived from TCP Vegas
	double rate = 0.0;
	update_base_RTT(RTT);

	if (baseRTT && RTT) {

			if (RTT < baseRTT)
				RTT = RTT;
			if (RTT > maxRTT)
				RTT = maxRTT;

		double expected_throughput = window / baseRTT;
		double actual_throughput = window / RTT;

		double diff_window = (expected_throughput - actual_throughput) * baseRTT;

		if (diff_window < m_alpha) {
			window += 0.5;
		}

		if (diff_window > m_beta && window > 0){
			window--;
		}

		rate = window * PKT_SIZE * 8 / RTT;

		cerr << "expected_throughput " << expected_throughput << ", actual_throughput " << actual_throughput << ", diff_window " << diff_window << ", avgRTT " << RTT << ", baseRTT " << baseRTT << ", maxRTT " << maxRTT << ", window " << window << ", rate " << rate << ", alpha " << m_alpha << ", beta " << m_beta << endl;
	}

	return static_cast<unsigned int>(rate);
}

void prrt_delay_cc::setAlpha(double alpha)
{
  m_alpha = alpha;
}
double prrt_delay_cc::getAlpha()
{
  return m_alpha;
}

void prrt_delay_cc::setBeta(double beta)
{
  m_beta = beta;
}
double prrt_delay_cc::getBeta()
{
  return m_beta;
}

void prrt_delay_cc::setGamma(double gamma)
{
  m_gamma = gamma;
}
double prrt_delay_cc::getGamma()
{
  return m_gamma;
}
