/* 
 * (c) 2010 Saarland University, Telecommunications Lab 
 * (prrt@nt.uni-saarland.de)
 *
 * The contents of this file are subject to the Mozilla Public License
 + Version 1.1 (the "License"); you may not use this file except in
 + compliance with the License. You may obtain a copy of the License at
 + http://www.mozilla.org/MPL/
 + 
 + Software distributed under the License is distributed on an "AS IS"
 + basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 + License for the specific language governing rights and limitations
 + under the License.
 + 
 + The Original Code is "Predictably Reliable Real-time Transport (PRRT)".
 + 
 + The Initial Developer of the Original Code is Saarland University.
 + All Rights Reserved.
 + 
 + Contributor(s): Jochen Gruen (Saarland University), Manuel Gorius
 + (Saarland University), Yongtao Shuai (Saarland University).
 + 
 + Disclaimer: This Alpha version of PRRT is not supported and stable yet.
 + Do not use it in a production environment: it may crash your device, 
 + so please make sure that you have a backup of your data before
 + installing PRRT.
 */

#include "socket_list.h"
#include "prrt_sockets.h"
#include "../prrt_module/kui/kui_error.h"
#include "udp_wrapper.h"

#include <sstream>
#include <string.h>
#include <stdlib.h>
#include <cerrno>       // for errnos (ETIMEDOUT)
#include <stdexcept>    // for runtime_error

#include <sys/socket.h>
#include <sys/types.h>
#include <arpa/inet.h>

using namespace media_transport_protocol;

socket_list prrt_socket::SocketList;

#define DEBUG 1
// #undef DEBUG

inline void DEBUG_MSG(const char *msg) {
#ifdef DEBUG
	std::cerr << msg << std::endl;
#endif
}

const std::string CONTROL_FILE = "control";
const std::string WRITE_FILE = "write";
const std::string READ_FILE = "read";
const std::string OPTION_FILE = "option";
const std::string STATS_FILE = "stats";

static const int MAX_BANDWIDTH = 20000; // in kbits/s.
const int prrt_socket::DEFAULT_PAYLOAD_SIZE = 1316;

prrt_socket::prrt_socket()
throw (module_error, list_error, prrt_error) :
boundSockAddr(""), boundPeerAddr(""), boundPort(0),
connectedSockAddr(""), connectedHostAddr(""), connectedPort(0),
rcv_buf_size(0), kui_kernel_sock_buf_size(0), kui_user_sock_buf_size(0),
m_adaptation_mode(0), m_update_interval(200), 
m_delay_target(200), m_delay_fec(2), m_k(1), m_n_p(),
m_bw_est_mode(0), /*m_filename_analysis("goodput.dat"), m_file_analysis(0),*/
m_recv_rate(100), m_avg_recv_rate(-0.0000001), m_est_bw(1000),
m_is_running(false), prrtAnalysis(), m_send_payload_size(DEFAULT_PAYLOAD_SIZE), m_streamrate(1000),
m_recv_payload_size(DEFAULT_PAYLOAD_SIZE) {
    std::string command;
    std::string feedback;
    int i = 0;
    int ret = 0;
    std::string register_file_path(PRRT_PATH + REGISTER_FILE);

    this->open = false;

    proc_file pf(register_file_path);

    create_kui_socket();

    std::cerr << "Kui port " << this->kui_port_read << std::endl;

    std::ostringstream stm;
    stm << this->kui_port_read;

    command.append("create "+ stm.str());
    //command.append("create");

    pf.write(command);

    pf.read(feedback);

    /*the return value of the register file should be finished with an 'a'*/
    std::stringstream sstream(feedback);
    sstream >> ret;
    sstream.str("");
    sstream << ret;

    if (ret < 0) {
        throw (module_error("Could not create socket", ret));
    }

    this->sockno = sstream.str();

    this->open = true;
    std::cerr << "Created socket no " << ret << std::endl;

    try {
        this->SocketList.add(this);
    } catch (...) {
        this->close();
        throw (list_error("Could not insert socket in list!"));
    }

    /* Create the UDP socket */
    if ((this->kui_sock_write = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0) {
	    perror("kui write socket failed");
    }
    
    // Get the default buffer size of the UDP sockets used in PRRT.
    this->setsockopt( option_RCVBUF( option_RCVBUF::PRRT_RECV, 8388608 ) );
    this->setsockopt( option_RCVBUF( option_RCVBUF::USER_RECV, 8388608 ) );
    this->setsockopt( option_RCVBUF( option_RCVBUF::KERNEL_RECV, 8388608 ) );

    std::cerr << "buffer size of PRRT sock:" << this->rcv_buf_size << std::endl;
    std::cerr << "buffer size of user sock:" << this->kui_user_sock_buf_size << std::endl;
    std::cerr << "buffer size of kernel sock:" << this->kui_kernel_sock_buf_size << std::endl;

    // Initialize timing parameters for TFRC
    this->lastGRTT = 50; 
    gettimeofday(&(this->tv_last_param_update), NULL);

    // Initialize mutex attributes
    pthread_mutexattr_t attr;
    pthread_mutexattr_init(&attr);
    pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_NORMAL);

    // Initialize mutex
    pthread_mutex_init(&m_mutex, &attr);
    pthread_mutex_init(&m_mutex_streamrate, &attr);
    pthread_mutexattr_destroy(&attr);
  
    pthread_cond_init(&m_condition, 0);

    interval.tv_sec = 0;
    interval.tv_usec = 200*1000;
    m_update_interval = 200;

    //m_file_analysis = fopen64(m_filename_analysis.c_str(), "w");
    //if (m_file_analysis == NULL) {
    //  std::cerr << "prrt_sockets : m_file_analysis open fail" << std::endl;
    //}
}

prrt_socket::~prrt_socket() {
    //	std::cerr << "Destructor called for socket no " << this->sockno << std::endl;
    this->close();
}

void prrt_socket::bind(std::string &sockaddr, unsigned short sockport, std::string &peeraddr)
throw (module_error, prrt_error) {
    std::string feedback;
    std::string command;
    std::stringstream cstream;
    int ret = 0;
    std::string control_file_path(PRRT_PATH + this->sockno + "/" + CONTROL_FILE);
    proc_file pf(control_file_path);

    cstream << "bind " << sockaddr << " " << sockport << " " << peeraddr;
    command = cstream.str();

    //	DEBUG_MSG("writing to procfile\n");
    pf.write(command);

    //	DEBUG_MSG("reading feedback from procfile\n");
    pf.read(feedback);

    std::stringstream sstream(feedback);
    sstream >> ret;

    if (ret < 0)
        throw (module_error("Unable to bind socket", ret));

    boundPort = sockport;
    boundSockAddr = sockaddr;
    boundPeerAddr = peeraddr;
}

void prrt_socket::connect(std::string &sockaddr, unsigned short sockport, std::string &hostaddr)
throw (module_error, prrt_error) {
    std::string feedback;
    std::string command;
    std::stringstream cstream;
    int ret = 0;
    std::string control_file_path(PRRT_PATH + this->sockno + "/" + CONTROL_FILE);
    proc_file pf(control_file_path);

    cstream << "connect " << sockaddr << " " << sockport << " " << hostaddr;
    command = cstream.str();

    //	DEBUG_MSG("writing to procfile");
    pf.write(command);

    //	DEBUG_MSG("reading feedback from procfile");
    pf.read(feedback);

    std::stringstream retstream(feedback);
    retstream >> ret;

    if (ret < 0) {
        throw (module_error("Unable to connect socket", ret));
    }
    this->kui_port_write = ret;
    //std::cerr << "Bound kui_socket_write to port " << this->kui_port_write << std::endl;
    connectedPort = sockport;
    connectedSockAddr = sockaddr;
    connectedHostAddr = hostaddr;
}

void prrt_socket::close()
throw () {

    if (m_adaptation_mode) {
      stopAdaptation();
    }

    std::string command;
    std::string feedback;
    int i = 0;
    int ret = 0;
    std::string register_file_path(PRRT_PATH + REGISTER_FILE);

    if (!this->open)
        return;

    proc_file pf(register_file_path);

    std::cerr << "Closing socket no " << this->sockno << std::endl;
    command.append("delete " + this->sockno);

    //	DEBUG_MSG("writing to procfile\n");
    pf.write(command);

    //	DEBUG_MSG("reading feedback from procfile\n");
    pf.read(feedback);

    /*the return value of the register file should be finished with an 'a'*/
    std::stringstream sstream(feedback);
    sstream >> ret;

    //	if(ret < 0)
    //		throw(module_error("Could not close socket", ret));

    if (this->kui_sock_write)
	    udp_wrapper::udp_close(this->kui_sock_write);
    if (this->kui_sock_read)
	    udp_wrapper::udp_close(this->kui_sock_read);

    this->SocketList.del(this);
    this->open = false;
    //fclose(m_file_analysis);
    std::cerr << "Done Closing socket no " << this->sockno << std::endl;
}

int prrt_socket::send(char *data, int len, int flags)
throw (prrt_error) {
    struct sockaddr_in write_sock;

    memset(&write_sock, 0, sizeof(write_sock));  
    write_sock.sin_family = AF_INET; 
    write_sock.sin_addr.s_addr = inet_addr("127.0.0.1");
    write_sock.sin_port = htons(this->kui_port_write);

    //static uint64_t sent_time = -1;  // in microseconds. (-1 as maxima for initialization.)
    //static timeval tv_start;
    //timeval tv_now;
    //timeval tv_diff;
    //if (sent_time == -1) {
    //  gettimeofday(&tv_start, NULL);
    //}
    //gettimeofday(&tv_now, NULL);
    //timersub(&tv_now, &tv_start, &tv_diff);
    //sent_time = tv_diff.tv_sec * 1000 + tv_diff.tv_usec / 1000;

    //static char buf[1500]; // TODO handle len > 1492.
    //memcpy(buf, &sent_time, 8);
    //memcpy(buf+8, data, len);

    int ret = 0;
    if ((ret = udp_wrapper::udp_sendto(this->kui_sock_write, data, len, 0, (struct sockaddr *) &write_sock, sizeof(write_sock))) < 0){
      perror("sendto");
      return -1;
    }

    // Compute the average payload size using exponential smoothing.
    int avg_len = (len + getSendPayloadSize()) / 2;
    setSendPayloadSize( avg_len );

    //std::cerr << "Wrote " << ret << " bytes to kernel space" << std::endl;

    return ret;
}

int prrt_socket::recv(char *data, int len, int flags)
throw(module_error, prrt_error)
{
    int ret = 0;
    int recv_bytes = 0;

    //static char buf[1500];  // TODO handle len > 1492.
    //uint64_t sent_time = 0;

    if ((ret = udp_wrapper::udp_recvfrom(this->kui_sock_read, data, len, 0, NULL, NULL)) < 0){
      perror("recvfrom");
      return -1;
    }
        
    //memcpy(&sent_time, buf, 8);
    //memcpy(buf, data, ret);

    //static uint64_t recv_time = -1;
    //static timeval tv_start;
    //timeval tv_now;
    //timeval tv_diff;
    //if (recv_time == -1) {
    //  gettimeofday(&tv_start, NULL);
    //}
    //gettimeofday(&tv_now, NULL);
    //timersub(&tv_now, &tv_start, &tv_diff);
    //recv_time = tv_diff.tv_sec * 1000 + tv_diff.tv_usec / 1000;

    // receive data till recv bytes equal to request len if there is data in buffer.
    //while (m_recv_payload_size <= len)
    //{
    //  if ((recv_bytes = udp_wrapper::udp_recvfrom(this->kui_sock_read, data, len, 0, NULL, NULL)) < 0){
    //    perror("recvfrom");
    //    return -1;
    //  }
    //  ret += recv_bytes;
    //  len -= recv_bytes;
    //  data += recv_bytes;
    //  if (recv_bytes == 0 || recv_bytes < m_recv_payload_size) {
    //    // receiving buffer is empty at the moment.
    //    break;
    //  }
    //}


    if (0 < m_bw_est_mode) {
      this->update_bw_est();
    }

    // Compute the goodput : the number of bytes of received data pre time unit.
    static int bytes = -1;  // -1 for initialize the tv_last.
    static timeval tv_last;
    //static int time = 0;
    if (bytes == -1) {
      gettimeofday(&tv_last, NULL);
      bytes = 0;
    }
    timeval tv_now;
    timeval tv_diff;
    timeval time_unit;
    time_unit.tv_sec = 1;
    time_unit.tv_usec = 0;

    gettimeofday(&tv_now, NULL);
    timersub(&tv_now, &tv_last, &tv_diff);
    bytes += ret;

    if ( tv_diff.tv_sec > time_unit.tv_sec || (tv_diff.tv_sec == time_unit.tv_sec && tv_diff.tv_usec > time_unit.tv_usec )) {
      double goodput = bytes * 8 / (tv_diff.tv_sec +  (double) tv_diff.tv_usec / (1000*1000));
      // FIXME : m_recv_rate thread unsafety.
      m_recv_rate = (goodput / 8) / m_recv_payload_size; // roughly in packets per second.
      if (m_avg_recv_rate < 0.0) {
        // initialize. init value -0.0000001
        m_avg_recv_rate = m_recv_rate;
      } else {
        m_avg_recv_rate = (255 * m_avg_recv_rate + m_recv_rate) / 256;
      }
      //std::cerr << "prrt_sockets recv() : goodput=" << bytes * 8 / (tv_diff.tv_sec + tv_diff.tv_usec / (1000*1000)) << " (bps); bytes=" << bytes << " recv_time-sent_time=" << recv_time << " - " << sent_time << " = " << recv_time-sent_time << std::endl;
      bytes = 0;
      gettimeofday(&tv_last, NULL);
      //fprintf(m_file_analysis, "%d %f %d\n", time++, goodput/1024, m_est_bw);
      // FIXME : we call getStats() here, it could lead to thread unsafety because another thread running for update_coding_params() uses it too.
	    SocketStatistics sockStats;
      this->getStats(&sockStats);
      //fprintf(m_file_analysis, "%d %f %d %d %d %d %d %d %d\n", time++, goodput/1024, m_est_bw, sockStats.lastGPLR, sockStats.lastGRTT, sockStats.numGaps, sockStats.numBursts, sockStats.gapLength, sockStats.burstLength);
    }

    //std::cerr << "Read " << ret << " bytes" << std::endl;
    return ret;
}

void prrt_socket::setsockopt(const prrt_option &option)
throw (prrt_error, module_error) {
    int ret = 0;
    std::string command;
    std::string feedback;
    std::stringstream cstream;
    unsigned int len = 0;

    // Determine if the adaptation is used.
    if (option.get_option() == "simple") {
      if (option.get_value() < 0) {
        throw (prrt_error("Unable to set option due to invalid update interval for adaptation"));
        return;
      }
      m_update_interval = option.get_value();
      m_adaptation_mode = 2;
      //startAdaptation();
      return;
    }

    // Determine if the feedback suppression is used.
    if (option.get_option() == "fb_supp_on") {
      if (option.get_value() < 0) {
        throw (prrt_error("Unable to set option due to invalid feedback sent delay for feedback suppression"));
        return;
      }
      //m_fb_sent_delay = option.get_value();
    }

    // Determine if the bandwidth estimation is used.
    if (option.get_option() == "bwest_tfrc") {
      if (option.get_value() <= 0) {
        throw (prrt_error("Unable to set option due to invalid update interval for bandwidth estimation"));
        return;
      }
      m_bw_est_mode = 1;
      interval.tv_sec = option.get_value() / 1000;
      interval.tv_usec = option.get_value() % 1000 * 1000;
      return;
    }
    else if (option.get_option() == "bwest_rtt") {
      if (option.get_value() <= 0) {
        throw (prrt_error("Unable to set option due to invalid update interval for bandwidth estimation"));
        return;
      }
      m_bw_est_mode = 2;
      interval.tv_sec = option.get_value() / 1000;
      interval.tv_usec = option.get_value() % 1000 * 1000;
      return;
    }

    // Set the max buffer size of the socket used in the user space of KUI communication.
    if (option.get_option() == "user_rcvbuf") {
      int size = option.get_value();
      if (size < 0) {
        throw (module_error("Unable to set option", ret));
      }
      else if (0 < size) {
        if ( ::setsockopt(this->kui_sock_read, SOL_SOCKET, SO_RCVBUF, &size, sizeof(size)) == -1 ) {
          throw (prrt_error("Unable to set option due to ::setsockopt() on SO_RCVBUF failed"));
        }
      }
      // If the option value is 0, we do nothing and just query the default/current max buffer size by getsockopt.
      len = sizeof(size);
      if ( ::getsockopt(this->kui_sock_read, SOL_SOCKET, SO_RCVBUF, &size, &len) == -1 ) {
        throw (prrt_error("Unable to set option due to ::getsockopt() on SO_RCVBUF failed"));
      }
      kui_user_sock_buf_size = size;
      return;
    }

    std::string option_file_path(PRRT_PATH + this->sockno + "/" + OPTION_FILE);
    proc_file pf(option_file_path);

    cstream << option.get_option() << " " << option.get_value();
    command = cstream.str();
    ret = pf.write(command);

    ret = pf.read(feedback);
    cstream.str(feedback);
    cstream >> ret;

    if (ret < 0)
        throw (module_error("Unable to set option", ret));

    // Stores the max buffer size of the sockets.
    // If the option value is 0, we just query the default/current max buffer size by return value from the proc file.
    if (option.get_option() == "prrt_rcvbuf") {
      rcv_buf_size = ret;
    }
    else if (option.get_option() == "kernel_rcvbuf") {
      kui_kernel_sock_buf_size = ret;
    }
}

int prrt_socket::getsockopt(const prrt_option &option)
throw (prrt_error, module_error) {
    if (option.get_option() == "prrt_rcvbuf") {
      return rcv_buf_size;
    }
    else if (option.get_option() == "kernel_rcvbuf") {
      return kui_kernel_sock_buf_size;
    }
    else if (option.get_option() == "user_rcvbuf") {
      return kui_user_sock_buf_size;
    }
    else {
      throw (prrt_error("Unable to get option due to not supported option."));
    }
}

/*template<class T>*/ 
void prrt_socket::set_param(const prrt_param/*<T>*/ &param)
throw (prrt_error) {

    int ret = 0;
    std::string write_file_path(PRRT_PATH + this->sockno + "/" + param.get_file_name());
    proc_file pf(write_file_path);

    /*DEBUG_MSG("Writing new value to param")*/;
    ret = pf.write(param.get_strvalue());

    if (param.get_file_name() == "d_target") {
      //m_delay_target = atoi(param.get_strvalue().c_str());
    }

    if (param.get_file_name() == "ahec") {
      const param_AHEC *ahec = dynamic_cast<const param_AHEC*>( &param );
      m_delay_target = ahec->get_d_target();	
      m_delay_fec = ahec->get_d_fec();
      m_delay_fb = ahec->get_d_fb();
      m_k = ahec->get_k();
      m_n_p = ahec->get_n_p();
    }
}

void prrt_socket::getStats(SocketStatistics *sockStats)
throw (prrt_error) {
    int ret = 0;
    std::string stats_file_path(PRRT_PATH + this->sockno + "/" + STATS_FILE);
    proc_file pf(stats_file_path, false);
    std::string buffer;

    /*DEBUG_MSG("Writing new value to param")*/;
    ret = pf.read(buffer);
    //std::cerr << buffer;

    unsigned long temp;

    std::stringstream ss(buffer);
    std::string tempString;

    ss >> tempString >> tempString >> temp;
    sockStats->deliveredData = temp;

    ss >> tempString >> tempString >> temp;
    sockStats->deliveredRedundancy = temp;

    ss >> tempString >> tempString >> temp;
    sockStats->lostPackets = temp;

    ss >> tempString >> tempString >> temp;
    sockStats->droppedPackets = temp;

    //Residual loss
    ss >> tempString >> tempString >> temp;
    sockStats->missedPackets = temp;

    ss >> tempString >> tempString >> temp;
    sockStats->receivedData = temp;

    ss >> tempString >> tempString >> temp;
    sockStats->receivedRedundancy = temp;

    ss >> tempString >> tempString >> temp;
    sockStats->lastGRTT = temp;

    ss >> tempString >> tempString >> temp;
    sockStats->lastFTT = temp;

    ss >> tempString >> tempString >> temp;
    sockStats->lastGFTT = temp;

    ss >> tempString >> tempString >> temp;
    sockStats->queueDelay = temp;

    ss >> tempString >> tempString >> temp;
    sockStats->lastPLR = temp;

    ss >> tempString >> tempString >> temp;
    sockStats->lastGPLR = temp;

    ss >> tempString >> tempString >> temp;
    sockStats->numGaps = temp;

    ss >> tempString >> tempString >> temp;
    sockStats->numBursts = temp;

    ss >> tempString >> tempString >> temp;
    sockStats->gapLength = temp;

    ss >> tempString >> tempString >> temp;
    sockStats->burstLength = temp;

    ss >> tempString >> tempString >> temp;
    sockStats->numGGaps = temp;

    ss >> tempString >> tempString >> temp;
    sockStats->numGBursts = temp;

    ss >> tempString >> tempString >> temp;
    sockStats->gapGLength = temp;

    ss >> tempString >> tempString >> temp;
    sockStats->burstGLength = temp;

    ss >> tempString >> tempString >> temp;
    sockStats->packetInterval = temp;

    ss >> tempString >> tempString >> temp;
    sockStats->bwEst = temp;

    ss >> tempString >> tempString >> temp;
    sockStats->buffb = temp;

    ss >> tempString >> tempString >> sockStats->lossPattern;

//	//std::cout << tempString << std::endl;
//
//	std::cerr << pattern.c_str() << " RTT: " << sockStats->lastGRTT << std::endl;
//	std::cout << pattern.c_str() << std::endl;
//	
//
}

///*template<class T>*/ void prrt_socket::get_param(prrt_param/*<T>*/ &param)
//throw(prrt_error)
//{
//	int ret = 0;
//	std::string read_file_path(PRRT_PATH+this->sockno+"/"+param.get_file_name());
//	proc_file pf(read_file_path);
//	std::string value;
//
//	DEBUG_MSG("Writing new value to param");
//	ret = pf.read(value);
//
//	param.set(value);
//}

bool prrt_socket::operator==(prrt_socket& s)
throw (prrt_error) {
    return sockno == s.sockno;
}

void prrt_socket::delete_all_sockets() {
    try {
        while (true) {
            prrt_socket *sock = SocketList.del_rnd();
            sock->close();
            std::cerr << "Deleted socket" << std::endl;
        }
    } catch (socket_list::list_error &le) {
        ;
    }
}

int prrt_socket::create_kui_socket()
throw (prrt_error)
{
	int err = 0;
	int receivelen;
	struct sockaddr_in receivesocket;

	/* Create the UDP socket */
	if ((this->kui_sock_read = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0) {
		throw (prrt_error("Could not create kui socket: create failed"));
	}

	memset(&receivesocket, 0, sizeof(receivesocket));  
	receivesocket.sin_family = AF_INET; 
	receivesocket.sin_addr.s_addr = inet_addr("127.0.0.1");
	receivelen = sizeof(receivesocket);
	int kui_port = 0;

	for (kui_port = 32000; kui_port < 32100; kui_port++){
		receivesocket.sin_port = htons(kui_port);

	//std::cerr << "trying port " << kui_port << std::endl;
		err = udp_wrapper::udp_bind(this->kui_sock_read, (struct sockaddr *) &receivesocket, receivelen);
		if (!(err < 0)) {
			break;
		}
	}

	if (err < 0)
		throw (prrt_error("Could not create kui socket: bind failed"));
	this->kui_port_read = kui_port;
	//std::cerr << "Bound kui socket to port " << kui_port << std::endl;
}

void prrt_socket::update_bw_est(){
	unsigned int bw = 0;
	prrt_socket::SocketStatistics sockStats;
	timeval tv_now;
	timeval tv_diff;

	gettimeofday(&tv_now, NULL);
	timersub(&tv_now, &(this->tv_last_param_update), &tv_diff);

  if ( tv_diff.tv_sec > interval.tv_sec || (tv_diff.tv_sec == interval.tv_sec && tv_diff.tv_usec > interval.tv_usec )){
		this->lastGRTT = sockStats.lastGRTT;
		gettimeofday(&(this->tv_last_param_update), NULL);

		this->getStats(&sockStats);

		if (1 == m_bw_est_mode) {
			if (sockStats.lossPattern.c_str())
			{
				if(sockStats.lastGRTT != 0)
					bw = tfrc.calc_T((float)sockStats.lastGRTT/1000.0, sockStats.lossPattern.c_str(),sockStats.packetInterval/1000.0 );
				else
					bw = tfrc.calc_T((float)1/1000.0, sockStats.lossPattern.c_str(), sockStats.packetInterval/1000.0);
			}
			//std::cerr << "Estimated bandwidth by TCP Friendly Rate Control : " << bw << std::endl;
		}
		else if (2 == m_bw_est_mode) {
      static unsigned long last_num_lost_packets = 0;
      double delta = sockStats.lostPackets - last_num_lost_packets;
      double loss = sockStats.lastPLR / 65535.0;
      static double avg_loss = loss;
      avg_loss = (127 * avg_loss + loss) / 128;
      last_num_lost_packets = sockStats.lostPackets;
			if (sockStats.lastGRTT != 0) {
				bw = m_delay_cc.calc_W(m_recv_rate, sockStats.lastGRTT, sockStats.lostPackets);
				//bw = m_delay_cc.calc_W(sockStats.lastGRTT * (1 + avg_loss));
			}
      //std::cerr.precision(4);
			//std::cerr << "Estimated bandwidth by RTT : " << bw  << " with loss=" << avg_loss << " lastGRTT=" << sockStats.lastGRTT << std::endl;
		}
		else {
			//std::cerr << "update_bw_est() unknown mode for bandwidth estimation." << std::endl;
		}
		this->set_param(param_BW_est(bw));
    m_est_bw = bw;
	}
}

void prrt_socket::update_coding_params()
{
  float channel_bandwidth = 8.0e3; // in kbits/s. The channel bandwidth we assume it to be twice than stream bandwidth.
  float stream_bandwidth = 4.0e3;  // in kbits/s. The actual bandwidth allowed for streaming.
  
  FILE *file_stats = fopen64("stats.dat", "w");
  timeval tv_last;
  int time = 0;
  gettimeofday(&tv_last, NULL);

  pthread_mutex_lock(&m_mutex);
  try {
    while ( m_is_running ) {

      prrt_socket::SocketStatistics sockStats;

      /* initilaization to optimizeSingleParamSet arguments  */
      float RTT = 10.0;
      unsigned int R =1;     /* number of Receivers*/
      long double Pf = 0;
      long double Pe = 0;
      long double Ptarget = 1.0e-6;
      float rho = 0.0;

      struct timeval current_time;
      ::gettimeofday(&current_time, 0);

      // Set up 'timespec' structure. Check for overrun on the 'nsec' field.
      long nsec = m_update_interval % 1000 * 1000000;
      long current_time_nsec = current_time.tv_usec * 1000;
      long nsec_from_max = 1000000000L - nsec;

      struct timespec timeout;
      timeout.tv_sec = current_time.tv_sec + m_update_interval / 1000;

      if(nsec_from_max > current_time_nsec) {
        timeout.tv_nsec = current_time_nsec + nsec;
      }
      else {
        // Handle nsec overrun that would cause pthread_cond_timedwait() to return EINVAL.
        ++timeout.tv_sec;
        timeout.tv_nsec = current_time_nsec - nsec_from_max;
      }

      int result = pthread_cond_timedwait(&m_condition, &m_mutex, &timeout);

      this->getStats(&sockStats);

      if (sockStats.bwEst > MAX_BANDWIDTH) {
        //sockStats.bwEst = MAX_BANDWIDTH;
      }

      if (stream_bandwidth < sockStats.bwEst) {
        // Logarithmic increasing.
        //stream_bandwidth = stream_bandwidth + (sockStats.bwEst-stream_bandwidth)/(0.005*stream_bandwidth);
        stream_bandwidth = sockStats.bwEst;
      }

      if (stream_bandwidth > sockStats.bwEst) {
        stream_bandwidth = sockStats.bwEst;
      }
      channel_bandwidth = 2 * stream_bandwidth;

      Pe = (float)sockStats.lastGPLR/65535.0;
      float plr = prrtAnalysis.smoothPLR(Pe, 0.001, 0.05, 0.02);
      //if (plr > 0.15) {
      //  plr = 0.15;
      //}
      RTT = (float)sockStats.lastGRTT;
	
	    //std::cerr<< "updateInterval=" << m_update_interval << " Pe=" << Pe << " m_delay_target="<<m_delay_target <<" plr=" << plr << " RTT=" << RTT << " channel_bw=" << channel_bandwidth << " stream_bw=" << stream_bandwidth << std::endl;

      unsigned int payload_size = getSendPayloadSize();    // This is used to compute the actual stream rate i.e. substracting the protocol overhead.
	    double stream_rate = prrtAnalysis.optimizeSingleParamSet(m_delay_target, Ptarget, channel_bandwidth, stream_bandwidth, R, plr, RTT, 0, rho, payload_size, 30, 20, 10, m_delay_fb, 0);
      //std::cerr << "stream_rate=" << stream_rate << " payload_size=" << payload_size << std::endl;
      setStreamrate( stream_rate );

      // Record the number of gaps and bursts, gap length, burst length, rtt, datarate, plr in a file called "stats.dat".
      timeval tv_now;
      timeval tv_diff;
      timeval time_unit;
      time_unit.tv_sec = 1;
      time_unit.tv_usec = 0;

      gettimeofday(&tv_now, NULL);
      timersub(&tv_now, &tv_last, &tv_diff);

      if ( tv_diff.tv_sec > time_unit.tv_sec || (tv_diff.tv_sec == time_unit.tv_sec && tv_diff.tv_usec > time_unit.tv_usec )) {
        //std::cerr << "prrt_sockets update_coding_params() : stream_rate=" << stream_rate << " (bps); stream_bw=" << stream_bandwidth << std::endl;
        gettimeofday(&tv_last, NULL);
        fprintf(file_stats, "%d %f %f %d %d %d %d %d %d\n", time++, stream_rate, stream_bandwidth, sockStats.lastGPLR, sockStats.lastGRTT, sockStats.numGaps, sockStats.numBursts, sockStats.gapLength, sockStats.burstLength);
      }

        this->set_param(prrtAnalysis.get_AHEC());      

      if (result == ETIMEDOUT) {
        
      }
      else if (result == 0) {
        
      }
      else {
        // This never happens.
        // As with most strange things that never happen, we would like to know when it happens.
        // In software, never can be any time, and we never want to miss it.
        //std::cerr << "> prrt_socket::update_coding_params() : Error in relativeTimedWait" << std::endl;
      }

    }
  } catch (prrt_error &e) {
    std::cerr << "prrt_socket::update_coding_params() prrt_error : " << e.what() << " !" << std::endl;
  } catch (...) {
    std::cerr << "prrt_socket::update_coding_params() something wrong!" << std::endl;
  }

  pthread_mutex_unlock(&m_mutex);
  fclose(file_stats);
  std::cout << "prrt_socket::update_coding_params() done & exit." << std::endl;
}

void prrt_socket::startAdaptation()
{
    if (m_adaptation_mode < 2) {
      // Online computation is not active.
      return;
    }

    pthread_mutex_lock(&m_mutex);
    // If the adaptation is running, do nothing.
    if ( m_is_running ) {
      pthread_mutex_unlock(&m_mutex);
      return;
    }
    
    m_is_running = true;

    // Start adaptation thread
    try {
      if (pthread_create(&m_thread, 0, static_update_coding_params, static_cast<void *>(this)) != 0) {
        std::cerr << "> prrt_socket::startAdaptation() pthread_create failed" << std::endl;
        throw std::runtime_error("not enough system resources to create a process for the new thread");
      }
    }
    catch(...) {
      std::cerr << "> PRRT ERROR (start) : Unable to activate thread for the adaptation of the PRRT." << std::endl;
      pthread_mutex_unlock(&m_mutex);
      throw prrt_error("Unable to activate thread for the adaptation of the PRRT.");
    }

    pthread_mutex_unlock(&m_mutex);
    pthread_cond_broadcast(&m_condition);
}

void prrt_socket::stopAdaptation()
{
    pthread_mutex_lock(&m_mutex);
    if (m_is_running) {
      m_is_running = false;
    }
    else {
      pthread_mutex_unlock(&m_mutex);
      return;
    }
    pthread_mutex_unlock(&m_mutex);

    pthread_cond_broadcast(&m_condition);

    // Join thread
    if (pthread_join(m_thread, 0) == 0) {
      
    }
    else {
      std::cerr << "> prrt_socket::stopAdaptation() pthread_join failed" << std::endl;
    }
}

int prrt_socket::getSendPayloadSize()
{
  pthread_mutex_lock(&m_mutex_streamrate);
  int size = m_send_payload_size;
  pthread_mutex_unlock(&m_mutex_streamrate);
  return size;
}

void prrt_socket::setSendPayloadSize( int size )
{
  pthread_mutex_lock(&m_mutex_streamrate);
  m_send_payload_size = size;
  pthread_mutex_unlock(&m_mutex_streamrate);
}

double prrt_socket::getStreamrate()
{
  pthread_mutex_lock(&m_mutex_streamrate);
  int rate = m_streamrate;
  pthread_mutex_unlock(&m_mutex_streamrate);
  return rate;
}

void prrt_socket::setStreamrate( double rate )
{
  pthread_mutex_lock(&m_mutex_streamrate);
  m_streamrate = rate;
  pthread_mutex_unlock(&m_mutex_streamrate);
}

double prrt_socket::evaluateStreamrate()
{
  double ri = 0.0;  // redundancy information.
  double stream_rate = 1000.0;
  unsigned int payload_size = getSendPayloadSize();
  unsigned int num_recv = 1; // the number of receivers.
  unsigned int *n_p = 0;   // the transmission rounds vector for parity packets.
  unsigned int n_rr = (unsigned int)m_n_p.size();
  prrt_analysis analysis; // used to calculate the stream rate.
	prrt_socket::SocketStatistics sockStats;
  this->getStats(&sockStats);   // FIXME thread-safety?
  if (n_rr > 0) {
    n_p = new unsigned int[n_rr];
    for (unsigned int i = 0; i < n_rr; i++) {
      n_p[i] = m_n_p[i];
    }
  }

  float plr = sockStats.lastGPLR / 65535.0;
  ri = analysis.computeRI(num_recv, m_k, n_p, n_rr, plr, 0.0);

  // 16 : PRRT DATA PACKET HEADER LENGTH; 8 : PRRT PARITY PACKET HEADER LENGTH
  stream_rate = sockStats.bwEst / ( 1 + ri + 16.0 / payload_size + 8.0 / payload_size * ri ); 

  //std::cerr << "evaluateStreamrate() : ri=" << ri << " stream_rate=" << stream_rate << " bwEst=" << sockStats.bwEst << " k=" << m_k << " n_p=[";
  //for (int i = 0; i < n_rr; i++) {
  //  std::cerr << n_p[i] << ",";
  //}
  //std::cerr << "] plr=" << plr << std::endl;
  
  delete [] n_p;

  return stream_rate;
}


double prrt_socket::getReceivingRate()
{
  return m_avg_recv_rate * 8 * m_recv_payload_size;
}

void prrt_socket::setReceivingRate( double rate )
{
  m_avg_recv_rate = rate / 8 / m_recv_payload_size;
}


unsigned int prrt_socket::getClientBufferInfo()
{
  prrt_socket::SocketStatistics sock_stats;
  this->getStats(&sock_stats);   // FIXME thread-safety?

  return sock_stats.buffb;
}

void prrt_socket::setClientBufferInfo(unsigned int buf_info)
{
  this->set_param(param_BUF_fb(buf_info));
}


void prrt_socket::setDelayCCAlpha(double alpha)
{
  m_delay_cc.setAlpha(alpha);
}
double prrt_socket::getDelayCCAlpha()
{
  return m_delay_cc.getAlpha();
}
void prrt_socket::setDelayCCBeta(double beta)
{
  m_delay_cc.setBeta(beta);
}
double prrt_socket::getDelayCCBeta()
{
  return m_delay_cc.getBeta();
}
void prrt_socket::setDelayCCGamma(double gamma)
{
  m_delay_cc.setGamma(gamma);
}
double prrt_socket::getDelayCCGamma()
{
  return m_delay_cc.getGamma();
}

