/* 
 * (c) 2010 Saarland University, Telecommunications Lab 
 * (prrt@nt.uni-saarland.de)
 *
 * The contents of this file are subject to the Mozilla Public License
 + Version 1.1 (the "License"); you may not use this file except in
 + compliance with the License. You may obtain a copy of the License at
 + http://www.mozilla.org/MPL/
 + 
 + Software distributed under the License is distributed on an "AS IS"
 + basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 + License for the specific language governing rights and limitations
 + under the License.
 + 
 + The Original Code is "Predictably Reliable Real-time Transport (PRRT)".
 + 
 + The Initial Developer of the Original Code is Saarland University.
 + All Rights Reserved.
 + 
 + Contributor(s): Jochen Gruen (Saarland University), Manuel Gorius
 + (Saarland University), Yongtao Shuai (Saarland University).
 + 
 + Disclaimer: This Alpha version of PRRT is not supported and stable yet.
 + Do not use it in a production environment: it may crash your device, 
 + so please make sure that you have a backup of your data before
 + installing PRRT.
 */

#include "kernel_user_interface.h"
#include "socket_manager.h"
#include "kui_option.h"
#include "kui_control.h"
#include "kui_read.h"
#include "kui_register.h"
#include "kui_relay.h"
#include "kui_write.h"
#include "kui_debug.h"
#include "kui_param.h"
#include "kui_stats.h"
#include "../prrt_packet.h"
#include "kui_error.h"
#include "relay_manager.h"

#define PARAM_NO 4

static const char *DEBUG_NAME = "KUI";

/*from below*/
static void kui_create_socket_dir(Socketno_t socketno);

/**
 * Registers files for a socket with the procfs.
 * \param flags see sm_register_socket()
 * \param msock see sm_register_socket()
 * \return Number of the socket that was created
 */
Socketno_t kui_register_socket(int flags, struct prrt_socket *msock, int kui_port)
{
	Socketno_t socketno = 0;
	void *priv_datas[5+PARAM_NO];
	int len = 5+PARAM_NO;

	/*Save internal that this socket was created and get socketno*/
	while( (socketno = sm_register_socket(flags, msock, kui_port)) == -ERESTARTSYS)
		;

	if(socketno <= 0) {
		PDEBUG(MSG_COMMON, "ERROR %s: Could not create new socket\n",
			DEBUG_NAME);
		return -1;
	}

	/*create the directory for the socket*/
	kui_create_socket_dir(socketno);

	/*Register a control file with the procfs*/
	priv_datas[0] = kui_create_control_file(socketno);

	/*Register a write file with the procfs*/
	priv_datas[1] = (void *)kui_create_write_file(socketno);

	/*Register a read file with the procfs*/
	priv_datas[2] = (void *)kui_create_read_file(socketno);

	/*Register an option file with the procfs*/
	priv_datas[3] = kui_create_option_file(socketno);

	/*Register an stats file with the procfs*/
	priv_datas[4] = kui_create_stats_file(socketno);

	/*Register param file for D_target*/
	//priv_datas[5] = kui_create_param_file(socketno, D_TARGET);

	/*Register param file for PLR*/
	priv_datas[5] = kui_create_param_file(socketno, PLR);

  /*Register param file for D_FEC*/
	//priv_datas[7] = kui_create_param_file(socketno, D_FEC);

	/*Register param file for k,r,n_p,code_start*/
	//priv_datas[8] = kui_create_param_file(socketno, CODINGPARAS);

	/*Register param file for n_rr and n_cc*/
	//priv_datas[9] = kui_create_param_file(socketno, RETRROUNDS);

	/*Register param file for bw_est*/
	priv_datas[6] = kui_create_param_file(socketno, BWEST);

  /*Register param file for AHEC*/
  priv_datas[7] = kui_create_param_file(socketno, AHEC);

	/*Register param file for buf_fb*/
	priv_datas[8] = kui_create_param_file(socketno, BUFFB);

	/*Save the private data for the different files*/
	while(sm_set_socket_privdata(socketno, priv_datas, len) == -ERESTARTSYS)
		;

	return socketno;
}

void kui_unregister_socket(Socketno_t socketno)
{
	char name[NAME_SIZE];

	struct prrt_socket *msock = 0;
	sm_get_prrt_sock(socketno, &msock);
	if(msock == 0)
		return;

	PDEBUG(MSG_COMMON, "%s: Deleting files of socket no %u\n", DEBUG_NAME, socketno);

	/*remove param files from procfs*/
	kui_delete_param_file(socketno, BUFFB);
	kui_delete_param_file(socketno, BWEST);
	kui_delete_param_file(socketno, PLR);
  kui_delete_param_file(socketno, AHEC);

	PDEBUG(MSG_COMMON, "%s: Param files deleted\n", DEBUG_NAME);

	/*remove option file from procfs*/
	kui_delete_stats_file(socketno);

	/*remove option file from procfs*/
	kui_delete_option_file(socketno);

	PDEBUG(MSG_COMMON, "%s: Option file deleted\n", DEBUG_NAME);

	/*remove read file from procfs*/
	kui_delete_read_file(socketno);

	PDEBUG(MSG_COMMON, "%s: Read file deleted\n", DEBUG_NAME);

	/*remove write file from procfs*/
	kui_delete_write_file(socketno);

	PDEBUG(MSG_COMMON, "%s: Write file deleted\n", DEBUG_NAME);

	/*remove control file from procfs*/
	kui_delete_control_file(socketno);

	PDEBUG(MSG_COMMON, "%s: Control file deleted\n", DEBUG_NAME);

	/*remove socket directory from procfs*/
	snprintf(name, NAME_SIZE, "prrt/%u", socketno);
	remove_proc_entry(name, NULL);

	PDEBUG(MSG_FINAL, "%s: Unregistering socket no %u\n", DEBUG_NAME, socketno);

	/*unregister socket internally*/
	while(sm_unregister_socket(socketno) == -ERESTARTSYS)
		;
}

Relayno_t kui_relay_sockets(Socketno_t *socketno_in, Socketno_t *socketno_out, char *relayname){
    struct prrt_socket *msock_in = 0;
    struct prrt_socket *msock_out = 0;
    struct prrt_relay *mrelay = 0;
    Relayno_t relayno = 0;
    int err = 0;

    sm_get_prrt_sock(*socketno_in, &msock_in);
    sm_get_prrt_sock(*socketno_out, &msock_out);

    if(msock_in == 0)
        *socketno_in = 0;

    if(msock_out == 0)
        *socketno_out = 0;

    if(msock_in == 0 || msock_out == 0) {
        return -ENOSOCK;
    }

    relayno = rm_register_relay(relayname);
    if(relayno < 0)
	    return relayno;

    err = rm_get_prrt_relay(relayno, &mrelay);
    if(err < 0)
	    return err;

    mrelay->msock_in = msock_in;
    mrelay->msock_out = msock_out;

    prrt_start_relay(mrelay);

    return relayno;
}

void kui_unrelay_sockets(Relayno_t relayno)
{
	struct prrt_relay *mrelay = 0;
	rm_get_prrt_relay(relayno, &mrelay);
	if(mrelay == 0)
		return;

	/*unregister relay internally*/
	while(rm_unregister_relay(relayno) == -ERESTARTSYS)
		;
}

void kui_init_proc_interface(void)
{
	struct proc_dir_entry *procentry;
	int file_created = 0;
	int err = 0;

	/*create directory for the module under /proc*/
	procentry = proc_mkdir("prrt", NULL);
	if(!procentry) {
		/*ERROR handling*/;
		PDEBUG(MSG_COMMON, "%s: could not create /proc/prrt\n", DEBUG_NAME);
	} else {
		PDEBUG(MSG_COMMON, "%s: /proc/prrt created\n", DEBUG_NAME);
	}

	/*create procfs file for registering sockets*/
	file_created = kui_create_register_file();
	if(!file_created) {
		/*ERROR handling*/;
		PDEBUG(MSG_COMMON, "%s: could not create /proc/prrt/register\n", DEBUG_NAME);
	} else {
		PDEBUG(MSG_COMMON, "%s: /proc/prrt/register created\n", DEBUG_NAME);
	}

        /*create procfs file for relay information*/
	file_created = kui_create_relay_file();
	if(!file_created) {
		/*ERROR handling*/;
		PDEBUG(MSG_COMMON, "%s: could not create /proc/prrt/relays\n", DEBUG_NAME);
	} else {
		PDEBUG(MSG_COMMON, "%s: /proc/prrt/relays created\n", DEBUG_NAME);
	}

	/*initialise socket_manager*/
	sm_init();
	rm_init();

	/*initialise packet cache*/
	/*TODO: handle error*/
	err = prrt_entry_cache_init();
	if(err < 0)
		PERROR("%s: Could not initialise entry cache", DEBUG_NAME);
	err = prrt_packet_cache_init();
	if(err < 0)
		PERROR("%s: Could not initialise packet cache", DEBUG_NAME);
	err = prrt_fbpacket_cache_init();
	if(err < 0)
		PERROR("%s: Could not initialise fbpacket cache", DEBUG_NAME);
	err = prrt_block_cache_init();
	if(err < 0)
		PERROR("%s: Could not initialise block cache", DEBUG_NAME);
	err = prrt_fbana_caches_init();
	if(err < 0)
		PERROR("%s: Could not initialise fbana caches", DEBUG_NAME);
}

void kui_uninit_proc_interface(void)
{
	Socketno_t socketno = 0;
  Relayno_t relayno = 0;

	PDEBUG(MSG_COMMON, "%s: Deleting prrt/relay\n", DEBUG_NAME);
	kui_delete_relay_file();
  PDEBUG(MSG_COMMON, "%s: Deleting prrt/register\n", DEBUG_NAME);
	kui_delete_register_file();

	PDEBUG(MSG_COMMON, "%s: Removing sockets\n", DEBUG_NAME);

	/*remove all open sockets*/
	while((socketno = sm_get_socketno()) == -ERESTARTSYS ||
		socketno > 0) {
		if(socketno > 0)
			kui_unregister_socket(socketno);
	}

  /*remove all open relays*/
	while((relayno = rm_get_relayno()) == -ERESTARTSYS ||
		relayno > 0) {
		if(relayno > 0)
			kui_unrelay_sockets(relayno);
	}

	sm_uninit();
	rm_uninit();

	prrt_entry_cache_uninit();
	prrt_packet_cache_uninit();
	prrt_fbpacket_cache_uninit();
	prrt_block_cache_uninit();
	prrt_fbana_caches_uninit();

	PDEBUG(MSG_COMMON, "%s: Deleting prrt\n", DEBUG_NAME);
	remove_proc_entry("prrt", NULL);
}

static void kui_create_socket_dir(Socketno_t socketno)
{
	char name[NAME_SIZE];
	struct proc_dir_entry *procentry;

	snprintf(name, NAME_SIZE, "prrt/%u", socketno);

	/*create directory for the socket under /proc/prrt */
	procentry = proc_mkdir(name, NULL);
	if(!procentry) {
		/*ERROR handling*/;
		PDEBUG(MSG_COMMON, "%s: could not create /proc/prrt/%u\n", DEBUG_NAME, socketno);
	} else {
		PDEBUG(MSG_COMMON, "%s: /proc/prrt/%u created\n", DEBUG_NAME, socketno);
	}
}

