/* 
 * (c) 2010 Saarland University, Telecommunications Lab 
 * (prrt@nt.uni-saarland.de)
 *
 * The contents of this file are subject to the Mozilla Public License
 + Version 1.1 (the "License"); you may not use this file except in
 + compliance with the License. You may obtain a copy of the License at
 + http://www.mozilla.org/MPL/
 + 
 + Software distributed under the License is distributed on an "AS IS"
 + basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 + License for the specific language governing rights and limitations
 + under the License.
 + 
 + The Original Code is "Predictably Reliable Real-time Transport (PRRT)".
 + 
 + The Initial Developer of the Original Code is Saarland University.
 + All Rights Reserved.
 + 
 + Contributor(s): Jochen Gruen (Saarland University), Manuel Gorius
 + (Saarland University), Yongtao Shuai (Saarland University).
 + 
 + Disclaimer: This Alpha version of PRRT is not supported and stable yet.
 + Do not use it in a production environment: it may crash your device, 
 + so please make sure that you have a backup of your data before
 + installing PRRT.
 */

#include "kernel_user_interface.h"

#include <linux/list.h>
#include "../lists/prrt_list.h"

static const char *DEBUG_NAME = "KUI_DEBUG";

/*type for different events on this socket type (only one atm but keeps it extensible)*/
enum debug_event_t {
	NO_EVENT,		/*this describes that no event has happened yet*/
	PRINT_SOCKETS,		/*print out all registered sockets*/
	PRINT_PARAMS,		/*print out all params of a specific socket*/
	PRINT_INCQUEUE,		/*print out packets in incoming packet queue*/
	PRINT_OUTQUEUE		/*print out packets in outgoing packet queue*/
};

struct debug_event {
	struct semaphore 	mutex;		/*mutex to lock this shared variable*/
	unsigned char		sem_init;	/*is the semaphore already initialised*/
	enum debug_event_t 	type;
	Socketno_t		socketno;
};

static struct debug_event last_event;

/*from socket_manager.c*/
int sm_get_socket_info(char **sinfo);
int sm_get_prrt_sock(Socketno_t socketno, struct prrt_socket **msock);

/*from below*/
static int kui_debug_read_proc(char *buf, char **start, off_t offset,
				int count, int *eof, void *data);
static int kui_debug_write_proc(struct file *file, const char __user *buffer,
				unsigned long count, void *data);

/*
 * This function creates a new file in the procfs for debugging.
 */
int kui_create_debug_file(void)
{
	struct proc_dir_entry *procentry;

	if(!last_event.sem_init) {
		init_MUTEX(&last_event.mutex);
		last_event.sem_init = 1;
	}

	procentry = create_proc_entry( "prrt/debug", S_IRUGO|S_IWUGO , NULL );
	if( procentry ) {
		procentry->read_proc  = kui_debug_read_proc;
		procentry->write_proc = kui_debug_write_proc;
		procentry->data = NULL;

		return 1;	/*SUCCESS*/
	}

	PDEBUG(MSG_COMMON, "%s: failure when creating prrt/debug\n", DEBUG_NAME);

	return 0;	/*FAILURE*/
}

/*
 * Function for deleting a procfs debug file.
 */
void kui_delete_debug_file(void)
{
	remove_proc_entry("prrt/debug", NULL);
}

/*
 * Read method for a debug file.
 * Gives feedback when a user process tries to gather
 * debug information.
 */
static int kui_debug_read_proc(char *buf, char **start, off_t offset,
				int count, int *eof, void *data)
{
	int len=0;
	int ret = 0;
	char *info = NULL;
	struct prrt_socket *msock = NULL;

	int interrupted = down_interruptible(&last_event.mutex);
	if(interrupted)
		return -ERESTARTSYS;

	switch(last_event.type) {
	case NO_EVENT:
		len = snprintf(buf, count, "-1\n");
		if(len < 0 || len > count) {
			PDEBUG(MSG_COMMON, "%s ERROR: could not write to proc"
				"file\n", DEBUG_NAME);
			goto free_mutex;
		}
		break;
	case PRINT_SOCKETS:
		while( (len = sm_get_socket_info(&info)) == -ERESTARTSYS )
			;
		if(info == NULL || len <= 0 || len >= count) {
			PDEBUG(MSG_COMMON, "%s ERROR: error when retrieving"
				"socket information\n", DEBUG_NAME);
			goto free_mutex;
		}
		len = strlcpy(buf, info, len);
		PDEBUG(MSG_COMMON, "%s: Feedback sent\n", DEBUG_NAME);
		last_event.type = NO_EVENT;	/* this assures for one call
						 * that feedback is only sent once*/
		break;
	case PRINT_PARAMS:
		if(sm_get_prrt_sock(last_event.socketno, &msock) == -ERESTARTSYS)
			goto free_mutex;
		if(!msock) {
			/*ERROR: socket not found*/
			len = strlcpy(buf, "-1", 2);
			goto out;
		}
		info = kmalloc(20, GFP_KERNEL);
// 		snprintf(info, 20, "%d\n", msock->test);
		len = strlcpy(buf, info, 20);
		buf[len] = '\0';
		break;
	case PRINT_INCQUEUE:
		if(sm_get_prrt_sock(last_event.socketno, &msock) == -ERESTARTSYS)
			goto free_mutex;
		if(!msock) {
			/*ERROR: socket not found*/
			len = strlcpy(buf, "-1", 2);
			goto out;
		}
		//prrtlist_print_packetdata(&msock->inc_packet_list.data_list);
		break;
	case PRINT_OUTQUEUE:
		if(sm_get_prrt_sock(last_event.socketno, &msock) == -ERESTARTSYS)
			goto free_mutex;
		if(!msock) {
			/*ERROR: socket not found*/
			len = strlcpy(buf, "-1", 2);
			goto out;
		}
		//prrtlist_print_packetdata(&msock->out_packet_list.list);
		break;
		
	default:
		/* Error handling*/
		break;
	}

out:
	*eof = 1;
	ret = len;

free_mutex:
	up(&last_event.mutex);
	kfree(info);

	return ret;
}

/*
 * Write method for a debug file.
 * Here the input from the user process is analysed and the proper
 * action is done.
 */
static int kui_debug_write_proc(struct file *file, const char __user *buffer,
				unsigned long count, void *data)
{
	char kbuffer[count+1];
	Socketno_t socketno = 0;

	int interrupted = down_interruptible(&last_event.mutex);
	if(interrupted)
		return -ERESTARTSYS;

	copy_from_user(kbuffer, buffer, count);
	kbuffer[count] = '\0';

	PDEBUG(MSG_COMMON, "%s: Parsing input(%lu bytes):\n%s\n", DEBUG_NAME,
		count, kbuffer);

	if(strncmp("print sockets", kbuffer, 13) == 0) {
		last_event.type = PRINT_SOCKETS;
		PDEBUG(MSG_COMMON, "%s: Sockets should be printed\n", DEBUG_NAME);
	} else if(strncmp("print params", kbuffer, 12) == 0) {
		last_event.type = PRINT_PARAMS;
		sscanf(&kbuffer[13], "%d", &last_event.socketno);
		PDEBUG(MSG_COMMON, "%s: Socket params of socket %d should be"
"				printed\n", DEBUG_NAME, socketno);
	} else if(strncmp("print incqueue", kbuffer, 14) == 0) {
		last_event.type = PRINT_INCQUEUE;
		sscanf(&kbuffer[15], "%d", &last_event.socketno);
		PDEBUG(MSG_COMMON, "%s: Inc queue of socket %d should be"
"				printed\n", DEBUG_NAME, socketno);
	} else if(strncmp("print outqueue", kbuffer, 14) == 0) {
		last_event.type = PRINT_OUTQUEUE;
		sscanf(&kbuffer[15], "%d", &last_event.socketno);
		PDEBUG(MSG_COMMON, "%s: Out queue of socket %d should be"
"				printed\n", DEBUG_NAME, socketno);
	} else {
		last_event.type = NO_EVENT;
		PDEBUG(MSG_COMMON, "%s: Illegal input\n", DEBUG_NAME);
	}

	up(&last_event.mutex);
	return count;
}

