/* 
 * (c) 2010 Saarland University, Telecommunications Lab 
 * (prrt@nt.uni-saarland.de)
 *
 * The contents of this file are subject to the Mozilla Public License
 + Version 1.1 (the "License"); you may not use this file except in
 + compliance with the License. You may obtain a copy of the License at
 + http://www.mozilla.org/MPL/
 + 
 + Software distributed under the License is distributed on an "AS IS"
 + basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 + License for the specific language governing rights and limitations
 + under the License.
 + 
 + The Original Code is "Predictably Reliable Real-time Transport (PRRT)".
 + 
 + The Initial Developer of the Original Code is Saarland University.
 + All Rights Reserved.
 + 
 + Contributor(s): Jochen Gruen (Saarland University), Manuel Gorius
 + (Saarland University), Yongtao Shuai (Saarland University).
 + 
 + Disclaimer: This Alpha version of PRRT is not supported and stable yet.
 + Do not use it in a production environment: it may crash your device, 
 + so please make sure that you have a backup of your data before
 + installing PRRT.
 */

#include "kernel_user_interface.h"
#include "kui_error.h"

static const char *DEBUG_NAME = "KUI_REGISTER";

/*type for different events on this socket type*/
enum register_event_t {
    NO_EVENT, /*this describes that no event has happened yet*/
    SOCKET_REGISTERED, /*this event happens when a socket is registered*/
    SOCKET_DELETED, /*this event is for deleting a socket*/
    RELAY_CREATED, /*this event is for relaying data from a socket to another socket*/
    RELAY_DELETED /*this event is for the deletion of a relay*/
};

struct register_event {
    struct semaphore mutex; /*mutex to lock this shared variable*/
    unsigned char sem_init; /*is the semaphore already initialised*/
    enum register_event_t type;
    Socketno_t socketno; /*this is the number of a socket*/
    Socketno_t socketno2; /*this is the number of a 2nd socket*/
    Relayno_t relayno; /*this is the number of the relay*/
};

static struct register_event last_event;

/*from kernel_user_interface.c*/
Socketno_t kui_register_socket(int flags, struct prrt_socket* msock, int kui_port);
void kui_unregister_socket(Socketno_t);
Relayno_t kui_relay_sockets(Socketno_t*, Socketno_t*, char*);
void kui_unrelay_sockets(Relayno_t relayno);

/*from below*/
static int kui_register_read_proc(char *buf, char **start, off_t offset,
        int count, int *eof, void *data);
static int kui_register_write_proc(struct file *file, const char __user *buffer,
        unsigned long count, void *data);

/*
 * This function creates a new file in the procfs for registering sockets.
 */
int kui_create_register_file(void) {
    struct proc_dir_entry *procentry;

    if (!last_event.sem_init) {
        init_MUTEX(&last_event.mutex);
        last_event.sem_init = 1;
        PDEBUG(MSG_COMMON, "%s: semaphore initialised\n", DEBUG_NAME);
    }

    procentry = create_proc_entry("prrt/register", S_IRUGO | S_IWUGO, NULL);
    if (procentry) {
        procentry->read_proc = kui_register_read_proc;
        procentry->write_proc = kui_register_write_proc;
        procentry->data = NULL;

        return 1; /*SUCCESS*/
    }

    PERROR("%s: failure when creating prrt/register\n", DEBUG_NAME);

    return 0; /*FAILURE*/
}

/*
 * Function for deleting a procfs register file.
 */
void kui_delete_register_file(void) {
    remove_proc_entry("prrt/register", NULL);
}

/*
 * Read method for a register file.
 * Gives feedback when a user process tries to creat a
 * a socket.
 */
static int kui_register_read_proc(char *buf, char **start, off_t offset,
        int count, int *eof, void *data) {
    int len = 0;
    int ret = 0;
    int interrupted = down_interruptible(&last_event.mutex);
    if (interrupted)
        return -ERESTARTSYS;

    switch (last_event.type) {
        case NO_EVENT:
            len = snprintf(buf, PAGE_SIZE, "-1a");
            break;
        case SOCKET_REGISTERED:
            len = snprintf(buf, PAGE_SIZE, "%da",
                    last_event.socketno);
            break;
        case SOCKET_DELETED:
            len = snprintf(buf, PAGE_SIZE, "%da",
                    last_event.socketno);
            break;
        case RELAY_CREATED:
            len = snprintf(buf, PAGE_SIZE, "%da %da %da",
                    last_event.relayno,
                    last_event.socketno,
                    last_event.socketno2
                    );
            break;
        case RELAY_DELETED:
            len = snprintf(buf, PAGE_SIZE, "%da",
                    last_event.relayno);
            break;
        default:
            /* Error handling*/
            break;
    }

    if (len < 0 || len > PAGE_SIZE) {
        PERROR("%s ERROR: Could not write to procfile\n", DEBUG_NAME);
        goto free_mutex;
    }

    last_event.type = NO_EVENT; /* this assures for a certain
					 * socket that feedback is only
					 * sent once*/

    *eof = 1;
    ret = len;

free_mutex:
    up(&last_event.mutex);

    return ret;
}

/*
 * Write method for a register file.
 * Here the input from the user process is analysed and the proper
 * action is done.
 */
static int kui_register_write_proc(struct file *file, const char __user *buffer,
        unsigned long count, void *data) {
    char kbuffer[count + 1];
    int args = 0;
    char command[20];
    char relayname[RELAYNAMELENGTH+1];
    unsigned long socketno = 0;
    unsigned long socketno2 = 0;
    int err = count;

    int interrupted = down_interruptible(&last_event.mutex);
    if (interrupted) {
        // 		last_event.error = -EINTERR;
        err = -ERESTARTSYS;
        goto out;
    }

    copy_from_user(kbuffer, buffer, count);
    kbuffer[count] = '\0';

    //PDEBUG(MSG_COMMON, "%s: buffer - %s\n", DEBUG_NAME, kbuffer);

    args = sscanf(kbuffer, "%s %lu %lu %s", command, &socketno, &socketno2, relayname);

    if (args == 2 && strncmp("create", command, 6) == 0) {
        last_event.socketno = kui_register_socket(0, 0, (int)socketno);
        last_event.type = SOCKET_REGISTERED;
        PDEBUG(MSG_FINAL, "%s: Socket no. %d created with kui_port %d\n", DEBUG_NAME, last_event.socketno, (int)socketno);
        PDEBUG(MSG_COMMON, "%s: Socket no. %d created with kui_port %d\n", DEBUG_NAME, last_event.socketno, (int)socketno);
    } else if (args == 2 && strncmp("delete", command, 6) == 0) {
        /*TODO validate socket number*/
        last_event.socketno = (Socketno_t) socketno;
        last_event.type = SOCKET_DELETED;
        kui_unregister_socket(last_event.socketno);
        PDEBUG(MSG_COMMON, "%s: Socket no. %d deleted\n", DEBUG_NAME, last_event.socketno);
    } else if (args == 4 && strncmp("relay", command, 5) == 0) {
        /*TODO validate socket number*/
        last_event.socketno = (Socketno_t) socketno;
        last_event.socketno2 = (Socketno_t) socketno2;
        last_event.type = RELAY_CREATED;
        PDEBUG(MSG_COMMON, "%s: Trying to create relay with sockets %lu, %lu\n", DEBUG_NAME, socketno, socketno2);
        last_event.relayno = kui_relay_sockets(&last_event.socketno, &last_event.socketno2, relayname);
        if (last_event.relayno > 0) {
            PDEBUG(MSG_COMMON, "%s: Relay no. %d created\n", DEBUG_NAME, last_event.relayno);
        } else {
            PDEBUG(MSG_COMMON, "%s: Error no %d when creating relay\n", DEBUG_NAME, last_event.relayno);
        }
    } else if (args == 2 && strncmp("unrelay", command, 7) == 0) {
        /*TODO validate socket number*/
        last_event.relayno = (Relayno_t) socketno;
        last_event.type = RELAY_DELETED;
        kui_unrelay_sockets(last_event.relayno);
        PDEBUG(MSG_COMMON, "%s: Relay no. %d deleted\n", DEBUG_NAME, last_event.relayno);
    } else {
        last_event.type = NO_EVENT;
        PERROR("%s: Illegal input: %s\n", DEBUG_NAME, kbuffer);
    }

    // done:
    up(&last_event.mutex);
out:
    return err;

    // nomem:
    up(&last_event.mutex);
    return -ENOMEM;
}
