/* 
 * (c) 2010 Saarland University, Telecommunications Lab 
 * (prrt@nt.uni-saarland.de)
 *
 * The contents of this file are subject to the Mozilla Public License
 + Version 1.1 (the "License"); you may not use this file except in
 + compliance with the License. You may obtain a copy of the License at
 + http://www.mozilla.org/MPL/
 + 
 + Software distributed under the License is distributed on an "AS IS"
 + basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 + License for the specific language governing rights and limitations
 + under the License.
 + 
 + The Original Code is "Predictably Reliable Real-time Transport (PRRT)".
 + 
 + The Initial Developer of the Original Code is Saarland University.
 + All Rights Reserved.
 + 
 + Contributor(s): Jochen Gruen (Saarland University), Manuel Gorius
 + (Saarland University), Yongtao Shuai (Saarland University).
 + 
 + Disclaimer: This Alpha version of PRRT is not supported and stable yet.
 + Do not use it in a production environment: it may crash your device, 
 + so please make sure that you have a backup of your data before
 + installing PRRT.
 */

#include "kernel_user_interface.h"
#include "kui_error.h"

#define DEBUG_NAME "KUI_STATS"

struct stats_priv_data {
	Socketno_t		socketno;	/*socketno of the stats file*/
	struct semaphore	mutex;		/*mutex to synchronize access
						 *to one file*/
};

/*from socket_manager.c*/
int sm_get_prrt_sock(Socketno_t socketno, struct prrt_socket **msock);

/*from below*/
int kui_stats_read_proc(char *buf, char **start, off_t offset,
				int count, int *eof, void *data);

/*
 * This function creates a new file in the procfs for stats of a socket.
 */
void *kui_create_stats_file(Socketno_t socketno)
{
	char name[NAME_SIZE];
	struct proc_dir_entry *procentry = 0;

	struct stats_priv_data *priv_data = 0;

	priv_data = kmalloc(sizeof(*priv_data), GFP_KERNEL);

	priv_data->socketno = socketno;
	init_MUTEX(&(priv_data->mutex));

	snprintf(name, NAME_SIZE, "prrt/%d/stats", socketno);

	procentry = create_proc_entry(name, S_IRUGO, NULL);
	if(procentry) {
		procentry->read_proc  = kui_stats_read_proc;
		procentry->write_proc = 0;
		procentry->data = priv_data;

		return priv_data;	/*SUCCESS*/
	}

	return 0;	/*FAILURE*/
}

/*
 * Function for deleting a procfs stats file.
 */
void kui_delete_stats_file(Socketno_t socketno)
{
	char name[NAME_SIZE];
	struct prrt_socket *msock = 0;
	struct stats_priv_data *pd;

	snprintf(name, NAME_SIZE, "prrt/%d/stats", socketno);

	while(sm_get_prrt_sock(socketno, &msock) == -ERESTARTSYS)
		;
	if(!msock)
		return;

	pd = (struct stats_priv_data *)msock->stats_pd;

	while(down_interruptible(&pd->mutex))
		;

	/*free private data of file*/
	kfree(msock->stats_pd);

	remove_proc_entry(name, NULL);

	up(&pd->mutex);
}

static void BuildStatsString(struct prrt_socket *msock, char *buffer);

/*
 * Read method for a stats file.
 */
int kui_stats_read_proc(char *buf, char **start, off_t offset,
				int count, int *eof, void *data)
{
	int len = 0;
	int ret = 0;
	struct stats_priv_data *pd = (struct stats_priv_data *)data;
	struct prrt_socket *msock = 0;
	// char buffer[PAGE_SIZE];
  // Solve warning of larger than frame size 1024 bytes.
  char *buffer = 0;

	int interrupted = 0;

  buffer = (char *) kmalloc(PAGE_SIZE, GFP_KERNEL);

  interrupted = down_interruptible(&pd->mutex);
	if(interrupted)
		return -ERESTARTSYS;

	sm_get_prrt_sock(pd->socketno, &msock);
	if(!msock)
		goto free_mutex;

	BuildStatsString(msock, buffer);

	len = snprintf(buf, PAGE_SIZE, "%s", buffer);

	if(len < 0 || len > PAGE_SIZE) {
		PERROR("%s: Could not write to proc file\n", DEBUG_NAME);
		goto free_mutex;
	}

	*eof = 1;
	ret = len;

free_mutex:
  kfree(buffer);
	up(&pd->mutex);

	return ret;
}

static void BuildStatsString(struct prrt_socket *msock, char *buffer) {
	int charsWritten = 0;
	//char loss_pattern[MAX_MEASURES];	/**<This array reflects the pattern of lost packets.*/
  char *loss_pattern = 0;
  loss_pattern = (char *) kmalloc(MAX_MEASURES_PLR, GFP_KERNEL);
	prrtnpar_get_LossPattern(&msock->inc_packet_list.net_params, loss_pattern);
	charsWritten += sprintf(buffer+charsWritten, "Sent data: %lu\n", msock->out_packet_list.send_replacements);
	charsWritten += sprintf(buffer+charsWritten, "Sent redundancy: %lu\n", msock->out_packet_list.send_packets);
	//charsWritten += sprintf(buffer+charsWritten, "Not recovered packets: %lu\n", 
	//			msock->inc_packet_list.missed_packets-msock->inc_packet_list.old_packets);
	charsWritten += sprintf(buffer+charsWritten, "Lost packets: %lu\n", 
				msock->inc_packet_list.lost_packets);
	charsWritten += sprintf(buffer+charsWritten, "Dropped packets: %lu\n", msock->inc_packet_list.old_packets);
	//Residual loss
	charsWritten += sprintf(buffer+charsWritten, "Missed packets: %lu\n", msock->inc_packet_list.missed_packets);
	charsWritten += sprintf(buffer+charsWritten, "Received data: %lu\n", msock->inc_packet_list.received_datapackets);
	charsWritten += sprintf(buffer+charsWritten, "Received redundancy: %lu\n", msock->inc_packet_list.received_redpackets);
//	charsWritten += sprintf(buffer+charsWritten, "Last replace time: %d\n", atomic_read(&msock->inc_packet_list.lastMeasured_t_replace));
//	charsWritten += sprintf(buffer+charsWritten, "Last decoding delay: %d\n", atomic_read(&msock->inc_packet_list.lastDecodingDelay));
//	charsWritten += sprintf(buffer+charsWritten, "Last RTT: %d\n", prrtnpar_get_ReplaceTime(&msock->inc_packet_list.net_params));
	charsWritten += sprintf(buffer+charsWritten, "Last GRTT: %d\n", prrtnpar_get_ReplaceTime(&msock->net_params));
	charsWritten += sprintf(buffer+charsWritten, "Last FTT: %d\n", prrtnpar_get_FTT(&msock->inc_packet_list.net_params));
	charsWritten += sprintf(buffer+charsWritten, "Last GFTT: %d\n", prrtnpar_get_FTT(&msock->net_params));
	charsWritten += sprintf(buffer+charsWritten, "Queue Delay: %d\n", prrtnpar_get_QueueDelay(&msock->inc_packet_list.net_params));
	prrtnpar_update_PLR(&msock->inc_packet_list.net_params);
	charsWritten += sprintf(buffer+charsWritten, "Last PLR: %d\n", prrtnpar_get_PLR(&msock->inc_packet_list.net_params));
	charsWritten += sprintf(buffer+charsWritten, "Last GPLR: %d\n", prrtnpar_get_PLR(&msock->net_params));
	charsWritten += sprintf(buffer+charsWritten, "Num Gaps: %d\n", prrtnpar_get_NumGaps(&msock->inc_packet_list.net_params));
	charsWritten += sprintf(buffer+charsWritten, "Num Bursts: %d\n", prrtnpar_get_NumBursts(&msock->inc_packet_list.net_params));
	charsWritten += sprintf(buffer+charsWritten, "Gap Length: %d\n", prrtnpar_get_GapLength(&msock->inc_packet_list.net_params));
	charsWritten += sprintf(buffer+charsWritten, "Burst Length: %d\n", prrtnpar_get_BurstLength(&msock->inc_packet_list.net_params));
	charsWritten += sprintf(buffer+charsWritten, "Num GGaps: %d\n", prrtnpar_get_NumGaps(&msock->net_params));
	charsWritten += sprintf(buffer+charsWritten, "Num GBursts: %d\n", prrtnpar_get_NumBursts(&msock->net_params));
	charsWritten += sprintf(buffer+charsWritten, "Gap GLength: %d\n", prrtnpar_get_GapLength(&msock->net_params));
	charsWritten += sprintf(buffer+charsWritten, "Burst GLength: %d\n", prrtnpar_get_BurstLength(&msock->net_params));
	charsWritten += sprintf(buffer+charsWritten, "Packet Interval: %d\n", prrtnpar_get_PacketInterval(&msock->inc_packet_list.net_params));
	//charsWritten += sprintf(buffer+charsWritten, "BW Estimate: %d\n", prrtlpar_get_BW_est(&msock->loc_params_out));
	charsWritten += sprintf(buffer+charsWritten, "BW Estimate: %d\n", prrtcpar_get_BW_est(&msock->cod_params));
	charsWritten += sprintf(buffer+charsWritten, "BUF feedback: %d\n", prrtlpar_get_BUF_fb(&msock->loc_params_in));
	charsWritten += sprintf(buffer+charsWritten, "Loss Pattern: %s\n", loss_pattern);
//	charsWritten += sprintf(buffer+charsWritten, "Last Alpha: %d\n", prrtnpar_get_Alpha(&msock->inc_packet_list.net_params));
//	charsWritten += sprintf(buffer+charsWritten, "Last Beta: %d\n", prrtnpar_get_Beta(&msock->inc_packet_list.net_params));

  kfree(loss_pattern);
}

