/* 
 * (c) 2010 Saarland University, Telecommunications Lab 
 * (prrt@nt.uni-saarland.de)
 *
 * The contents of this file are subject to the Mozilla Public License
 + Version 1.1 (the "License"); you may not use this file except in
 + compliance with the License. You may obtain a copy of the License at
 + http://www.mozilla.org/MPL/
 + 
 + Software distributed under the License is distributed on an "AS IS"
 + basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 + License for the specific language governing rights and limitations
 + under the License.
 + 
 + The Original Code is "Predictably Reliable Real-time Transport (PRRT)".
 + 
 + The Initial Developer of the Original Code is Saarland University.
 + All Rights Reserved.
 + 
 + Contributor(s): Jochen Gruen (Saarland University), Manuel Gorius
 + (Saarland University), Yongtao Shuai (Saarland University).
 + 
 + Disclaimer: This Alpha version of PRRT is not supported and stable yet.
 + Do not use it in a production environment: it may crash your device, 
 + so please make sure that you have a backup of your data before
 + installing PRRT.
 */

#include "prrt_block_list.h"

#define DEBUG_NAME "PRRT_BLOCK_LIST"

void prrtblist_init(struct prrt_block_list *blist)
{
	INIT_LIST_HEAD(&blist->head);

	blist->next_packet = 0;
	blist->next_packet_block = &blist->head;
	blist->next_packet_no = -1;
	init_MUTEX_LOCKED(&blist->packet_count);
	PDEBUG(MSG_OUTLIST, "%s: Initialised count, now %u left\n", DEBUG_NAME,
		blist->packet_count.count);

	atomic_set(&blist->block_count, 0);
}

int prrtblist_reinit(struct prrt_block_list *blist)
{
	int err = 0;
	/*delete all packets in queues*/
	prrtblist_destroy_all_blocks(blist);

	blist->next_packet = 0;
	blist->next_packet_block = &blist->head;
	blist->next_packet_no = -1;
	init_MUTEX_LOCKED(&blist->packet_count);
	PDEBUG(MSG_OUTLIST, "%s: Initialised count, now %u left\n", DEBUG_NAME,
		blist->packet_count.count);

	atomic_set(&blist->block_count, 0);

	return err;
}

void prrtblist_destroy_all_blocks(struct prrt_block_list *blist)
{
	struct list_head *iter = 0,
			 *next = 0;
	struct prrt_block *mblock = 0;

	list_for_each_safe(iter, next, &blist->head) {
		list_del(iter);
		mblock = list_entry(iter, struct prrt_block, list);

		prrt_block_destroy(&mblock);

		atomic_dec(&blist->block_count);
	}
}


void ResetNextPacketForNewlyInsertedBlock(struct prrt_block_list *blist, struct prrt_block *mblock);

int prrtblist_insert_block(struct prrt_block_list *blist, struct prrt_block *mblock)
{
	if(!mblock)
		return -1;

	PDEBUG(MSG_BLOCKLIST, "%s: Inserting block(%p) in list\n", DEBUG_NAME, mblock);
	list_add_tail(&mblock->list, &blist->head);

	ResetNextPacketForNewlyInsertedBlock(blist, mblock);

	atomic_inc(&blist->block_count);

	return 0;
}

static int NextPacketInNewBlock(struct prrt_block_list *blist, struct packet_list_entry *first_entry);
static void ResetNextPacketIfFirstInsertedPacket(struct prrt_block_list *blist, struct prrt_block *mblock,
	  					 struct packet_list_entry *first_entry);

void ResetNextPacketForNewlyInsertedBlock(struct prrt_block_list *blist, struct prrt_block *mblock)
{
	int err = 0;
	struct packet_list_entry *first_entry = 0;

	err = prrt_block_get_first_data(mblock, &first_entry);
	if(err < 0)
		return;

	if(NextPacketInNewBlock(blist, first_entry)) {
		blist->next_packet_block = &mblock->list;
	}

	ResetNextPacketIfFirstInsertedPacket(blist, mblock, first_entry);
}

static int NextPacketInNewBlock(struct prrt_block_list *blist, struct packet_list_entry *first_entry)
{
	int first_pno = 0;

	first_pno = first_entry->pcont->pack.data.header.seqnum;

	if(prrt_comp_packetnos(first_pno, blist->next_packet_no) <= 0)
		return 1;

	return 0;
}

void SetNextPacket(struct prrt_block_list *blist, struct prrt_block *mblock,
				   struct packet_list_entry *p);

static void ResetNextPacketIfFirstInsertedPacket(struct prrt_block_list *blist, struct prrt_block *mblock,
	  					 struct packet_list_entry *first_entry)
{
	if(blist->next_packet_no == -1) {
		SetNextPacket(blist, mblock, first_entry);
	}
}

void SetNextPacket(struct prrt_block_list *blist, struct prrt_block *mblock, struct packet_list_entry *p)
{
	blist->next_packet = &p->list;
	blist->next_packet_block = &mblock->list;
	blist->next_packet_no = p->pcont->pack.data.header.seqnum;
}

int LastBlockAlreadyCoded(struct prrt_block_list *blist);

int prrtblist_can_insert_packet(struct prrt_block_list *blist)
{
	if(list_empty(&blist->head)) {
		PERROR("%s: Block list is empty, cannot insert new packet\n", DEBUG_NAME);
		return 0;
	}

	if(LastBlockAlreadyCoded(blist)) {
		PERROR("%s: Last block is already coded, may not insert packet\n", DEBUG_NAME);
		return 0;
	}

	return 1;
}

int LastBlockAlreadyCoded(struct prrt_block_list *blist)
{
	struct list_head *iter = 0;
	struct prrt_block *last_block = 0;

	iter = blist->head.prev;

	last_block = container_of(iter, struct prrt_block, list);

	if(prrt_block_is_already_coded(last_block)) {
		return 1;
	}

	return 0;
}

//int InsertDataPacketInLastBlock(struct prrt_block_list *blist, struct prrt_packet_container *pcont, struct params_code *cpars, struct params_local *lpars);
int InsertDataPacketInLastBlock(struct prrt_block_list *blist, struct prrt_packet_container *pcont, struct params_code *cpars);

//int prrtblist_insert_packet(struct prrt_block_list *blist, struct prrt_packet_container *pcont, struct params_code *cpars, struct params_local *lpars)
int prrtblist_insert_packet(struct prrt_block_list *blist, struct prrt_packet_container *pcont, struct params_code *cpars)
{
	int err = 0;

	if(!prrtblist_can_insert_packet(blist))
		return -1;

	//err = InsertDataPacketInLastBlock(blist, pcont, cpars, lpars);
	err = InsertDataPacketInLastBlock(blist, pcont, cpars);
	if(err < 0)
		return err;

	return err;
}

int GetLastBlock(struct prrt_block_list *blist, struct prrt_block **mblock);
static void ResetNextPacketForInsertedPacket(struct prrt_block_list *blist, struct prrt_block *last_block,
					     struct packet_list_entry *p);

//int InsertDataPacketInLastBlock(struct prrt_block_list *blist, struct prrt_packet_container *pcont, struct params_code *cpars, struct params_local *lpars)
int InsertDataPacketInLastBlock(struct prrt_block_list *blist, struct prrt_packet_container *pcont, struct params_code *cpars)
{
	int err = 0;
	struct prrt_block *last_block = 0;
	struct packet_list_entry *p = 0;

	err = GetLastBlock(blist, &last_block);
	if(err < 0)
		return err;

	//err = prrt_block_insert_data_packet(last_block, pcont, &p, cpars, lpars);
	err = prrt_block_insert_data_packet(last_block, pcont, &p, cpars);
	if(err < 0)
		return err;

	ResetNextPacketForInsertedPacket(blist, last_block, p);

	up(&blist->packet_count);
	PDEBUG(MSG_BLOCKLIST, "%s: Inserted packet(%d) in block(%p), position %d, k %d, now %u left\n", DEBUG_NAME,
		pcont->pack.data.header.seqnum, last_block, last_block->no_data, prrtcpar_get_k(&(last_block->cpars)), blist->packet_count.count);

	return err;
}

int GetLastBlock(struct prrt_block_list *blist, struct prrt_block **mblock)
{
	struct list_head *iter = 0;

	*mblock = 0;

	if(list_empty(&blist->head))
		return -1;

	iter = blist->head.prev;

	*mblock = container_of(iter, struct prrt_block, list);

	return 0;
}

static void ResetNextPacketForInsertedPacket(struct prrt_block_list *blist, struct prrt_block *last_block,
					     struct packet_list_entry *p)
{
	if(blist->next_packet == 0)
		SetNextPacket(blist, last_block, p);
	
	ResetNextPacketIfFirstInsertedPacket(blist, last_block, p);
}

int prrtblist_block_was_send_completely(struct prrt_block_list *blist, struct prrt_block *mblock) {
	struct packet_list_entry *p = 0;
	int err = 0;

	err = prrt_block_get_last_data(mblock, &p);
	if(err < 0) {
		PDEBUG(MSG_BLOCKLIST, "%s: Could not retrieve last data of block(%p)\n", DEBUG_NAME, mblock);
		goto not;	/*no packets in block => not send yet*/
	}

	if(prrt_comp_packetnos(p->pcont->pack.data.header.seqnum, blist->next_packet_no) < 0)
		return 1;

	PDEBUG(MSG_BLOCKLIST, "%s: last data no = %u\n", DEBUG_NAME, p->pcont->pack.data.header.seqnum);
	PDEBUG(MSG_BLOCKLIST, "%s: blist next no = %ld\n", DEBUG_NAME, blist->next_packet_no);

not:
	PDEBUG(MSG_BLOCKLIST, "%s: Block(%p) was not send completely yet\n", DEBUG_NAME, mblock);
	return 0;
}

int prrtblist_get_block(struct prrt_block_list *blist, u16 packetno, struct prrt_block **mblock)
{
	struct list_head *iter = 0;

	list_for_each(iter, &blist->head) {
		*mblock = list_entry(iter, struct prrt_block, list);

		if(prrt_block_contains_packet(*mblock, packetno)) {
			PDEBUG(MSG_SENDFBDEB, "%s: Found block for replacement (packetno %d):\n", DEBUG_NAME, packetno);
			PDEBUG(MSG_SENDFBDEB, "%s: #data = %d\n", DEBUG_NAME, (*mblock)->no_data);
			PDEBUG(MSG_SENDFBDEB, "%s: #red = %d\n", DEBUG_NAME, (*mblock)->no_red);
			/*this is the right block*/
			return 0;
		}
	}

	*mblock = 0;

	return -1;
}

int prrtblist_get_last_block(struct prrt_block_list *blist, struct prrt_block **mblock)
{
	return GetLastBlock(blist, mblock);
}

int NoDataPacketsThatMustBeSend(struct prrt_block_list *blist);

int prrtblist_get_next_packet(struct prrt_block_list *blist, struct prrt_packet_container **pcont)
{
	struct packet_list_entry *p = 0;

	if(NoDataPacketsThatMustBeSend(blist))
		return -1;

	p = container_of(blist->next_packet, struct packet_list_entry, list);
	*pcont = p->pcont;

	return 0;
}

int NoDataPacketsThatMustBeSend(struct prrt_block_list *blist)
{
	PDEBUG(MSG_BLOCKLIST, "%s: next_packet = %p\n", DEBUG_NAME, blist->next_packet);
	PDEBUG(MSG_BLOCKLIST, "%s: next_packet_no = %ld\n", DEBUG_NAME, blist->next_packet_no);
	return (blist->next_packet == 0);
}

void AdvanceNextPacketPointer(struct prrt_block_list *blist);

void prrtblist_confirm_packet_sending(struct prrt_block_list *blist)
{
	AdvanceNextPacketPointer(blist);
}

int NextPacketIsNotLastPacketInBlock(struct prrt_block_list *blist);
int SetNextPacketToNextPacketInBlock(struct prrt_block_list *blist);
int TryToGetBlockForNextPacket(struct prrt_block_list *blist, struct prrt_block **mblock);
int SetNextPacketThatIsNotAvailableYet(struct prrt_block_list *blist);
int TryToSetNextPacketToFirstPacketOfTheNextBlock(struct prrt_block_list *blist,
						  struct prrt_block *next_block);

void AdvanceNextPacketPointer(struct prrt_block_list *blist)
{
	int err = 0;
	struct prrt_block *next_block = 0;

	if(NextPacketIsNotLastPacketInBlock(blist)) {
		PDEBUG(MSG_BLOCKLIST, "%s: Moving next packet pointer in block\n", DEBUG_NAME);
		SetNextPacketToNextPacketInBlock(blist);
	} else {
		PDEBUG(MSG_BLOCKLIST, "%s: Retrieving next block to advance next packet\n", DEBUG_NAME);
		err = TryToGetBlockForNextPacket(blist, &next_block);
		if(err < 0) {
			PDEBUG(MSG_BLOCKLIST, "%s: Could not retrieve\n", DEBUG_NAME);
			SetNextPacketThatIsNotAvailableYet(blist);
			return;
		}

		PDEBUG(MSG_BLOCKLIST, "%s: Retrieved, retrieve first packet\n", DEBUG_NAME);
		TryToSetNextPacketToFirstPacketOfTheNextBlock(blist, next_block);
	}
}

int NextPacketIsNotLastPacketInBlock(struct prrt_block_list *blist)
{
	struct prrt_block *next_packet_block = 0;

	next_packet_block = container_of(blist->next_packet_block, struct prrt_block, list);

	return (blist->next_packet->next != &next_packet_block->data_head);
}

int SetNextPacketToNextPacketInBlock(struct prrt_block_list *blist)
{
	struct packet_list_entry *next_packet_entry = 0;
	struct prrt_block *block_of_next_packet = 0;

	PDEBUG(MSG_BLOCKLIST, "%s: next packetno = %ld\n", DEBUG_NAME, blist->next_packet_no);
	next_packet_entry = container_of(blist->next_packet->next, struct packet_list_entry, list);
	block_of_next_packet = container_of(blist->next_packet_block, struct prrt_block, list);

	SetNextPacket(blist, block_of_next_packet, next_packet_entry);
	PDEBUG(MSG_BLOCKLIST, "%s: new next packetno = %ld\n", DEBUG_NAME, blist->next_packet_no);

	return 0;
}

int TryToGetBlockForNextPacket(struct prrt_block_list *blist, struct prrt_block **mblock)
{
	struct list_head *next_block = 0;

	*mblock = 0;
	next_block = blist->next_packet_block->next;

	if(next_block == &blist->head) {
		return -1;
	}

	*mblock = container_of(next_block, struct prrt_block, list);

	return 0;
}

int SetNextPacketThatIsNotAvailableYet(struct prrt_block_list *blist)
{
	blist->next_packet = 0;
	blist->next_packet_block = &blist->head;
	blist->next_packet_no++;

	return 0;
}

int TryToSetNextPacketToFirstPacketOfTheNextBlock(struct prrt_block_list *blist,
						  struct prrt_block *next_block)
{
	struct packet_list_entry *next_packet_p = 0;
	int err = 0;

	if(prrt_block_is_empty(next_block)) {
		PDEBUG(MSG_BLOCKLIST, "%s: Block empty unsetting packet\n", DEBUG_NAME);
		SetNextPacketThatIsNotAvailableYet(blist);
		return 0;
	}

	err = prrt_block_get_first_data(next_block, &next_packet_p);
	if(err < 0) {
		SetNextPacketThatIsNotAvailableYet(blist);
		return 0;
	}

	PDEBUG(MSG_BLOCKLIST, "%s: Setting next packet\n", DEBUG_NAME);
	SetNextPacket(blist, next_block, next_packet_p);

	return 0;
}

int GetLastPacket(struct prrt_block_list *blist, struct prrt_packet_container **pcont);

int prrtblist_get_last_packetno(struct prrt_block_list *blist, int *packetno)
{
	struct prrt_packet_container *pcont= 0;
	u16 last_pno = 0;
	int err = 0;

	err = GetLastPacket(blist, &pcont);
	if (err < 0)
		return -1;

	last_pno = pcont->pack.data.header.seqnum;

	*packetno = last_pno;

	return 0;
}

int GetLastPacket(struct prrt_block_list *blist, struct prrt_packet_container **pcont)
{
	struct prrt_block *last_block = container_of(blist->head.prev, struct prrt_block, list);
	struct packet_list_entry *last = 0;
	int err = 0;

	*pcont = 0;

	if(list_empty(&blist->head))
		return -1;

	err = prrt_block_get_last_data(last_block, &last);
	if(err < 0) {
		if(last_block->list.prev == &blist->head)
			goto out;
		/*try the block before the last block, if last block is empty*/
		last_block = container_of(last_block->list.prev, struct prrt_block, list);

		err = prrt_block_get_last_data(last_block, &last);
		if(err < 0)
			goto out;
	}

	*pcont = last->pcont;

out:
	return err;
}

int prrtblist_get_timestamp_of_last_packet(struct prrt_block_list *blist, u32 *timestamp)
{
	int err = 0;
	struct prrt_packet_container *pcont = 0;

	err = GetLastPacket(blist, &pcont);
	if (err < 0)
		return -1;

	*timestamp = (u32)pcont->time.transmit_time;

	return 0;
}

int prrtblist_get_first_block(struct prrt_block_list *blist, struct prrt_block **mblock)
{
	*mblock = 0;

	if(!list_empty(&blist->head)) {
		*mblock = container_of(blist->head.next, struct prrt_block, list);
		return 0;
	}

	return -1;
}

int BlockContainsNextPacket(struct prrt_block_list *blist, struct prrt_block *mblock);

int prrtblist_try_to_delete_first_block(struct prrt_block_list *blist)
{
	struct prrt_block *first_block = 0;
	int err = 0;

	err = prrtblist_get_first_block(blist, &first_block);
	if(err < 0) {
		PERROR("%s: Could not retrieve first block\n", DEBUG_NAME);
		return -1;
	}

	PDEBUG(MSG_BLOCKLIST, "%s: Trying to delete block(%p)\n", DEBUG_NAME, first_block);

	if(!prrtblist_block_was_send_completely(blist, first_block)) {
		//PERROR("%s: Block must still be send, may not delete\n", DEBUG_NAME);
		return -1;
	}

	if(first_block->first_red_send == 0) {
		//PERROR("%s: Redundancy of first round must still be send, may not delete\n", DEBUG_NAME);
		return -1;
	}

	list_del(&first_block->list);
	prrt_block_destroy(&first_block);
	PDEBUG(MSG_BLOCKLIST, "%s: Deleted old block\n", DEBUG_NAME);

	atomic_dec(&blist->block_count);

	return 0;
}

void prrtblist_print_stats(struct prrt_block_list *blist) {
	PDEBUG(MSG_LISTSTATS, "%s: Blocks in block list: %d\n", DEBUG_NAME,
		atomic_read(&blist->block_count));
}
