/* 
 * (c) 2010 Saarland University, Telecommunications Lab 
 * (prrt@nt.uni-saarland.de)
 *
 * The contents of this file are subject to the Mozilla Public License
 + Version 1.1 (the "License"); you may not use this file except in
 + compliance with the License. You may obtain a copy of the License at
 + http://www.mozilla.org/MPL/
 + 
 + Software distributed under the License is distributed on an "AS IS"
 + basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 + License for the specific language governing rights and limitations
 + under the License.
 + 
 + The Original Code is "Predictably Reliable Real-time Transport (PRRT)".
 + 
 + The Initial Developer of the Original Code is Saarland University.
 + All Rights Reserved.
 + 
 + Contributor(s): Jochen Gruen (Saarland University), Manuel Gorius
 + (Saarland University), Yongtao Shuai (Saarland University).
 + 
 + Disclaimer: This Alpha version of PRRT is not supported and stable yet.
 + Do not use it in a production environment: it may crash your device, 
 + so please make sure that you have a backup of your data before
 + installing PRRT.
 */

#include "prrt_data_list.h"

#define DEBUG_NAME "PRRT_DATA_LIST"

int prrtdlist_init(struct prrt_data_list *dlist, struct prrt_time_state *time_state) {
    INIT_LIST_HEAD(&dlist->head);
    init_MUTEX_LOCKED(&dlist->count);
    prrtstats_init_list_stats(&dlist->stats);
    dlist->last_read_pno = -1;
    dlist->last_packet_inserted_at_the_end = 0;

	  dlist->WorkingBlockStart = -1;
    dlist->WorkingBlockEnd = -1;
    dlist->WorkingBlocksLastData = 0;
	  dlist->time_state = time_state;
	
    return 0;
}

int prrtdlist_reinit(struct prrt_data_list *dlist) {
    prrtdlist_free(dlist);

    init_MUTEX_LOCKED(&dlist->count);
    prrtstats_init_list_stats(&dlist->stats);
    dlist->last_read_pno = -1;
    dlist->last_packet_inserted_at_the_end = 0;

	  dlist->WorkingBlockStart = -1;
    dlist->WorkingBlockEnd = -1;
    dlist->WorkingBlocksLastData = 0;
	
    return 0;
}

void prrtdlist_free(struct prrt_data_list *dlist) {
    prrt_list_delete_all_packets_in_list(&dlist->head);
}

int prrtdlist_is_empty(struct prrt_data_list *dlist) {
    return list_empty(&dlist->head);
}

int prrtdlist_get_last_read_packetnumber(struct prrt_data_list *dlist, u16 *pno) {
    *pno = 0;

    if (dlist->last_read_pno == -1)
        return -1;

    *pno = dlist->last_read_pno;

    return 0;
}

int PacketIsOlderThanLastReadPacket(struct prrt_data_list *dlist, struct prrt_packet_container *pcont);
int PacketGoesToTheEnd(struct prrt_data_list *dlist, struct prrt_packet_container *pcont);
int InsertPacketAtTheEnd(struct prrt_data_list *dlist, struct prrt_packet_container *pcont);
int InsertPacketAtRightPosition(struct prrt_data_list *dlist, struct prrt_packet_container *pcont);

int prrtdlist_insert_datapacket(struct prrt_data_list *dlist, struct prrt_packet_container *pcont) {
    int err = 0;

    PDEBUG(MSG_DATALIST, "%s: Checking if packet is too old\n", DEBUG_NAME);
    if (PacketIsOlderThanLastReadPacket(dlist, pcont)) {
        return -1;
    }

    if (PacketGoesToTheEnd(dlist, pcont)) {
        //PDEBUG(MSG_DATALIST, "%s: Packet goes to the end\n", DEBUG_NAME);
        err = InsertPacketAtTheEnd(dlist, pcont);
        if (err < 0) {
            PDEBUG(MSG_DATALIST, "%s: Could not insert packet in data list\n", DEBUG_NAME);
            return err;
        }

        dlist->last_packet_inserted_at_the_end = 1;
    } else {
        PDEBUG(MSG_DATALIST, "%s: Packet goes to the middle\n", DEBUG_NAME);
        err = InsertPacketAtRightPosition(dlist, pcont);
        if (err < 0) {
            PDEBUG(MSG_DATALIST, "%s: Could not insert packet in data list\n", DEBUG_NAME);
            return err;
        }

        dlist->last_packet_inserted_at_the_end = 0;
    }

    PDEBUG(MSG_DATALIST, "%s: Updating list stats\n", DEBUG_NAME);
    prrtstats_update_list_stats_add(&dlist->stats, pcont);

    return err;
}

int PacketIsOlderThanLastReadPacket(struct prrt_data_list *dlist, struct prrt_packet_container *pcont) {
    if (list_empty(&dlist->head)){
	    dlist->last_read_pno = -1;
	    return 0;
    }

    if (dlist->last_read_pno == -1)
        return 0;

    PDEBUG(MSG_DATALIST, "%s: New packet old? last_read_pno %ld, packetno %d \n", DEBUG_NAME, dlist->last_read_pno, pcont->pack.data.header.seqnum);
    return (prrt_comp_packetnos(dlist->last_read_pno, pcont->pack.data.header.seqnum) >= 0);
}

int PacketGoesToTheEnd(struct prrt_data_list *dlist, struct prrt_packet_container *pcont) {
    struct packet_list_entry *lastPacketEntry = 0;
    int lastPacketNumber = 0;

    if (list_empty(&dlist->head))
        return 1;

    lastPacketEntry = container_of(dlist->head.prev, struct packet_list_entry, list);
    lastPacketNumber = lastPacketEntry->pcont->pack.data.header.seqnum;

    return (prrt_comp_packetnos(pcont->pack.data.header.seqnum, lastPacketNumber) > 0);
}

int InsertPacketBehindPos(struct prrt_data_list *dlist, struct prrt_packet_container *pcont,
        struct list_head *position);

int InsertPacketAtTheEnd(struct prrt_data_list *dlist, struct prrt_packet_container *pcont) {
    int err = 0;
    struct list_head *position = dlist->head.prev;

    PDEBUG(MSG_DATALIST, "%s: Inserting packet at the end\n", DEBUG_NAME);
    err = InsertPacketBehindPos(dlist, pcont, position);

    return err;
}

int InsertPacketBehindPos(struct prrt_data_list *dlist, struct prrt_packet_container *pcont,
        struct list_head *position) {
    int err = 0;

    err = prrt_list_insert_packet_behind_position(pcont, position);
    if (err < 0){
        return err;
}

    up(&dlist->count);

    return err;
}

int SearchRightPosition(struct prrt_data_list *dlist, struct prrt_packet_container *pcont,
        struct list_head **position);

int InsertPacketAtRightPosition(struct prrt_data_list *dlist, struct prrt_packet_container *pcont) {
    int err = 0;
    struct list_head *position = 0;

    err = SearchRightPosition(dlist, pcont, &position);
    if (err < 0)
        return err;

    PDEBUG(MSG_DATALIST, "%s: Inserting packet at right position\n", DEBUG_NAME);
    err = InsertPacketBehindPos(dlist, pcont, position);

    return err;
}

int SearchRightPosition(struct prrt_data_list *dlist, struct prrt_packet_container *pcont,
        struct list_head **position) {
    int err = 0;

    err = prrt_list_find_position_for_sorted_insertion(&dlist->head, pcont->pack.data.header.seqnum,
            position);

    return err;
}

int prrtdlist_last_packet_inserted_at_the_end(struct prrt_data_list *dlist) {
    return dlist->last_packet_inserted_at_the_end;
}

int prrtdlist_wait_for_and_grab_next_datapacket(struct prrt_data_list *dlist) {
    if (down_timeout(&dlist->count, POLL_MAXWAITTIME) == -ETIME) {
	//reset last read packet number since queue is empty.
        PDEBUG(MSG_PACKRETURN, "%s: Waiting for next data packet timed out.", DEBUG_NAME);
        return -ENODATA;
    }

    return 0;
}

int prrtdlist_next_datapacket_is_readable(struct prrt_data_list *dlist, u32 d_target) {
    struct packet_list_entry *p = 0;

    p = container_of(dlist->head.next, struct packet_list_entry, list);

    return (prrt_packet_may_read(p->pcont, d_target, dlist->time_state));
}

int prrtdlist_show_next_datapacket(struct prrt_data_list *dlist, struct prrt_packet_container **pcont) {
    struct list_head *iter = 0;
    struct packet_list_entry *p = 0;

    *pcont = 0;

    if (list_empty(&dlist->head)) {
        //PERROR("%s: Data list is empty but we are reading packet\n", DEBUG_NAME);
        return -ENODATA;
    }

    iter = dlist->head.next;
    p = list_entry(iter, struct packet_list_entry, list);
    *pcont = p->pcont;

    return 0;
}

int prrtdlist_get_next_datapacket(struct prrt_data_list *dlist, struct prrt_packet_container **pcont) {
    struct list_head *iter = 0;
    struct packet_list_entry *p = 0;

    *pcont = 0;

    if (list_empty(&dlist->head)) {
        PERROR("%s: Data list is empty but we are reading packet\n", DEBUG_NAME);
        return -ENODATA;
    }

    iter = dlist->head.next;
    p = list_entry(iter, struct packet_list_entry, list);
    *pcont = p->pcont;

    list_del(iter);
    prrt_list_entry_free(&p);

    prrtstats_update_list_stats_del(&dlist->stats, *pcont);

    dlist->last_read_pno = (*pcont)->pack.data.header.seqnum;

    return 0;
}

static int GetLastPacketInList(struct prrt_data_list *dlist, struct prrt_packet_container **pcont);
static int CouldGrabAPacketForDeletion(struct prrt_data_list *dlist);
static int MayDeleteFirstPacket(struct prrt_data_list *dlist, u32 d_target);
static long int DeleteFirstPacket(struct prrt_data_list *dlist);
static int LastReadPacketNumberMustBeRefreshed(struct prrt_data_list *dlist, int deletedPacketNumber);

long int prrtdlist_delete_old_packet(struct prrt_data_list *dlist, u32 d_target) {
    int err = 0;
    long int packetNumberDeleted = -1;
    struct prrt_packet_container *pcont = 0;

    err = GetLastPacketInList(dlist, &pcont);
    if (err < 0)
        return err;

    if (!CouldGrabAPacketForDeletion(dlist))
        return -1;

    if (!MayDeleteFirstPacket(dlist, d_target)) {
        prrtdlist_release_a_grabbed_packet(dlist);
        return -1;
    }

    packetNumberDeleted = DeleteFirstPacket(dlist);

    if (LastReadPacketNumberMustBeRefreshed(dlist, packetNumberDeleted)) {
        dlist->last_read_pno = packetNumberDeleted;
    }

    return packetNumberDeleted;
}

static int GetLastPacketInList(struct prrt_data_list *dlist, struct prrt_packet_container **pcont) {
    struct packet_list_entry *p = 0;

    if (list_empty(&dlist->head))
        return -1;

    p = list_entry(dlist->head.prev, struct packet_list_entry, list);
    *pcont = p->pcont;

    return 0;
}

static int CouldGrabAPacketForDeletion(struct prrt_data_list *dlist) {
    if (down_trylock(&dlist->count))
        return 0;

    return 1;
}

static int GetFirstPacketInList(struct prrt_data_list *dlist, struct prrt_packet_container **pcont);

static int MayDeleteFirstPacket(struct prrt_data_list *dlist, u32 d_target) {
    int err = 0;
    struct prrt_packet_container *pcont = 0;

    err = GetFirstPacketInList(dlist, &pcont);
    if (err < 0)
        return 0;

    if (!prrt_packet_may_delete(pcont, d_target, dlist->time_state)) {
        return 0;
    }

    return 1;
}

static int GetFirstPacketInList(struct prrt_data_list *dlist, struct prrt_packet_container **pcont) {
    struct packet_list_entry *p = 0;

    if (list_empty(&dlist->head))
        return -1;

    p = list_entry(dlist->head.next, struct packet_list_entry, list);
    *pcont = p->pcont;

    return 0;
}

void prrtdlist_release_a_grabbed_packet(struct prrt_data_list *dlist) {
    up(&dlist->count);
}

static long int DeleteFirstPacket(struct prrt_data_list *dlist) {
    struct packet_list_entry *p = 0;
    u16 firstPacketNumber = 0;

    p = container_of(dlist->head.next, struct packet_list_entry, list);
    firstPacketNumber = p->pcont->pack.data.header.seqnum;

    PDEBUG(MSG_PACKOMMIT, "%s: Ommiting packet number %u:\n", DEBUG_NAME, firstPacketNumber);
    PDEBUG(MSG_PACKOMMIT, "%s: -----------------------------\n", DEBUG_NAME);
    prrt_packet_log_omission(p->pcont);
    PDEBUG(MSG_PACKOMMIT, "%s: -----------------------------\n", DEBUG_NAME);

    prrt_list_delete_packet_from_list(dlist->head.next);

    return firstPacketNumber;
}

static int LastReadPacketNumberMustBeRefreshed(struct prrt_data_list *dlist, int deletedPacketNumber) {
    return (prrt_comp_packetnos(dlist->last_read_pno, deletedPacketNumber) <= 0);
}

static int StartToGetPacketsOfNewBlock(long WorkingBlockStart, u16 BlockStart, long WorkingBlockEnd,
        u16 BlockEnd);
static int FindADataPacketForBlock(struct prrt_data_list *dlist, u16 BlockStart, u16 BlockEnd,
        struct list_head **LastDataFound);

int prrtdlist_get_data_packet_by_block(struct prrt_data_list *dlist, struct packet_list_entry **p,
        u16 BlockStart, u16 BlockEnd) {
    int err = 0;

    if (StartToGetPacketsOfNewBlock(dlist->WorkingBlockStart, BlockStart, dlist->WorkingBlockEnd, BlockEnd)) {
        dlist->WorkingBlockStart = BlockStart;
        dlist->WorkingBlockEnd = BlockEnd;
        dlist->WorkingBlocksLastData = 0;
    }

    err = FindADataPacketForBlock(dlist, dlist->WorkingBlockStart, dlist->WorkingBlockEnd, &dlist->WorkingBlocksLastData);
    if (err < 0) {
        dlist->WorkingBlockStart = -1;
        dlist->WorkingBlockEnd = -1;
        dlist->WorkingBlocksLastData = 0;
        *p = 0;
    } else {
        *p = container_of(dlist->WorkingBlocksLastData, struct packet_list_entry, list);
    }

    return err;
}

static int StartToGetPacketsOfNewBlock(long WorkingBlockStart, u16 BlockStart, long WorkingBlockEnd,
        u16 BlockEnd) {
    return (WorkingBlockStart == -1 ||
            (WorkingBlockStart != BlockStart || WorkingBlockEnd != BlockEnd));
}

static int NoPacketOfBlockSearchedYet(struct list_head *LastDataFound);
static int SearchCompleteListForPacket(struct prrt_data_list *dlist, u16 BlockStart,
        u16 BlockEnd, struct list_head **LastDataFound);
static int ContinueWithSearchAtLastPacketFound(struct prrt_data_list *dlist, u16 BlockStart,
        u16 BlockEnd, struct list_head **LastDataFound);

static int FindADataPacketForBlock(struct prrt_data_list *dlist, u16 BlockStart, u16 BlockEnd,
        struct list_head **LastDataFound) {
    int err = 0;

    if (NoPacketOfBlockSearchedYet(*LastDataFound)) {
        err = SearchCompleteListForPacket(dlist, BlockStart, BlockEnd, LastDataFound);
    } else {
        err = ContinueWithSearchAtLastPacketFound(dlist, BlockStart, BlockEnd, LastDataFound);
    }

    return err;
}

static int NoPacketOfBlockSearchedYet(struct list_head *LastDataFound) {
    return (LastDataFound == 0);
}

static int DataBelongsToBlock(struct list_head *Data, u16 BlockStart, u16 BlockEnd);

static int SearchCompleteListForPacket(struct prrt_data_list *dlist, u16 BlockStart,
        u16 BlockEnd, struct list_head **LastDataFound) {
    struct list_head *iter = 0;

    list_for_each(iter, &dlist->head) {
        if (DataBelongsToBlock(iter, BlockStart, BlockEnd)) {
            *LastDataFound = iter;
            return 0;
        }
    }

    *LastDataFound = 0;
    return -1;
}

static int DataBelongsToBlock(struct list_head *Data, u16 BlockStart, u16 BlockEnd) {
    struct packet_list_entry *p = container_of(Data, struct packet_list_entry, list);
    u16 packetno = p->pcont->pack.data.header.seqnum;

    return (prrt_packetno_lies_in_limits(packetno, BlockStart, BlockEnd));
}

static int NoMoreDataIsAvailable(struct prrt_data_list *dlist, struct list_head *LastDataFound);

static int ContinueWithSearchAtLastPacketFound(struct prrt_data_list *dlist, u16 BlockStart,
        u16 BlockEnd, struct list_head **LastDataFound) {
    struct list_head *NextData = 0;

    if (NoMoreDataIsAvailable(dlist, *LastDataFound)) {
        goto error;
    }

    NextData = (*LastDataFound)->next;

    if (DataBelongsToBlock(NextData, BlockStart, BlockEnd)) {
        *LastDataFound = NextData;
        return 0;
    }

error:
    *LastDataFound = 0;
    return -1;
}

static int NoMoreDataIsAvailable(struct prrt_data_list *dlist, struct list_head *LastDataFound) {
    return (LastDataFound->next == &dlist->head);
}
