/* 
 * (c) 2010 Saarland University, Telecommunications Lab 
 * (prrt@nt.uni-saarland.de)
 *
 * The contents of this file are subject to the Mozilla Public License
 + Version 1.1 (the "License"); you may not use this file except in
 + compliance with the License. You may obtain a copy of the License at
 + http://www.mozilla.org/MPL/
 + 
 + Software distributed under the License is distributed on an "AS IS"
 + basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 + License for the specific language governing rights and limitations
 + under the License.
 + 
 + The Original Code is "Predictably Reliable Real-time Transport (PRRT)".
 + 
 + The Initial Developer of the Original Code is Saarland University.
 + All Rights Reserved.
 + 
 + Contributor(s): Jochen Gruen (Saarland University), Manuel Gorius
 + (Saarland University), Yongtao Shuai (Saarland University).
 + 
 + Disclaimer: This Alpha version of PRRT is not supported and stable yet.
 + Do not use it in a production environment: it may crash your device, 
 + so please make sure that you have a backup of your data before
 + installing PRRT.
 */

#include "prrt_inc_list.h"
#include <linux/sched.h>

#define DEBUG_NAME "PRRT_INC_LIST"

void prrtilist_init(struct prrt_inc_list *ilist, struct prrt_time_state *time_state)
{
	prrtdlist_init(&ilist->data_list, time_state);
	prrtredlist_init(&ilist->red_list);
	init_MUTEX(&ilist->data_lock);
	init_MUTEX(&ilist->red_lock);
	ilist->decoder = 0;
	ilist->biggest_inserted_data_packet = -1;
	ilist->current_sn_base = 0;
	ilist->dec_timeout_min = 0;
	ilist->dec_timeout_max = 0;

	init_MUTEX(&ilist->opt_lock);
	ilist->feedback = ON;
	ilist->decoding = ON;

	ilist->received_datapackets = 0;
	ilist->received_redpackets = 0;
	ilist->received_overall = 0;
	ilist->ommited_at_start_packets = 0;
	ilist->old_packets = 0;
	ilist->lost_packets = 0;
	ilist->missed_packets = 0;
	ilist->used_redpackets = 0;
	ilist->useful_redpackets = 0;

	ilist->feedbacks_sent = 0;
	ilist->fbpackets_sent = 0;

	ilist->lastPacketDelivered = -1;
	atomic_set(&ilist->readingAlreadyStarted, 0);
	atomic_set(&ilist->lastMeasured_t_replace, 0);
	atomic_set(&ilist->lastDecodingDelay, 0);
	atomic_set(&ilist->lastPacketTs, 0);
	atomic_set(&ilist->lastPacketRecvTs, 0);
	atomic_set(&ilist->lastFBSentTs, 0);

	prrt_lossgath_init(&ilist->lossgath, time_state);
	prrtnpar_init_netparams(&ilist->net_params);
	
	ilist->last_retrieved_packetno_plus_one = 0;
	ilist->time_state = time_state;
}

void prrtilist_uninit(struct prrt_inc_list *ilist)
{
	prrtdlist_free(&ilist->data_list);
	prrtredlist_free(&ilist->red_list);
	prrt_lossgath_free(&ilist->lossgath);
	prrt_fec_free(ilist->decoder);
}

static int insert_datapacket(struct prrt_inc_list *ilist, struct prrt_packet_container *pcont);

int prrtilist_insert_datapacket(struct prrt_inc_list *ilist, struct prrt_packet_container *pcont)
{
	int err = 0;
	u32 packet_interval = 0;
	u32 time_us = 0;
	u32 last_packet_recv_ts = 0;

	down(&ilist->data_lock);

	err = insert_datapacket(ilist, pcont);

	if (prrt_packet_is_redata(pcont)){
		ilist->received_redpackets++;
	}else{
		ilist->received_datapackets++;
		prrtnpar_insert_Sent(&ilist->net_params, pcont->pack.data.header.seqnum);
		//PERROR("%s: seqnum %u\n", DEBUG_NAME, pcont->pack.data.data_header.seqnum);
	}
	ilist->received_overall++;

	up(&ilist->data_lock);

	if (!prrt_packet_is_redata(pcont)){
		time_us = prrt_get_ustime(ilist->time_state);
		// Next if changed = to == ...Check!!
		if (last_packet_recv_ts == atomic_read(&ilist->lastPacketRecvTs))
			packet_interval = prrt_diff_timestamps(ilist->time_state, time_us, last_packet_recv_ts);

		atomic_set(&ilist->lastPacketTs, pcont->pack.data.data_header.timestamp);
		atomic_set(&ilist->lastPacketTimeout, pcont->pack.data.data_header.pack_timeout);
		//PDEBUG(MSG_FINAL, "%s: Timestamp %u, at %u, prrt time %u, timeout %u\n", DEBUG_NAME, 
		//	pcont->pack.data.data_header.timestamp, time_us, prrt_get_prrttime(ilist->time_state),
		//	pcont->pack.data.data_header.pack_timeout);
		//atomic_set(&ilist->lastPacketRecvTs, pcont->time.creation_time);
		atomic_set(&ilist->lastPacketRecvTs, time_us);
		atomic_set(&ilist->lastPacketRecvTsLocal, prrt_get_mstime(ilist->time_state));

		prrtnpar_update_PacketInterval(&ilist->net_params, packet_interval);
		prrtnpar_update_FTT(&ilist->net_params, prrt_diff_timestamps(ilist->time_state, prrt_get_prrttime(ilist->time_state), pcont->pack.data.data_header.timestamp));
		prrtnpar_update_QueueDelay(&ilist->net_params, (prrt_get_mstime(ilist->time_state) - pcont->pack.data.data_header.timestamp), pcont->pack.data.header.seqnum);
	}

	return err;
}

//static int HandleNetworkParameterUpdatingStuff(struct prrt_inc_list *ilist,
//					       struct prrt_packet_container *pcont,
//					       char listWasEmptyBeforeInsertion);
static int HandleLossGatheringStuff(struct prrt_inc_list *ilist, struct prrt_packet_container *pcont);
static int DeleteLossForRetransmittedPacket(struct prrt_inc_list *ilist,
					    struct prrt_packet_container *pcont);
static int UpdateBiggestInsertedPacket(struct prrt_inc_list *ilist, struct prrt_packet_container *pcont);

static int insert_datapacket(struct prrt_inc_list *ilist, struct prrt_packet_container *pcont)
{
	/*this is an internal function that is not locked, lock should already be hold*/

	int err = 0;
	u32 timestamp = 0;
	u32 group_rtt = pcont->pack.data.data_header.group_rtt;

	//char listWasEmptyBeforeInsertion = prrtdlist_is_empty(&ilist->data_list);

	err = prrtdlist_insert_datapacket(&ilist->data_list, pcont);
	if(err < 0){
		return err;
	}

	if(prrtdlist_last_packet_inserted_at_the_end(&ilist->data_list) && prrt_packet_is_data(pcont)) {
		HandleLossGatheringStuff(ilist, pcont);

		//HandleNetworkParameterUpdatingStuff(ilist, pcont, listWasEmptyBeforeInsertion);

		atomic_set(&ilist->lastMeasured_t_replace, group_rtt);
		prrtnpar_set_ReplaceTime(&ilist->net_params, group_rtt);

		timestamp = pcont->pack.data.data_header.timestamp;
		if (timestamp)
			prrt_update_prrttime(ilist->time_state, timestamp, (group_rtt >> 1));

		//PDEBUG(MSG_FINAL, "%s: Group_rtt %u, propagation_delay %u\n", DEBUG_NAME, group_rtt, (group_rtt >> 1));
		//PDEBUG(MSG_INCLIST, "%s: Updating prrt_time at %u, timestamp %u, prrt_time %u \n", DEBUG_NAME, prrt_get_mstime(), timestamp, prrt_get_prrttime());
		//PERROR("%s: Group_rtt %u, propagation_delay %u\n", DEBUG_NAME, group_rtt, (group_rtt >> 1));
		//PERROR("%s: Updating prrt_time at %u, timestamp %u, prrt_time %u meanD %d \n", DEBUG_NAME, prrt_get_mstime(ilist->time_state), timestamp, prrt_get_prrttime(ilist->time_state), atomic_read(&ilist->time_state->mean_deviation));

	}

	/*losses that are replaced with redundancy are deleted when receiving redundancy*/
	DeleteLossForRetransmittedPacket(ilist, pcont);

	UpdateBiggestInsertedPacket(ilist, pcont);

	return 0;
}

//static int InsertMeasurementsForPLREstimation(struct prrt_inc_list *ilist,
//					      struct prrt_packet_container *pcont,
//					      char listWasEmptyBeforeInsertion);
//
//static int HandleNetworkParameterUpdatingStuff(struct prrt_inc_list *ilist,
//					       struct prrt_packet_container *pcont,
//					       char listWasEmptyBeforeInsertion)
//{
//	int err = 0;
//
//	InsertMeasurementsForPLREstimation(ilist, pcont, listWasEmptyBeforeInsertion);
//
//	prrtnpar_update_PLR(&ilist->net_params);
//
//	return err;
//}

static int ThisIsTheFirstPacketToInsert(struct prrt_inc_list *ilist);
//static int InsertNewGoodMeasurmentForArrivedPacket(struct prrt_inc_list *ilist,
//						   struct prrt_packet_container *pcont);
static int ThereIsAGapBetweenBiggestPacketAndNewPacket(struct prrt_inc_list *ilist,
						       struct prrt_packet_container *pcont);
//static int InsertNewMeasurementsForPacket(struct prrt_inc_list *ilist, struct prrt_packet_container *pcont);
//static int UpdateGoodMeasurementForPacket(struct prrt_inc_list *ilist, struct prrt_packet_container *pcont);

//static int InsertMeasurementsForPLREstimation(struct prrt_inc_list *ilist,
//					      struct prrt_packet_container *pcont,
//					      char listWasEmptyBeforeInsertion)
//{
//	int err = 0;
//
//	if(listWasEmptyBeforeInsertion && ThisIsTheFirstPacketToInsert(ilist)) {
//		InsertNewGoodMeasurmentForArrivedPacket(ilist, pcont);
//		return err;
//	}
//
//	if(ThereIsAGapBetweenBiggestPacketAndNewPacket(ilist, pcont))
//		InsertNewMeasurementsForPacket(ilist, pcont);
//	else
//		UpdateGoodMeasurementForPacket(ilist, pcont);
//
//	return err;
//}

static int ThisIsTheFirstPacketToInsert(struct prrt_inc_list *ilist)
{
	return (ilist->biggest_inserted_data_packet == -1);
}

//static int InsertNewGoodMeasurmentForArrivedPacket(struct prrt_inc_list *ilist,
//						   struct prrt_packet_container *pcont)
//{
//	prrtnpar_insert_GoodMeas(&ilist->net_params, 1);
//
//	return 0;
//}
//
static int getGapWidth(struct prrt_inc_list *ilist, struct prrt_packet_container *pcont);

static int ThereIsAGapBetweenBiggestPacketAndNewPacket(struct prrt_inc_list *ilist,
						       struct prrt_packet_container *pcont)
{
	int gapWidth = 0;

	if(ThisIsTheFirstPacketToInsert(ilist)) /*no packets inserted yet => no gap*/
		return 0;

	gapWidth = getGapWidth(ilist, pcont);
	return (gapWidth > 0);
}

static int getGapWidth(struct prrt_inc_list *ilist, struct prrt_packet_container *pcont)
{
	return prrt_dist_packetnos(pcont->pack.data.header.seqnum, ilist->biggest_inserted_data_packet)-1;
}

//static int InsertBadMeasurementForTheGap(struct prrt_inc_list *ilist, struct prrt_packet_container *pcont);
//
//static int InsertNewMeasurementsForPacket(struct prrt_inc_list *ilist, struct prrt_packet_container *pcont)
//{
//	int err = 0;
//
//	InsertBadMeasurementForTheGap(ilist, pcont);
//	InsertNewGoodMeasurmentForArrivedPacket(ilist, pcont);
//
//	return err;
//}
//
//static int InsertBadMeasurementForTheGap(struct prrt_inc_list *ilist, struct prrt_packet_container *pcont)
//{
//	int gapWidth = getGapWidth(ilist, pcont);
//
//	prrtnpar_insert_BadMeas(&ilist->net_params, gapWidth);
//
//	return 0;
//}
//
//static int UpdateGoodMeasurementForPacket(struct prrt_inc_list *ilist, struct prrt_packet_container *pcont)
//{
//	prrtnpar_update_last_GoodMeas(&ilist->net_params);
//	return 0;
//}

static int InsertNewLossesInGathererList(struct prrt_inc_list *ilist, struct prrt_packet_container *pcont);

static int HandleLossGatheringStuff(struct prrt_inc_list *ilist, struct prrt_packet_container *pcont)
{
	if (ilist->dec_timeout_max)
		ilist->dec_timeout_min = ilist->dec_timeout_max;
	else
		ilist->dec_timeout_min = pcont->pack.data.data_header.timestamp + pcont->pack.data.data_header.dec_timeout;

	ilist->dec_timeout_max = pcont->pack.data.data_header.timestamp + pcont->pack.data.data_header.dec_timeout;

	InsertNewLossesInGathererList(ilist, pcont);

	//ilist->dec_timeout = pcont->pack.data.data_header.fbtimeout;

	return 0;
}

static int InsertLossForEachPacketInTheGap(struct prrt_inc_list *ilist, u16 gapStart, u16 gapEnd, u16 sn_base,
					   u32 pack_timeout, int first_due, u16 fb_timeout);

static int InsertNewLossesInGathererList(struct prrt_inc_list *ilist, struct prrt_packet_container *pcont)
{
	//int time = jif_to_ms(jiffies);
	u32 pack_timeout = pcont->pack.data.data_header.timestamp + pcont->pack.data.data_header.pack_timeout;
	u16 fb_timeout = pcont->pack.data.data_header.fb_timeout;

	int gapStart = prrt_increment_packetno(ilist->biggest_inserted_data_packet, 1);
	int gapEnd = prrt_decrement_packetno(pcont->pack.data.header.seqnum, 1);
	int blockLimit = 0;
	int blockLimitPlusOne = 0;
	int lowerSNBase = ilist->current_sn_base;
	int upperSNBase = 0;

	if(ThereIsAGapBetweenBiggestPacketAndNewPacket(ilist, pcont)) {
		PDEBUG(MSG_INCLIST, "%s: gapStart = %d\n", DEBUG_NAME, gapStart);
		PDEBUG(MSG_INCLIST, "%s: gapEnd = %d\n", DEBUG_NAME, gapEnd);
		if(!(prrt_comp_packetnos(gapStart+MAX_BURST, gapEnd)<0)){
			blockLimit = prrt_decrement_packetno(gapEnd, pcont->pack.data.header.index);
			blockLimitPlusOne = prrt_increment_packetno(blockLimit, 1);
			upperSNBase = blockLimitPlusOne;
			PDEBUG(MSG_INCLIST, "%s: blockLimit = %d\n", DEBUG_NAME, blockLimit);
			PDEBUG(MSG_INCLIST, "%s: blockLimitPlusOne = %d\n", DEBUG_NAME, blockLimitPlusOne);
			//TODO: does this work reliably in case gap > MAX_BURST?
			//TODO: What if gap is > blocklength?
			if (prrt_packetno_lies_in_limits(blockLimit, gapStart, gapEnd)){
				InsertLossForEachPacketInTheGap(ilist, gapStart, blockLimit, (u16)lowerSNBase, pack_timeout, ilist->dec_timeout_min, fb_timeout);
				InsertLossForEachPacketInTheGap(ilist, blockLimitPlusOne, gapEnd, (u16)upperSNBase, pack_timeout, ilist->dec_timeout_max, fb_timeout);
			PDEBUG(MSG_INCLIST, "%s: Within limits\n", DEBUG_NAME);
			}
			else{
				InsertLossForEachPacketInTheGap(ilist, gapStart, gapEnd, (u16)upperSNBase, pack_timeout, ilist->dec_timeout_max, fb_timeout);
				PDEBUG(MSG_INCLIST, "%s: Outside limits\n", DEBUG_NAME);
			}
		}
	else
		PDEBUG(MSG_INCLIST, "%s: Ignored gap, too long\n", DEBUG_NAME);
	}

	return 0;
}

static int InsertLossForEachPacketInTheGap(struct prrt_inc_list *ilist, u16 gapStart, u16 gapEnd,
					   u16 sn_base, u32 pack_timeout, int first_due, u16 fb_timeout)
{
	int i = 0;
	//u32 t_replace = (u32)prrtnpar_get_ReplaceTime(&ilist->net_params);
	u32 t_replace = fb_timeout;

	for(i=gapStart; prrt_comp_packetnos(i, gapEnd) <= 0; i=prrt_increment_packetno(i, 1)) {
		prrt_lossgath_insertfb(&ilist->lossgath, i, sn_base, pack_timeout, first_due, t_replace);
		prrtnpar_insert_Sent(&ilist->net_params, i);
		prrtnpar_insert_Loss(&ilist->net_params, i);
		ilist->lost_packets++;
		PDEBUG(MSG_INCLIST, "%s: Inserting loss %d, first_due %u, fb_timeout %u\n", DEBUG_NAME, i, first_due, t_replace);
	}

	return 0;
}

static int DeleteLossForRetransmittedPacket(struct prrt_inc_list *ilist,
					    struct prrt_packet_container *pcont)
{
	int measured_t_replace = 0;
	if(prrt_packet_is_redata(pcont))
		prrt_lossgath_delete1fb(&ilist->lossgath, pcont->pack.data.header.seqnum, 0, &measured_t_replace);

	//if(measured_t_replace)
	//	atomic_set(&ilist->lastMeasured_t_replace, measured_t_replace);

	return 0;
}

static int UpdateBiggestInsertedPacket(struct prrt_inc_list *ilist, struct prrt_packet_container *pcont)
{
	if(ilist->biggest_inserted_data_packet == -1 ||
	   prrt_comp_packetnos(pcont->pack.data.header.seqnum, ilist->biggest_inserted_data_packet) > 0) {
		PDEBUG(MSG_PRRTRECV, "%s: Updating last packet number %ld to %d\n", DEBUG_NAME,
			ilist->biggest_inserted_data_packet, pcont->pack.data.header.seqnum);
		ilist->biggest_inserted_data_packet = pcont->pack.data.header.seqnum;
		ilist->current_sn_base = prrt_decrement_packetno(pcont->pack.data.header.seqnum, pcont->pack.data.header.index);
	}

	return 0;
}

static int SleepSomeTimeToWaitForPacketGettingReadable(int waitTime);

int TryDecode(struct prrt_inc_list *ilist, struct prrt_packet_container *pcont);

int prrtilist_get_first_packet(struct prrt_inc_list *ilist, struct prrt_packet_container **pcont,
			      u32 d_target)
{
	int interrupted = 0;
	int err = 0;
	int i = 0;
	u32 time = 0;
	int waiting_time = 0;
	int missed_packets = 0;
	unsigned long retrieved_packetno = 0;

	*pcont = 0;

	down(&ilist->data_lock);
	err = prrtdlist_show_next_datapacket(&ilist->data_list, pcont);
        PDEBUG(MSG_COMMONRECV, "%s: After show_next_datapacket.\n", DEBUG_NAME);
	if(err == 0) {
		retrieved_packetno = (*pcont)->pack.data.header.seqnum;
		
		ilist->last_retrieved_packetno_plus_one = retrieved_packetno+1;
	}
	up(&ilist->data_lock);


	*pcont = 0;
	retrieved_packetno = 0;

        PDEBUG(MSG_COMMONRECV, "%s: before wait for next datapacket.\n", DEBUG_NAME);
	err = prrtdlist_wait_for_and_grab_next_datapacket(&ilist->data_list);
        PDEBUG(MSG_COMMONRECV, "%s: After wait for next datapacket.\n", DEBUG_NAME);
	if(err < 0)
		return err;

        PDEBUG(MSG_COMMONRECV, "%s: After error return of wait for next datapacket.\n", DEBUG_NAME);

	interrupted = down_interruptible(&ilist->data_lock);
	if(interrupted) {
		prrtdlist_release_a_grabbed_packet(&ilist->data_list);
		return -ERESTARTSYS;
	}

        PDEBUG(MSG_COMMONRECV, "%s: After interruptible down 1.\n", DEBUG_NAME);

	for(i=0; i<20; i++) {
		if(prrtdlist_next_datapacket_is_readable(&ilist->data_list, d_target))
			break;

		up(&ilist->data_lock);

		waiting_time = ms_to_jif((MAYREAD_DELAY_MULTI*d_target+MAYREAD_DELAY_ADD)/10);
		err = SleepSomeTimeToWaitForPacketGettingReadable(waiting_time);
		/*sleep some time to wait for packet getting ready to read*/
		if(err < 0) {
			prrtdlist_release_a_grabbed_packet(&ilist->data_list);
			goto out_nolock;
		}

		interrupted = down_interruptible(&ilist->data_lock);
		if(interrupted) {
			/*increase the counter again when we were interrupted*/
			prrtdlist_release_a_grabbed_packet(&ilist->data_list);
			return -ERESTARTSYS;
		}

		PDEBUG(MSG_PACKRETURN, "%s: Check if packet may be read\n", DEBUG_NAME);
	}

        PDEBUG(MSG_COMMONRECV, "%s: After wait loop.\n", DEBUG_NAME);

	/*this means we waited long enough*/
	if(i==20) {
		PERROR("%s: Packet may not be read yet\n", DEBUG_NAME);
		err = -ENODATA;
		goto out;
	}

	err = prrtdlist_get_next_datapacket(&ilist->data_list, pcont);
	if(err == 0) {
		retrieved_packetno = (*pcont)->pack.data.header.seqnum;
		down(&ilist->red_lock);
		prrtredlist_inform_of_packet_deletion(&ilist->red_list, retrieved_packetno);
		prrtredlist_delete_redundancy_packets_by_block_start(&ilist->red_list, retrieved_packetno);
		up(&ilist->red_lock);

		if(ilist->lastPacketDelivered != -1) {
			missed_packets = prrt_dist_packetnos((u16)retrieved_packetno,
						(u16)ilist->lastPacketDelivered)-1;

			if(missed_packets > 0) {
			    PDEBUG(MSG_TRANSPORT, "%s: Missed packets %lu to %lu\n", DEBUG_NAME,
				ilist->lastPacketDelivered+1, retrieved_packetno-1);
			    time = prrt_get_prrttime(ilist->time_state);
			    PDEBUG(MSG_TRANSPORT, "%s: Adding %d to missed packets at %u\n", DEBUG_NAME,
					    missed_packets, time);
			} else if(missed_packets < 0) {
			    PDEBUG(MSG_PACKMISSES, "%s: Woot something terrible happened between %lu and %lu\n",
				    DEBUG_NAME, ilist->lastPacketDelivered, retrieved_packetno);
			}
			

			ilist->missed_packets += missed_packets;
		} else {
			atomic_set(&ilist->readingAlreadyStarted, 1);
		}
		ilist->lastPacketDelivered = retrieved_packetno;
	}

out:
	PDEBUG(MSG_COMMONRECV, "%s: Releasing lock\n", DEBUG_NAME);
	up(&ilist->data_lock);

out_nolock:
	return err;
}

static int SleepSomeTimeToWaitForPacketGettingReadable(int waitTime)
{
	int err = 0;
        PDEBUG(MSG_COMMONRECV, "%s: Sleeping to wait for packet getting readable.\n", DEBUG_NAME);

	set_current_state(TASK_INTERRUPTIBLE);
	if(schedule_timeout(waitTime) != 0)
		err = -ERESTARTSYS;

	return err;
}

void prrtilist_delete_old_packets(struct prrt_inc_list *ilist, u32 d_target)
{
	down(&ilist->data_lock);

	while(prrtdlist_delete_old_packet(&ilist->data_list, d_target) >= 0) {
		//PDEBUG(MSG_COMMONRECV, "%s: Deleted old packet\n", DEBUG_NAME);

		if(atomic_read(&ilist->readingAlreadyStarted))
			ilist->old_packets++;
		else
			ilist->ommited_at_start_packets++;
	}

	up(&ilist->data_lock);
}

static int FillFeedbackPacket(struct prrt_inc_list *ilist, struct prrt_fbpacket_container *mfbcont, int d_target, u16 *max_repetitions);

int prrtilist_gather_feedback_packet(struct prrt_inc_list *ilist, struct prrt_fbpacket_container **mfbcont,
				    int d_target, u16 *repetitions)
{
	int err = 0;

	if(ilist->feedback == OFF)
		return -1;

	err = prrt_fbpacket_alloc(mfbcont);
	if(err < 0) {
		PERROR("%s: Could not allocate feedback packet\n", DEBUG_NAME);
		return err;
	}

	err = FillFeedbackPacket(ilist, *mfbcont, d_target, repetitions);
	if(err < 0)
		prrt_fbpacket_free(mfbcont);

	return err;
}

static int FillBlockInFBPacket(struct prrt_inc_list *ilist, struct prrt_fbpacket_container *mfbcont, int d_target, u16 *repetitions);

static int FillFeedbackPacket(struct prrt_inc_list *ilist, struct prrt_fbpacket_container *mfbcont, int d_target, u16 *max_repetitions)
{
	int err = 0;
	u16 repetitions = 1;
	*max_repetitions = 1;

	ilist->first_packet_fb_gathering = -1;
	mfbcont->block_count = 0;

	while(FillBlockInFBPacket(ilist, mfbcont, d_target, &repetitions) == 0) {
		if (repetitions > *max_repetitions)
			*max_repetitions = repetitions;
	}

	if(mfbcont->block_count == 0)
		err = -1;

	return err;
}

static int FillBlockInFBPacket(struct prrt_inc_list *ilist, struct prrt_fbpacket_container *mfbcont, int d_target, u16 *repetitions)
{
	int err = 0;
	//u32 time = jif_to_ms(jiffies);
	u32 time = prrt_get_prrttime(ilist->time_state);

	int t_replace = prrtnpar_get_ReplaceTime(&ilist->net_params);
	struct prrt_fbblock bl;

	*repetitions = 1;

	//PDEBUG(MSG_FINAL, "%s: Need replacement? %u, %d, %d\n", DEBUG_NAME, time, t_replace, d_target);
	if(prrt_fbpacket_is_full(mfbcont)) {
		PDEBUG(MSG_INCLIST, "%s: Feedback packet full\n", DEBUG_NAME);
		return -1;
	}

	err = prrt_lossgath_getnextfb(&ilist->lossgath, time, t_replace, d_target, &(bl.packno), &(bl.retr_round), repetitions);
	if(err < 0 || ilist->first_packet_fb_gathering == bl.packno) {
		PDEBUG(MSG_INCLIST, "%s: Didn't get next feedback (err %d)\n", DEBUG_NAME, err);
		return -1;
	}

	PDEBUG(MSG_INCLIST, "%s: Found loss of packet %d in round %d\n", DEBUG_NAME,
		bl.packno, bl.retr_round);

	if(ilist->first_packet_fb_gathering == -1)
		ilist->first_packet_fb_gathering = bl.packno;

	err = prrt_fbpacket_add_block(mfbcont, &bl);
	if(err < 0)
		PERROR("%s: Tried to insert block in full feedback packet!\n", DEBUG_NAME);

	return err;
}

struct decodingHelperContainer {
	int n;
	int k;
	u16 BlockStart;
	u16 BlockEnd;

	struct prrt_packet_container *newRedundancy;
	struct packet_list_entry **foundPackets;
	int numberDataPacketsFound;
	int numberRedPacketsFound;

	gf **PacketsToDecode;
	int *PacketIndexes;

	int lastRedundancyIndexUsed;
	char newRedundancyAlreadyUsed;
	char newRedundancyAlreadyInQueue;
};

static void setupDecodingHelperContainer(struct decodingHelperContainer *dhcont,
					 struct packet_list_entry *foundPackets[],
					 gf *PacketsToDecode[], int PacketIndexes[],
					 struct prrt_packet_container *pcont)
{
	int i = 0;

	dhcont->k = pcont->pack.red.red_header.k;
	dhcont->n = pcont->pack.red.red_header.n;
	dhcont->BlockStart = pcont->pack.red.red_header.sn_base;
	dhcont->BlockEnd = dhcont->BlockStart+dhcont->k-1;

	dhcont->newRedundancy = pcont;
	dhcont->foundPackets = foundPackets;
	dhcont->numberDataPacketsFound = 0;
	dhcont->numberRedPacketsFound = 1; /*the new redundancy that arrived*/

	dhcont->PacketsToDecode = PacketsToDecode;
	dhcont->PacketIndexes = PacketIndexes;

	dhcont->lastRedundancyIndexUsed = dhcont->k-1;
	dhcont->newRedundancyAlreadyUsed = 0;
	dhcont->newRedundancyAlreadyInQueue = 0;

	for(i=0; i<dhcont->k; ++i) {
		dhcont->foundPackets[i] = 0;
		dhcont->PacketsToDecode[i] = 0;
		dhcont->PacketIndexes[i] = -1;
	}

	for(i=dhcont->k; i<dhcont->n; ++i) {
		dhcont->foundPackets[i] = 0;
	}
}

int prrtilist_try_decode(struct prrt_inc_list *ilist, struct prrt_packet_container *pcont)
{
	int err = 0;

	if(!prrt_packet_is_redundancy(pcont)) {
		PERROR("%s: Packet is no redundancy packet\n", DEBUG_NAME);
		return -1;
	}

	//FIXME:
	//prrtnpar_set_ReplaceTime(&ilist->net_params, pcont->pack.red.red_header.t_replace);
	//prrtnpar_set_ReplaceTime(&ilist->net_params, 10);

	down(&ilist->data_lock);

	err = TryDecode(ilist, pcont);

	ilist->received_redpackets++;
	ilist->received_overall++;

	up(&ilist->data_lock);

	return err;

}

static int insert_redpacket(struct prrt_inc_list *ilist, struct prrt_packet_container *pcont);
static int GatherDataPacketsOfBlock(struct prrt_inc_list *ilist, struct decodingHelperContainer *dhcont);
static int DecodingShouldBeDone(struct prrt_inc_list *ilist, struct decodingHelperContainer *dhcont);
static int StartDecodingProcess(struct prrt_inc_list *ilist, struct decodingHelperContainer *dhcont);
static int DeleteDecodingRelatedData(struct prrt_inc_list *ilist, struct decodingHelperContainer *dhcont);

int TryDecode(struct prrt_inc_list *ilist, struct prrt_packet_container *pcont)
{
	int err = 0;

	int k = pcont->pack.red.red_header.k;
	int n = pcont->pack.red.red_header.n;

	struct decodingHelperContainer dhcont;
	struct packet_list_entry *found_packets[n];
	gf *PacketsToDecode[k];
	int PacketIndexes[k];

	if(ilist->decoding == OFF)
		return -1;

	setupDecodingHelperContainer(&dhcont, found_packets, PacketsToDecode,
					PacketIndexes, pcont);

	err = GatherDataPacketsOfBlock(ilist, &dhcont);
	if(err < 0)
		return err;

	if(DecodingShouldBeDone(ilist, &dhcont)) {
		//PDEBUG(MSG_TRANSPORT, "%s: Should decode\n", DEBUG_NAME);
		err = StartDecodingProcess(ilist, &dhcont);
	} else {
		//PDEBUG(MSG_TRANSPORT, "%s: Should not decode\n", DEBUG_NAME);
		DeleteDecodingRelatedData(ilist, &dhcont);
		err = -1;
	}

	//if(err == 0)
	//	PDEBUG(MSG_TRANSPORT, "%s: Decoding successful (block %u)\n", DEBUG_NAME, pcont->pack.red.red_header.sn_base)
	//else
	//	PDEBUG(MSG_FINAL, "%s: Decoding unsuccessful (block %u)\n", DEBUG_NAME, pcont->pack.red.red_header.sn_base);

	return err;
}

static u16 GetDataPacketIndex(struct prrt_packet_container *pcont, struct decodingHelperContainer *dhcont);

static int GatherDataPacketsOfBlock(struct prrt_inc_list *ilist, struct decodingHelperContainer *dhcont)
{
	int BlockStart = dhcont->BlockStart;
	int BlockEnd = dhcont->BlockEnd;
	struct packet_list_entry *p = 0;

	u16 indexDataPacket = 0;

	while(prrtdlist_get_data_packet_by_block(&ilist->data_list, &p, BlockStart, BlockEnd) == 0) {

		indexDataPacket = GetDataPacketIndex(p->pcont, dhcont);

		//PDEBUG(MSG_TRANSPORT, "%s: IndexDataPacket %u\n", DEBUG_NAME, indexDataPacket);
		dhcont->foundPackets[indexDataPacket] = p;
		dhcont->numberDataPacketsFound++;
	}

	return 0;
}

static u16 GetDataPacketIndex(struct prrt_packet_container *pcont, struct decodingHelperContainer *dhcont)
{
	u16 packetno = pcont->pack.data.header.seqnum;

	return (packetno-dhcont->BlockStart);
}

static int PacketsAreMissing(struct decodingHelperContainer *dhcont);
static int NotTooLateForDecoding(struct prrt_inc_list *ilist, struct decodingHelperContainer *dhcont);

static int DecodingShouldBeDone(struct prrt_inc_list *ilist, struct decodingHelperContainer *dhcont)
{
	return (PacketsAreMissing(dhcont) && NotTooLateForDecoding(ilist, dhcont));
}

static int PacketsAreMissing(struct decodingHelperContainer *dhcont)
{
	return (dhcont->numberDataPacketsFound < dhcont->k);
}

static int NotTooLateForDecoding(struct prrt_inc_list *ilist, struct decodingHelperContainer *dhcont)
{
	int err = 0;
	u16 last_read_pno = 0;

	err = prrtdlist_get_last_read_packetnumber(&ilist->data_list, &last_read_pno);

	return (err < 0 || prrt_comp_packetnos(last_read_pno, dhcont->BlockStart) < 0);
}

static int DecodeSingleMissAndCleanup(struct prrt_inc_list *ilist, struct decodingHelperContainer *dhcont);
static int GatherRedundancyPacketsOfBlock(struct prrt_inc_list *ilist,
					  struct decodingHelperContainer *dhcont);
static int TryDecodingMultipleMissesAndCleanup(struct prrt_inc_list *ilist,
					       struct decodingHelperContainer *dhcont);

static int StartDecodingProcess(struct prrt_inc_list *ilist, struct decodingHelperContainer *dhcont)
{
	int err = 0;
	//int interrupted = 0;
	int missing_packets = dhcont->k - dhcont->numberDataPacketsFound;

		PDEBUG(MSG_TRANSPORT, "%s: Missing %d packets (k = %d, received = %d)\n", DEBUG_NAME, missing_packets, dhcont->k, dhcont->numberDataPacketsFound);
	if(missing_packets == 1){
		   down(&ilist->red_lock);
		   err = DecodeSingleMissAndCleanup(ilist, dhcont);
		   up(&ilist->red_lock);
	}
	else {

		//TODO: Why down_interruptible?
		//interrupted = down_interruptible(&ilist->red_lock);
		//if(interrupted) {
		//	err = -EINTR;
		//	goto out;
		//}
		down(&ilist->red_lock);

		err = GatherRedundancyPacketsOfBlock(ilist, dhcont);
		if(err < 0)
			goto release_lock;

		err = TryDecodingMultipleMissesAndCleanup(ilist, dhcont);
		if(err < 0)
			goto release_lock;

release_lock:
		up(&ilist->red_lock);
	}

//out:
	return err;
}

static int PrepareDecodingAndDecode(struct prrt_inc_list *ilist, struct decodingHelperContainer *dhcont);
static int CleanupAfterDecodingFailure(struct prrt_inc_list *ilist,
				       struct decodingHelperContainer *dhcont);
static int CleanupOnSuccessfulDecoding(struct prrt_inc_list *ilist,
				       struct decodingHelperContainer *dhcont);
static int DecodeSingleMissAndCleanup(struct prrt_inc_list *ilist, struct decodingHelperContainer *dhcont)
{
	int err = 0;

	err = PrepareDecodingAndDecode(ilist, dhcont);
	if(err < 0) {
		CleanupAfterDecodingFailure(ilist, dhcont);
	} else {
		CleanupOnSuccessfulDecoding(ilist, dhcont);
	}


	return err;
}

static int PreparePacketsForDecoding(struct prrt_inc_list *ilist,
				     struct decodingHelperContainer *dhcont);
static int DoDecoding(struct prrt_inc_list *ilist, struct decodingHelperContainer *dhcont);

static int PrepareDecodingAndDecode(struct prrt_inc_list *ilist, struct decodingHelperContainer *dhcont)
{
	int err = 0;

	err = PreparePacketsForDecoding(ilist, dhcont);
	if(err < 0){
		//PDEBUG(MSG_TRANSPORT, "%s: Error preparing for decoding\n", DEBUG_NAME);
		return err;
}

	err = DoDecoding(ilist, dhcont);
	if(err < 0){
		//PDEBUG(MSG_TRANSPORT, "%s: Error on decoding\n", DEBUG_NAME);
		return err;
}

	return err;
}

static int InsertDataPacketInDecodingArray(struct decodingHelperContainer *dhcont, int index);
static int InsertNextRedundancyPacketInDecodingArray(struct decodingHelperContainer *dhcont, int index);

static int PreparePacketsForDecoding(struct prrt_inc_list *ilist,
				     struct decodingHelperContainer *dhcont)
{
	int index = 0;
	int err = 0;

	for(index=0; index<dhcont->k; ++index) {
		err = InsertDataPacketInDecodingArray(dhcont, index);
		//PDEBUG(MSG_TRANSPORT, "%s: Inserting data index %d\n", DEBUG_NAME, index);
		if(err < 0) {
			err = InsertNextRedundancyPacketInDecodingArray(dhcont, index);
		//PDEBUG(MSG_TRANSPORT, "%s: Inserting redundancy index %d\n", DEBUG_NAME, index);
			if(err < 0){
				//PDEBUG(MSG_TRANSPORT, "%s: Error inserting redundancy index %d\n", DEBUG_NAME, index);
				break;
			}
		}
	}

	return err;
}

static int PacketWithIndexWasFound(struct decodingHelperContainer *dhcont, int index);

static int InsertDataPacketInDecodingArray(struct decodingHelperContainer *dhcont, int index)
{
	struct prrt_packet_container *PacketToDecode = 0;
	if(PacketWithIndexWasFound(dhcont, index)) {
		PacketToDecode = (dhcont->foundPackets[index])->pcont;
		dhcont->PacketsToDecode[index] = prrt_packet_get_data(PacketToDecode);
		dhcont->PacketIndexes[index] = index;
		return 0;
	} else {
		return -1;
	}
}

static int PacketWithIndexWasFound(struct decodingHelperContainer *dhcont, int index)
{
	return (dhcont->foundPackets[index] != 0);
}

static int InsertNewRedundancyInDecodingArray(struct decodingHelperContainer *dhcont, int index);
static int InsertNextRedundancyPacketFoundInDecodingArray(struct decodingHelperContainer *dhcont, int index);

static int InsertNextRedundancyPacketInDecodingArray(struct decodingHelperContainer *dhcont, int index)
{
	int err = 0;

	if(!dhcont->newRedundancyAlreadyUsed) {
		err = InsertNewRedundancyInDecodingArray(dhcont, index);
	} else {
		err = InsertNextRedundancyPacketFoundInDecodingArray(dhcont, index);
	}

	return err;
}

static int InsertNewRedundancyInDecodingArray(struct decodingHelperContainer *dhcont, int index)
{
	dhcont->PacketsToDecode[index] = (gf *)prrt_packet_get_data(dhcont->newRedundancy);
	dhcont->PacketIndexes[index] = dhcont->newRedundancy->pack.red.header.index;
	dhcont->newRedundancyAlreadyUsed = 1;
/*
	PDEBUG(PRINT_PACKETHEX, "%s: New redundancy before decoding\n", DEBUG_NAME);
	prrt_packet_print_hexdata(dhcont->newRedundancy);
*/

	return 0;
}

static int InsertNextRedundancyPacketFoundInDecodingArray(struct decodingHelperContainer *dhcont, int index)
{
	struct packet_list_entry *nextRedundancyToUse = 0;
	char *nextDataToUse = 0;

	while(!nextRedundancyToUse && dhcont->lastRedundancyIndexUsed < dhcont->n -1)
		nextRedundancyToUse = dhcont->foundPackets[++dhcont->lastRedundancyIndexUsed];

	if(!nextRedundancyToUse)
		return -1;

	nextDataToUse = prrt_packet_get_data(nextRedundancyToUse->pcont);

	dhcont->PacketsToDecode[index] = (gf *)nextDataToUse;
	dhcont->PacketIndexes[index] = nextRedundancyToUse->pcont->pack.red.header.index;
	//prrt_packet_print_hexdata(nextRedundancyToUse->pcont);

	return 0;
}

static int DoDecoding(struct prrt_inc_list *ilist, struct decodingHelperContainer *dhcont)
{
	int err = 0;

	err = prrt_fec_get_coder(&ilist->decoder, dhcont->n, dhcont->k);
	if(err < 0) {
		PERROR("%s: Could not retrieve coder for decoding\n", DEBUG_NAME);
		return err;
	}

	err = prrt_fec_decode(ilist->decoder, dhcont->PacketsToDecode, dhcont->PacketIndexes,
				dhcont->newRedundancy->payload_size);
	if(err < 0) {
		PERROR("%s: Decoding failed\n", DEBUG_NAME);
		return err;
	}

	return err;
}

static void UpdateDecodingDelay(struct prrt_inc_list *ilist, struct decodingHelperContainer *dhcont);
static int InsertDecodedPacketsInDataList(struct prrt_inc_list *ilist,
					  struct decodingHelperContainer *dhcont);
static int DeleteNotNeededRedundancy(struct prrt_inc_list *ilist,
				     struct decodingHelperContainer *dhcont);

static int CleanupOnSuccessfulDecoding(struct prrt_inc_list *ilist,
					  struct decodingHelperContainer *dhcont)
{
	UpdateDecodingDelay(ilist, dhcont);

	InsertDecodedPacketsInDataList(ilist, dhcont);

	prrt_lossgath_deleteAllFB(&ilist->lossgath, dhcont->BlockStart, dhcont->BlockEnd);
	DeleteNotNeededRedundancy(ilist, dhcont);

	return 0;
}

static int GetFirstDataPacketFound(struct decodingHelperContainer *dhcont,
				   struct prrt_packet_container **pcont);

static void UpdateDecodingDelay(struct prrt_inc_list *ilist, struct decodingHelperContainer *dhcont)
{
	int err = 0;
	u32 now = prrt_get_prrttime(ilist->time_state);
	int lastDecodingDelay = 0;
	struct prrt_packet_container *pcont = 0;

	err = GetFirstDataPacketFound(dhcont, &pcont);
	if(err < 0)
		return;

	//PERROR("Setting decoding delay\n");
	lastDecodingDelay = prrt_diff_timestamps(ilist->time_state, now, pcont->time.creation_time);
	atomic_set(&ilist->lastDecodingDelay, lastDecodingDelay);
}

static int GetFirstDataPacketFound(struct decodingHelperContainer *dhcont,
				   struct prrt_packet_container **pcont)
{
	int i = 0;
	for(i=0; i<dhcont->k; ++i) {
		if(dhcont->foundPackets[i] != 0) {
			*pcont = dhcont->foundPackets[i]->pcont;
			return 0;
		}
	}

	*pcont = 0;
	return -1;
}

static int DecodedPacketAtIndexWasRedundancy(struct decodingHelperContainer *dhcontdhcont, int index);
static int InsertDecodedRedundancyinDataList(struct prrt_inc_list *ilist,
					     struct decodingHelperContainer *dhcont, int index);

static int InsertDecodedPacketsInDataList(struct prrt_inc_list *ilist,
					  struct decodingHelperContainer *dhcont)
{
	int err = 0;
	int index = 0;

	for(index=0; index<dhcont->k; ++index) {
		if(DecodedPacketAtIndexWasRedundancy(dhcont, index)) {
			PDEBUG(MSG_INCLIST, "%s: Inserting red packet with index %d in data list via "
				"data index %d\n",
				DEBUG_NAME, dhcont->PacketIndexes[index], index);
/*
			if(dhcont->PacketIndexes[index] == dhcont->newRedundancy->pack.red.red_header.index) {
				PDEBUG(PRINT_PACKETHEX, "%s: New redundancy after decoding\n", DEBUG_NAME);
				prrt_packet_print_hexdata(dhcont->newRedundancy);
			}
*/
			err = InsertDecodedRedundancyinDataList(ilist, dhcont, index);
		}
	}

	return err;
}

static int DecodedPacketAtIndexWasRedundancy(struct decodingHelperContainer *dhcont, int index)
{
	if(index < 0 || index >= dhcont->k)
		return 0;

	return (dhcont->PacketIndexes[index] >= dhcont->k);
}

//static int InsertNewRedundancyInDataList(struct prrt_inc_list *ilist,
//					 struct decodingHelperContainer *dhcont);
static int InsertOldRedundancyInDataList(struct prrt_inc_list *ilist,
					 struct decodingHelperContainer *dhcont, struct prrt_packet_container *pcont, int index, u16 pack_timeout);
//static void GiveRedPacketDataPacketNumber(struct prrt_packet_container *pcont, int index);

static void ConvertRedundancyToData(struct prrt_packet_container *redcont, struct prrt_packet_container *pcont, int seqnum, u16 pack_timeout);

static int InsertRecoveredPacketInDataList(struct prrt_inc_list *ilist,
					 struct prrt_packet_container *pcont);

static int InsertDecodedRedundancyinDataList(struct prrt_inc_list *ilist,
					     struct decodingHelperContainer *dhcont, int index)
{
	int err = 0;
	struct prrt_packet_container *pcont = 0;
	//u16 dataPacketNumber = 0;
	u16 pack_timeout = (u16)atomic_read(&ilist->lastPacketTimeout);

	pcont = prrt_packet_alloc(ilist->time_state);
	if (!pcont){
		PERROR("%s: Could not insert decoded packet in data list\n", DEBUG_NAME);
		return -1;
	}

	if((dhcont->PacketIndexes[index] == dhcont->newRedundancy->pack.red.header.index) && !dhcont->newRedundancyAlreadyInQueue) {
		//PDEBUG(MSG_TRANSPORT, "%s: Red packet is the new one\n", DEBUG_NAME);
		ConvertRedundancyToData(dhcont->newRedundancy, pcont, dhcont->newRedundancy->pack.red.red_header.sn_base+index, pack_timeout);
		//prrt_packet_destroy(&(dhcont->newRedundancy));
		//GiveRedPacketDataPacketNumber(dhcont->newRedundancy, index);
		//err = InsertNewRedundancyInDataList(ilist, pcont);
		//PDEBUG(MSG_TRANSPORT, "%s: Inserting Recovered Packet\n", DEBUG_NAME);
		//PDEBUG(MSG_TRANSPORT, "%s: Inserting Recovered Packet (new redundancy) (%u)\n", DEBUG_NAME, pcont->pack.data.header.seqnum);
		err = InsertRecoveredPacketInDataList(ilist, pcont);
		if(err == 0) {
			ilist->used_redpackets++;
		}
	} else {
		//PDEBUG(MSG_TRANSPORT, "%s: Inserting Recovered Packet (old redundancy) (%u)\n", DEBUG_NAME, pcont->pack.data.header.seqnum);
		//PDEBUG(MSG_TRANSPORT, "%s: Red packet is already in red list\n", DEBUG_NAME);
		err = InsertOldRedundancyInDataList(ilist, dhcont, pcont, index, pack_timeout);
	}

	if(err < 0)
		PERROR("%s: Could not insert decoded packet in data list\n", DEBUG_NAME);
	return err;
}

//static void GiveRedPacketDataPacketNumber(struct prrt_packet_container *pcont, int index)
//{
//	u16 dataPacketNumber = pcont->pack.red.red_header.sn_base+index;
//
//	pcont->pack.red.header.seqnum = dataPacketNumber;
//}

// because vdmcode writes back into parity packet's memory
static void ConvertRedundancyToData(struct prrt_packet_container *redcont, struct prrt_packet_container *pcont, int seqnum, u16 pack_timeout)
{
	
	//PDEBUG(MSG_TRANSPORT, "%s: Converting RedPacket to DataPacket: seqnum: %u, payload_size %u, creation_time %u \n", DEBUG_NAME, (u16)seqnum, redcont->payload_size, redcont->time.creation_time);
	//pcont->pack.data.data_header.pack_timeout = pack_timeout;
	//FIXME: Insert correct packet timeout (maybe from previous or successive packet). Need to include source packet header into coding process ...
	pcont->pack.data.data_header.pack_timeout = 0;
	prrt_packet_set_fields(pcont, (u16)seqnum, redcont->pack.red.header.index, redcont->time.creation_time, 0, redcont->payload_size, packtype_data);

	memcpy(&(pcont->pack.data.data), &(redcont->pack.red.data), redcont->payload_size);
}

//static int InsertNewRedundancyInDataList(struct prrt_inc_list *ilist,
//					 struct decodingHelperContainer *dhcont)
//{
//	int err = 0;
//
//	err = insert_datapacket(ilist, dhcont->newRedundancy);
//
//	if(err < 0)
//		prrt_packet_destroy(&(dhcont->newRedundancy));
//	else
//		ilist->useful_redpackets++;
//
//
//	return err;
//}

static int InsertRecoveredPacketInDataList(struct prrt_inc_list *ilist,
					 struct prrt_packet_container *pcont)
{
	int err = 0;

	err = insert_datapacket(ilist, pcont);

	if(err < 0){
		prrt_packet_destroy(&pcont);
	}
	else{
		ilist->useful_redpackets++;
	}

	return err;
}


static int InsertOldRedundancyInDataList(struct prrt_inc_list *ilist,
					 struct decodingHelperContainer *dhcont, struct prrt_packet_container *pcont, int index, u16 pack_timeout)
{
	int err = 0;
	struct packet_list_entry *DecodedEntry = 0;
	//struct prrt_packet_container *PacketToInsert = 0;

	DecodedEntry = dhcont->foundPackets[dhcont->PacketIndexes[index]];
	//PDEBUG(MSG_INCLIST, "%s: Decoded Entry = %p, index = %d\n", DEBUG_NAME, DecodedEntry,
	//	dhcont->PacketIndexes[index]);

	//PacketToInsert = DecodedEntry->pcont;

	//GiveRedPacketDataPacketNumber(PacketToInsert, index);
	ConvertRedundancyToData(DecodedEntry->pcont, pcont, DecodedEntry->pcont->pack.red.red_header.sn_base+index, pack_timeout);
	//prrt_packet_destroy(&(DecodedEntry->pcont));

	err = insert_datapacket(ilist, pcont);
	if(err < 0) {
		return err;
	} else {
		ilist->useful_redpackets++;
	}

	/*this is done after successful insertion because we can leave it in the list
	 *where it will be deleted at the end anyway if it was not inserted succesfully*/
	prrtredlist_remove_entry_from_list(&ilist->red_list, DecodedEntry);

	return err;
}

static int DeleteNotNeededRedundancy(struct prrt_inc_list *ilist,
				     struct decodingHelperContainer *dhcont)
{
	int err = 0;

	prrtredlist_delete_redundancy_packets_by_block_start(&ilist->red_list, dhcont->BlockStart);

	return err;
}

static int SaveNewRedundancyForLaterUse(struct prrt_inc_list *ilist,
					struct decodingHelperContainer *dhcont);

static int CleanupAfterDecodingFailure(struct prrt_inc_list *ilist,
				       struct decodingHelperContainer *dhcont)
{
	int err = 0;

	err = SaveNewRedundancyForLaterUse(ilist, dhcont);
	if(err < 0)
		prrt_packet_destroy(&dhcont->newRedundancy);

	return 0;
}

static int SaveNewRedundancyForLaterUse(struct prrt_inc_list *ilist,
					struct decodingHelperContainer *dhcont)
{
	int err = 0;

	err = insert_redpacket(ilist, dhcont->newRedundancy);

	return err;
}

static int GatherRedundancyPacketsOfBlock(struct prrt_inc_list *ilist,
					  struct decodingHelperContainer *dhcont)
{
	struct packet_list_entry *p = 0;
	int index = 0;

	while(prrtredlist_get_redundancy_packet_by_block_start(&ilist->red_list, &p, dhcont->BlockStart) == 0) {
		index = p->pcont->pack.red.header.index;

		dhcont->foundPackets[index] = p;
		dhcont->numberRedPacketsFound++;
	PDEBUG(MSG_TRANSPORT, "%s: Found red packet: %u\n", DEBUG_NAME, dhcont->numberRedPacketsFound);
	}

	return 0;
}

static int TryDecodingMultipleMisses(struct prrt_inc_list *ilist, struct decodingHelperContainer *dhcont);

static int TryDecodingMultipleMissesAndCleanup(struct prrt_inc_list *ilist,
					       struct decodingHelperContainer *dhcont)
{
	int err = 0;

	err = TryDecodingMultipleMisses(ilist, dhcont);
	if(err == 0) {
		CleanupOnSuccessfulDecoding(ilist, dhcont);
	} else {
		CleanupAfterDecodingFailure(ilist, dhcont);
	}

	return err;
}

static int FoundEnoughRedundancyToDecode(struct decodingHelperContainer *dhcont);

static int TryDecodingMultipleMisses(struct prrt_inc_list *ilist, struct decodingHelperContainer *dhcont)
{
	int err = 0;

	if(FoundEnoughRedundancyToDecode(dhcont)) {
		PDEBUG(MSG_FBADEB, "%s: Found enough redundancy to decode block %d\n", DEBUG_NAME, dhcont->BlockStart);
		err = PrepareDecodingAndDecode(ilist, dhcont);
	} else {
		PDEBUG(MSG_FBADEB, "%s: Did not find enough redundancy to decode block %d\n", DEBUG_NAME, dhcont->BlockStart);
		err = -1;
	}

	return err;
}

static int FoundEnoughRedundancyToDecode(struct decodingHelperContainer *dhcont)
{
	int numberMisses = dhcont->k - dhcont->numberDataPacketsFound;
		PDEBUG(MSG_FBADEB, "%s: Number misses %d, number red packets %d\n", DEBUG_NAME, numberMisses, dhcont->numberRedPacketsFound);

	return (dhcont->numberRedPacketsFound >= numberMisses);
}

static int DeleteDecodingRelatedData(struct prrt_inc_list *ilist, struct decodingHelperContainer *dhcont)
{
	prrt_lossgath_deleteAllFB(&ilist->lossgath, dhcont->BlockStart, dhcont->BlockEnd);
	prrtredlist_delete_redundancy_packets_by_block_start(&ilist->red_list, dhcont->BlockStart);
	if (!dhcont->newRedundancyAlreadyInQueue)
		prrt_packet_destroy(&dhcont->newRedundancy);

	return 0;
}



static int insert_redpacket(struct prrt_inc_list *ilist, struct prrt_packet_container *pcont)
{
	long block_start = pcont->pack.red.red_header.sn_base;
	long block_end = prrt_increment_packetno(block_start, pcont->pack.red.red_header.k-1);
	//u8 round = pcont->pack.red.red_header.round;
	int err = 0;
	int measured_t_replace = 0;

	PDEBUG(MSG_FBADEB, "%s: Inserting parity for block %ld for later use\n", DEBUG_NAME, block_start);
 	err = prrtredlist_insert_redundancy_packet(&ilist->red_list, pcont);
	if(err < 0){
		PDEBUG(MSG_TRANSPORT, "%s: Error inserting parity for block %ld\n", DEBUG_NAME, block_start);
		return err;
	}

	ilist->used_redpackets++;

	//printk(KERN_EMERG "Deleting loss entry\n");
	//PDEBUG(MSG_TRANSPORT, "%s: Deleting feedback for block %lu\n", DEBUG_NAME, block_start);
	prrt_lossgath_deletefb(&ilist->lossgath, block_start, block_end, 0, &measured_t_replace);
	//PDEBUG(MSG_FINAL, "%s: Deleting feedback\n", DEBUG_NAME);
	//if(measured_t_replace){
	////PDEBUG(MSG_TRANSPORT, "%s: Updating replace time = %d\n", DEBUG_NAME, measured_t_replace);
	//	atomic_set(&ilist->lastMeasured_t_replace, measured_t_replace);
	//	prrtnpar_update_ReplaceTime(&ilist->net_params, measured_t_replace);
	//}

	return err;
}

int prrtilist_set_option(struct prrt_inc_list *ilist, enum option_t opt, int val)
{
	int interrupted = 0;
	int err = 0;

	interrupted = down_interruptible(&ilist->opt_lock);
	if(interrupted) {
		err = -EINTR;
		goto out;
	}

	switch(opt) {
	case FEEDBACK:
		ilist->feedback = val;
		break;
	case DECODE:
		ilist->decoding = val;
		break;
  case FB_SUPP_ON:
    ilist->lossgath.max_fb_send_delay = val;
    break;
	default:
		PERROR("%s: Option not recognized", DEBUG_NAME);
		err = -1;
		break;
	}

	up(&ilist->opt_lock);

out:
	return err;
}

int prrtilist_get_option(struct prrt_inc_list *ilist, enum option_t opt)
{
	int interrupted = 0;
	int err = 0;

	interrupted = down_interruptible(&ilist->opt_lock);
	if(interrupted) {
		err = -EINTR;
		goto out;
	}

	switch(opt) {
	case FEEDBACK:
		err = ilist->feedback;
		break;
	case DECODE:
		err = ilist->decoding;
		break;
  case FB_SUPP_ON:
    err = ilist->lossgath.max_fb_send_delay;
    break;
	default:
		PERROR("%s: Option not recognized", DEBUG_NAME);
		err = -1;
		break;
	}

	up(&ilist->opt_lock);

	out:
	return err;
}

