/* 
 * (c) 2010 Saarland University, Telecommunications Lab 
 * (prrt@nt.uni-saarland.de)
 *
 * The contents of this file are subject to the Mozilla Public License
 + Version 1.1 (the "License"); you may not use this file except in
 + compliance with the License. You may obtain a copy of the License at
 + http://www.mozilla.org/MPL/
 + 
 + Software distributed under the License is distributed on an "AS IS"
 + basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 + License for the specific language governing rights and limitations
 + under the License.
 + 
 + The Original Code is "Predictably Reliable Real-time Transport (PRRT)".
 + 
 + The Initial Developer of the Original Code is Saarland University.
 + All Rights Reserved.
 + 
 + Contributor(s): Jochen Gruen (Saarland University), Manuel Gorius
 + (Saarland University), Yongtao Shuai (Saarland University).
 + 
 + Disclaimer: This Alpha version of PRRT is not supported and stable yet.
 + Do not use it in a production environment: it may crash your device, 
 + so please make sure that you have a backup of your data before
 + installing PRRT.
 */

#include "prrt_out_list.h"

#include "../kernel_wrappers.h"
#include <linux/random.h>
#include <linux/sched.h>

#define DEBUG_NAME "PRRT_OUT_LIST"

// TBF for smooth sending rate
#define RATE 4
#define LIMIT 7000

static int get_next_replacement_packet(struct prrt_out_list *olist, struct prrt_block *mblock,
        int packetno, struct prrt_packet_container **pcont);
// static int get_next_packet(struct prrt_out_list *olist, struct prrt_packet_container **pcont);
#if SIMULATION_ON_SEND
#if DROP_PACKS_ON_SEND
static int get_random_percent(void);
#endif
#endif

void prrtolist_init(struct prrt_out_list *olist, struct prrt_time_state *time_state)
{
    prrtblist_init(&olist->block_list);
    olist->next_block = &olist->block_list.head;
    init_MUTEX(&olist->list_lock);
    init_MUTEX_LOCKED(&olist->coded_block_count);
    olist->out_cod = 0;

    init_MUTEX(&olist->opt_lock);
    olist->replace = ON;
    olist->coding = ON;
    olist->gather_feedback = ON;

    olist->send_packets = 0;
    olist->send_replacements = 0;
    olist->old_blocks = 0;

    olist->out_pno_red = 0;
    init_MUTEX(&olist->out_pno_lock);

    prrtrlist_init(&olist->recv_list);
    prrt_fbana_init(&olist->fbana, time_state);

    olist->time_state = time_state;
    olist->last_sent = prrt_get_ustime(time_state);
    olist->send_tokens = 0;
    olist->traffic_shaping_datarate = SENDING_RATE; // Default 10 MBytes/s see defines.h
}

int prrtolist_reinit(struct prrt_out_list *olist)
{
    int err = 0;

    PDEBUG(MSG_INITSEND, "%s: Reinitialising out_packet_list\n", DEBUG_NAME);

    if (down_interruptible(&olist->list_lock))
    {
        err = -EINTR;
        goto out;
    }

    err = prrtblist_reinit(&olist->block_list);
    if (err < 0)
        goto out;

    olist->next_block = &olist->block_list.head;
    init_MUTEX_LOCKED(&olist->coded_block_count);

    up(&olist->list_lock);

    if (down_interruptible(&olist->opt_lock))
    {
        err = -EINTR;
        goto out;
    }

    olist->replace = ON;
    olist->coding = ON;
    olist->gather_feedback = ON;

    up(&olist->opt_lock);

    olist->send_packets = 0;
    olist->send_replacements = 0;
    olist->old_blocks = 0;

    if (down_interruptible(&olist->out_pno_lock))
    {
        err = -EINTR;
        goto out;
    }

    olist->out_pno_red = 0;

    up(&olist->out_pno_lock);

    err = prrtrlist_reinit(&olist->recv_list);
    if (err < 0)
        goto out;

    prrt_fbana_reinit(&olist->fbana);

    olist->last_sent = prrt_get_ustime(olist->time_state);
    olist->send_tokens = 0;

out:
    return err;
}

void prrtolist_free(struct prrt_out_list *olist)
{
    prrtblist_destroy_all_blocks(&olist->block_list);
    prrt_fec_free(olist->out_cod);
    prrt_fbana_free(&olist->fbana);
}

int AllocateAndInsertNewBlock(struct prrt_out_list *olist, struct params_code *cpars);

int prrtolist_insert_packet(struct prrt_out_list *olist, struct prrt_packet_container *pcont, struct params_code *cpars)
{
    int err = 0;

    if (down_interruptible(&olist->list_lock))
        return -1;

    if (!prrtblist_can_insert_packet(&olist->block_list))
    {
        err = AllocateAndInsertNewBlock(olist, cpars);
        if (err < 0)
            goto error;
    }

    err = prrtblist_insert_packet(&olist->block_list, pcont, cpars);
    if (err < 0)
        goto error;

error:
    up(&olist->list_lock);

    return err;
}

int AllocateAndInsertNewBlock(struct prrt_out_list *olist, struct params_code *cpars)
{
    struct prrt_block *new_block = 0;
    int err = 0;

    err = prrt_block_alloc(&new_block, cpars);
    if (err < 0)
    {
        PERROR("%s: Could not allocate new block\n", DEBUG_NAME);
        return -1;
    }

    PDEBUG(MSG_OUTLIST, "%s: Allocated new block %p\n", DEBUG_NAME, new_block);

    err = prrtblist_insert_block(&olist->block_list, new_block);
    if (err < 0)
    {
        PERROR("%s: Could not insert new block\n", DEBUG_NAME);
        return -1;
    }

    return 0;
}

int GetNextUncodedBlock(struct prrt_out_list *olist, struct prrt_block **next_block);
int CodeNextBlock(struct prrt_out_list *olist);
void SetNextBlockIfNecessary(struct prrt_out_list *olist, struct prrt_block *next_block);

int prrtolist_try_to_code_next_block(struct prrt_out_list *olist, struct params_code *cpars)
{
    int err = 0;
    struct prrt_block *next_block = 0;
    struct prrt_block *new_block = 0;

    if (down_interruptible(&olist->list_lock))
        return -1;

    err = GetNextUncodedBlock(olist, &next_block);
    if (err < 0)
        goto error;
    PDEBUG(MSG_BLOCKDEB, "%s: List head %p\n", DEBUG_NAME, &olist->block_list.head);
    PDEBUG(MSG_CODING, "%s: next_uncoded_block=%p\n", DEBUG_NAME, next_block);

    PDEBUG(MSG_SENDBLOCK, "%s: new_block = %p, should be zero\n", DEBUG_NAME, new_block);
    err = prrt_block_prepare(next_block, cpars, &new_block, olist->time_state);
    if (err < 0)
    {
        //PERROR("%s: Preparing block failed\n", DEBUG_NAME);
        goto error;
    }
    PDEBUG(MSG_SENDBLOCK, "%s: Preparing block successful\n", DEBUG_NAME);
    PDEBUG(MSG_SENDBLOCK, "%s: new_block = %p, should be new block\n", DEBUG_NAME, new_block);
    PDEBUG(MSG_SENDBLOCK, "%s: block to code = %p\n", DEBUG_NAME, next_block);
    PDEBUG(MSG_SENDBLOCK, "%s: no_data = %d\n", DEBUG_NAME, next_block->no_data);
    PDEBUG(MSG_SENDBLOCK, "%s: no_red = %d\n", DEBUG_NAME, next_block->no_red);

    err = CodeNextBlock(olist);

    PDEBUG(MSG_SENDBLOCK, "%s: new_block = %p\n", DEBUG_NAME, new_block);

    prrtblist_insert_block(&olist->block_list, new_block);

    SetNextBlockIfNecessary(olist, next_block);
    next_block = container_of(olist->next_block, struct prrt_block, list);
    PDEBUG(MSG_CODING, "%s: next_block=%p\n", DEBUG_NAME, next_block);
    next_block = 0;
    GetNextUncodedBlock(olist, &next_block);
    PDEBUG(MSG_CODING, "%s: next_uncoded_block=%p\n", DEBUG_NAME, next_block);

error:
    up(&olist->list_lock);

    return err;
}

int ThereIsABlockToCode(struct prrt_out_list *olist);

int GetNextUncodedBlock(struct prrt_out_list *olist, struct prrt_block **next_block)
{
    if (!ThereIsABlockToCode(olist))
    {
        //PERROR("%s: No codable blocks found\n", DEBUG_NAME);
        return -1;
    }

    return prrtblist_get_last_block(&olist->block_list, next_block);
}

int ThereIsABlockToCode(struct prrt_out_list *olist)
{
    int err = 0;
    struct prrt_block *last_block;

    err = prrtblist_get_last_block(&olist->block_list, &last_block);
    if (err < 0)
        return 0;

    return (!prrt_block_is_already_coded(last_block));
}

int CodeNextBlock(struct prrt_out_list *olist)
{
    int err = 0;
    struct prrt_block *next_block = 0;
    struct packet_list_entry *first_data = 0;

    err = prrtblist_get_last_block(&olist->block_list, &next_block);
    if (err < 0)
        return -1;

    err = prrt_block_get_coder(next_block, &olist->out_cod);
    if (err < 0)
    {
        //PERROR("%s: Could not retrieve coder for block\n", DEBUG_NAME);
        return -1;
    }

    err = prrt_block_code(next_block, olist->out_cod, &olist->out_pno_red, &olist->out_pno_lock, olist->time_state);
    if (err < 0)
    {
        PERROR("%s: Could not code block\n", DEBUG_NAME);
        return -1;
    }

    up(&olist->coded_block_count);
    prrt_block_get_first_data(next_block, &first_data);

    PDEBUG(MSG_SENDFBDEB, "%s: Block coded and counted(start %u) (at %u)\n", DEBUG_NAME,
            first_data->pcont->pack.data.header.seqnum, prrt_get_prrttime(olist->time_state));

    return 0;
}

int ThereIsACodedBlock(struct prrt_out_list *olist);
void SetNextBlock(struct prrt_out_list *olist, struct prrt_block *next_block);

void SetNextBlockIfNecessary(struct prrt_out_list *olist, struct prrt_block *next_block)
{
    if (!ThereIsACodedBlock(olist))
        SetNextBlock(olist, next_block);
}

void SetNextBlock(struct prrt_out_list *olist, struct prrt_block *next_block)
{
    PDEBUG(MSG_SENDBLOCK, "%s: Setting next_block to %p\n", DEBUG_NAME, next_block);
    olist->next_block = &next_block->list;
}

int ThereIsACodedBlock(struct prrt_out_list *olist)
{
    struct prrt_block *next_block = 0;

    if (olist->next_block == &olist->block_list.head)
        return 0;

    next_block = container_of(olist->next_block, struct prrt_block, list);

    return (prrt_block_is_already_coded(next_block));
}

static int send_packet(struct prrt_out_list *olist, struct socket *sock,
        struct sockaddr_in *paddr, struct prrt_packet_container *pcont);

int getSendTokens(struct prrt_out_list *olist, int msglen);

int prrtolist_send_next_packet(struct prrt_out_list *olist, struct socket *sock,
        struct sockaddr_in *paddr)
{
    int err = 0;
    int msglen = 0;
    int send_tokens = 0;
    struct prrt_packet_container *pcont = 0;

    down(&olist->list_lock);

    err = prrtblist_get_next_packet(&olist->block_list, &pcont);
    if (err < 0)
    {
        PERROR("%s: Why do we send a packet although there are none\n", DEBUG_NAME);
        err = -1;
        goto done;
    }

    msglen = PRRT_HEADER_SIZE + PRRT_DATA_HEADER_SIZE + pcont->payload_size;
    PDEBUG(MSG_SENDDEB, "%s: pcont = %p\n", DEBUG_NAME, pcont);

    //PDEBUG(MSG_FINAL, "%s: Sending data packet ...\n", DEBUG_NAME);
	send_tokens = getSendTokens(olist, msglen);
//	if (!send_tokens){
//		up(&olist->list_lock);
//
//		//PERROR("%s: Sleeping some time (data)\n", DEBUG_NAME);
//		set_current_state(TASK_INTERRUPTIBLE);
//		schedule_timeout(ms_to_jif(1));
//
//		down(&olist->list_lock);
//
//		send_tokens = getSendTokens(olist, msglen);
//	}

    err = send_packet(olist, sock, paddr, pcont);
    if (err < 0)
        goto done;
    else
        prrtblist_confirm_packet_sending(&olist->block_list);

done:
    up(&olist->list_lock);

    return err;
}

static int send_packet(struct prrt_out_list *olist, struct socket *sock,
        struct sockaddr_in *paddr, struct prrt_packet_container *pcont)
{
    /*need to hold lock here*/
    int err = 0;
    // 	struct prrt_packet_container *pcont = 0;
    int packetlen = 0;
#if SIMULATION_ON_SEND
#if DROP_PACKS_ON_SEND
    int sendit = 0;
#endif /*DROP_PACKS_ON_SEND*/
#endif

    // 	down(&olist->list_lock);

    PRINT_TRACE("SEND", "S/OBL", "get_nextpack");
    // 	err = get_next_packet(olist, &pcont);
    // 	if(err < 0)
    // 		goto out;

    packetlen = PRRT_HEADER_SIZE + PRRT_DATA_HEADER_SIZE + pcont->payload_size;
    //PDEBUG(MSG_FINAL, "%s: Packetlength: %d\n", DEBUG_NAME, packetlen);

    /*this is only to get a good printk result*/
    *(&pcont->pack.data.data + pcont->payload_size) = '\0';
    PDEBUG(MSG_PRINTDATA, "%s: Sending message: %s\n", DEBUG_NAME,
            &pcont->pack.data.data);

    /*this simulates a channel with LOSS_RATE% lost packets on
     *localhost*/
    PRINT_TRACE("SEND", "S", "send_pack");

    /*set actual timestamp, needed for RTT estimation*/
    pcont->time.transmit_time = prrt_get_prrttime(olist->time_state);
    //PDEBUG(MSG_TRANSPORT, "%s: Transmit_time: %lu\n", DEBUG_NAME, pcont->time.transmit_time);

#if SIMULATION_ON_SEND
#if DROP_PACKS_ON_SEND
//    sendit = get_random_percent();
//    if (sendit > LOSS_RATE)
//    {
//        /*convert to network byteorder only for sending*/
//        prrt_packet_hton(pcont);
//        err = wrap_kernel_sendmsg(sock, paddr, &pcont->pack, packetlen);
//        prrt_packet_ntoh(pcont);
//	prrt_packet_print_header(pcont, MSG_FINAL);
//    }
//    else
//    {
//        err = 0;
//        PDEBUG(MSG_PRINTDATA, "%s: Channel dropped message: %s\n", DEBUG_NAME,
//                &pcont->pack.data.data);
//    }
#else
//    /*convert to network byteorder only for sending*/
//    prrt_packet_hton(pcont);
//    err = wrap_kernel_sendmsg(sock, paddr, &pcont->pack, packetlen);
//    prrt_packet_ntoh(pcont);
//    prrt_packet_print_header(pcont, MSG_FINAL);
#endif /*DROP_PACKS_ON_SEND*/
#else
    prrt_packet_hton(pcont);
    err = wrap_kernel_sendmsg(sock, paddr, &(pcont->pack), packetlen);
    prrt_packet_ntoh(pcont);
    //prrt_packet_print_header(pcont, MSG_FINAL);
#endif /*SIMULATION_ON_SEND*/

    if (err < 0)
    {
        PERROR("%s: Error on send (no %d)\n", DEBUG_NAME, err);
    }
    else
    {
        //PDEBUG(MSG_SENDFBDEB, "%s: Data Packet no %u was sent\n", DEBUG_NAME,
        //pcont->pack.data.header.packetno);
        olist->send_packets++;
    }

    // out:
    // 	up(&olist->list_lock);

    return err;
}

int prrtolist_wait_for_next_packet(struct prrt_out_list *olist)
{
    int ret = 0;

    //down(&olist->block_list.packet_count);
    ret = down_timeout(&olist->block_list.packet_count, ms_to_jif(5));

    PDEBUG(MSG_SENDBLOCK, "%s: Grabbed packet, now only %u left\n", DEBUG_NAME,
            olist->block_list.packet_count.count);

    return ret;
}

void AdvanceNextBlockPointer(struct prrt_out_list *olist);
static int send_pre_red_packet(struct prrt_out_list *olist, struct socket *sock,
        struct sockaddr_in *paddr, struct prrt_packet_container *pcont);

int SendRedundancyOfNextCodedBlock(struct prrt_out_list *olist, struct socket *sock,
        struct sockaddr_in *paddr);

int prrtolist_try_to_send_redundancy_of_next_coded_block(struct prrt_out_list *olist, struct socket *sock,
        struct sockaddr_in *paddr)
{
    int err = 0;

    if (down_trylock(&olist->coded_block_count) == 0)
    {
        err = SendRedundancyOfNextCodedBlock(olist, sock, paddr);
        if (err < 0)
        {
            if (err != -2)
                //PERROR("%s: Could not send redundancy of next block\n", DEBUG_NAME);
            up(&olist->coded_block_count);
            return err;
        }
    }
    else
    {
        //PDEBUG(MSG_SENDFBDEB, "%s: Did not find block that was ready for preredundancy "
        //"sending\n", DEBUG_NAME);
        err = -1;
    }

    return err;
}

void PrepareRedundancyForSending(struct prrt_out_list *olist, struct prrt_packet_container *pcont);

int SendRedundancyOfNextCodedBlock(struct prrt_out_list *olist, struct socket *sock,
        struct sockaddr_in *paddr)
{
    struct prrt_block *next_block = 0;
    struct packet_list_entry *p = 0;
    int n_p = 0;
    int i = 0;
    int err = 0;
    int msglen = 0;
    int send_tokens = 0;

    down(&olist->list_lock);

    if (!ThereIsACodedBlock(olist))
    {
        PERROR("%s: No coded block found\n", DEBUG_NAME);
        err = -1;
        goto done;
    }

    next_block = container_of(olist->next_block, struct prrt_block, list);

    prrt_block_get_first_data(next_block, &p);

    PDEBUG(MSG_SENDBLOCK, "%s: next_block = %p\n", DEBUG_NAME, next_block);
    PDEBUG(MSG_SENDBLOCK, "%s: next_block->no_data = %d\n", DEBUG_NAME, next_block->no_data);
    PDEBUG(MSG_SENDBLOCK, "%s: next_block->no_red = %d\n", DEBUG_NAME, next_block->no_red);

    n_p = next_block->cpars.n_p;
    PDEBUG(MSG_SENDBLOCK, "%s: next_block->n_p = %d\n", DEBUG_NAME, n_p);

    if (!prrtblist_block_was_send_completely(&olist->block_list, next_block))
    {
        err = -2;
        PERROR("%s: Block was not sent completely yet(no %u)\n", DEBUG_NAME, p->pcont->pack.data.header.seqnum);
        goto done;
    }

    i = 0;
    /*only send n_p packets at most*/
    while (i < n_p && prrt_block_get_next_redundancy_packet(next_block, &p) == 0)
    {

        //PrepareRedundancyForSending(olist, p->pcont);
        msglen = PRRT_HEADER_SIZE + PRRT_RED_HEADER_SIZE + p->pcont->payload_size;

	//PDEBUG(MSG_FINAL, "%s: Sending parity packet ...\n", DEBUG_NAME);
	send_tokens = getSendTokens(olist, msglen);
	if (!send_tokens){
		up(&olist->list_lock);

		//PERROR("%s: Sleeping some time\n", DEBUG_NAME);
		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(ms_to_jif(1));

		down(&olist->list_lock);

		send_tokens = getSendTokens(olist, msglen);
	}

        err = send_pre_red_packet(olist, sock, paddr, p->pcont);
        if (err < 0)
        {
            goto done;
        }

        ++i;
    }

    next_block->first_red_send = 1;
    AdvanceNextBlockPointer(olist);

    //PDEBUG(MSG_SENDFBDEB, "%s: Pre Redundancy of block with start %u sent\n", DEBUG_NAME,
    //	firstpackno);

done:
    up(&olist->list_lock);

    return err;
}

void PrepareRedundancyForSending(struct prrt_out_list *olist, struct prrt_packet_container *pcont)
{
    //pcont->pack.red.red_header.t_replace = prrtnpar_get_ReplaceTime(&olist->recv_list.worstNetParams);
}

int SendPreRedundancyOrReplacement(struct prrt_out_list *olist, struct prrt_packet_container *pcont, struct socket *sock,
        struct sockaddr_in *paddr);

/*send replacement packet in the first round*/
static int send_pre_red_packet(struct prrt_out_list *olist, struct socket *sock,
        struct sockaddr_in *paddr, struct prrt_packet_container *pcont)
{
    /*need to hold lock here*/
    int err = 0;

    PDEBUG(MSG_RECVFBDEB, "%s: Sending pre redundancy packet no %d\n", DEBUG_NAME,
            pcont->pack.red.header.index);
    pcont->pack.red.header.packtype = (pcont->pack.red.header.packtype & ~0xF) | ((u8)packtype_prered & 0xF);
    //pcont->pack.red.red_header.round = 0;
    err = SendPreRedundancyOrReplacement(olist, pcont, sock, paddr);
	//FIXME: Why to set packtype again?
    pcont->pack.red.header.packtype = (pcont->pack.red.header.packtype & ~0xF) | ((u8)packtype_red & 0xF);
    if (err < 0)
    {
        PERROR("%s: Failed to send a pre redundancy packet of coded block\n", DEBUG_NAME);
    }
    else
    {
        olist->send_replacements++;
    }

    return err;
}

int SendPreRedundancyOrReplacement(struct prrt_out_list *olist, struct prrt_packet_container *pcont, struct socket *sock,
        struct sockaddr_in *paddr)
{
    int err = 0;
    int msglen = 0;
    u32 time = 0;

#if SIMULATION_ON_SEND
#if DROP_REPACKS_ON_SEND
    int sendit = 0;
#endif /*DROP_FBPACKS_ON_SEND*/
#endif /*SIMULATION_ON_SEND*/

    /* we have 2 options for retransmissions
     * 1. we send data packets if there are no coded packets
     * 2. we send redundancy packets if there are some*/
    if (!prrt_packet_is_redundancy(pcont))
    {
	pcont->pack.data.header.packtype = (pcont->pack.data.header.packtype & ~0xF) | ((u8)packtype_redata & 0xF);
        msglen = PRRT_HEADER_SIZE + PRRT_DATA_HEADER_SIZE + pcont->payload_size;
    }
    else
    {
        msglen = PRRT_HEADER_SIZE + PRRT_RED_HEADER_SIZE + pcont->payload_size;
    }

#if SIMULATION_ON_SEND
#if DROP_REPACKS_ON_SEND
//    sendit = get_random_percent();
//    if (sendit > LOSS_RATE)
//    {
//        /*convert to network byteorder only for sending*/
//        prrt_packet_hton(pcont);
//        err = wrap_kernel_sendmsg(sock, paddr, &pcont->pack, msglen);
//        prrt_packet_ntoh(pcont);
//	prrt_packet_print_header(pcont, MSG_FINAL);
//    }
//    else
//    {
//        err = 0;
//        PDEBUG(MSG_COMMONSEND, "%s: Channel dropped replacement packet no %u\n", DEBUG_NAME,
//                pcont->pack.data.data_header.seqnum);
//    }
#else
//    prrt_packet_hton(pcont);
//    err = wrap_kernel_sendmsg(sock, paddr, &pcont->pack, msglen);
//    prrt_packet_ntoh(pcont);
//    prrt_packet_print_header(pcont, MSG_FINAL);
#endif /*DROP_REPACKS_ON_SEND*/
#else
    prrt_packet_hton(pcont);
    err = wrap_kernel_sendmsg(sock, paddr, &pcont->pack, msglen);
    prrt_packet_ntoh(pcont);
    //prrt_packet_print_header(pcont, MSG_FINAL);
#endif /*SIMULATION_ON_SEND*/

    if (!prrt_packet_is_redundancy(pcont))
	pcont->pack.data.header.packtype = (pcont->pack.data.header.packtype & ~0xF) | ((u8)packtype_data & 0xF);
    else
    {

        time = prrt_get_prrttime(olist->time_state);

        PDEBUG(MSG_FBANADEB, "%s: Redundancy Packet no %d was sent (at %u)\n", DEBUG_NAME,
                pcont->pack.red.red_header.sn_base, time);
    }

    if (err < 0)
    {
        PERROR("%s: Error on send (no %d)\n", DEBUG_NAME, err);
    }

    return err;
}

void AdvanceNextBlockPointer(struct prrt_out_list *olist)
{
    struct prrt_block *next_block = 0;

    if (olist->next_block == &olist->block_list.head)
    {
        PERROR("%s: Already at the end of block list, can not advance next_block\n", DEBUG_NAME);
        return;
    }

    olist->next_block = olist->next_block->next;

    if (olist->next_block == &olist->block_list.head)
    {
        PDEBUG(MSG_SENDBLOCK, "%s: Setting next_block to list_head\n", DEBUG_NAME);
    }
    else
    {
        next_block = container_of(olist->next_block, struct prrt_block, list);
        PDEBUG(MSG_SENDBLOCK, "%s: Setting next_block to %p\n", DEBUG_NAME, next_block);
    }
}

int prrtolist_delete_old_blocks(struct prrt_out_list *olist, u32 d_target)
{
    struct prrt_block *mblock_first = 0;
    u32 timestamp_last = 0;
    int err = 0;

    if (down_interruptible(&olist->list_lock))
        return -1;

    PDEBUG(MSG_COMMONSEND, "%s: Trying to get timestamp of last block\n", DEBUG_NAME);
    err = prrtblist_get_timestamp_of_last_packet(&olist->block_list, &timestamp_last);
    if (err < 0)
    {
        //PERROR("%s: Could not retrieve timestamp of last packet\n", DEBUG_NAME);
        goto out;
    }

    PDEBUG(MSG_COMMONSEND, "%s: Deleting old blocks\n", DEBUG_NAME);
    while (prrtblist_get_first_block(&olist->block_list, &mblock_first) == 0)
    {
        if (!prrt_block_too_old_than_timestamp(mblock_first, timestamp_last, d_target, olist->time_state))
            break;

        err = prrtblist_try_to_delete_first_block(&olist->block_list);
        if (err < 0)
            break;
    }

    PDEBUG(MSG_COMMONSEND, "%s: Deleting blocks done\n", DEBUG_NAME);

out:
    up(&olist->list_lock);

    return 0;
}

static int send_replacement(struct prrt_out_list *olist, struct socket *sock,
        struct sockaddr_in *paddr, struct fbana_replacement_entry *rple);

int prrtolist_send_replacements(struct prrt_out_list *olist, struct socket *sock,
        struct sockaddr_in *paddr)
{
    struct fbana_replacement_entry *rple = 0;

    if (olist->replace == OFF)
    {
        prrt_fbana_free_feedbacks(&olist->fbana);
        PERROR("%s: Replacement sending is OFF, will not send replacements\n", DEBUG_NAME);
        return 0;
    }

    down(&olist->list_lock);


    prrt_fbana_compute_replacements(&olist->fbana, &olist->block_list, &olist->recv_list);

    //PDEBUG(MSG_SENDREPL, "%s: Starting to send replacements\n", DEBUG_NAME);
    while (prrt_fbana_get_next_replacement(&olist->fbana, &rple) == 0)
    {
        //PDEBUG(MSG_SENDREPL, "%s: Feedback found, going to start processing (first_packno %u)\n", DEBUG_NAME, rple->first_packno);
        send_replacement(olist, sock, paddr, rple);
        prrt_fbana_free_replacement_entry(&rple);
    }

    up(&olist->list_lock);

    return 0;
}

int get_block(struct prrt_out_list *olist, int packetno, struct prrt_block **mblock);

static int send_replacement(struct prrt_out_list *olist, struct socket *sock,
        struct sockaddr_in *paddr, struct fbana_replacement_entry *rple)
{
    int err = 0;
    struct prrt_block *mblock = 0;
    struct prrt_packet_container *pcont = 0;
    int i = 0;
    int n_cc = 1;

    if (get_block(olist, rple->first_packno, &mblock) < 0)
    {
        PERROR("%s: Could not find block for replacements\n", DEBUG_NAME);
        err = -1;
        goto out;
    }

    n_cc = rple->max_n_cc;

    //PDEBUG(MSG_SENDREPL, "%s: number replacements needed %d\n", DEBUG_NAME, rple->no_repl);
    //PDEBUG(MSG_SENDREPL, "%s: replacement multiplier %d\n", DEBUG_NAME, n_cc);

    /*send the number of replacements needed*/
    for (i = 0; i < rple->no_repl * n_cc; i++)
    {
        //PDEBUG(MSG_SENDREPL, "%s: Getting next replacement packet\n", DEBUG_NAME);
        if (get_next_replacement_packet(olist, mblock, rple->first_packno, &pcont) < 0)
        {
            PERROR("%s: Did not find replacement packet\n", DEBUG_NAME);
            err = -1;
            goto out;
        }

        //PDEBUG(MSG_SENDREPL, "%s: Sending replacement packet no %d in round %d\n", DEBUG_NAME,
        //        pcont->pack.red.header.index, rple->retr_round);
	//pcont->pack.red.red_header.round = (u8) rple->retr_round;
        err = SendPreRedundancyOrReplacement(olist, pcont, sock, paddr);
        if (err < 0)
            goto out;

        olist->send_replacements++;
    }

out:
    if (err < 0)
    {
        PERROR("%s: Could not send replacement(no %d)\n", DEBUG_NAME, err);
        return err;
    }

    return 0;
}

int get_block(struct prrt_out_list *olist, int packetno, struct prrt_block **mblock)
{
    /*we should be locked, entering this function*/
    int err = 0;

    err = prrtblist_get_block(&olist->block_list, packetno, mblock);

    return err;
}

int get_packet(struct prrt_out_list *olist, struct prrt_block *mblock, int packetno,
        struct prrt_packet_container **pcont);
int getAndPrepareNextRedundancy(struct prrt_out_list *olist, struct prrt_block *mblock,
        struct prrt_packet_container **pcont);

/*returns 0 on success, <0 on failure*/
static int get_next_replacement_packet(struct prrt_out_list *olist, struct prrt_block *mblock,
        int packetno, struct prrt_packet_container **pcont)
{
    /*we should be locked, entering this function*/
    int err = 0;
    int coding = 0;
    int k = 0;
    int n = 0;
    struct prrt_coder *replacement_cod = 0;

    k = prrtcpar_get_k(&mblock->cpars);
    n = prrtcpar_get_n(&mblock->cpars);

    *pcont = 0;
    /* mblock->next_red = block->red_head if we have send already all redundancy packets
     * that were build*/
    if (n == k)
    {
        /*send old data packets if there are no coded packets and we
         *can't build them*/
        PDEBUG(MSG_RECVFBDEB, "%s: No coded packets retrieve data packet\n", DEBUG_NAME);
        err = get_packet(olist, mblock, packetno, pcont);

        return err;


    }
    else if (prrt_block_can_and_must_code(mblock))
    {
        /*this means we haven't build all redundancy packets yet*/
        coding = prrtolist_get_option(olist, CODING);
        if (!coding)
            return -1;

        /*therefore we have to code a new replacement packet*/
        //err = prrt_block_get_coder(mblock, &olist->out_cod);
        err = prrt_block_get_coder(mblock, &replacement_cod);
        if (err < 0)
        {
            PERROR("%s: Could not create coder for coding\n", DEBUG_NAME);
	    prrt_fec_free(replacement_cod);
            return -1;
        }

        //err = prrt_block_code_next_red(mblock, olist->out_cod, olist->time_state);
        err = prrt_block_code_next_red(mblock, replacement_cod, olist->time_state);
        if (err < 0)
        {
            PERROR("%s: Encoding of next replacement failed\n", DEBUG_NAME);
	    prrt_fec_free(replacement_cod);
            return -1;
        }

        getAndPrepareNextRedundancy(olist, mblock, pcont);
	prrt_fec_free(replacement_cod);

        return 0;

    }
    else
    {
        if (prrt_block_all_redundancy_was_sent(mblock))
        {
            PERROR("%s: The replacement packets for this block are used up"
                    "(%d redundancy packets in block)\n",
                    DEBUG_NAME, mblock->no_red);
            return -1;
        }

        getAndPrepareNextRedundancy(olist, mblock, pcont);

        return 0;
    }

    PERROR("%s: Block of this packet too old\n", DEBUG_NAME);
    return -1;
}

int get_packet(struct prrt_out_list *olist, struct prrt_block *mblock, int packetno,
        struct prrt_packet_container **pcont)
{
    /*we should be locked, entering this function*/
    struct packet_list_entry *p = 0;
    int err = 0;

    err = prrt_block_get_packet(mblock, packetno, &p);
    if (err < 0)
        return -1;

    *pcont = p->pcont;

    return 0;
}

int getAndPrepareNextRedundancy(struct prrt_out_list *olist, struct prrt_block *mblock,
        struct prrt_packet_container **pcont)
{
    int err = 0;
    struct packet_list_entry *p = 0;

    err = prrt_block_get_next_redundancy_packet(mblock, &p);
    if (err < 0)
        return err;

    *pcont = p->pcont;

    PrepareRedundancyForSending(olist, *pcont);

    return 0;
}

int getSendTokens(struct prrt_out_list *olist, int msglen)
{
	// Has to be locked: list_lock
	u32 time = prrt_get_ustime(olist->time_state);
	u32 time_diff = prrt_diff_timestamps(olist->time_state, time, olist->last_sent);

	olist->last_sent = time;
// 	olist->send_tokens += time_diff * RATE;
  olist->send_tokens += time_diff * olist->traffic_shaping_datarate;

	//TODO: obtain bw estimate here
	//prrtlpar_set_BW_est(&msock->loc_params_out, mfbcont->fb.fbheader.bw_est);

	if (olist->send_tokens > LIMIT)
		olist->send_tokens = LIMIT;

	//PDEBUG(MSG_FINAL, "%s: Tokens available: %u, time: %u\n", DEBUG_NAME, olist->send_tokens, time_diff);

	if (olist->send_tokens >= (u32)msglen){
		olist->send_tokens -= msglen;
		//PDEBUG(MSG_FINAL, "%s: Returning %u send_tokens\n", DEBUG_NAME, msglen);
		return msglen;
	}

	//PDEBUG(MSG_FINAL, "%s: Returning no send_tokens\n", DEBUG_NAME);
	return 0;
}

/*
 * Returns a random number between 1 and 100.
 */
#if SIMULATION_ON_SEND
#if DROP_PACKS_ON_SEND

static int get_random_percent(void)
{
    u8 rnd = 0;
    get_random_bytes(&rnd, 1);
    return rnd * 100 / 255 + 1;
}
#endif
#endif

int prrtolist_insert_feedback(struct prrt_out_list *olist, u16 miss, u32 receiver, int retr_round)
{
  struct prrt_block *bl = 0;
  int n_rr = 0;
  down(&olist->list_lock);
  prrtblist_get_block(&olist->block_list, miss, &bl);
  up(&olist->list_lock);
  if (bl) {
    n_rr = prrtcpar_get_n_rr(&bl->cpars);
  }

  if (0 < n_rr) {
    return prrt_fbana_insert_feedback(&olist->fbana, miss, receiver, retr_round);
  }
  return 0;
}

int prrtolist_set_option(struct prrt_out_list *olist, enum option_t opt, int val)
{
    int interrupted = 0;
    int err = 0;

    interrupted = down_interruptible(&olist->opt_lock);
    if (interrupted)
    {
        err = -EINTR;
        goto out;
    }

    switch (opt)
    {
    case REPLACE:
        olist->replace = val;
        if (val == ON)
            olist->coding = ON;
        break;
    case CODING:
        olist->coding = val;
        if (val == OFF)
            olist->replace = OFF;
        break;
    case GATHER_FEEDBACK:
        olist->gather_feedback = val;
        break;
    case TRAFFICSHAPING:
        olist->traffic_shaping_datarate = val;
        break;
    default:
        PERROR("%s: Option not recognized", DEBUG_NAME);
        err = -1;
        break;
    }

    up(&olist->opt_lock);

out:
    return err;
}

int prrtolist_get_option(struct prrt_out_list *olist, enum option_t opt)
{
    int interrupted = 0;
    int err = 0;

    interrupted = down_interruptible(&olist->opt_lock);
    if (interrupted)
    {
        err = -EINTR;
        goto out;
    }

    switch (opt)
    {
    case REPLACE:
        err = olist->replace;
        break;
    case CODING:
        err = olist->coding;
        break;
    case GATHER_FEEDBACK:
        err = olist->gather_feedback;
        break;
    case TRAFFICSHAPING:
        err = olist->traffic_shaping_datarate;
        break;
    default:
        PERROR("%s: Option not recognized", DEBUG_NAME);
        err = -1;
        break;
    }

    up(&olist->opt_lock);

out:
    return err;
}
