#ifndef LOSS_GATHERER_H
#define LOSS_GATHERER_H

#include "defines.h"
#include "prrt_time.h"
#include "prrt_packet.h"
#include <linux/list.h>

#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,26)
#include <asm/semaphore.h>
#else
#include <linux/semaphore.h>
#endif
#include <linux/slab.h>

/**
 * This structure represents an entry in the loss list.
 * The timestamps in this structure are based on the receiver clock and in ms.
 */
struct fbgath_fb_entry {
	struct list_head list;
	u16 miss_packno;	/**< The number of the missing packet.*/
	u16 sn_base;		/**< Base sequence number of the corresponding block.*/
	u32 first_inserted;	/**< This timestamp is set when this loss is created.*/
	u32 pack_timeout;	/**< This timestamp determines the expiration date of the corresponding data packet.*/
	u32 first_due;		/**< First time at which feedback is expected by the sender.*/
	u32 last_send;		/**< The timestamp when this was last checked.*/
	u16 retr_round;		/**< How often was the feedback for this packet transmitted.*/
  u16 fb_send_delay;  /**< A random delay for sending the feedback used in feedback suppression (ms). */
};

/**
 * This pseudo class manages the sending of feedback by analyzing incoming and missing packets.
 * The feedback analysis works as follows:
 * - When a data packet arrives and is inserted in the packet list we also gather the missing
 *   packets in front of it to add them to the loss_list and perhaps delete itself from the loss_list.
 * - When we receive a redundancy packet we delete a packet of the same block from the loss_list
 *   if any are on it. If not we can discard the redundancy packet.
 *
 * The implementation of rounds works like this:
 * - Every FEEDBACKDELAYMS ms (see defines.h) we gather all packetnumbers in the fb_list whose
 *   timestamps+t_replace (in jiffies) are bigger than the receiver clock and
 *   send a feedback packet for them.
 * - Additionally we check if the first_timestamp+D_target-t_replace(in jiffies) is bigger
 *   than the receiver clock:
 *   - If it is bigger then we have to delete the feedback from the list because it will not arrive
 *     in time.
 *   - if it is less we refresh the timestamp of the feedback.
 *
 * In this way we repeat asking for replacements until a packet will not arrive in time anymore.
 *
 * INVARIANTS:
 * - The fb_list is ordered according to due-date.
 */
__PCLASS __THREAD_SAFE struct prrt_loss_gatherer {
	struct list_head loss_list;	/**< This is a list of losses that were seen.*/
	struct semaphore lock;		/**< Mutex to lock access to the list.*/
	int count;			/**< Counts elements in the list.*/
	long max_fb_seqno;		/**< The biggest data sequence number to send feedback for.*/
	char use_fb_limit;		/**< Should max_fb_seqno be used to limit the range of feedback.*/
	int no_data_last_block;		/**< The number of data packets in the last block checked for feedback.*/
	struct prrt_time_state *time_state; /**< The state of time in this socket. */
  unsigned int max_fb_send_delay; /**< Maximum delay for sending the feedback used in feedback suppression (ms). */
};

/**
 * This initialises all variables in the struct.
 */
void prrt_lossgath_init(struct prrt_loss_gatherer *fbgath, struct prrt_time_state *time_state);

/**
 * This frees all allocated memory in this struct.
 */
void prrt_lossgath_free(struct prrt_loss_gatherer *fbgath);

/**
 * Insert feedback in list.
 * This can be done packet per packet because new feedback always goes to the start of the list.
 * \param packetno the packetnumber of the missing packet.
 * \param timestamp the actual receiver time in ms.
 * \return 0 on success, -ENOMEM if no memory
 */
int prrt_lossgath_insertfb(struct prrt_loss_gatherer *fbgath, long packetno, u16 sn_base, u32 pack_timeout, u32 first_due, u32 replace_time);

/**
 * Delete a certain feedback from the list if present.
 * \param packetno the packetno to delete.
 * \param retransmission round the redundancy belongs to
 * \param measured_t_replace this will hold the the time that was needed to replace the packet or 0
 *				 if it is not in the first retransmission round.
 * \return 0 on success, -1 if it was not there.
 */
int prrt_lossgath_delete1fb(struct prrt_loss_gatherer *fbgath, long packetno, u8 round, int *measured_t_replace);

/**
 * Deletes one of the feedbacks for a given block from the list.
 * It deletes the next entry that will be due. This is because replacements for feedbacks that
 * need longer for the next round should arrive later.
 * \param block_start packetnumber of the first packet in the block.
 * \param block_end packetnumber of the last packet in the block.
 * \param retransmission round the redundancy belongs to
 * \param measured_t_replace this will hold the the time that was needed to replace the packet or 0
 *				 if it is not in the first retransmission round.
 * \return 0 if deleted, -1 if nothing was found to delete.
 */
int prrt_lossgath_deletefb(struct prrt_loss_gatherer *fbgath, long block_start, long block_end, u8 round,
				int *measured_t_replace);

/**
 * Deletes all feedback for a given block.
 * \param block_start packetnumber of the first packet in the block.
 * \param block_end packetnumber of the last packet in the block.
 * \return 0 if deleted, -1 if nothing was found to delete.
 */
int prrt_lossgath_deleteAllFB(struct prrt_loss_gatherer *fbgath, long block_start, long block_end);

/**
 * Gives the number of the next packet where feedback is needed for.
 * It checks for the timestamps of the feedbacks. If the first entry is due it returns the packetno,
 * moves the entry to the end of the list and refreshs the timestamp.
 * This function also deletes entries that are to old.
 * \param timestamp the actual receiver time in ms.
 * \param RTT Round Trip Time in ms.
 * \param D_target Target delay of receiver in ms.
 * \param packno The packetnumber of the missing packet.
 * \param retr_round The retransmission round of the packet.
 * \return 0 on success, -1 if no entry is due or list empty.
 */
int prrt_lossgath_getnextfb(struct prrt_loss_gatherer *fbgath, u32 time_now, int replace_time, int D_target,
			   u16 *packno, u16 *retr_round, u16 *repetitions);

/**
 * This enables/disables the feedback suspension by checking the given packet.
 * \param pcont the packet that arrived and must be checked for feedback suspension.
 */
void prrt_lossgath_check_fb_limit(struct prrt_loss_gatherer *fbgath, struct prrt_packet_container *pcont);

#endif /*FEEDBACK_GATHERER_H*/
