/* 
 * (c) 2010 Saarland University, Telecommunications Lab 
 * (prrt@nt.uni-saarland.de)
 *
 * The contents of this file are subject to the Mozilla Public License
 + Version 1.1 (the "License"); you may not use this file except in
 + compliance with the License. You may obtain a copy of the License at
 + http://www.mozilla.org/MPL/
 + 
 + Software distributed under the License is distributed on an "AS IS"
 + basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 + License for the specific language governing rights and limitations
 + under the License.
 + 
 + The Original Code is "Predictably Reliable Real-time Transport (PRRT)".
 + 
 + The Initial Developer of the Original Code is Saarland University.
 + All Rights Reserved.
 + 
 + Contributor(s): Jochen Gruen (Saarland University), Manuel Gorius
 + (Saarland University), Yongtao Shuai (Saarland University).
 + 
 + Disclaimer: This Alpha version of PRRT is not supported and stable yet.
 + Do not use it in a production environment: it may crash your device, 
 + so please make sure that you have a backup of your data before
 + installing PRRT.
 */

#include "param_code.h"
#include <linux/slab.h>

// First two lines are from the original param_local.c
//#define DEBUG_NAME "PARAM_LOCAL"

static const int delta = 3; /*division is implemented with rightshifts, this is the number
				 *which is used to divide when smoothing the results, e.g. 3 means
				 *division by 8*/

#define DEBUG_NAME "PARAM_CODE"


void prrtcpar_reinit_locparams(struct params_code *params)
{
    atomic_set(&params->D_target, D_TARGET_STARTMS);
    atomic_set(&params->D_Fec, D_FEC_STARTMS);
    atomic_set(&params->D_C, D_C_STARTMS);
    atomic_set(&params->D_Req, D_REQ_STARTMS);
    atomic_set(&params->PLR_target, PLR_TARGET_START);
    atomic_set(&params->BW_est, BW_EST_START);
    atomic_set(&params->Scheduled_RI, 0);
}

int prrtcpar_init_codeparams(struct params_code *params)
{
	int i = 0;

	init_MUTEX(&params->lock);

  atomic_set(&params->D_target, D_TARGET_STARTMS);
  atomic_set(&params->D_Fec, D_FEC_STARTMS);
  atomic_set(&params->D_C, D_C_STARTMS);
  atomic_set(&params->D_Req, D_REQ_STARTMS);
  atomic_set(&params->PLR_target, PLR_TARGET_START);
  atomic_set(&params->BW_est, BW_EST_START);
  atomic_set(&params->Scheduled_RI, 0);

	params->n = N_START;
	params->k = K_START;
	params->r = N_START-K_START;
	params->n_p = MIN(N_P_START, N_START-K_START);
	params->code_start = MAX(N_P_START, CODE_START);

	params->n_rr_max = 3;
	params->n_rr = params->n_rr_max;
	params->n_cc = kmalloc(sizeof(*params->n_cc)*params->n_rr_max, GFP_KERNEL);
	if(!(params->n_cc)) {
		params->n_rr = 0;
		params->n_rr_max = 0;
		return -ENOMEM;
	}

	for(i=0; i<params->n_rr; ++i)
		params->n_cc[i] = N_CC_START;

	params->n_cc[0] = 9;
	params->n_cc[1] = 9;
	params->n_cc[2] = 9;

	return 0;
}

void prrtcpar_zero_codeparams(struct params_code *params)
{
	bzero(params, sizeof(*params));
	init_MUTEX(&params->lock);
}

void prrtcpar_free_codeparams(struct params_code *params)
{
	kfree(params->n_cc);
}

int prrtcpar_already_set(struct params_code *params)
{
	return (params->k != 0);
}

int prrtcpar_copy_code(struct params_code *params, struct params_code *cpar)
{
	int err = 0;
	int i = 0;

	down(&params->lock);
	down(&cpar->lock);

	cpar->n = params->n;
	cpar->k = params->k;
	cpar->r = params->r;
	cpar->n_p = params->n_p;
	cpar->code_start = params->code_start;
	cpar->n_rr = params->n_rr;

	cpar->n_rr_max = params->n_rr_max;
	if(cpar->n_cc)
		kfree(cpar->n_cc);

	cpar->n_cc = kmalloc(sizeof(*params->n_cc)*params->n_rr_max, GFP_KERNEL);
	if(!cpar->n_cc) {
		err = -ENOMEM;
		goto out;
	}

	for(i=0; i<params->n_rr_max; i++) {
		cpar->n_cc[i] = params->n_cc[i];
	}

  atomic_set( &cpar->D_target, atomic_read(&params->D_target) );
  atomic_set( &cpar->D_Fec, atomic_read(&params->D_Fec) );
  atomic_set( &cpar->D_C, atomic_read(&params->D_C) );
  atomic_set( &cpar->D_Req, atomic_read(&params->D_Req) );
  atomic_set( &cpar->PLR_target, atomic_read(&params->PLR_target) );
  atomic_set( &cpar->D_fb, atomic_read(&params->D_fb) );
  //atomic_set( &cpar->BW_est, atomic_read(&params->BW_est) );
  //atomic_set( &cpar->Scheduled_RI, atomic_read(&params->Scheduled_RI) );

out:
	up(&cpar->lock);
	up(&params->lock);

	return err;
}

static int CheckCodingParamsValidity(int k, int r, int n_p, int code_start);

int prrtcpar_set_coding_params(struct params_code *params, int k, int r, int n_p, int code_start)
{
	int err = 0;

	//down(&params->lock);

	err = CheckCodingParamsValidity(k, r, n_p, code_start);
	if(err < 0) {
		PERROR("%s: Inputted coding parameters not valid\n", DEBUG_NAME);
		goto error;
	}

	//prrtcpar_init_codeparams(params);

	params->k = k;

	if (k == 1){
		params->n = 1;
		params->r = 0;
		params->n_p = 0;
	}
	else {
		params->n = k+r;
		params->r = r;
		params->n_p = n_p;
	}

	params->code_start = code_start;

	PDEBUG(MSG_KUIPARAM, "%s: Set coding parameters:\n", DEBUG_NAME);
	PDEBUG(MSG_KUIPARAM, "%s: k: %d\n", DEBUG_NAME, params->k);
	PDEBUG(MSG_KUIPARAM, "%s: n: %d\n", DEBUG_NAME, params->n);
	PDEBUG(MSG_KUIPARAM, "%s: r: %d\n", DEBUG_NAME, params->r);
	PDEBUG(MSG_KUIPARAM, "%s: n_p: %d\n", DEBUG_NAME, params->n_p);
	PDEBUG(MSG_KUIPARAM, "%s: code_start: %d\n", DEBUG_NAME, params->code_start);

error:
	//up(&params->lock);

	return err;
}


int prrtcpar_set_ahec_params(struct params_code *params,
                             int d_target, int d_fec, int d_c, int d_req, int d_fb, int k, int n_p, int n_cc, int *n_ccs)
{
  int err = 0;
  int r = 0;
  int code_start = 0;
  int i = 0;

  if (k == 1)
	  n_p = 0;

  r = n_p;
  code_start = n_p; 

	PDEBUG(MSG_KUIPARAM, "%s: Set AHEC parameters:\n", DEBUG_NAME);
	PDEBUG(MSG_KUIPARAM, "%s: d_target: %d\n", DEBUG_NAME, d_target);
	PDEBUG(MSG_KUIPARAM, "%s: d_fec: %d\n", DEBUG_NAME, d_fec);
	PDEBUG(MSG_KUIPARAM, "%s: d_c: %d\n", DEBUG_NAME, d_c);
	PDEBUG(MSG_KUIPARAM, "%s: d_req: %d\n", DEBUG_NAME, d_req);
	PDEBUG(MSG_KUIPARAM, "%s: d_fb: %d\n", DEBUG_NAME, d_fb);
	PDEBUG(MSG_KUIPARAM, "%s: k: %d\n", DEBUG_NAME, k);
	PDEBUG(MSG_KUIPARAM, "%s: n_p: %d\n", DEBUG_NAME, n_p);
	PDEBUG(MSG_KUIPARAM, "%s: n_cc: %d\n", DEBUG_NAME, n_cc);
	for(i=0; i<n_cc; ++i) {
		PDEBUG(MSG_KUIPARAM, "%s: n_ccs[%d]: %d\n", DEBUG_NAME, i, n_ccs[i]);
	}

  down(&params->lock);

  atomic_set(&params->D_target, d_target);
  atomic_set(&params->D_Fec, d_fec);
  atomic_set(&params->D_C, d_c);
  atomic_set(&params->D_Req, d_req);
  atomic_set(&params->D_fb, d_fb);

  // Set retransmission rounds first, because we need to determine the N_rr in params_code.
  err = prrtcpar_set_retransmission_rounds(params, n_cc, n_ccs);
  if (err < 0) {
    goto error;
  }

  for (i=0; i<params->n_rr; i++) {
    r += n_ccs[i];
  }
  //code_start = r;
  err = prrtcpar_set_coding_params(params, k, r, n_p, code_start);

error:
  up(&params->lock);

  return err;
}


static int CheckCodingParamsValidity(int k, int r, int n_p, int code_start)
{
	if(code_start > r)
		return -1;
	
	if(n_p > code_start)
		return -1;

	if(n_p > r)
		return -1;

	return 0;
}

static void SetN_RR_MAX(struct params_code *params, int n_rr_max);
static void SetN_RR(struct params_code *params, int n_rr);
static int ResizeN_CC(struct params_code *params);
static int SetN_CC(struct params_code *params, int n_rr, int* n_ccs);

int prrtcpar_set_retransmission_rounds(struct params_code *params, int n_rr, int *n_ccs)
{
	int err = 0;
	int i = 0;

	//down(&params->lock);

	SetN_RR_MAX(params, n_rr);
	SetN_RR(params, n_rr);
	err = ResizeN_CC(params);
	if(err < 0) {
		SetN_RR_MAX(params, 0);
		goto error;
	}

	err = SetN_CC(params, n_rr, n_ccs);
	if(err < 0) {
		SetN_RR_MAX(params, 0);
		kfree(params->n_cc);
		params->n_cc = 0;
		goto error;
	}

	PDEBUG(MSG_KUIPARAM, "%s: Set retransmission rounds:\n", DEBUG_NAME);
	PDEBUG(MSG_KUIPARAM, "%s: n_rr: %d\n", DEBUG_NAME, params->n_rr);
	PDEBUG(MSG_KUIPARAM, "%s: n_rr_max: %d\n", DEBUG_NAME, params->n_rr_max);
	for(i=0; i<params->n_rr_max; ++i) {
		PDEBUG(MSG_KUIPARAM, "%s: n_cc[%d]: %d\n", DEBUG_NAME, i, params->n_cc[i]);
	}

error:
	//up(&params->lock);

	return err;
}

static void SetN_RR_MAX(struct params_code *params, int n_rr_max)
{
	params->n_rr_max = MIN(MAX_NUMBER_RETRROUNDS, n_rr_max);
	params->n_rr = MIN(params->n_rr, params->n_rr_max);
}

static void SetN_RR(struct params_code *params, int n_rr)
{
	params->n_rr = MIN(n_rr, params->n_rr_max);
}

static int ResizeN_CC(struct params_code *params)
{
	int *new_n_cc = 0;
	int i = 0;

	kfree(params->n_cc);

	params->n_cc = kmalloc(sizeof(*new_n_cc)*params->n_rr_max, GFP_KERNEL);
	if(!params->n_cc)
		goto error;

	for(i=0; i<params->n_rr_max; i++) {
		params->n_cc[i] = 1;
	}

	return 0;

error:
	params->n_rr = 0;
	params->n_rr_max = 0;
	return -ENOMEM;
}

static int SetN_CC(struct params_code *params, int n_rr, int* n_ccs)
{
	int i = 0;

	if(params->n_rr > params->n_rr_max)
		return -1;

	for(i=0; i<n_rr; ++i) {
		params->n_cc[i] = n_ccs[i];
	}

	return 0;
}

static void UpdateN_RR_MAX(struct params_code *params, int t_replace, int d_target);
static void update_params(struct params_code *params, struct params_network *netparams);

/* NOTE : THIS IS NOT YET READY TO USE*/
void prrtcpar_update_params(struct params_code *params, struct params_network *netparams)
{
	int t_replace = prrtnpar_get_ReplaceTime(netparams);
	int d_target = atomic_read(&params->D_target);

	down(&params->lock);

	UpdateN_RR_MAX(params, t_replace, d_target);

	update_params(params, netparams);

	up(&params->lock);
}

static void UpdateN_RR_MAX(struct params_code *params, int t_replace, int d_target)
{
	if(t_replace != 0) {
		params->n_rr_max = (d_target - (t_replace/2))/t_replace;
	}

	params->n_rr_max = MIN(MAX_NUMBER_RETRROUNDS, params->n_rr_max);
}

/*
 * TODO: Implement dynamic parameter estimation.
 * This is the function that can be implemented to enable dynamic parameter estimation.
 * The parameters estimated in this function will be used to code outgoing blocks.
 */
static void update_params(struct params_code *params, struct params_network *netparams)
{

}

int prrtcpar_get_k(struct params_code *params)
{
	int k;

	down(&params->lock);
	k = params->k;
	up(&params->lock);

	return k;
}

int prrtcpar_get_r(struct params_code *params)
{
	int r;

	down(&params->lock);
	r = params->r;
	up(&params->lock);

	return r;
}

int prrtcpar_get_n(struct params_code *params)
{
	int n;

	down(&params->lock);
	n = params->n;
	up(&params->lock);

	return n;
}

int prrtcpar_get_n_p(struct params_code *params)
{
	int n_p;

	down(&params->lock);
	n_p = params->n_p;
	up(&params->lock);

	return n_p;
}

int prrtcpar_get_n_rr(struct params_code *params)
{
	int n_rr;

	down(&params->lock);
	n_rr = params->n_rr;
	up(&params->lock);

	return n_rr;
}

int prrtcpar_get_n_cc(struct params_code *params, int retr_round)
{
	int n_cc = 0;

	down(&params->lock);

	if(retr_round < 1 || retr_round > params->n_rr)
		n_cc = 0;
	else
		n_cc = params->n_cc[retr_round-1];

	up(&params->lock);
	
	return n_cc;
}

int prrtcpar_get_code_start(struct params_code *params)
{
	int code_start;

	down(&params->lock);
	code_start = params->code_start;
	up(&params->lock);

	return code_start;
}


// The following code are from the param_local.c

int prrtcpar_get_D_target(struct params_code *params)
{
    return atomic_read(&params->D_target);
}

void prrtcpar_set_PLR_target(struct params_code *params, int plr_target)
{
    atomic_set(&params->PLR_target, plr_target);
}

int prrtcpar_get_PLR_target(struct params_code *params)
{
    return atomic_read(&params->PLR_target);
}

int prrtcpar_get_D_Fec(struct params_code *params)
{
    return atomic_read(&params->D_Fec);
}

int prrtcpar_get_D_C(struct params_code *params)
{
    return atomic_read(&params->D_C);
}

int prrtcpar_get_D_Req(struct params_code *params)
{
    return atomic_read(&params->D_Req);
}

int prrtcpar_get_D_fb(struct params_code *params)
{
    return atomic_read(&params->D_fb);
}

void prrtcpar_set_BW_est(struct params_code *params, int bw_est)
{
    atomic_set(&params->BW_est, bw_est);
}

int prrtcpar_get_BW_est(struct params_code *params)
{
    return atomic_read(&params->BW_est);
}

void prrtcpar_update_Scheduled_RI(struct params_code *params, int RI_increment)
{
//TODO: atomic_inc?
    atomic_set(&params->Scheduled_RI, RI_increment);
}

int prrtcpar_get_Scheduled_RI(struct params_code *params)
{
    return atomic_read(&params->Scheduled_RI);
}



