/* 
 * (c) 2010 Saarland University, Telecommunications Lab 
 * (prrt@nt.uni-saarland.de)
 *
 * The contents of this file are subject to the Mozilla Public License
 + Version 1.1 (the "License"); you may not use this file except in
 + compliance with the License. You may obtain a copy of the License at
 + http://www.mozilla.org/MPL/
 + 
 + Software distributed under the License is distributed on an "AS IS"
 + basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 + License for the specific language governing rights and limitations
 + under the License.
 + 
 + The Original Code is "Predictably Reliable Real-time Transport (PRRT)".
 + 
 + The Initial Developer of the Original Code is Saarland University.
 + All Rights Reserved.
 + 
 + Contributor(s): Jochen Gruen (Saarland University), Manuel Gorius
 + (Saarland University), Yongtao Shuai (Saarland University).
 + 
 + Disclaimer: This Alpha version of PRRT is not supported and stable yet.
 + Do not use it in a production environment: it may crash your device, 
 + so please make sure that you have a backup of your data before
 + installing PRRT.
 */

#ifndef PARAM_NETWORK_H
#define PARAM_NETWORK_H

#include "../defines.h"
#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,26)
#include <asm/semaphore.h>
#else
#include <linux/semaphore.h>
#endif
#include <asm/atomic.h>

#define MAX_MEASURES_RTT 8
#define MAX_MEASURES_PLR 512
#define MAX_PARAM 65535
/**
 * Start value for fix_t_replace
 */
#define FIX_T_REPLACE 500

/**
 * This structure holds the network specific parameters for a connection between
 * two hosts, e.g. one sender and one receiver.
 * Computation of the Packet Loss Rate is done based on the Gilbert Elliot Model.
 * This means each receiver holds exactly one of this structures for the connection
 * to the sender.
 * Each sender has to hold a list of these structures, one for each receiver, because we want
 * to use this protocol in multicast scenarios.
 * All times in ms, all probabilities in percent.
 * This pseudo class is thread safe when private elements are only accessed with it's functions.
 */
__THREAD_SAFE __PCLASS struct params_network {
	__PRIVATE
	struct semaphore lock;	/**<Lock to make the whole thing thread safe.*/
	atomic_t t_replace;	/**<The time from detecting packet miss until replacement arrives in ms.*/
	atomic_t fix_t_replace;	/**<Only for testing purposes when using a fixed t_r.*/
	atomic_t pack_interval;	/**<The time between the arrivals of subsequent packets from the
				*   application.*/
	atomic_t FTT;		/**<Forward Trip Time*/
	atomic_t QueueDelay;	/**<Network queue delay*/

  int rtt_samples[MAX_MEASURES_RTT];  /**<This array is used to smooth the RTT (i.e. t_replace).*/
  int rtt_samples_sum;                /**<The sum of all RTT samples in the array.*/
  int rtt_samples_idx;                /**<The index of the current RTT sample in the array.*/

	//int last_bad_meas;	/**<The oldest measurement in the array bad_meas.*/
	//int last_good_meas;	/**<The oldest measurement in the array good_meas.*/
	//int bad_meas[MAX_MEASURES];	/**<This array contains samples for burst error lengths.*/
	//int good_meas[MAX_MEASURES];	/**<This array contains samples for correct interval lengths.*/
	//TODO: pattern could be a bit map
	char loss_pattern[MAX_MEASURES_PLR];	/**<This array reflects the pattern of lost packets.*/
	u16 numberOfSamples;
	u16 lastSampleIndex;
	int alpha;	/**<This is the probability to stay in G state of the GE model*/
	int beta;	/**<This is the probability to stay in B state of the GE model*/
	u16 numberOfGaps;  /**<Number of loss-free periods*/
	u16 numberOfBursts;/**<Number of loss periods*/
	u16 gapLength;	/**<Aggregated length of packet receptions*/
	u16 burstLength;/**<Aggregated length of packet losses*/
	int PLR;	/**<The estimated packet loss rate.*/
	int maxOffset;  /**<Max time offset between sender and receiver clock*/
	int minOffset;  /**<Min time offset between sender and receiver clock*/
	int lastOffset; /**<Latest time offset between sender and receiver clock*/
	//int clockSkew;	/**<Skew between sender and receiver clock*/
	int clock_offsets[MAX_MEASURES_PLR]; /**History of reference clock samples*/
};

/**
 * Initialise the parameter values.
 */
void prrtnpar_init_netparams(struct params_network *params);
void prrtnpar_reinit_netparams(struct params_network *params);

/**
 * Updates the replace time for the the given connection.
 */
void prrtnpar_update_ReplaceTime(struct params_network *params, int sample_t_replace);

/**
 * Set the replace time to the given value.
 */
int prrtnpar_set_ReplaceTime(struct params_network *params, int t_replace);

/**
 * Returns an estimated value for the replace time.
 */
int prrtnpar_get_ReplaceTime(struct params_network *params);

/**
 * This function updates the PLR with the actual values of the good and bad measurements.
 */
void prrtnpar_update_PLR(struct params_network *params);

/**
 * This function sets the PLR to the given value.
 * \param PLR value that the PLR should be set to (0<=PLR<=100).
 * \return 0 on success, -1 for invalid input.
 */
int prrtnpar_set_PLR(struct params_network *params, int PLR);

/**
 * Insert a loss notification.
 * \param seqnum Sequence number of the lost packet.
 */
void prrtnpar_insert_Loss(struct params_network *params, u16 seqnum);

/**
 * Insert a sending notification.
 * \param seqnum Sequence number of the sent packet.
 */
void prrtnpar_insert_Sent(struct params_network *params, u16 seqnum);

/**
 * Get the current loss pattern.
 */
void prrtnpar_get_LossPattern(struct params_network *params, char* pattern);

///**
// * Insert a good measurement into the corresponding array.
// * \param meas The measurement that should be inserted.
// */
//void prrtnpar_insert_GoodMeas(struct params_network *params, int meas);
//
///**
// * This increases the last good measurement by 1.
// */
//void prrtnpar_update_last_GoodMeas(struct params_network *params);
//
///**
// * Insert a bad measurement into the corresponding array.
// * \param meas The measurement that should be inserted.
// */
//void prrtnpar_insert_BadMeas(struct params_network *params, int meas);

/**
 * Get the current estimated value of the PLR.
 */
int prrtnpar_get_PLR(struct params_network *params);

/**
 * Set the current estimated value of alpha.
 */
void prrtnpar_set_Alpha(struct params_network *params, int alpha);

/**
 * Get the current estimated value of alpha.
 */
int prrtnpar_get_Alpha(struct params_network *params);

/**
 * Get the current estimated value of beta.
 */
void prrtnpar_set_Beta(struct params_network *params, int beta);

/**
 * Get the current estimated value of beta.
 */
int prrtnpar_get_Beta(struct params_network *params);

/**
 * Set the current packet interval
 */
void prrtnpar_update_PacketInterval(struct params_network *params, int sample_pack_interval);

/**
 * Get the current packet interval
 */
int prrtnpar_get_PacketInterval(struct params_network *params);

/**
 * Set number of gaps
 */
void prrtnpar_set_NumGaps(struct params_network *params, int numGaps);

/**
 * Set number of bursts
 */
void prrtnpar_set_NumBursts(struct params_network *params, int numBursts);

/**
 * Set aggregated gap length 
 */
void prrtnpar_set_GapLength(struct params_network *params, int gapLength);

/**
 * Set aggregated burst length
 */
void prrtnpar_set_BurstLength(struct params_network *params, int burstLength);

/**
 * Get number of gaps
 */
int prrtnpar_get_NumGaps(struct params_network *params);

/**
 * Get number of bursts
 */
int prrtnpar_get_NumBursts(struct params_network *params);

/**
 * Get aggregated burst length
 */
int prrtnpar_get_BurstLength(struct params_network *params);

/**
 * Get aggregated gap length 
 */
int prrtnpar_get_GapLength(struct params_network *params);

/**
 * Set current FTT 
 */
void prrtnpar_update_FTT(struct params_network *params, int FTT_sample);

/**
 * Get current FTT 
 */
int prrtnpar_get_FTT(struct params_network *params);

/**
 * Set current QueueDelay 
 */
void prrtnpar_update_QueueDelay(struct params_network *params, int QueueDelay_sample, int seqnum);

/**
 * Get current QueueDelay 
 */
int prrtnpar_get_QueueDelay(struct params_network *params);

#endif /*PARAM_NETWORK_H*/

