/* 
 * (c) 2010 Saarland University, Telecommunications Lab 
 * (prrt@nt.uni-saarland.de)
 *
 * The contents of this file are subject to the Mozilla Public License
 + Version 1.1 (the "License"); you may not use this file except in
 + compliance with the License. You may obtain a copy of the License at
 + http://www.mozilla.org/MPL/
 + 
 + Software distributed under the License is distributed on an "AS IS"
 + basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 + License for the specific language governing rights and limitations
 + under the License.
 + 
 + The Original Code is "Predictably Reliable Real-time Transport (PRRT)".
 + 
 + The Initial Developer of the Original Code is Saarland University.
 + All Rights Reserved.
 + 
 + Contributor(s): Jochen Gruen (Saarland University), Manuel Gorius
 + (Saarland University), Yongtao Shuai (Saarland University).
 + 
 + Disclaimer: This Alpha version of PRRT is not supported and stable yet.
 + Do not use it in a production environment: it may crash your device, 
 + so please make sure that you have a backup of your data before
 + installing PRRT.
 */

#include "prrt_block.h"
#include <linux/slab.h>
#include "defines.h"
#include "prrt_socket.h"
#include "lists/prrt_out_list.h"
#include "kui/kui_param.h"
#include "prrt_time.h"

#define DEBUG_NAME "PRRT_BLOCK"

static struct kmem_cache *block_cache;
static struct semaphore block_cache_lock;
int prrtblocks = 0;



inline int prrt_block_get_first_data(struct prrt_block *mblock, struct packet_list_entry **first)
{
    *first = 0;

    if (list_empty(&mblock->data_head))
    {
        //PERROR("%s: Tried to retrieve first data entry of empty block\n", DEBUG_NAME);
        return -1;
    }

    *first = container_of(mblock->data_head.next, struct packet_list_entry, list);

    return 0;
}

inline u16 prrt_block_get_first_packetno(struct prrt_block *mblock)
{
    int err = 0;
    struct packet_list_entry *firstData = 0;

    err = prrt_block_get_first_data(mblock, &firstData);
    if (err < 0)
        return 0;

    return firstData->pcont->pack.data.header.seqnum;
}

inline int prrt_block_get_last_data(struct prrt_block *mblock, struct packet_list_entry **last)
{
    *last = 0;

    if (list_empty(&mblock->data_head))
    {
        //PERROR("%s: Tried to retrieve last data entry of empty block\n", DEBUG_NAME);
        return -1;
    }

    *last = container_of(mblock->data_head.prev, struct packet_list_entry, list);

    return 0;
}

inline int prrt_block_get_first_red(struct prrt_block *mblock, struct packet_list_entry **first)
{
    *first = 0;

    if (list_empty(&mblock->red_head))
    {
        //PERROR("%s: Tried to retrieve first redundancy entry of empty block\n", DEBUG_NAME);
        return -1;
    }

    *first = container_of(mblock->red_head.next, struct packet_list_entry, list);

    return 0;
}


int prrt_block_cache_init(void)
{
    init_MUTEX(&block_cache_lock);
    block_cache = kmem_cache_create("block_cache",
            sizeof (struct prrt_block),
            0, 0,
            NULL);

    if (!block_cache)
    {
        PERROR("%s: Could not create block cache\n", DEBUG_NAME);
        return -1;
    }

    return 0;
}

void prrt_block_cache_uninit(void)
{
    kmem_cache_destroy(block_cache);
}

int prrt_block_alloc(struct prrt_block **mblock, struct params_code *cpar)
{
    int err = 0;

    *mblock = 0;

    down(&block_cache_lock);
    *mblock = kmem_cache_alloc(block_cache, GFP_KERNEL);
    prrtblocks++;
    //PDEBUG(MSG_BLOCKDEB, "%s: Allocating block (count %d)\n", DEBUG_NAME, prrtblocks);
    up(&block_cache_lock);
    if (!(*mblock))
        return -ENOMEM;

    PDEBUG(MSG_BLOCKLIST, "%s: Allocated block(%p)\n", DEBUG_NAME, *mblock);

    /*intitialise block*/
    prrtcpar_zero_codeparams(&(*mblock)->cpars);

    err = prrtcpar_copy_code(cpar, &(*mblock)->cpars);
    if (err < 0)
    {
	    PERROR("%s: Not enough memory to copy code to block\n", DEBUG_NAME);
	    return -ENOMEM;
    }

    (*mblock)->no_data = 0;
    (*mblock)->no_red = 0;
    (*mblock)->biggest_data = 0;
    (*mblock)->next_red_pno = 0;
    INIT_LIST_HEAD(&(*mblock)->data_head);
    INIT_LIST_HEAD(&(*mblock)->red_head);
    // 	(*mblock)->first = 0;
    // 	(*mblock)->last = 0;
    // 	(*mblock)->first_red = 0;
    // 	(*mblock)->last_red = 0;
    (*mblock)->next_red = &(*mblock)->red_head;
    (*mblock)->first_red_send = 0;
    (*mblock)->is_coded = 0;
    (*mblock)->retr_round = 0;

    return err;
}

void prrt_block_free(struct prrt_block **mblock)
{
    if (!(*mblock))
        return;

    prrtcpar_free_codeparams(&(*mblock)->cpars);

    PDEBUG(MSG_BLOCKLIST, "%s: Freeing block(%p)\n", DEBUG_NAME, *mblock);

    down(&block_cache_lock);
    kmem_cache_free(block_cache, *mblock);
    prrtblocks--;
    up(&block_cache_lock);
    *mblock = 0;
}

void free_data_packets(struct prrt_block *mblock);
void free_redundancy_packets(struct prrt_block *mblock);

void prrt_block_destroy(struct prrt_block **mblock)
{
    PDEBUG(MSG_BLOCKDEB, "%s: Destroying block\n", DEBUG_NAME);

    if (*mblock == 0)
        return;

    PDEBUG(MSG_BLOCKDEB, "%s: Deleting data packets\n", DEBUG_NAME);

    free_data_packets(*mblock);


    PDEBUG(MSG_BLOCKDEB, "%s: Deleting redundancy packets\n", DEBUG_NAME);

    free_redundancy_packets(*mblock);


    prrt_block_free(mblock);

    PDEBUG(MSG_ALLOC, "%s: Block destroyed\n", DEBUG_NAME);
}

void free_packet_in_block_list(struct prrt_block *mblock, struct list_head *iter);

void free_data_packets(struct prrt_block *mblock)
{
    struct list_head *iter = 0;
    struct list_head *next = 0;

    list_for_each_safe(iter, next, &mblock->data_head)
    {
        PDEBUG(MSG_BLOCKDEB, "%s: Deleting data packet\n", DEBUG_NAME);
        free_packet_in_block_list(mblock, iter);

        mblock->no_data--;
    }
}

void free_packet_in_block_list(struct prrt_block *mblock, struct list_head *iter)
{
    struct packet_list_entry *p = 0;

    p = container_of(iter, struct packet_list_entry, list);
    PDEBUG(MSG_BLOCKDEB, "%s: Deleting packet p = %p\n", DEBUG_NAME, p);
    PDEBUG(MSG_BLOCKDEB, "%s: Deleting packet pno = %d\n", DEBUG_NAME, p->pcont->pack.data.header.seqnum);

    list_del(iter);
    prrt_packet_destroy(&p->pcont);
    prrt_list_entry_free(&p);
}

void free_redundancy_packets(struct prrt_block *mblock)
{
    struct list_head *iter = 0;
    struct list_head *next = 0;

    list_for_each_safe(iter, next, &mblock->red_head)
    {
        PDEBUG(MSG_BLOCKDEB, "%s: Deleting redundancy packet\n", DEBUG_NAME);
        free_packet_in_block_list(mblock, iter);

        mblock->no_red--;
    }
}

int prrt_block_is_already_coded(struct prrt_block *mblock)
{
    return mblock->is_coded;
}

int prrt_block_is_empty(struct prrt_block *mblock)
{
    return (list_empty(&mblock->data_head));
}

int space_for_redundancy_left(struct prrt_block *mblock);

int prrt_block_can_and_must_code(struct prrt_block *mblock)
{
    return (prrt_block_all_redundancy_was_sent(mblock) && space_for_redundancy_left(mblock));
}

int prrt_block_all_redundancy_was_sent(struct prrt_block *mblock)
{
    return (&mblock->red_head == mblock->next_red);
}

int prrt_block_too_old_than_timestamp(struct prrt_block *mblock, u32 time_now, u32 d_target, struct prrt_time_state *time_state)
{
    struct packet_list_entry *p = 0;
    u32 last_packets_transmit_time = 0;
    u32 sum = 0;
    u32 delay;
    int comp = 0;

    p = container_of(mblock->data_head.prev, struct packet_list_entry, list);

    delay = p->pcont->pack.data.data_header.pack_timeout+MAYDELBLOCK_DELAY_ADD;

    last_packets_transmit_time = p->pcont->time.transmit_time;

    sum = (last_packets_transmit_time + delay) % MAX_TIMESTAMP;

    comp = prrt_compare_timestamps(time_state, sum, time_now);
    //PDEBUG(MSG_BLOCKDEB, "%s: Comparing: %u and %u\n", DEBUG_NAME, sum, time_now);

    if (comp == -1)
        return 1;
    else
        return 0;

    return 0;
}

int space_for_redundancy_left(struct prrt_block *mblock)
{
    int n = prrtcpar_get_n(&mblock->cpars);

    return (mblock->no_red + mblock->no_data < n);
}

int prrt_block_insert_data_packet(struct prrt_block *mblock, struct prrt_packet_container *pcont,
        struct packet_list_entry **p, struct params_code *cpars)
        //struct packet_list_entry **p, struct params_code *cpars, struct params_local *lpars)
{
    int err = 0;
    struct packet_list_entry *new = 0;
    struct packet_list_entry *pFirst = 0;
    struct prrt_data_packet *firstPack = 0;
    u32 timestampFirstPacket = 0;
    u32 timestampCurrentPacket = 0;
    u32 timestampDiff = 0;
    u32 pack_timeout = 0;
    u32 dec_timeout = 0;
    //u32 d_fec = (u32)prrtlpar_get_D_Fec(lpars);
    u32 d_fec = (u32)prrtcpar_get_D_Fec(cpars);
    u32 d_req = (u32)prrtcpar_get_D_Req(cpars);
    //u32 d_target = (u32)prrtlpar_get_D_target(lpars);
    u32 d_target = (u32)prrtcpar_get_D_target(cpars);
    int n_rr = 1;
    //int n_cc = prrtcpar_get_n_cc(lpars, 1);

    *p = 0;
    prrt_list_entry_alloc(&new);
    if (!new)
        return -ENOMEM;

    new->pcont = pcont;

    *p = new;
    list_add_tail(&new->list, &mblock->data_head);

    err = prrt_block_get_first_data(mblock, &pFirst);
    //TODO: Is this possible?
    //if (err < 0)
    //    return;

    firstPack = &pFirst->pcont->pack.data;
    timestampFirstPacket = firstPack->data_header.timestamp;
    timestampCurrentPacket = pcont->pack.data.data_header.timestamp;

    // Shine : I dont think we need prrt_time_state here, so set it to null in the following.
    timestampDiff = prrt_diff_timestamps(0, timestampCurrentPacket, timestampFirstPacket);

    n_rr = prrtcpar_get_n_rr(&mblock->cpars);
    //PDEBUG(MSG_FINAL, "%s: Pure FEC or HEC? nrr=%d (from cpars: nrr=%d)\n", DEBUG_NAME, n_rr, prrtcpar_get_n_rr(cpars));
    if (n_rr){
	    //PDEBUG(MSG_FINAL, "%s: HEC!\n", DEBUG_NAME, n_rr);
		//FIXME: still required to compensate forward trip time?
		d_fec -= pcont->pack.data.data_header.group_rtt/2;
	    if (d_fec > timestampDiff)
		    dec_timeout = prrt_diff_timestamps(0, d_fec,timestampDiff);
    }
    else{
	    //PDEBUG(MSG_FINAL, "%s: Pure FEC!\n", DEBUG_NAME, n_rr);
	    if (d_target > timestampDiff)
		    dec_timeout = prrt_diff_timestamps(0, d_target,timestampDiff);
    }

//    if (d_target > timestampDiff)
//	    pack_timeout = prrt_diff_timestamps(d_target,timestampDiff);
    pack_timeout = d_target;

    //just transmit the differences between timeouts and packet timestamp
    dec_timeout += prrtcpar_get_D_fb(cpars);
    pcont->pack.data.data_header.pack_timeout = (u16)pack_timeout;
    pcont->pack.data.data_header.dec_timeout = (u16)dec_timeout;
    pcont->pack.data.data_header.fb_timeout = (u16)d_req;
    pcont->pack.data.header.index = (u16)(mblock->no_data);

    mblock->no_data++;

    //PERROR("%s: Setting dectimeout to %d at %d, ", DEBUG_NAME, dec_timeout, prrt_get_mstime(0));
    PDEBUG(MSG_BLOCKDEB, "%s: Setting dectimeout to %d at %d, ", DEBUG_NAME, dec_timeout, prrt_get_mstime(0));
    PDEBUG(MSG_BLOCKDEB, "%s: inserted packet in block(%p), now %d packets (packetno %d)\n", DEBUG_NAME,
            mblock, mblock->no_data, pcont->pack.data.header.seqnum);

    mblock->biggest_data = MAX(pcont->payload_size, mblock->biggest_data);

    return 0;
}

void insert_red_in_block(struct prrt_block *mblock, struct packet_list_entry *p)
{
    list_add_tail(&p->list, &mblock->red_head);

    if (prrt_block_all_redundancy_was_sent(mblock))
        mblock->next_red = &p->list;

    mblock->no_red++;
    PDEBUG(MSG_BLOCKDEB, "%s: Inserted red %d in block - %p\n", DEBUG_NAME, mblock->no_red, p);
}

int block_at_least_full(struct prrt_block *mblock, int k);
int ShortenBlockCodeIfNecessary(struct prrt_block *mblock, struct params_code *cpars, struct prrt_time_state *time_state);
int block_too_full(struct prrt_block *mblock);
int move_packet_to_new_block(struct prrt_block *mblock, struct prrt_block *new_block, struct list_head *iter);
void set_dectimeouts_of_packets(struct prrt_block *mblock, struct params_code *cpars);

int prrt_block_prepare(struct prrt_block *mblock, struct params_code *cpar, struct prrt_block **new_block, struct prrt_time_state *time_state)
{
    struct list_head *iter = 0;
    struct list_head *next = 0;
    int err = 0;
    int k = 1;
    int shortened_k = ShortenBlockCodeIfNecessary(mblock, cpar, time_state);

    *new_block = 0;

    k = MAX(1, MIN(prrtcpar_get_k(&mblock->cpars), prrtcpar_get_k(cpar)));
    //k = MAX(1,prrtcpar_get_k(&mblock->cpars));
    k = (shortened_k && (shortened_k < k)) ? shortened_k : k;
    mblock->cpars.k = k;
    //PDEBUG(MSG_FINAL, "%s: cpars n_rr = %d, k = %d, n = %d, no_data = %d\n", DEBUG_NAME, prrtcpar_get_n_rr(&mblock->cpars), k, mblock->cpars.n, mblock->no_data);

    if (!mblock->cpars.n){
	    mblock->cpars.n++;
	    PDEBUG(MSG_FINAL, "%s: refined cpars n_rr = %d, k = %d, n = %d\n", DEBUG_NAME, prrtcpar_get_n_rr(&mblock->cpars), k, mblock->cpars.n);
    }

    if (shortened_k){
	PDEBUG(MSG_FINAL, "%s: Shortening block due to delay: shortened_k = %d, k = %d\n", DEBUG_NAME, shortened_k, k);
    }

    if (!block_at_least_full(mblock, k))
    {
        //PERROR("%s: Not enough packets in block for coding (k = %d)\n", DEBUG_NAME, k);
        return -1;
    }

    /*allocate new block*/
    err = prrt_block_alloc(new_block, cpar);
    if (err < 0)
    {
        PERROR("%s: Not enough memory for new block\n", DEBUG_NAME);
        return -ENOMEM;
    }

    /*move supernumerous packets to new block*/
    list_for_each_prev_safe(iter, next, &mblock->data_head)
    {
        if (!block_too_full(mblock))
            break; /*only move until right number of packets are in block*/

        /*remove packet from block*/
	PDEBUG(MSG_BLOCKDEB, "%s: Move packet to new block\n", DEBUG_NAME);
        err = move_packet_to_new_block(mblock, *new_block, iter);
        if (err < 0)
        {
            PERROR("%s: Could not move packet to new block\n", DEBUG_NAME);
            break;
        }

    }

    //set_dectimeouts_of_packets(mblock, lpars);

    return err;
}

int block_at_least_full(struct prrt_block *mblock, int k)
{
    //int k = prrtcpar_get_k(cpar);

    PDEBUG(MSG_BLOCKDEB, "%s: Testing if block(%p) is full no_data == %d, k == %d\n", DEBUG_NAME,
            mblock, mblock->no_data, k);

    if (mblock->no_data >= k)
    {
        return 1;
    }

    return 0;
}

int BlockCodingDelayExceeded(u32 timeDifference, int d_fec, struct prrt_time_state *time_state);

int ShortenBlockCodeIfNecessary(struct prrt_block *mblock, struct params_code *cpars, struct prrt_time_state *time_state)
{
    int err = 0;
    struct packet_list_entry *pFirst = 0;
    struct prrt_data_packet *firstPack = 0;
    struct packet_list_entry *pLast = 0;
    //struct prrt_data_packet *lastPack = 0;
    //u32 timestampLastPacket = 0;
    u32 timestampFirstPacket = 0;
    u32 timestampDiff = 0;
    int blockCodingDelayExceeded = 0;

    err = prrt_block_get_first_data(mblock, &pFirst);
    if (err < 0)
    {
        return 0;
    }

    err = prrt_block_get_last_data(mblock, &pLast);
    if (err < 0)
    {
        return 0;
    }

    firstPack = &pFirst->pcont->pack.data;

    timestampFirstPacket = firstPack->data_header.timestamp;

    timestampDiff = prrt_diff_timestamps(time_state, prrt_get_mstime(time_state), timestampFirstPacket);

    blockCodingDelayExceeded = BlockCodingDelayExceeded(timestampDiff, prrtcpar_get_D_C(cpars), time_state);
    PDEBUG(MSG_BLOCKDEB, "%s: BlockCodingDelayExceeded? %d\n", DEBUG_NAME, blockCodingDelayExceeded);
    if (blockCodingDelayExceeded)
        return mblock->no_data;

    return 0;
}

int BlockCodingDelayExceeded(u32 timeDifference, int d_fec, struct prrt_time_state *time_state)
{
    return prrt_compare_timestamps(time_state, timeDifference, d_fec) == 1;
}

int block_too_full(struct prrt_block *mblock)
{
    int k = prrtcpar_get_k(&mblock->cpars);

    if (mblock->no_data > k)
    {
        return 1;
    }

    return 0;
}

int insert_data_packet_entry_at_front(struct prrt_block *mblock, struct packet_list_entry *p);

int move_packet_to_new_block(struct prrt_block *mblock, struct prrt_block *new_block, struct list_head *iter)
{
    struct packet_list_entry *p = 0;
    int err = 0;

    p = container_of(iter, struct packet_list_entry, list);
    list_del(iter);
    mblock->no_data--;

    err = insert_data_packet_entry_at_front(new_block, p);
    if (err < 0)
        return -1;

    return 0;
}

int insert_data_packet_entry_at_front(struct prrt_block *mblock, struct packet_list_entry *p)
{
    list_add(&p->list, &mblock->data_head);

    mblock->no_data++;
    mblock->biggest_data = MAX(p->pcont->payload_size, mblock->biggest_data);

    return 0;
}

void set_dectimeouts_of_packets(struct prrt_block *mblock, struct params_code *cpars)
{
    int err = 0;
    int d_fec = prrtcpar_get_D_Fec(cpars);
    int d_target = prrtcpar_get_D_target(cpars);
    int n_rr = prrtcpar_get_n_rr(&mblock->cpars);
    int n_cc = prrtcpar_get_n_cc(&mblock->cpars, 0);
    struct packet_list_entry *pFirst = 0;
    struct prrt_data_packet *firstPack = 0;
    struct packet_list_entry *p = 0;
    struct prrt_packet_container *pcont = 0;
    struct prrt_data_packet *pack = 0;
    u32 timestampFirstPacket = 0;
    u32 dectimeoutBlock = 0;
    struct list_head *iter = 0;

    err = prrt_block_get_first_data(mblock, &pFirst);
    if (err < 0)
        return;

    firstPack = &pFirst->pcont->pack.data;
    timestampFirstPacket = firstPack->data_header.timestamp;

    if (n_rr && n_cc){
	    dectimeoutBlock = timestampFirstPacket + (u32) d_fec;
    }
    else{
	    dectimeoutBlock = timestampFirstPacket + (u32) d_target;
    }

    list_for_each(iter, &mblock->data_head)
    {
        p = container_of(iter, struct packet_list_entry, list);
        pcont = p->pcont;

        pack = &pcont->pack.data;
        pack->data_header.dec_timeout = dectimeoutBlock;
    PDEBUG(MSG_BLOCKDEB, "%s: Setting dectimeout to %d at %d\n", DEBUG_NAME, dectimeoutBlock, prrt_get_mstime(0));
    }
}

static int gather_data_to_code(struct prrt_block *mblock, char **data_to_code, int k)
{
    int i = 0;
    struct packet_list_entry *p = 0;
    struct list_head *iter = 0;

    PDEBUG(MSG_BLOCKDEB, "%s: Gathering data to code\n", DEBUG_NAME);

    /*gather data to code*/
    list_for_each(iter, &mblock->data_head)
    {
        p = container_of(iter, struct packet_list_entry, list);
        data_to_code[i] = prrt_packet_get_data(p->pcont);
        i++;
    }
    if (i < k)
    {
        PERROR("%s: Did not find all data packets in block, only %d\n", DEBUG_NAME, i);
        return -1;
    }

    return 0;
}

int prrt_block_get_coder(struct prrt_block *mblock, struct prrt_coder **cod)
{
    int err = 0;
    int k = prrtcpar_get_k(&mblock->cpars);
    int n = prrtcpar_get_n(&mblock->cpars);

    if (k > 1 && n > k) {
      err = prrt_fec_get_coder(cod, n, k);
    }

    return err;
}

int code_startup_redundancy(struct prrt_block *mblock, char **data_to_code, struct prrt_coder *cod, struct prrt_time_state *time_state);

int prrt_block_code(struct prrt_block *mblock, struct prrt_coder *cod, u16 *next_red_packno,
        struct semaphore *pno_lock, struct prrt_time_state *time_state)
{
    int err = 0;
    int k = prrtcpar_get_k(&mblock->cpars);
    int n = prrtcpar_get_n(&mblock->cpars);
    // 	unsigned long timestamp = jiffies;
    char *data_to_code[k];

    mblock->next_red_pno = prrt_increment_packetno_locked(next_red_packno, n - k, pno_lock);

    /*this means we don't need to code anything*/
	//TODO: does this also hold true for k == 1?
    if (k == 1 || n == k){
      mblock->is_coded = 1;
      return 0;
    }

    err = gather_data_to_code(mblock, data_to_code, k);
    if (err < 0)
        return -1;

    mblock->no_red = 0;
    PDEBUG(MSG_BLOCKDEB, "%s: Coding replacement packets\n", DEBUG_NAME);
    /*code code_start replacement packets*/
    err = code_startup_redundancy(mblock, data_to_code, cod, time_state);

    mblock->is_coded = 1;

    return err;
}

int setup_redundancy_packet(struct prrt_block *mblock, struct prrt_red_packet *rpack, int index);

int code_startup_redundancy(struct prrt_block *mblock, char **data_to_code, struct prrt_coder *cod, struct prrt_time_state *time_state)
{
    struct prrt_red_packet *rpack = 0;
    int i = 0;
    int err = 0;
    struct packet_list_entry *p = 0;
    struct prrt_packet_container *pcont = 0;
    int code_start = prrtcpar_get_code_start(&mblock->cpars);
    int k = prrtcpar_get_k(&mblock->cpars);

    for (i = k; i < k + code_start; i++)
    {
        pcont = prrt_packet_alloc(time_state);
        prrt_list_entry_alloc(&p);
        if (!pcont || !p)
        {
            err = -ENOMEM;
            PERROR("%s: Not enough memory when coding packets\n", DEBUG_NAME);
            goto free_mem;
        }

        p->pcont = pcont;
        rpack = &pcont->pack.red;

        prrt_fec_encode(cod, (gf **) data_to_code, (gf *) & rpack->data, i, mblock->biggest_data);

	pcont->payload_size = (u16)(mblock->biggest_data);
        setup_redundancy_packet(mblock, rpack, i);

        PDEBUG(MSG_BLOCKDEB, "%s: Inserting %dth redundancy packet in block\n", DEBUG_NAME,
                i - k + 1);
        insert_red_in_block(mblock, p);
    }

    return err;

free_mem:
    prrt_packet_destroy(&pcont);
    prrt_list_entry_free(&p);
    return err;
}

int setup_redundancy_packet(struct prrt_block *mblock, struct prrt_red_packet *rpack, int index)
{
    int k = prrtcpar_get_k(&mblock->cpars);
    int n = prrtcpar_get_n(&mblock->cpars);
    struct packet_list_entry *first = 0;

    prrt_block_get_first_data(mblock, &first);

    /*set header information of redundancy packet*/
    rpack->header.packtype = (rpack->header.packtype & ~0xF) | ((u8)packtype_red & 0xF);
    //rpack->data_header.packetno = mblock->next_red_pno++;
    //rpack->header.payload_size = (u16) (mblock->biggest_data);
    // 		rpack->header.timestamp = (u32)jif_to_ms(timestamp);

    rpack->red_header.sn_base = (u16) (first->pcont->pack.data.header.seqnum);
    rpack->red_header.k = (u8) k;
    rpack->red_header.n = (u8) n;
    rpack->header.index = (u8) index;
    //rpack->red_header.round = 0;

    return 0;
}

int prrt_block_code_next_red(struct prrt_block *mblock, struct prrt_coder *cod, struct prrt_time_state *time_state)
{
    int err = 0;
    int k = prrtcpar_get_k(&mblock->cpars);
    int n_r = mblock->no_red;
    int n = prrtcpar_get_n(&mblock->cpars);
    // 	unsigned long timestamp = jiffies;
    char *data_to_code[k];
    struct packet_list_entry *p = 0;
    struct packet_list_entry *first = 0;
    struct prrt_packet_container *pcont = 0;
    struct prrt_red_packet *rpack = 0;
    int index = k + n_r;

    prrt_block_get_first_data(mblock, &first);

    /*this means we don't need to code anything*/
    if (n == k || k + n_r >= n)
        return 0;

    err = gather_data_to_code(mblock, data_to_code, k);
    if (err < 0)
        return -1;

    /*code the next redundancy packet*/
    pcont = prrt_packet_alloc(time_state);
    prrt_list_entry_alloc(&p);
    if (!pcont || !p)
    {
        err = -ENOMEM;
        PERROR("%s: Not enough memory when coding packets\n", DEBUG_NAME);
        goto free_mem;
    }

    p->pcont = pcont;
    rpack = &pcont->pack.red;

    if (cod)
    prrt_fec_encode(cod, (gf **) data_to_code, (gf *) & rpack->data, index, mblock->biggest_data);
	else
        PERROR("%s: Coder is not valid\n", DEBUG_NAME);

    pcont->payload_size = (u16)(mblock->biggest_data);
    setup_redundancy_packet(mblock, rpack, index);

    /*insert redundancy packet in block*/
    insert_red_in_block(mblock, p);

    return 0;

free_mem:
    prrt_packet_destroy(&pcont);
    prrt_list_entry_free(&p);
    return err;
}


int prrt_block_get_packet(struct prrt_block *mblock, int packetno, struct packet_list_entry **p)
{
    struct list_head *iter = 0;
    int found = 0;

    list_for_each(iter, &mblock->data_head)
    {
        *p = container_of(iter, struct packet_list_entry, list);

        if (prrt_comp_packetnos((*p)->pcont->pack.data.header.seqnum, packetno) == 0)
        {
            found = 1;
        }
    }

    if (!found)
    {
        *p = 0;
        return -1;
    }

    return 0;
}

int prrt_block_get_next_redundancy_packet(struct prrt_block *mblock, struct packet_list_entry **p)
{
    *p = 0;

    if (prrt_block_all_redundancy_was_sent(mblock))
        return -1;

    *p = container_of(mblock->next_red, struct packet_list_entry, list);
    mblock->next_red = mblock->next_red->next;

    return 0;
}

int prrt_block_contains_packet(struct prrt_block *mblock, u16 packetno)
{
    int err = 0;
    u16 first_pno = 0;
    u16 last_pno = 0;
    struct packet_list_entry *first = 0;

    err = prrt_block_get_first_data(mblock, &first);
    if (err < 0)
        return 0;

    first_pno = first->pcont->pack.data.header.seqnum;
    last_pno = prrt_increment_packetno(first_pno, mblock->no_data - 1);

    if (prrt_comp_packetnos(first_pno, packetno) <= 0 &&
            prrt_comp_packetnos(packetno, last_pno) <= 0)
    {
        return 1;
    }

    return 0;
}

