/* 
 * (c) 2010 Saarland University, Telecommunications Lab 
 * (prrt@nt.uni-saarland.de)
 *
 * The contents of this file are subject to the Mozilla Public License
 + Version 1.1 (the "License"); you may not use this file except in
 + compliance with the License. You may obtain a copy of the License at
 + http://www.mozilla.org/MPL/
 + 
 + Software distributed under the License is distributed on an "AS IS"
 + basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 + License for the specific language governing rights and limitations
 + under the License.
 + 
 + The Original Code is "Predictably Reliable Real-time Transport (PRRT)".
 + 
 + The Initial Developer of the Original Code is Saarland University.
 + All Rights Reserved.
 + 
 + Contributor(s): Jochen Gruen (Saarland University), Manuel Gorius
 + (Saarland University), Yongtao Shuai (Saarland University).
 + 
 + Disclaimer: This Alpha version of PRRT is not supported and stable yet.
 + Do not use it in a production environment: it may crash your device, 
 + so please make sure that you have a backup of your data before
 + installing PRRT.
 */

#ifndef PRRT_BLOCK_H
#define PRRT_BLOCK_H

#include <linux/list.h>
#include "lists/prrt_list.h"
#include "vdmcode/prrt_block_code.h"
#include "parameters/param_code.h"

struct prrt_socket;

/**
 * This structure represents a block of coded informationen.
 * Such a block contains two partial lists:
 * - The list with the pure data packets.
 * - The list with the created redundancy packets.
 * It also saves which redundany packet of the block is to be send next.
 */
struct prrt_block {
	struct list_head list;	/**< This makes the blocks listable.*/
	int no_data;		/**< #data packets in the block at the moment*/
	int no_red;		/**< #redundancy packets in the block at the moment*/
	struct params_code cpars;/**< This struct contains the coding parameters for this block.*/
	int biggest_data;	/**< Size of the biggest data in this block in bytes.*/
	struct list_head data_head;	/**< Head of the list for the data packets.*/
	struct list_head red_head;	/**< Head od the list for the redundancy packets.*/
	struct list_head *next_red;	/**< Next redundancy packet to be send*/
	long next_red_pno;		/**< The next number for the redundancy packets in this block.*/
	char first_red_send;	/**< Was the precautionary redundancy of this block already sent.*/
        short is_coded;         /**< Is this block already coded for the first time.*/
	u8 retr_round;		/**< current (re-)transmission round*/
};

int prrt_block_cache_init(void);
void prrt_block_cache_uninit(void);

/**
 * Allocate space for a block.
 */
int prrt_block_alloc(struct prrt_block **mblock, struct params_code *cpar);

/**
 * Frees the prrt_block data structure.
 */
void prrt_block_free(struct prrt_block **mblock);

/**
 * Frees the prrt_block data structure and deletes all packets in it.
 * It should remove the packets from their lists before deleting them.
 */
void prrt_block_destroy(struct prrt_block **mblock);

/**
 * Is the redundancy of this block already coded.
 */
int prrt_block_is_already_coded(struct prrt_block *mblock);

/**
 * Are there any data packets in this block.
 */
int prrt_block_is_empty(struct prrt_block *mblock);

/**
 * Have we sent all available redundancy and can we code further redundancy for this block or
 * was all redundancy already build.
 */
int prrt_block_can_and_must_code(struct prrt_block *mblock);

/**
 * All redundancy that was build until now was already sent.
 */
int prrt_block_all_redundancy_was_sent(struct prrt_block *mblock);

/**
 * Has more time than delay passed since the last packet from this block was sent.
 * \param time_now actual time.
 * \param delay the allowed maximum time difference between the sending of the last packet and now.
 */
int prrt_block_too_old_than_timestamp(struct prrt_block *mblock, u32 time_now, u32 delay, struct prrt_time_state *time_state);

/**
 * Insert a data packet into a block.
 * \param pcont the packet container that should be added to the block.
 * \param p this will point to the list entry that was generated.
 * \return 0 on success.
 */
int prrt_block_insert_data_packet(struct prrt_block *mblock, struct prrt_packet_container *pcont,
				 struct packet_list_entry **p, struct params_code *cpars);

/**
 * Prepares a block for coding.
 * Due to changing coding parameters it can happen that there are too many data packets in a block.
 * This function moves the supernumerous data packets to a new block and returns a pointer to it.
 * \param k maximum number of data packets in the block.
 * \return Pointer to the newly created block, 0 if no packets were moved.
 */
int prrt_block_prepare(struct prrt_block *mblock, struct params_code *cpar,
		       struct prrt_block **new_block, struct prrt_time_state *time_state);

/**
 * Get a coder object matching the coding parameters of the block.
 * \param cod the coder that is retrieved.
 */
int prrt_block_get_coder(struct prrt_block *mblock, struct prrt_coder **cod);

/**
 * Create redundancy packets for a block.
 * \param cod the coder that is used to create the redundancy packets.
 * \param next_red_packno the number of the next redundancy packet.
 */
int prrt_block_code(struct prrt_block *mblock, struct prrt_coder *cod, u16 *next_red_packno,
		   struct semaphore *pno_lock, struct prrt_time_state *time_state);

/**
 * This creates one redundancy packet for a block.
 * \param pcont this is the packet that is created.
 * \return 0 on success, <0 on failure.
 */
int prrt_block_code_next_red(struct prrt_block *mblock, struct prrt_coder *cod, struct prrt_time_state *time_state);

/**
 * Retrieve the packetno of the first data packet in the block.
 */
extern u16 prrt_block_get_first_packetno(struct prrt_block *mblock);

/**
 * Retrieve the first data packet in the block.
 */
extern int prrt_block_get_first_data(struct prrt_block *mblock, struct packet_list_entry **first);

/**
 * Retrieve the last data packet in the block.
 */
extern int prrt_block_get_last_data(struct prrt_block *mblock, struct packet_list_entry **last);

/**
 * Retrieve the first redundancy packet in the block.
 */
extern int prrt_block_get_first_red(struct prrt_block *mblock, struct packet_list_entry **first);

/**
 * Retrieve a data packet from a block.
 * \param packetno packet number of the packet.
 * \param p this will point to the list entry of the packet.
 * \return 0 on success.
 */
int prrt_block_get_packet(struct prrt_block *mblock, int packetno, struct packet_list_entry **p);

/**
 * Retrieve the next redundancy packet from this block that was not yet send.
 */
int prrt_block_get_next_redundancy_packet(struct prrt_block *mblock, struct packet_list_entry **p);

/**
 * Does this block contain the data packet with the given sequence number.
 * \param packetno sequence number to check for.
 */
int prrt_block_contains_packet(struct prrt_block *mblock, u16 packetno);

#endif /*PRRT_BLOCK_H*/

