/* 
 * (c) 2010 Saarland University, Telecommunications Lab 
 * (prrt@nt.uni-saarland.de)
 *
 * The contents of this file are subject to the Mozilla Public License
 + Version 1.1 (the "License"); you may not use this file except in
 + compliance with the License. You may obtain a copy of the License at
 + http://www.mozilla.org/MPL/
 + 
 + Software distributed under the License is distributed on an "AS IS"
 + basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 + License for the specific language governing rights and limitations
 + under the License.
 + 
 + The Original Code is "Predictably Reliable Real-time Transport (PRRT)".
 + 
 + The Initial Developer of the Original Code is Saarland University.
 + All Rights Reserved.
 + 
 + Contributor(s): Jochen Gruen (Saarland University), Manuel Gorius
 + (Saarland University), Yongtao Shuai (Saarland University).
 + 
 + Disclaimer: This Alpha version of PRRT is not supported and stable yet.
 + Do not use it in a production environment: it may crash your device, 
 + so please make sure that you have a backup of your data before
 + installing PRRT.
 */

#include "prrt_fbpacket.h"
#include "prrt_socket.h"
#include <linux/slab.h>
#include "defines.h"

#define DEBUG_NAME "PRRT_FBPACKET"

static struct kmem_cache *fbpacket_cache;
static struct semaphore fbpacket_cache_lock;

int prrt_fbpacket_cache_init(void)
{
	init_MUTEX(&fbpacket_cache_lock);
	fbpacket_cache = kmem_cache_create("fbpacket_cache",
					  MAX_FB_PACKET_CONT_SIZE,
					  0, 0,
					  NULL);

	if(!fbpacket_cache) {
		PERROR("%s: Could not create feedback packet cache\n", DEBUG_NAME);
		return -1;
	}

	return 0;
}

void prrt_fbpacket_cache_uninit(void)
{
	kmem_cache_destroy(fbpacket_cache);
}

int prrt_fbpacket_alloc(struct prrt_fbpacket_container **mfbcont)
{
	PDEBUG(MSG_ALLOC, "%s: Allocating fbpacket\n", DEBUG_NAME);
	down(&fbpacket_cache_lock);
	*mfbcont = kmem_cache_alloc(fbpacket_cache, GFP_KERNEL);
	up(&fbpacket_cache_lock);

	PDEBUG(MSG_ALLOC, "%s: Fbpacket allocated\n", DEBUG_NAME);
	/*zero out packet for coding purposes*/
	if(!(*mfbcont)) {
		PERROR("%s: Could not allocate memory for fbpacket\n", DEBUG_NAME);
		return -1;
	} else {
		bzero(*mfbcont, MAX_FB_PACKET_CONT_SIZE);
	}

	return 0;
}

//int prrt_chanfbpacket_alloc(struct prrt_chanfbpacket **mfbcont)
//{
//	PDEBUG(MSG_ALLOC, "%s: Allocating chanfbpacket\n", DEBUG_NAME);
//	down(&fbpacket_cache_lock);
//	*mfbcont = kmem_cache_alloc(fbpacket_cache, GFP_KERNEL);
//	up(&fbpacket_cache_lock);
//
//	PDEBUG(MSG_ALLOC, "%s: Chanfbpacket allocated\n", DEBUG_NAME);
//	/*zero out packet for coding purposes*/
//	if(!(*mfbcont)) {
//		PERROR("%s: Could not allocate memory for chanfbpacket\n", DEBUG_NAME);
//		return -1;
//	} else {
//		bzero(*mfbcont, MAX_FB_PACKET_CONT_SIZE);
//	}
//
//	return 0;
//}

void prrt_fbpacket_free(struct prrt_fbpacket_container **mfbcont)
{
	if(!(*mfbcont))
		return;

	down(&fbpacket_cache_lock);
	kmem_cache_free(fbpacket_cache, *mfbcont);
	up(&fbpacket_cache_lock);
	*mfbcont = 0;
	PDEBUG(MSG_ALLOC, "%s: Fbpacket deleted\n", DEBUG_NAME);
}

//void prrt_chanfbpacket_free(struct prrt_chanfbpacket_container **mfbcont)
//{
//	if(!(*mfbcont))
//		return;
//
//	down(&fbpacket_cache_lock);
//	kmem_cache_free(fbpacket_cache, *mfbcont);
//	up(&fbpacket_cache_lock);
//	*mfbcont = 0;
//	PDEBUG(MSG_ALLOC, "%s: Chanfbpacket deleted\n", DEBUG_NAME);
//}

int prrt_fbpacket_get_block(struct prrt_fbpacket_container *mfbcont, struct prrt_fbblock *fbblock, int blockno)
{
	struct prrt_fbblock *bl = 0;

	/*this means this block does not exist*/
	if(blockno >= mfbcont->block_count || blockno < 0) {
		PERROR("%s: Error when requesting block in fb packet\n", DEBUG_NAME);
		return -1;
	}

	bl = &((&mfbcont->fb.blocks)[blockno]);

	fbblock->packno = bl->packno;
	fbblock->retr_round = bl->retr_round;

	return 0;
}

int prrt_fbpacket_add_block(struct prrt_fbpacket_container *mfbcont, struct prrt_fbblock *fbblock)
{
	struct prrt_fbblock *bl = 0;

	/*this means this block is full*/
	if(prrt_fbpacket_is_full(mfbcont)) {
		PERROR("%s: Error when trying to add block to fb packet\n", DEBUG_NAME);
		return -1;
	}

	bl = &((&mfbcont->fb.blocks)[mfbcont->block_count++]);

	bl->packno = fbblock->packno;
	bl->retr_round = fbblock->retr_round;

	mfbcont->payload_size += PRRT_FB_BLOCK_SIZE;

	return 0;
}

int prrt_fbpacket_is_full(struct prrt_fbpacket_container *mfbcont)
{
	//if (TYPEII_HEC_MODE){
	//	// just one fb block per fb packet
	//	return (mfbcont->payload_size >= MAX_FB_SIZE);
	//}
	//else{
		return (mfbcont->payload_size >= PRRT_FB_BLOCK_SIZE);
	//}
}

void prrt_fbpacket_hton(struct prrt_fbpacket_container *mfbcont)
{
	int i = 0;
	struct prrt_fbblock *bl = 0;

	for(i=0; i<mfbcont->block_count; i++) {
		bl = &((&mfbcont->fb.blocks)[i]);
		bl->packno = htons(bl->packno);
		bl->retr_round = htons(bl->retr_round);
	}

	//mfbcont->fb.header.packetno = htons(mfbcont->fb.header.packetno);
	//mfbcont->fb.header.payload_size = htons(mfbcont->fb.header.payload_size);
	mfbcont->fb.header.index = htons(mfbcont->fb.header.index);
  // NOTE : we don't need hton(), because receiver_addr is already network-order.
	//mfbcont->fb.fbheader.receiver_addr = htonl(mfbcont->fb.fbheader.receiver_addr);
	mfbcont->fb.fbheader.rtt_probe = htonl(mfbcont->fb.fbheader.rtt_probe);
	mfbcont->fb.fbheader.ftt = htonl(mfbcont->fb.fbheader.ftt);
	mfbcont->fb.fbheader.plr = htonl(mfbcont->fb.fbheader.plr);
	mfbcont->fb.fbheader.gap = htons(mfbcont->fb.fbheader.gap);
	mfbcont->fb.fbheader.ngap = htons(mfbcont->fb.fbheader.ngap);
	mfbcont->fb.fbheader.burst = htons(mfbcont->fb.fbheader.burst);
	mfbcont->fb.fbheader.nburst = htons(mfbcont->fb.fbheader.nburst);
	mfbcont->fb.fbheader.bw_est = htonl(mfbcont->fb.fbheader.bw_est);
	mfbcont->fb.fbheader.buf_fb = htonl(mfbcont->fb.fbheader.buf_fb);
}

void prrt_fbpacket_ntoh(struct prrt_fbpacket_container *mfbcont)
{
	int i = 0;
	struct prrt_fbblock *bl = 0;

	//mfbcont->fb.header.packetno = ntohs(mfbcont->fb.header.packetno);
	//mfbcont->fb.header.payload_size = ntohs(mfbcont->fb.header.payload_size);
	mfbcont->fb.header.index = ntohs(mfbcont->fb.header.index);
  // NOTE : The address in kernel space should be network ordered, host ordered in user space.
	//mfbcont->fb.fbheader.receiver_addr = ntohl(mfbcont->fb.fbheader.receiver_addr);
	mfbcont->fb.fbheader.rtt_probe = ntohl(mfbcont->fb.fbheader.rtt_probe);
	mfbcont->fb.fbheader.ftt = ntohl(mfbcont->fb.fbheader.ftt);
	mfbcont->fb.fbheader.plr = ntohl(mfbcont->fb.fbheader.plr);
	mfbcont->fb.fbheader.gap = ntohs(mfbcont->fb.fbheader.gap);
	mfbcont->fb.fbheader.ngap = ntohs(mfbcont->fb.fbheader.ngap);
	mfbcont->fb.fbheader.burst = ntohs(mfbcont->fb.fbheader.burst);
	mfbcont->fb.fbheader.nburst = ntohs(mfbcont->fb.fbheader.nburst);
	mfbcont->fb.fbheader.bw_est = ntohl(mfbcont->fb.fbheader.bw_est);
	mfbcont->fb.fbheader.buf_fb = ntohl(mfbcont->fb.fbheader.buf_fb);

	for(i=0; i<mfbcont->block_count; i++) {
		bl = &((&mfbcont->fb.blocks)[i]);
		bl->packno = ntohs(bl->packno);
		bl->retr_round = ntohs(bl->retr_round);
	}
}

//void prrt_chanfbpacket_hton(struct prrt_chanfbpacket_container *mfbcont)
//{
//	mfbcont->fb.header.packetno = htons(mfbcont->fb.header.packetno);
//	mfbcont->fb.header.payload_size = htons(mfbcont->fb.header.payload_size);
//	mfbcont->fb.fbheader.receiver_addr = htonl(mfbcont->fb.fbheader.receiver_addr);
//	mfbcont->fb.fbheader.rtt_probe = htonl(mfbcont->fb.fbheader.rtt_probe);
//	mfbcont->fb.chanfb.res_plr = htonl(mfbcont->fb.chanfb.res_plr);
//	mfbcont->fb.chanfb.plr = htons(mfbcont->fb.chanfb.plr);
//	mfbcont->fb.chanfb.cor = htons(mfbcont->fb.chanfb.cor);
//}
//
//void prrt_chanfbpacket_ntoh(struct prrt_chanfbpacket_container *mfbcont)
//{
//	mfbcont->fb.header.packetno = ntohs(mfbcont->fb.header.packetno);
//	mfbcont->fb.header.payload_size = ntohs(mfbcont->fb.header.payload_size);
//	mfbcont->fb.fbheader.receiver_addr = ntohl(mfbcont->fb.fbheader.receiver_addr);
//	mfbcont->fb.fbheader.rtt_probe = ntohl(mfbcont->fb.fbheader.rtt_probe);
//	mfbcont->fb.chanfb.res_plr = ntohl(mfbcont->fb.chanfb.res_plr);
//	mfbcont->fb.chanfb.plr = ntohs(mfbcont->fb.chanfb.plr);
//	mfbcont->fb.chanfb.cor = ntohs(mfbcont->fb.chanfb.cor);
//}

void LogFeedbackPacket(struct prrt_fbpacket_container *mfbcont);

void prrt_fbpacket_log_sending(struct prrt_fbpacket_container *mfbcont)
{
	PLOGPACK("Sending Feedback Packet:\n");
	LogFeedbackPacket(mfbcont);
}

void LogFeedbackHeader(struct prrt_fbheader *fbheader);
void LogFeedbackPacketContent(struct prrt_fbblock *blocks, int number_blocks);

void LogFeedbackPacket(struct prrt_fbpacket_container *mfbcont)
{
	prrt_header_log_general(&mfbcont->fb.header);
	LogFeedbackHeader(&mfbcont->fb.fbheader);
	LogFeedbackPacketContent(&mfbcont->fb.blocks, mfbcont->block_count);
}

void LogFeedbackHeader(struct prrt_fbheader *fbheader)
{
	PLOGPACK("Receiver Address: %d\n", fbheader->receiver_addr);
	PLOGPACK("RTT probe: %d\n", fbheader->rtt_probe);
	PLOGPACK("------------------------------------------\n");
}

void LogFeedbackPacketContent(struct prrt_fbblock *blocks, int number_blocks)
{
	int i = 0;

	PLOGPACK("Feedback as - packetnumber/retransmission round:\n");
	for(i=0; i<number_blocks; ++i) {
		PLOGPACK("%d/%d, \n", blocks[i].packno, blocks[i].retr_round);
	}
	PLOGPACK("------------------------------------------\n");
}

void prrt_fbpacket_log_receiving(struct prrt_fbpacket_container *mfbcont)
{
	PLOGPACK("Receiving Feedback Packet:\n");
	LogFeedbackPacket(mfbcont);
}

