/* 
 * (c) 2010 Saarland University, Telecommunications Lab 
 * (prrt@nt.uni-saarland.de)
 *
 * The contents of this file are subject to the Mozilla Public License
 + Version 1.1 (the "License"); you may not use this file except in
 + compliance with the License. You may obtain a copy of the License at
 + http://www.mozilla.org/MPL/
 + 
 + Software distributed under the License is distributed on an "AS IS"
 + basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 + License for the specific language governing rights and limitations
 + under the License.
 + 
 + The Original Code is "Predictably Reliable Real-time Transport (PRRT)".
 + 
 + The Initial Developer of the Original Code is Saarland University.
 + All Rights Reserved.
 + 
 + Contributor(s): Jochen Gruen (Saarland University), Manuel Gorius
 + (Saarland University), Yongtao Shuai (Saarland University).
 + 
 + Disclaimer: This Alpha version of PRRT is not supported and stable yet.
 + Do not use it in a production environment: it may crash your device, 
 + so please make sure that you have a backup of your data before
 + installing PRRT.
 */

#ifndef PRRT_FBPACKET_H
#define PRRT_FBPACKET_H

#include <linux/types.h>
#include "prrt_packet.h"

extern struct prrt_socket msock;

/**
 * The header for feedback packets that are send by the receiver.
 */
struct prrt_fbheader {
	u32	receiver_addr;	/**< The ip address of the receiver.*/
	u32	rtt_probe;	/**< Timestamp of most recent data packet + delay since reception of this packet.*/
	u32 	ftt;		/**< Forward trip time*/
	u32 	plr;		/**< Packet loss rate measured at the receiver.*/
	u16 	gap;		/**< Channel statistics*/
	u16 	ngap;		/**< Channel statistics*/
	u16 	burst;		/**< Channel statistics*/
	u16 	nburst;		/**< Channel statistics*/
	u32 	bw_est;		/**< Bandwidth estimate measured at the receiver.*/
	u32 	buf_fb;		/**< Buffer information from receiver application, sent to server as a feedback for flow control or scheduling packet sending. The unit can be various, specified by application.*/
};

/**
 * Such a block represents the loss notification for one packet.
 */
struct prrt_fbblock {
	u16 packno;	/**< The packetno of the first miss.*/
	u16 retr_round;	/**< The retransmission round of the first miss.(first transmission is 0)*/
};

/**
 * Such a block represents the channel feedback.
 */
struct prrt_chanfb {
	u32 res_plr;	/**< Observed residual PLR */
	u16 plr;	/**< Observed original PLR */
	u16 cor;	/**< Correlation coefficient (for GE channel model) */
};

#define MAX_NUMBER_RETRROUNDS 65535	/**< Maximum number of retransmission rounds, given through
					 *   size of packet header field.*/

/**
 * An structure to represent feedback packets.
 * The payload_size in the header of these packets gives the number of blocks at the end of the packet.
 * The block at the end is repeated header.payload_size times which means we have the numbers and
 * retransmission rounds of that many packets in the feedback packet.
 * ATTENTION: Do not assume anything about the order of the numbers in the packets!!
 */
struct prrt_fbpacket {
	struct prrt_header	header;
	struct prrt_fbheader	fbheader;
	struct prrt_fbblock	blocks;	/**< This block represents the feedback for one packet.*/
};

/**
 * PRRT fbpacket container
 */
struct prrt_fbpacket_container {
	u16 payload_size;
	u16 block_count;
	struct prrt_fbpacket fb;
};

#define PRRT_FB_HEADER_SIZE sizeof(struct prrt_fbheader)
#define PRRT_FB_BLOCK_SIZE sizeof(struct prrt_fbblock)
#define MAX_FB_SIZE (MAXFBS) * PRRT_FB_BLOCK_SIZE
#define MAX_FB_PACKET_SIZE PRRT_HEADER_SIZE + PRRT_FB_HEADER_SIZE + MAX_FB_SIZE
#define MAX_FB_PACKET_CONT_SIZE 2*sizeof(u16) + MAX_FB_PACKET_SIZE

/**
 * An structure to represent channel feedback packets.
 */
struct prrt_chanfbpacket {
	struct prrt_header	header;
	struct prrt_fbheader	fbheader;
	struct prrt_chanfb	chanfb;	/**< This block represents the feedback for one packet.*/
};

#define CHANFBPACKETSIZE sizeof(struct prrt_chanfbpacket)

/**
 * Initialise the cache for this structure.
 */
int prrt_fbpacket_cache_init(void);
void prrt_fbpacket_cache_uninit(void);

/**
 * Allocate a new feedback packet.
 * \param mfbcont this variable will hold a pointer to the allocate packet if successful.
 * \return 0 on success, -1 if allocation failed.
 */
int prrt_fbpacket_alloc(struct prrt_fbpacket_container **mfbcont);

/**
 * Allocate a new channel feedback packet.
 * \param mfbcont this variable will hold a pointer to the allocate packet if successful.
 * \return 0 on success, -1 if allocation failed.
 */
//int prrt_chanfbpacket_alloc(struct prrt_chanfbpacket **mfbcont);

/**
 * Free the memory that was allocated for a feedback packet.
 * \param mfbcont the feedback packet that should be freed.
 */
void prrt_fbpacket_free(struct prrt_fbpacket_container **mfbcont);

/**
 * Free the memory that was allocated for a channel feedback packet.
 * \param mfbcont the feedback packet that should be freed.
 */
//void prrt_chanfbpacket_free(struct prrt_chanfbpacket **mfbcont);

/**
 * Get the values of a block in the packet.
 * \param fbblock the block that holds the values on success.
 * \param blockno the number of the block in the packet.(first block is 0)
 * \return 0 on success, -1 else.
 */
int prrt_fbpacket_get_block(struct prrt_fbpacket_container *mfbcont, struct prrt_fbblock *fbblock, int blockno);

/**
 * Sets a new block at the end of the packet to a certain value.
 * \param fbblock the block that contains the values to set.
 * \return 0 on success, -1 if the packet is full.
 */
int prrt_fbpacket_add_block(struct prrt_fbpacket_container *mfbcont, struct prrt_fbblock *fbblock);

/**
 * Checks if the maximum number of blocks is already contained in the packet.
 */
int prrt_fbpacket_is_full(struct prrt_fbpacket_container *mfbcont);

/* functions to convert packet contents to network/host byte order*/
void prrt_fbpacket_hton(struct prrt_fbpacket_container *mfbcont);
void prrt_fbpacket_ntoh(struct prrt_fbpacket_container *mfbcont);
//void prrt_chanfbpacket_hton(struct prrt_chanfbpacket *mfbcont);
//void prrt_chanfbpacket_ntoh(struct prrt_chanfbpacket *mfbcont);

void prrt_fbpacket_log_sending(struct prrt_fbpacket_container *mfbcont);
void prrt_fbpacket_log_receiving(struct prrt_fbpacket_container *mfbcont);

#endif /*PRRT_FBPACKET_H*/

