/* 
 * (c) 2010 Saarland University, Telecommunications Lab 
 * (prrt@nt.uni-saarland.de)
 *
 * The contents of this file are subject to the Mozilla Public License
 + Version 1.1 (the "License"); you may not use this file except in
 + compliance with the License. You may obtain a copy of the License at
 + http://www.mozilla.org/MPL/
 + 
 + Software distributed under the License is distributed on an "AS IS"
 + basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 + License for the specific language governing rights and limitations
 + under the License.
 + 
 + The Original Code is "Predictably Reliable Real-time Transport (PRRT)".
 + 
 + The Initial Developer of the Original Code is Saarland University.
 + All Rights Reserved.
 + 
 + Contributor(s): Jochen Gruen (Saarland University), Manuel Gorius
 + (Saarland University), Yongtao Shuai (Saarland University).
 + 
 + Disclaimer: This Alpha version of PRRT is not supported and stable yet.
 + Do not use it in a production environment: it may crash your device, 
 + so please make sure that you have a backup of your data before
 + installing PRRT.
 */

#include <linux/net.h>
#include <linux/slab.h>
#include <linux/in.h>
#include <linux/syscalls.h>
#include <linux/types.h>
#include <linux/inet.h>
#include <linux/sched.h>
#include <linux/delay.h>


#include "defines.h"
#include "prrt_socket.h"
#include "prrt_packet.h"
#include "receive_data.h"
#include "send_data.h"
#include "send_feedback.h"
#include "recv_feedback.h"
#include "lists/prrt_block_list.h"
#include "kui/kui_error.h"
#include "kernel_wrappers.h"


static const char *DEBUG_NAME = "PRRT_SOCK";

/*from below*/
static int build_addr(struct sockaddr_in *addr_s, char* ipaddr, u_short port);
static int build_multi_req(struct ip_mreq *req, char* ipaddr);
static int isMulticastAddress(char *sock_addr);
// static int in4_ntoa(__be32 baddr, char* caddr);
static void print_stats(struct prrt_socket *msock);
static int create_feedback_socket(struct prrt_socket *msock);
static int init_sending(struct prrt_socket *msock);

// static int init_sock_sending(struct prrt_socket *msock)
// {
//
// }
//
// static int init_sock_receiving(struct prrt_socket *msock)
// {
//
// }

int prrt_socket(struct prrt_socket **msock, int kui_port) {
    int err = 0;
    struct timeval totime;
    struct timeval totimefb;

    totime.tv_sec = POLLTIMEOUT;
    totime.tv_usec = POLLTIMEOUTUS;

    totimefb.tv_sec = POLLTIMEOUTFB;
    totimefb.tv_usec = 0;

    PDEBUG(MSG_COMMON, "%s: Creating new msock\n", DEBUG_NAME);

    /*TODO: Mem management*/
    *msock = kmalloc(sizeof (**msock), GFP_KERNEL);
    if (!(*msock)) {
        err = -ENOMEM;
        goto error;
    }

    //  init_sock_sending(*msock);
    //  init_sock_receiving(*msock);

    (*msock)->kui_port_read = kui_port;

    /*initialise socket connection infos*/
    (*msock)->sock = 0;
    (*msock)->sockfb = 0;
    (*msock)->host = 0;
    (*msock)->hostfb = 0;
    (*msock)->peer = 0;
    (*msock)->peerfb = 0;
    (*msock)->peerdbg = 0;
    (*msock)->host_addr.s_addr = 0;

    /*initialise lists*/
	  prrttime_init(&(*msock)->send_time_state);
	  prrttime_init(&(*msock)->receive_time_state);
    prrtilist_init(&(*msock)->inc_packet_list, &(*msock)->receive_time_state);
    prrtolist_init(&(*msock)->out_packet_list, &(*msock)->send_time_state);

    /*initialise work queues to null*/
    (*msock)->recv_queue = 0;
    (*msock)->recvfb_queue = 0;
    (*msock)->send_queue = 0;
    (*msock)->sendfb_queue = 0;
    (*msock)->sendrepl_queue = 0;
    (*msock)->readpkt_queue = 0;
    (*msock)->writepkt_queue = 0;

    /*initialise variables for packet shipment*/
    (*msock)->out_status = DOWN;
    (*msock)->out_pno = 0;
    INIT_DELAYED_WORK(&(*msock)->out_work, prrt_send_packet);
    init_MUTEX(&(*msock)->out_lock);
	
	  (*msock)->time_last_packet = 0;

    (*msock)->out_cod = 0;
    init_MUTEX(&(*msock)->out_cod_lock);

    /*initialise variables for receiving packets*/
    (*msock)->poll_status = DOWN;
    PDEBUG(MSG_COMMON, "%s: Poll status set to %d\n", DEBUG_NAME, (*msock)->poll_status);
    INIT_DELAYED_WORK(&(*msock)->poll_work, prrt_poll_packet);
    init_MUTEX(&(*msock)->poll_lock);
    (*msock)->poll_pno_fb = 0;

    /*initialise variables for sending feedback(receiver side)*/
    INIT_DELAYED_WORK(&(*msock)->sendfb_work, prrt_send_feedback);
    init_MUTEX(&(*msock)->sendfb_lock);

    /* For feedback suppression, we redo the initialation of receiving feedback on bind at the receiver. */
    /*initialise variables for receiving feedback(sender side)*/
    INIT_DELAYED_WORK(&(*msock)->recvfb_work, prrt_recv_feedback);
    init_MUTEX(&(*msock)->recvfb_lock);

    /*initialise variables for sending replacements(sender side)*/
    INIT_DELAYED_WORK(&(*msock)->sendrepl_work, prrt_send_replacements);

    /*initialise variables for packet forwarding to user space (receiver side)*/
    INIT_DELAYED_WORK(&(*msock)->readpkt_work, prrt_read_packet);

    /*initialise variables for packet forwarding to user space (sender side)*/
    INIT_DELAYED_WORK(&(*msock)->writepkt_work, prrt_write_packet);

    /*initialise parameter*/
    err = prrtcpar_init_codeparams(&(*msock)->cod_params);
    if (err < 0)
        goto error;

    //prrtlpar_init_locparams(&(*msock)->loc_params_out);
    prrtlpar_init_locparams(&(*msock)->loc_params_in);

    prrtnpar_init_netparams(&(*msock)->net_params);

    /*initialise stats*/
    (*msock)->delivered_packets = 0;
    (*msock)->delivered_data = 0;
    (*msock)->delivered_red = 0;

    err = create_feedback_socket(*msock);
    if (err < 0)
        goto error;

    //err = sock_create_kern(PF_INET, SOCK_DGRAM, IPPROTO_UDP, &((*msock)->sockdbg));
    //if (err < 0)
    //    goto error;

    /*allocate address structure for data socket*/
    if (!(*msock)->peerdbg) {
        /*no memory allocated yet*/
        (*msock)->peerdbg = kmalloc(sizeof (*((*msock)->peerdbg)), GFP_KERNEL);
        if (!(*msock)->peerdbg) {
            err = -ENOMEM;
            goto error;
        }
    }

    PDEBUG(MSG_FINAL, "%s: Before address allocation\n", DEBUG_NAME);
    /*build address for data connection, same for uni- and multicast*/

  memset(((*msock)->peerdbg), 0, sizeof((*msock)->peerdbg));
  (*msock)->peerdbg->sin_family = AF_INET;
  (*msock)->peerdbg->sin_addr.s_addr = in_aton("127.0.0.1");  
  (*msock)->peerdbg->sin_port = htons(5014);
    PDEBUG(MSG_FINAL, "%s: After address allocation\n", DEBUG_NAME);

    /*We use normal udp sockets from the ip suite for prrt sockets*/
    err = sock_create_kern(PF_INET, SOCK_DGRAM, IPPROTO_UDP, &((*msock)->sock));
    if (err < 0)
        goto error;

    /* sockets to user space */
    err = sock_create_kern(PF_INET, SOCK_DGRAM, IPPROTO_UDP, &((*msock)->kui_sock_read));
    if (err < 0)
        goto error;
    /* sockets to user space */
    err = sock_create_kern(PF_INET, SOCK_DGRAM, IPPROTO_UDP, &((*msock)->kui_sock_write));
    if (err < 0)
        goto error;
    err = kernel_setsockopt((*msock)->kui_sock_write, SOL_SOCKET, SO_RCVTIMEO,
        (char *) &totime, sizeof (totime));
    if (err < 0)
        goto error;

    /*set a timeout for the socket*/
    err = kernel_setsockopt((*msock)->sock, SOL_SOCKET, SO_RCVTIMEO,
        (char *) &totime, sizeof (totime));
    if (err < 0)
        goto error;

    (*msock)->timeout = ON;

    return 0;

error:
    if ((*msock)->kui_sock_read)
      sock_release((*msock)->kui_sock_read);
    if ((*msock)->kui_sock_write)
      sock_release((*msock)->kui_sock_write);
    kfree(*msock);

    return err;
}

      /******************************************************/
      /******** bind() and connect() on the receiver ********/
      /******************************************************/
  
/*
 * sock_addr should be in the form "x.x.x.x" and be null terminated
 */
int prrt_bind(struct prrt_socket *msock, char *sock_addr, u_short sock_port, char *peer_addr) {
    int err = 0;
    struct ip_mreq multi_req;
    struct sockaddr_in *bind_addr;

    /*initialise variables for receiving feedback(receiver side)*/
    INIT_DELAYED_WORK(&msock->recvfb_work, prrt_recv_feedback_at_receiver);
    //init_MUTEX(&msock->recvfb_lock);

    /*create address struct for the incoming packets*/
    if (!msock->host) {
        msock->host = kmalloc(sizeof (*(msock->host)), GFP_KERNEL);
        if (!msock->host) {
            err = -ENOMEM;
            goto error;
        }
    }

    /*create address struct for the target of feedback*/
    if (!msock->peerfb) {
        msock->peerfb = kmalloc(sizeof (*(msock->peerfb)), GFP_KERNEL);
        if (!msock->peerfb) {
            err = -ENOMEM;
            goto error;
        }
    }

    /*sock_addr must be safed in an extra place because we need it for receiver identification
     *and the s_addr in the host structure will be set to zero through the next line.*/
    build_addr(msock->host, sock_addr, sock_port);
    msock->host_addr.s_addr = msock->host->sin_addr.s_addr;
    msock->host->sin_addr.s_addr = htonl(INADDR_ANY);
    bind_addr = msock->host;

    build_addr(msock->peerfb, peer_addr, sock_port + 1);

    PDEBUG(MSG_COMMONRECV, "%s: Socket binding...\n", DEBUG_NAME);

    if (isMulticastAddress(peer_addr)) {
        build_addr(msock->host, peer_addr, sock_port);
        PDEBUG(MSG_COMMON, "%s: Address to bind to is %s\n", DEBUG_NAME, peer_addr);
    } else {
        build_addr(msock->host, sock_addr, sock_port);
        PDEBUG(MSG_COMMON, "%s: Address to bind to is %s\n", DEBUG_NAME, sock_addr);
    }

    err = kernel_bind(msock->sock, (struct sockaddr *) bind_addr, sizeof (*bind_addr));
    if (err < 0)
        goto error;

    /*add multicast membership*/
    if (isMulticastAddress(peer_addr)) {
        build_multi_req(&multi_req, peer_addr);
        err = kernel_setsockopt(msock->sock, IPPROTO_IP, IP_ADD_MEMBERSHIP,
                (char *) &multi_req, sizeof (multi_req));
        if (err < 0) {
            /*TODO:What to do here?*/
            PDEBUG(MSG_COMMONRECV, "%s: Could not establish multicast membership for receiving sock\n", DEBUG_NAME);
        }

        // TODO : do we need mutex to protect the binding and running queue?
        // Bind the multicast address to receive feedback for feedback suppression.
        err = kernel_bind(msock->sockfb, (struct sockaddr *) msock->peerfb, sizeof (*(msock->peerfb)));
        //PERROR("ERROR=%d %s: Socket bound to address %s and port %d\n", err, DEBUG_NAME, peer_addr, sock_port + 1);
        if (err < 0) {
          goto error;
        }

        /*Add multicast membership*/
        build_multi_req(&multi_req, sock_addr);
        err = kernel_setsockopt(msock->sockfb, IPPROTO_IP, IP_ADD_MEMBERSHIP, (char *) &multi_req, sizeof (multi_req));
        if (err < 0) {
            /*TODO:What to do here?*/
            PDEBUG(MSG_COMMONRECV, "%s: Could not establish multicast membership for receiving feedback sock\n", DEBUG_NAME);
        }
    }

    PDEBUG(MSG_FINAL, "%s: Socket bound to address %s and port %d\n",
            DEBUG_NAME, sock_addr, sock_port);

    /*initialise polling*/
    if (msock->recv_queue == 0)
        msock->recv_queue = create_singlethread_workqueue("prrt_recv");

    if (msock->sendfb_queue == 0)
        msock->sendfb_queue = create_singlethread_workqueue("prrt_sendfb");

    if (msock->readpkt_queue == 0) {
        PDEBUG(MSG_FINAL, "%s: Creating prrt_readpkt workqueue\n", DEBUG_NAME);
        msock->readpkt_queue = create_singlethread_workqueue("prrt_readpkt");
    }

    if (msock->recvfb_queue == 0) {
        PDEBUG(MSG_FINAL, "%s: Creating prrt_recvfb workqueue\n", DEBUG_NAME);
        msock->recvfb_queue = create_singlethread_workqueue("prrt_recvfb");
    }

    msock->poll_status = WORKING;
    PDEBUG(MSG_COMMONRECV, "%s: Poll status set to %d\n", DEBUG_NAME, msock->poll_status);

    queue_delayed_work(msock->recv_queue, &msock->poll_work, POLLDELAY);
    queue_delayed_work(msock->sendfb_queue, &msock->sendfb_work, FEEDBACKDELAY);
    queue_delayed_work(msock->readpkt_queue, &msock->readpkt_work, READPKTDELAY);
    queue_delayed_work(msock->recvfb_queue, &msock->recvfb_work, FEEDBACKDELAY);

    return 0;

error:
    PDEBUG(MSG_COMMONRECV, "ERROR %s: Error on bind (no %d)\n", DEBUG_NAME, err);
    kfree(msock->host);
    kfree(msock->peerfb);
    msock->host = 0;
    msock->peerfb = 0;
    return -1;
}

/*
 * sock_addr should be in the form "x.x.x.x" and be null terminated
 */
int prrt_bind_as_recv(struct prrt_socket *msock, char *sock_addr, u_short sock_port)
{
    int err = 0;
    struct sockaddr_in *bind_addr;

    /*create address struct for the incoming packets*/
    if (!msock->host) {
        msock->host = kmalloc(sizeof (*(msock->host)), GFP_KERNEL);
        if (!msock->host) {
            err = -ENOMEM;
            goto error;
        }
    }

    /*sock_addr must be safed in an extra place because we need it for receiver identification
     *and the s_addr in the host structure will be set to zero through the next line.*/
    build_addr(msock->host, sock_addr, sock_port);
    msock->host_addr.s_addr = msock->host->sin_addr.s_addr;
    msock->host->sin_addr.s_addr = htonl(INADDR_ANY);
    bind_addr = msock->host;

    PDEBUG(MSG_COMMONRECV, "%s: Socket of receiver binding...\n", DEBUG_NAME);

    build_addr(msock->host, sock_addr, sock_port);
    PDEBUG(MSG_COMMON, "%s: Address of host to bind to is %s\n", DEBUG_NAME, sock_addr);
    
    err = kernel_bind(msock->sock, (struct sockaddr *) bind_addr, sizeof (*bind_addr));
    if (err == 0) {
      PDEBUG(MSG_COMMONRECV, "%s: Socket bound to address %s and port %d\n", DEBUG_NAME, sock_addr, sock_port);
      return 0;
    }

error:
    PERROR("ERROR %s: Error on bind as recv (no %d)\n", DEBUG_NAME, err);
    kfree(msock->host);
    msock->host = 0;
    return -1;
}

/*
 * sock_addr should be in the form "x.x.x.x" and be null terminated
 */
int prrt_connect_to_feedback(struct prrt_socket *msock, char *peerfb_addr, u_short peerfb_port)
{
    int err = 0;
    struct ip_mreq multi_req;

    /*initialise variables for receiving feedback(receiver side)*/
    INIT_DELAYED_WORK(&msock->recvfb_work, prrt_recv_feedback_at_receiver);
    //init_MUTEX(&msock->recvfb_lock);

    PDEBUG(MSG_FINAL, "%s: prrt_connect_to_feedback ...\n", DEBUG_NAME);
    /*create address struct for the target of feedback*/
    if (!msock->peerfb) {
        msock->peerfb = kmalloc(sizeof (*(msock->peerfb)), GFP_KERNEL);
        if (!msock->peerfb) {
            err = -ENOMEM;
            goto error;
        }
    }

    build_addr(msock->peerfb, peerfb_addr, peerfb_port);

    PDEBUG(MSG_COMMONRECV, "%s: Socket of receiver connecting...\n", DEBUG_NAME);

    if (isMulticastAddress(peerfb_addr)) {
        // FIXME xxx shine: why we need to change host to peer, see original prrt_bind()? Because we want to use the same port of receiving data on the receiver for multicast feedback. But how to deal with the case where peerfb_port!=sock_port? actually we don't make receiver recv feedbacks at the moment.
        build_addr(msock->host, peerfb_addr, ntohs(msock->host->sin_port));
        PDEBUG(MSG_COMMON, "%s: Address of host for feedback to bind to is %s\n", DEBUG_NAME, peerfb_addr);
        
        // add multicast membership
        build_multi_req(&multi_req, peerfb_addr);
        err = kernel_setsockopt(msock->sock, IPPROTO_IP, IP_ADD_MEMBERSHIP, (char *) &multi_req, sizeof (multi_req));
        if (err < 0) {
            /*TODO:What to do here?*/
            PDEBUG(MSG_COMMONRECV, "%s: Could not establish multicast membership for receiving sock\n", DEBUG_NAME);
        }

        // TODO : do we need mutex to protect the binding and running queue?
        // Bind the multicast address to receive feedback for feedback suppression.
        err = kernel_bind(msock->sockfb, (struct sockaddr *) msock->peerfb, sizeof (*(msock->peerfb)));
        //PERROR("ERROR=%d %s: Socket bound to address %s and port %d\n", err, DEBUG_NAME, peerfb_addr, peerfb_port);
        if (err < 0) {
          goto error;
        }

        /*Add multicast membership*/
        err = kernel_setsockopt(msock->sockfb, IPPROTO_IP, IP_ADD_MEMBERSHIP, (char *) &multi_req, sizeof (multi_req));
        if (err < 0) {
            /*TODO:What to do here?*/
            PDEBUG(MSG_COMMONRECV, "%s: Could not establish multicast membership for receiving feedback sock\n", DEBUG_NAME);
        }
    }

    /*initialise polling*/
    if (msock->recv_queue == 0)
        msock->recv_queue = create_singlethread_workqueue("prrt_recv");

    if (msock->sendfb_queue == 0)
        msock->sendfb_queue = create_singlethread_workqueue("prrt_sendfb");

    if (msock->readpkt_queue == 0) {
        PDEBUG(MSG_FINAL, "%s: Creating prrt_readpkt workqueue\n", DEBUG_NAME);
        msock->readpkt_queue = create_singlethread_workqueue("prrt_readpkt");
    }

    if (msock->recvfb_queue == 0) {
        PDEBUG(MSG_FINAL, "%s: Creating prrt_recvfb workqueue\n", DEBUG_NAME);
        msock->recvfb_queue = create_singlethread_workqueue("prrt_recvfb");
    }

    msock->poll_status = WORKING;
    PDEBUG(MSG_COMMONRECV, "%s: Poll status set to %d\n", DEBUG_NAME, msock->poll_status);

    queue_delayed_work(msock->recv_queue, &msock->poll_work, POLLDELAY);
    queue_delayed_work(msock->sendfb_queue, &msock->sendfb_work, FEEDBACKDELAY);
    queue_delayed_work(msock->readpkt_queue, &msock->readpkt_work, READPKTDELAY);
    queue_delayed_work(msock->recvfb_queue, &msock->recvfb_work, FEEDBACKDELAY);

    return 0;

error:
    PERROR("ERROR %s: Error on connect feedback (no %d)\n", DEBUG_NAME, err);
    kfree(msock->host); // release host which is already bound, if connect fails.
    msock->host = 0;
    kfree(msock->peerfb);
    msock->peerfb = 0;
    return -1;
}

      /****************************************************/
      /******** bind() and connect() on the sender ********/
      /****************************************************/

int prrt_connect(struct prrt_socket *msock, char *peer_addr, u_short peer_port, char *host_addr, int *kui_port) {
    int err = 0;
    struct ip_mreq multi_req;
    //unsigned short kui_port_write = htons(msock->kui_port_write);
    struct sockaddr_in kui_bind_addr;
    int port;

    /*initialise variables for receiving feedback(sender side)*/
    //INIT_DELAYED_WORK(&msock->recvfb_work, prrt_recv_feedback);
    //init_MUTEX(&msock->recvfb_lock);

    if (down_interruptible(&msock->out_lock))
        return -ERESTARTSYS;

    /*Initialise sending*/
    err = init_sending(msock);
    if (err < 0)
        goto error;

    /*allocate address structure for feedback socket*/
    if (!msock->hostfb) {
        msock->hostfb = kmalloc(sizeof (*(msock->hostfb)), GFP_KERNEL);
        if (!msock->hostfb) {
            err = -ENOMEM;
            goto error;
        }
    }

    /*allocate address structure for data socket*/
    if (!msock->peer) {
        /*no memory allocated yet*/
        msock->peer = kmalloc(sizeof (*(msock->peer)), GFP_KERNEL);
        if (!msock->peer) {
            err = -ENOMEM;
            goto error;
        }
    }

    /*build address for data connection, same for uni- and multicast*/
    build_addr(msock->peer, peer_addr, peer_port);

    /*unicast or multicast*/
    if (host_addr != 0) {
        /*unicast case*/
        /*build address for feedback connection*/
        build_addr(msock->hostfb, host_addr, peer_port + 1);
        msock->hostfb->sin_addr.s_addr = htonl(INADDR_ANY);
    } else {
        /*multicast case, peer_addr is the multicast address for feedback and data*/
        build_addr(msock->hostfb, peer_addr, peer_port + 1);
        msock->hostfb->sin_addr.s_addr = htonl(INADDR_ANY);
    }

    err = kernel_bind(msock->sockfb, (struct sockaddr *) msock->hostfb, sizeof (*(msock->hostfb)));
    if (err < 0) {
        goto error;
    }

    if (host_addr == 0) {
        /*Add multicast membership*/
        build_multi_req(&multi_req, peer_addr);
        err = kernel_setsockopt(msock->sockfb, IPPROTO_IP, IP_ADD_MEMBERSHIP,
                (char *) &multi_req, sizeof (multi_req));
        //    if(err < 0) {
        //      /*TODO:What to do here?*/
        //      PDEBUG(MSG_COMMONSEND, "%s: Could not establish multicast membership\n",
        //        DEBUG_NAME);
        //      goto error_multi;
        //    }
    }

    memset(&kui_bind_addr,0, sizeof(kui_bind_addr));
    kui_bind_addr.sin_family = AF_INET;
    kui_bind_addr.sin_addr.s_addr = htonl(INADDR_ANY);  
    kui_bind_addr.sin_port = htons(0);

    for (port = 32000; port < 32100; port++){
      kui_bind_addr.sin_port = htons(port);
      err = kernel_bind(msock->kui_sock_write, (struct sockaddr *) &kui_bind_addr, sizeof (kui_bind_addr));
      if (!(err < 0))
        break;
    }

    if (err < 0){
      PDEBUG(MSG_FINAL, "%s: Bind failed\n", DEBUG_NAME);
      goto error;
    }

    msock->kui_port_write = port;
    *kui_port = port;

    PDEBUG(MSG_FINAL, "%s: Bind successful with kui_port %d\n", DEBUG_NAME, *kui_port);

    PDEBUG(MSG_COMMONSEND, "%s: Socket connected to %s:%d\n", DEBUG_NAME, peer_addr, peer_port);
    /*PDEBUG(MSG_COMMONSEND, "%s:sockaddr_in:\nsin_family=%d\nsin_port=%d\n"
            "sin_addr=%u\n", DEBUG_NAME, msock->peer->sin_family,
            msock->peer->sin_port, msock->peer->sin_addr.s_addr);*/

    /*cannot connect on udp sockets*/

    up(&msock->out_lock);

    return 0;

error:
    PDEBUG(MSG_COMMONSEND, "ERROR %s: Error on connect (no %d)\n", DEBUG_NAME,
            err);
    kfree(msock->peer);
    kfree(msock->hostfb);
    msock->peer = 0;
    msock->hostfb = 0;
    // error_multi:
    up(&msock->out_lock);
    return -1;
}

int prrt_bind_as_send(struct prrt_socket *msock, char *sock_addr, u_short sock_port)
{
    int err = 0;
    struct ip_mreq multi_req;

    /*initialise variables for receiving feedback(sender side)*/
    //INIT_DELAYED_WORK(&msock->recvfb_work, prrt_recv_feedback);
    //init_MUTEX(&msock->recvfb_lock);

    if (down_interruptible(&msock->out_lock))
        return -ERESTARTSYS;

    /*Initialise sending*/
    err = init_sending(msock);
        //PERROR("ERROR=%d %s: init_sending Socket bound to address %s and port %d\n", err, DEBUG_NAME, sock_addr, sock_port);
    if (err < 0)
        goto error;

    /*allocate address structure for feedback socket*/
    if (!msock->hostfb) {
        msock->hostfb = kmalloc(sizeof (*(msock->hostfb)), GFP_KERNEL);
        if (!msock->hostfb) {
            err = -ENOMEM;
        PERROR("ERROR=%d %s: allocate Socket bound to address %s and port %d\n", err, DEBUG_NAME, sock_addr, sock_port);
            goto error;
        }
    }
    PDEBUG(MSG_COMMONRECV, "%s: Socket of sender binding...\n", DEBUG_NAME);

    // NOTE This might violate the original bind/connect at the sender side regarding to the multicast feedback.
    // Because the mutlicast address of feedback here can be different to the multicast address of peer.
    // The original bind/connect uses the same peer address in the case of the multicast feedback.
    PDEBUG(MSG_COMMON, "%s: Address to send feedback to %s...\n", DEBUG_NAME, sock_addr);
    build_addr(msock->hostfb, sock_addr, sock_port);
    // We dont bind to INADDR_ANY, which accept all incoming connections on any local IP address. Because it will cause bind failure in the case of prrt relay application.
    //msock->hostfb->sin_addr.s_addr = htonl(INADDR_ANY);

    err = kernel_bind(msock->sockfb, (struct sockaddr *) msock->hostfb, sizeof (*(msock->hostfb)));
    //PERROR("ERROR=%d %s: Socket bound to address %s and port %d\n", err, DEBUG_NAME, sock_addr, sock_port);
    if (err == 0) {
        PDEBUG(MSG_COMMONRECV, "%s: Socket bound to address %s and port %d\n", DEBUG_NAME, sock_addr, sock_port);
        
        /* unicast or multicast feedback */
        if (isMulticastAddress(sock_addr)) {
            /*Add multicast membership*/
            build_multi_req(&multi_req, sock_addr);
            err = kernel_setsockopt(msock->sockfb, IPPROTO_IP, IP_ADD_MEMBERSHIP,
                    (char *) &multi_req, sizeof (multi_req));
            //    if(err < 0) {
            //      /*TODO:What to do here?*/
            //      PDEBUG(MSG_COMMONSEND, "%s: Could not establish multicast membership\n",
            //        DEBUG_NAME);
            //      goto error_multi;
            //    }
        }
    
        up(&msock->out_lock);
        return 0;
    }

error:
    PERROR("ERROR %s: Error on bind as send (no %d)\n", DEBUG_NAME, err);
    kfree(msock->hostfb);
    msock->hostfb = 0;
    // error_multi:
    up(&msock->out_lock);
    return -1;
}

int prrt_connect_to_receiver(struct prrt_socket *msock, char *peer_addr, u_short peer_port, int *kui_port)
{
    int err = 0;
    struct sockaddr_in kui_bind_addr;
    int port;

    if (down_interruptible(&msock->out_lock))
        return -ERESTARTSYS;

    /*allocate address structure for data socket*/
    if (!msock->peer) {
        /*no memory allocated yet*/
        msock->peer = kmalloc(sizeof (*(msock->peer)), GFP_KERNEL);
        if (!msock->peer) {
            err = -ENOMEM;
            goto error;
        }
    }

    PDEBUG(MSG_COMMON, "%s: Socket Connecting to %s on port %d...\n", DEBUG_NAME, peer_addr, peer_port);
    
    /*build address for data connection, same for uni- and multicast*/
    build_addr(msock->peer, peer_addr, peer_port);

    PDEBUG(MSG_COMMONSEND, "%s: Socket connected to %s:%d\n", DEBUG_NAME, peer_addr, peer_port);

    // Looking for a free port for kernel user communication and binding it.
    //unsigned short kui_port_write = htons(msock->kui_port_write);
    memset(&kui_bind_addr,0, sizeof(kui_bind_addr));
    kui_bind_addr.sin_family = AF_INET;
    kui_bind_addr.sin_addr.s_addr = htonl(INADDR_ANY);  
    kui_bind_addr.sin_port = htons(0);

    for (port = 32000; port < 32100; port++){
        kui_bind_addr.sin_port = htons(port);
        err = kernel_bind(msock->kui_sock_write, (struct sockaddr *) &kui_bind_addr, sizeof (kui_bind_addr));
        if (!(err < 0))
            break;
    }

    if (err < 0){
        PDEBUG(MSG_FINAL, "%s: Bind failed due to kui_port occupied\n", DEBUG_NAME);
        goto error;
    }

    msock->kui_port_write = port;
    if (kui_port ==  0) {
        PDEBUG(MSG_FINAL, "%s: Bind failed due to null pointer of kui_port\n", DEBUG_NAME);
        goto error;
    }
    *kui_port = port;

    PDEBUG(MSG_FINAL, "%s: Bind successful with kui_port %d\n", DEBUG_NAME, *kui_port);

    up(&msock->out_lock);

    return 0;

error:
    PERROR("ERROR %s: Error on connect (no %d)\n", DEBUG_NAME, err);
    kfree(msock->peer);
    msock->peer = 0;
    kfree(msock->hostfb); // release the host feedback which is already bound, if connect fails.
    msock->hostfb = 0;
    // error_multi:
    up(&msock->out_lock);
    return -1;
}


// ===============================================================================

/*
 *
 */
int prrt_accept(struct prrt_socket *msock, struct prrt_socket **new_msock) {
    int err = 0;

    /*TODO: Mem management*/
    *new_msock = kmalloc(sizeof (**new_msock), GFP_KERNEL);
    if (!(*new_msock)) {
        err = -ENOMEM;
        goto error;
    }

    PDEBUG(MSG_COMMON, "%s: Socket accepting\n", DEBUG_NAME);

    /*cannot connect on udp sockets*/
    //err = kernel_accept(msock->sock, &((*new_msock)->sock), 0);
    if (err < 0)
        goto error;

    return err;

error:
    PDEBUG(MSG_COMMON, "ERROR %s: Error on accept (no %d)\n", DEBUG_NAME,
            err);
    kfree(*new_msock);
    return -1;
}

int prrt_listen(struct prrt_socket *msock, int backlog) {
    int err = 0;

    PDEBUG(MSG_COMMON, "%s: Socket tries to listen...\n", DEBUG_NAME);

    /*cannot listen on udp sockets*/
    err = kernel_listen(msock->sock, backlog);
    if (err < 0) {
        PDEBUG(MSG_COMMON, "ERROR %s: Error on listen (no %d)\n",
                DEBUG_NAME, err);
        return -1;
    }

    return 0;
}

void prrt_shutdown(struct prrt_socket *msock) {
    PDEBUG(MSG_COMMON, "%s: Deleting msock\n", DEBUG_NAME);
    PDEBUG(MSG_COMMON, "%s: Shutting down poll activity\n", DEBUG_NAME);
    down(&msock->poll_lock);
    msock->poll_status = DOWN;
    PDEBUG(MSG_COMMON, "%s: Poll status set to %d\n", DEBUG_NAME, msock->poll_status);
    up(&msock->poll_lock);
    PDEBUG(MSG_COMMON, "%s: Shutting down send activity\n", DEBUG_NAME);
    down(&msock->out_lock);
    msock->out_status = DOWN;
    up(&msock->out_lock);

    /*this has to be done to free the send thread out of its loop*/
    up(&msock->out_packet_list.block_list.packet_count);
    PDEBUG(MSG_OUTLIST, "%s: Upped count to release thread, now %u left\n", DEBUG_NAME,
            msock->out_packet_list.block_list.packet_count.count);

    PDEBUG(MSG_COMMON, "%s: Clearing workqueues\n", DEBUG_NAME);
    /*clean up work queues before doing something else*/

    PDEBUG(MSG_COMMON, "%s: Clearing poll queue\n", DEBUG_NAME);
    cancel_delayed_work(&msock->poll_work);
    PDEBUG(MSG_COMMON, "%s: Flushing poll queue\n", DEBUG_NAME);
    if (msock->recv_queue != 0)
        flush_workqueue(msock->recv_queue);

    PDEBUG(MSG_COMMON, "%s: Clearing send queue\n", DEBUG_NAME);
    cancel_delayed_work(&msock->out_work);
    PDEBUG(MSG_COMMON, "%s: Flushing send queue\n", DEBUG_NAME);
    if (msock->send_queue != 0)
        flush_workqueue(msock->send_queue);

    PDEBUG(MSG_COMMON, "%s: Clearing sendrepl queue\n", DEBUG_NAME);
    cancel_delayed_work(&msock->sendrepl_work);
    PDEBUG(MSG_COMMON, "%s: Flushing sendrepl queue\n", DEBUG_NAME);
    if (msock->sendrepl_queue != 0)
        flush_workqueue(msock->sendrepl_queue);

    PDEBUG(MSG_COMMON, "%s: Clearing writepkt queue\n", DEBUG_NAME);
    cancel_delayed_work(&msock->writepkt_work);
    PDEBUG(MSG_COMMON, "%s: Flushing writepkt queue\n", DEBUG_NAME);
    if (msock->writepkt_queue != 0)
        flush_workqueue(msock->writepkt_queue);

    PDEBUG(MSG_COMMON, "%s: Clearing readpkt queue\n", DEBUG_NAME);
    cancel_delayed_work(&msock->readpkt_work);
    PDEBUG(MSG_COMMON, "%s: Flushing readpkt queue\n", DEBUG_NAME);
    if (msock->readpkt_queue != 0)
        flush_workqueue(msock->readpkt_queue);

    PDEBUG(MSG_COMMON, "%s: Clearing sendfb queue\n", DEBUG_NAME);
    cancel_delayed_work(&msock->sendfb_work);
    PDEBUG(MSG_COMMON, "%s: Flushing sendfb queue\n", DEBUG_NAME);
    if (msock->sendfb_queue != 0)
        flush_workqueue(msock->sendfb_queue);

    PDEBUG(MSG_COMMON, "%s: Clearing recvfb queue\n", DEBUG_NAME);
    cancel_delayed_work(&msock->recvfb_work);
    PDEBUG(MSG_COMMON, "%s: Flushing recvfb queue\n", DEBUG_NAME);
    if (msock->recvfb_queue != 0)
        flush_workqueue(msock->recvfb_queue);
    //  cancel_work_sync(msock->recvfb_queue);

    PDEBUG(MSG_COMMON, "%s: Acquiring locks\n", DEBUG_NAME);
    down(&msock->poll_lock);
    PDEBUG(MSG_COMMON, "%s: poll_lock down\n", DEBUG_NAME);
    down(&msock->out_lock);
    PDEBUG(MSG_COMMON, "%s: out_lock down\n", DEBUG_NAME);
    down(&msock->sendfb_lock);
    PDEBUG(MSG_COMMON, "%s: sendfb_lock down\n", DEBUG_NAME);
    down(&msock->recvfb_lock);
    PDEBUG(MSG_COMMON, "%s: recvfb_lock down\n", DEBUG_NAME);
    PRINT_TRACE("RECV", "P/DI/FBO", "going_down");
    PRINT_TRACE("SEND", "I/DO/FBI", "going_down");
    PDEBUG(MSG_COMMON, "%s: Deleting workqueues\n", DEBUG_NAME);
    if (msock->send_queue != 0)
        destroy_workqueue(msock->send_queue);
    if (msock->recv_queue != 0)
        destroy_workqueue(msock->recv_queue);
    if (msock->recvfb_queue != 0)
        destroy_workqueue(msock->recvfb_queue);
    if (msock->sendfb_queue != 0)
        destroy_workqueue(msock->sendfb_queue);
    if (msock->sendrepl_queue != 0)
        destroy_workqueue(msock->sendrepl_queue);
    if (msock->readpkt_queue != 0)
        destroy_workqueue(msock->readpkt_queue);
    if (msock->writepkt_queue != 0)
        destroy_workqueue(msock->writepkt_queue);

    PDEBUG(MSG_COMMON, "%s: Freeing resources\n", DEBUG_NAME);

    print_stats(msock);

    /*free list contents*/
    prrtilist_uninit(&msock->inc_packet_list);
    prrtolist_free(&msock->out_packet_list);

    /*TODO(is this needed?): drop multicast memberships*/

    /*free sockets*/
    kernel_sock_shutdown(msock->sock, 2);
    kernel_sock_shutdown(msock->sockfb, 2);
    sock_release(msock->sock);
    sock_release(msock->sockfb);
    sock_release(msock->kui_sock_read);
    sock_release(msock->kui_sock_write);

    /*free addresses*/
    kfree(msock->host);
    kfree(msock->hostfb);
    kfree(msock->peer);
    kfree(msock->peerfb);

    /*free coder*/
    prrt_fec_free(msock->out_cod);

    up(&msock->recvfb_lock);
    up(&msock->sendfb_lock);
    up(&msock->out_lock);
    up(&msock->poll_lock);

    kfree(msock);
}

int prrt_send(struct prrt_socket *msock, struct prrt_packet_container *pcont, size_t len) {
    int err = 0;

    int packtype;
    int coding = 0;
    int d_target;

    int index = 0;
    long time_diff = 0;
    u32 time = 0;
    u32 time_us = 0;
    //u32 gathering_time = 0;
    int group_rtt = 0;
    //unsigned char wait_intervals = 0;

    struct prrt_block_list *block_list = 0;
    struct timespec ts;

    INIT_MSGHDR(msg);

    /*Error when no destination address is given*/
    if (!msock->peer) {
        err = -EDESTADDRREQ;
        PERROR("%s: Can not send, socket not connected\n", DEBUG_NAME);
        goto free_packet;
    }

    /*Error when socked is about to be shut down*/
    if (down_interruptible(&msock->out_lock)) {
        err = -ERESTARTSYS;
        goto free_packet;
    }

    if (msock->out_status == DOWN) {
        err = -ESHUTDOWN;
        up(&msock->out_lock);
        PERROR("%s: May not send any further packets, socket down\n", DEBUG_NAME);
        goto free_packet;
    }

    up(&msock->out_lock);

    prrtblist_print_stats(&msock->out_packet_list.block_list);
    prrtrlist_print_stats(&msock->out_packet_list.recv_list);

    /*create packet that is to be send*/
    //time = jiffies;
    packtype = packtype_data;
    PRINT_TRACE("SEND", "I", "input_pack");

    block_list = &msock->out_packet_list.block_list;
    if (block_list) {
        struct prrt_block *last_block = 0;

        if (!prrtblist_get_last_block(block_list, &last_block))
            index = last_block->no_data;
    }
    //PDEBUG(MSG_FINAL, "%s: Index %d, k %d\n", DEBUG_NAME, index, prrtcpar_get_k(&msock->cod_params));

    getnstimeofday(&ts);

    time_us = ts.tv_sec * 1000 + ts.tv_nsec / 1000;
    time = prrt_get_mstime(&msock->send_time_state);

    //PDEBUG(MSG_FINAL, "%s: Time %u\n", DEBUG_NAME, time);

    //PDEBUG(MSG_FINAL, "%s: Decoding_time %u, Gathering_time %u\n", DEBUG_NAME, fb_timeout, gathering_time);

    group_rtt = prrtnpar_get_ReplaceTime(&(msock->net_params));

    //PDEBUG(MSG_FINAL, "%s: group_rtt = %u\n", DEBUG_NAME, group_rtt);

    err = prrt_packet_set_fields(pcont, msock->out_pno, index, time, group_rtt, len, packtype);
    if (err < 0) {
        PERROR("%s: Could not create packet for sending\n", DEBUG_NAME);
        goto free_packet;
    }

    //PDEBUG(MSG_COMMONSEND, "%s: Updating gathering time\n", DEBUG_NAME);
    if (msock->time_last_packet == 0) {
        prrtnpar_update_PacketInterval(&msock->net_params, msock->time_last_packet);
    } else {
        time_diff = (long) time_us - (long) msock->time_last_packet;
        prrtnpar_update_PacketInterval(&msock->net_params, time_diff);
        //PDEBUG(MSG_FINAL, "%s: Updating packet interval %u\n", DEBUG_NAME, time_diff);
    }
    msock->time_last_packet = time_us;
    d_target = prrtcpar_get_D_target(&msock->cod_params);

    /*update and retrieve the coding parameters*/
    PDEBUG(MSG_COMMONSEND, "%s: Getting network parameters\n", DEBUG_NAME);

    if (!USE_FIXED_T_REPLACE) {
        prrtrlist_update_WorstNetParams(&msock->out_packet_list.recv_list);

        PDEBUG(MSG_COMMONSEND, "%s: Updating coding parameters\n", DEBUG_NAME);
    }
    /*
            prrtcpar_update_params(&msock->cod_params, &msock->out_packet_list.recv_list.worstNetParams,
                                  &msock->loc_params_out);
     */

    PDEBUG(MSG_COMMONSEND, "%s: Getting coding option\n", DEBUG_NAME);
    coding = prrtolist_get_option(&msock->out_packet_list, CODING);

    //err = prrtolist_insert_packet(&msock->out_packet_list, pcont, &msock->cod_params, &msock->loc_params_out);
    err = prrtolist_insert_packet(&msock->out_packet_list, pcont, &msock->cod_params);
    if (err < 0)
        goto free_packet;
    
    //Notify network statistics about sent packet
    prrtnpar_insert_Sent(&(msock->net_params), pcont->pack.data.header.seqnum);  

    if (coding)
        //prrtolist_try_to_code_next_block(&msock->out_packet_list, &msock->cod_params, &msock->loc_params_out);
        prrtolist_try_to_code_next_block(&msock->out_packet_list, &msock->cod_params);

    prrtolist_delete_old_blocks(&msock->out_packet_list, d_target);

    msock->out_pno = prrt_increment_packetno(msock->out_pno, 1);

    PDEBUG(MSG_COMMONSEND, "%s: Done with sending\n", DEBUG_NAME);
    return err;

free_packet:
    prrt_packet_destroy(&pcont);
    return err;
}

void handle_read_packet(struct prrt_socket *msock);
void handle_write_packet(struct prrt_socket *msock);

void prrt_read_packet(struct work_struct *work)
{
  struct prrt_socket *msock = 0;
  struct delayed_work *dw = 0;
  enum worker_thread_status_t status;

  PDEBUG(MSG_FINAL, "%s: prrt_read_packet called!\n", DEBUG_NAME);
  dw = container_of(work, struct delayed_work, work);
  msock = container_of(dw, struct prrt_socket, readpkt_work);

  down(&msock->poll_lock);
  status = msock->poll_status;
  up(&msock->poll_lock);

  while( status == WORKING ){
    handle_read_packet(msock);

    down(&msock->poll_lock);
    status = msock->poll_status;
    up(&msock->poll_lock);

    if(status == DOWN){
      PDEBUG(MSG_FINAL, "%s: Poll status is down!\n", DEBUG_NAME);
    }
  }
}

void prrt_write_packet(struct work_struct *work)
{
  struct prrt_socket *msock = 0;
  struct delayed_work *dw = 0;
  enum worker_thread_status_t status;

  PDEBUG(MSG_FINAL, "%s: prrt_write_packet called!\n", DEBUG_NAME);
  dw = container_of(work, struct delayed_work, work);
  msock = container_of(dw, struct prrt_socket, writepkt_work);

  down(&msock->out_lock);
  status = msock->out_status;
  up(&msock->out_lock);

  while( status == WORKING ){
    handle_write_packet(msock);

    down(&msock->out_lock);
    status = msock->out_status;
    up(&msock->out_lock);

    if(status == DOWN){
      PDEBUG(MSG_FINAL, "%s: Out status is down!\n", DEBUG_NAME);
    }
  }
}

int write_to_user(int kui_port_read, struct socket *kui_sock_read, char *msg, int msg_size);
int read_from_user(int kui_port_write, struct socket *kui_sock_write, char *msg, int msg_size);

void handle_read_packet(struct prrt_socket *msock)
{
  struct prrt_packet_container *pcont = 0;
  int bytes_recv = 0;
  int err = 0;
  //int count = 0;
  //int count2 = 0;

  //PDEBUG(MSG_FINAL, "%s: prrt_read_packet called!\n", DEBUG_NAME);

  //while(sm_get_prrt_sock(priv->socketno, &msock) == -ERESTARTSYS)
  //  ;
  if(!msock){
    PDEBUG(MSG_FINAL, "%s: didn't get msock!\n", DEBUG_NAME);
    goto out;
  }

  bytes_recv = prrt_recv(msock, &pcont);
  if(bytes_recv < 0) {
    //PERROR("%s: read_packet failed with error %d!\n", DEBUG_NAME, bytes_recv);
    err = bytes_recv;
    goto out;
  }

  //PDEBUG(MSG_FINAL, "%s: Returning %d bytes of data to port %d!\n", DEBUG_NAME, bytes_recv, msock->kui_port_read);
  if(write_to_user(msock->kui_port_read, msock->kui_sock_read, prrt_packet_get_data(pcont), bytes_recv) < 0) {
    PDEBUG(MSG_FINAL, "%s: Could not write to user space!\n", DEBUG_NAME);
  }

  //PDEBUG(MSG_FINAL, "%s: done!\n", DEBUG_NAME, bytes_recv, msock->kui_port_read);
  prrt_packet_destroy(&pcont);

out:
  return;
}

void handle_write_packet(struct prrt_socket *msock)
{
  struct prrt_packet_container *pcont = 0;
  int bytes_recv = 0;
  int err = 0;

  //PDEBUG(MSG_FINAL, "%s: prrt_write_packet called!\n", DEBUG_NAME);

  //while(sm_get_prrt_sock(priv->socketno, &msock) == -ERESTARTSYS)
  //  ;
  if(!msock){
    PDEBUG(MSG_FINAL, "%s: didn't get msock!\n", DEBUG_NAME);
    return;
  }


  pcont = prrt_packet_alloc(&msock->send_time_state);
  if(!pcont)
    return;

  /*IMPORTANT: Timeout for this socket has to be set or the module will
   *           deadlock when this part is waiting for messages that
   *           do not arrive.*/

  if((bytes_recv = read_from_user(msock->kui_port_write, msock->kui_sock_write, &(pcont->pack.data.data), MAX_DATA_SIZE)) < 0) {
    PDEBUG(MSG_FINAL, "%s: Could not read from user space!\n", DEBUG_NAME);
    prrt_packet_destroy(&pcont);
    return;
  }
  //PDEBUG(MSG_FINAL, "%s: Read %d from user space!\n", DEBUG_NAME, bytes_recv);

  err = prrt_send(msock, pcont, bytes_recv);

  if(err < 0) {
    PDEBUG(MSG_FINAL, "%s: prrt_send failed!\n", DEBUG_NAME);
    return;
  }
}

int write_to_user(int kui_port_read, struct socket *kui_sock_read, char *msg, int msg_size)
{
  int len;
  struct sockaddr_in to;

  memset(&to, 0, sizeof(to));
  to.sin_family = AF_INET;
  to.sin_addr.s_addr = in_aton("127.0.0.1");  
  to.sin_port = htons(kui_port_read);

  len = wrap_kernel_sendmsg(kui_sock_read, &to, msg, msg_size);
  
  return len;
}

int read_from_user(int kui_port_write, struct socket *kui_sock_write, char *msg, int msg_size)
{
  int len = 0;

  len = wrap_kernel_recvmsg(kui_sock_write, msg, msg_size);
  
  return len;
}

int prrt_recv(struct prrt_socket *msock, struct prrt_packet_container **pcont) {
    int err = 0;
    int interrupted = 0;
    int data_size = 0;
    int d_target = 0;
    u16 packetno = 0;

    *pcont = 0;
    PDEBUG(MSG_PRRTRECV, "%s: At start of prrt_recv\n", DEBUG_NAME);

    //  PDEBUG(MSG_COMMONRECV, "%s: Taking on spinlock in recv1\n", DEBUG_NAME);
    interrupted = down_interruptible(&msock->poll_lock);
    if (interrupted)
        goto interrupted;

    if (msock->poll_status == DOWN) {
        up(&msock->poll_lock);
        PERROR("%s: Socket is down, receiving not possible\n", DEBUG_NAME);
        err = -ESHUTDOWN;
        goto error; /*TODO: explicit error message for this*/
    }

    /*these are only queued if not yet queued*/
    /*socket can not sleep atm, finishes right away*/
    //  msock->poll_status = WORKING;
    //  PDEBUG(MSG_COMMONRECV, "%s: Poll status set to %d\n", DEBUG_NAME, msock->poll_status);
    //  if(queue_delayed_work(msock->recv_queue, &msock->poll_work, 1) != 0)
    //    PRINT_TRACE("RECV", "G/P", "end_sleep");
    //  queue_delayed_work(msock->work_queue, &msock->sendfb_work, 1);

    up(&msock->poll_lock);

    PDEBUG(MSG_PACKRETURN, "%s: Trying to get first packet\n", DEBUG_NAME);
    d_target = prrtlpar_get_D_target(&msock->loc_params_in);
    /*this function waits for POLL_MAXWAITTIME*/
    //PDEBUG(MSG_FINAL, "%s: trying to obtain first packet from queue at %u ...\n", DEBUG_NAME, prrt_get_prrttime());
    err = prrtilist_get_first_packet(&msock->inc_packet_list, pcont, d_target);
    //PDEBUG(MSG_FINAL, "%s: ... done at %u\n", DEBUG_NAME, prrt_get_prrttime());

    if (!err && *pcont) {
        PRINT_TRACE("RECV", "G", "got_pack");
        PRINT_TRACE("RECV", "G", "output_pack");
        PDEBUG(MSG_PRINTDATA, "%s: Copying packet data: %s\n", DEBUG_NAME,
                prrt_packet_get_data(*pcont));
        /*TODO: remove this: copy packet data to buffer*/
        data_size = (*pcont)->payload_size;
        packetno = (*pcont)->pack.data.header.seqnum;

        //PDEBUG(MSG_TRANSPORT, "%s: Delivering packet of size %d\n", DEBUG_NAME, data_size);
        PDEBUG(MSG_PACKRETURN, "%s: Delivering packet no %d\n", DEBUG_NAME, packetno);


        if (prrt_packet_is_redundancy(*pcont)) {
            PDEBUG(PRINT_PACKETHEX, "%s: Delivering redundancy packet no %d:\n",
                    DEBUG_NAME, packetno);
        } else {
            PDEBUG(PRINT_PACKETHEX, "%s: Delivering data packet no %d:\n",
                    DEBUG_NAME, packetno);
        }
        /*
                        if(!prrt_packet_data_starts_with_sync_byte(*pcont)) {
                                PDEBUG(PRINT_PACKETHEX, "%s: No sync byte found!\n", DEBUG_NAME);
                        }
         */
        prrt_packet_print_hexdata(*pcont);


        msock->delivered_packets++;
        if (prrt_packet_is_redundancy(*pcont)) {
            msock->delivered_red++;
        } else {
            msock->delivered_data++;
        }

    } else if (err == -ERESTARTSYS) {
        PDEBUG(MSG_PACKRETURN, "%s: Packet was not ready yet, try again\n", DEBUG_NAME);
        goto interrupted;
    } else {
        PDEBUG(MSG_PACKRETURN, "%s: Timeout, no packet arrived in time (err = %d)\n", DEBUG_NAME, err);
        err = -ETIMEDOUT;
        goto error;
    }


    return data_size;

error:
    return err;

//nodata:
    //  up(&msock->poll_lock);
    //  PDEBUG(MSG_COMMONRECV, "%s: Released lock in recv\n", DEBUG_NAME);
    PRINT_TRACE("RECV", "G", "output_failure");
    err = -ENODATA;

    return err;

interrupted:
    err = -EINTR;

    return err;
}

int prrt_set_option(struct prrt_socket *msock, enum option_t opt, int val) {
    int err = 0;
    int len = 0;

    switch (opt) {
        case FEEDBACK:
            prrtilist_set_option(&msock->inc_packet_list, opt, val);
            break;
        case REPLACE:
        case CODING:
        case TIMEOUT:
            msock->timeout = val;
            break;
        case PRRT_RCVBUF:
            if (0 < val) {
              err = kernel_setsockopt(msock->sock, SOL_SOCKET, SO_RCVBUF, (char *)&val, sizeof(val));
              if (err < 0) {
                PERROR("%s: Option PRRT_RCVBUF : kernel_setsockopt() failed", DEBUG_NAME);
                return err;
              }
            }
            // HACK we return the default/current max buffer size to the user space by err flag.
            // TODO fix it by providing query function for the user space.
            len = sizeof(val);
            err = kernel_getsockopt(msock->sock, SOL_SOCKET, SO_RCVBUF, (char *)&val, &len);
            if (err < 0) {
              PERROR("%s: Option PRRT_RCVBUF : kernel_getsockopt() failed", DEBUG_NAME);
              return err;
            }
            err = val;
            break;
        case KERNEL_RCVBUF:
            if (0 < val) {
              err = kernel_setsockopt(msock->kui_sock_write, SOL_SOCKET, SO_RCVBUF, (char *)&val, sizeof(val));
              if (err < 0) {
                PERROR("%s: Option KERNEL_RCVBUF : kernel_setsockopt() failed", DEBUG_NAME);
                return err;
              }
            }
            // HACK we return the default/current value of buffer size to the user space by err flag.
            // TODO fix it by providing query function for the user space.
            len = sizeof(val);
            err = kernel_getsockopt(msock->kui_sock_write, SOL_SOCKET, SO_RCVBUF, (char *)&val, &len);
            if (err < 0) {
              PERROR("%s: Option KERNEL_RCVBUF : kernel_getsockopt() failed", DEBUG_NAME);
              return err;
            }
            err = val;
            break;
        case USER_RCVBUF:
            err = 0;    // We do nothing because it is set in the socket API, see prrt/prrt_sockets.cpp.
            break;
        case TRAFFICSHAPING:
            prrtolist_set_option(&msock->out_packet_list, opt, val);
            break;
        case FB_SUPP_ON:
            prrtilist_set_option(&msock->inc_packet_list, opt, val);
            break;
        default:
            PERROR("%s: Option not recognized", DEBUG_NAME);
            err = -1;
            break;
    }

    return err;
}

int prrt_get_option(struct prrt_socket *msock, enum option_t opt) {
    int err = 0;

    switch (opt) {
        case FEEDBACK:
            err = prrtilist_get_option(&msock->inc_packet_list, opt);
            break;
        case REPLACE:
        case CODING:
            err = prrtolist_get_option(&msock->out_packet_list, opt);
            break;
        case TIMEOUT:
            err = msock->timeout;
            break;
        case TRAFFICSHAPING:
            err = prrtolist_get_option(&msock->out_packet_list, opt);
            break;
        case FB_SUPP_ON:
            err = prrtilist_get_option(&msock->inc_packet_list, opt);
            break;
        default:
            PERROR("%s: Option not recognized", DEBUG_NAME);
            err = -1;
            break;
    }

    return err;
}

/*
 *-----------------------------------------------------------------------------
 * local functions
 *-----------------------------------------------------------------------------
 */
static int build_addr(struct sockaddr_in *addr_s, char* ipaddr, u_short port) {
    int i = 0;
    unsigned char *z = addr_s->__pad;

    /*size of __pad: taken from include/linux/in.h*/
    int zeroes = __SOCK_SIZE__ - sizeof (short int) -
            sizeof (unsigned short int) - sizeof (struct in_addr);

    /*zero out address space*/
    for (i = 0; i < zeroes; i++) {
        z[i] = 0;
    }

    addr_s->sin_family = AF_INET;
    addr_s->sin_port = htons(port);
    addr_s->sin_addr.s_addr = in_aton(ipaddr);

    return 0;
}

static int GetAddrPort(struct sockaddr_in *addr_s) {
	return ntohs(addr_s->sin_port);
}

int prrtsocket_get_receiveon_port(struct prrt_socket *msock) {
	return GetAddrPort(msock->host);
}

int prrtsocket_get_sendto_port(struct prrt_socket *msock) {
	return GetAddrPort(msock->peer);
}

static int build_multi_req(struct ip_mreq *req, char* ipaddr) {
    req->imr_multiaddr.s_addr = in_aton(ipaddr);
    req->imr_interface.s_addr = htonl(INADDR_ANY);

    return 0;
}

static int isMulticastAddress(char *sock_addr) {
    int ret = 0;
    unsigned int first_block;

    ret = sscanf(sock_addr, "%u.", &first_block);
    if (ret != 1)
        return 0;

    if (224 <= first_block && first_block <= 239)
        return 1;

    return 0;
}

/*
 * Converts the binary ip baddr to a human readable string.
 * caddr is allocated by this function and must be freed by the caller.
 */
// static int in4_ntoa(__be32 baddr, char* caddr)
// {
//  return 0;
// }

static void print_stats(struct prrt_socket *msock) {
    unsigned long part = 0;
    unsigned long part2 = 0;
    unsigned long packetsThatShouldBeDelivered = 0;

    packetsThatShouldBeDelivered = msock->delivered_packets +
            msock->inc_packet_list.missed_packets;

    PDEBUG(MSG_FINAL, "%s: <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<\n", DEBUG_NAME);
    PDEBUG(MSG_FINAL, "%s: Printing socket stats:\n", DEBUG_NAME);
    PDEBUG(MSG_FINAL, "%s: ======================\n", DEBUG_NAME);
    PDEBUG(MSG_FINAL, "%s: Send packet stats:\n", DEBUG_NAME);
    PDEBUG(MSG_FINAL, "%s: ----------------------\n", DEBUG_NAME);
    PDEBUG(MSG_FINAL, "%s: * send - %lu\n", DEBUG_NAME, msock->out_packet_list.send_packets);
    PDEBUG(MSG_FINAL, "%s: * replacements sent - %lu\n", DEBUG_NAME,
            msock->out_packet_list.send_replacements);
    PDEBUG(MSG_FINAL, "%s: * deleted blocks - %lu\n", DEBUG_NAME,
            msock->out_packet_list.old_blocks);
    PDEBUG(MSG_FINAL, "%s: ----------------------\n", DEBUG_NAME);
    PDEBUG(MSG_FINAL, "%s: Received packet stats:\n", DEBUG_NAME);
    PDEBUG(MSG_FINAL, "%s: ----------------------\n", DEBUG_NAME);

    PDEBUG(MSG_FINAL, "%s: * ommited at start - %lu \n", DEBUG_NAME,
            msock->inc_packet_list.ommited_at_start_packets);

    if (packetsThatShouldBeDelivered != 0)
        part = (msock->inc_packet_list.old_packets)*100 / (packetsThatShouldBeDelivered);
    PDEBUG(MSG_FINAL, "%s: * ommited - %lu (%lu%%)\n", DEBUG_NAME,
            msock->inc_packet_list.old_packets, part);

    if (msock->delivered_packets != 0)
        part = (msock->delivered_data)*100 / (msock->delivered_packets);
    PDEBUG(MSG_FINAL, "%s: * delivered data - %lu (%lu%%)\n", DEBUG_NAME,
            msock->delivered_data, part);

    if (msock->delivered_packets != 0)
        part = (msock->delivered_red)*100 / (msock->delivered_packets);
    PDEBUG(MSG_FINAL, "%s: * delivered redundancy - %lu (%lu%%)\n", DEBUG_NAME,
            msock->delivered_red, part);

    if (packetsThatShouldBeDelivered != 0)
        part = (msock->delivered_packets)*100 / (packetsThatShouldBeDelivered);
    PDEBUG(MSG_FINAL, "%s: * delivered overally - %lu (%lu%%)\n", DEBUG_NAME,
            msock->delivered_packets, part);

    if (packetsThatShouldBeDelivered != 0)
        part = (msock->inc_packet_list.missed_packets - msock->inc_packet_list.old_packets)*100
        / (packetsThatShouldBeDelivered);
    PDEBUG(MSG_FINAL, "%s: * missed packets - %lu (%lu%%)\n", DEBUG_NAME,
            msock->inc_packet_list.missed_packets - msock->inc_packet_list.old_packets, part);

    if (packetsThatShouldBeDelivered != 0)
        part = (msock->inc_packet_list.missed_packets)*100 / (packetsThatShouldBeDelivered);
    PDEBUG(MSG_FINAL, "%s: * overally missed packets - %lu (%lu%%)\n", DEBUG_NAME,
            msock->inc_packet_list.missed_packets, part);

    if (packetsThatShouldBeDelivered != 0)
        part = (msock->inc_packet_list.received_datapackets)*100 / (packetsThatShouldBeDelivered);
    PDEBUG(MSG_FINAL, "%s: * received data - %lu (%lu%%)\n", DEBUG_NAME,
            msock->inc_packet_list.received_datapackets, part);

    if (packetsThatShouldBeDelivered != 0)
        part = (msock->inc_packet_list.received_redpackets)*100 / (packetsThatShouldBeDelivered);
    PDEBUG(MSG_FINAL, "%s: * received redundancy - %lu (%lu%%)\n", DEBUG_NAME,
            msock->inc_packet_list.received_redpackets, part);

    part = 0;
    if (msock->inc_packet_list.received_redpackets != 0)
        part = (msock->inc_packet_list.used_redpackets)*100 / (msock->inc_packet_list.received_redpackets);
    PDEBUG(MSG_FINAL, "%s: * used redundancy - %lu (%lu%%)\n", DEBUG_NAME,
            msock->inc_packet_list.used_redpackets, part);

    part = 0;
    part2 = 0;
    if (msock->inc_packet_list.received_redpackets != 0)
        part = (msock->inc_packet_list.useful_redpackets)*100 / (msock->inc_packet_list.received_redpackets);
    if (packetsThatShouldBeDelivered != 0)
        part2 = (msock->inc_packet_list.useful_redpackets)*100 / (packetsThatShouldBeDelivered);
    PDEBUG(MSG_FINAL, "%s: * decoded redundancy - %lu (%lu%%, %lu%%)\n", DEBUG_NAME,
            msock->inc_packet_list.useful_redpackets, part, part2);

    PDEBUG(MSG_FINAL, "%s: * packets that should be delivered - %lu (100%%)\n", DEBUG_NAME,
            packetsThatShouldBeDelivered);
    PDEBUG(MSG_FINAL, "%s: * received overall - %lu \n", DEBUG_NAME,
            msock->inc_packet_list.received_overall);

    PDEBUG(MSG_FINAL, "%s: ----------------------\n", DEBUG_NAME);
    PDEBUG(MSG_FINAL, "%s: Inc feedback stats:\n", DEBUG_NAME);
    PDEBUG(MSG_FINAL, "%s: ----------------------\n", DEBUG_NAME);

    PDEBUG(MSG_FINAL, "%s: * sent feedbacks - %lu \n", DEBUG_NAME,
            msock->inc_packet_list.feedbacks_sent);

    PDEBUG(MSG_FINAL, "%s: * sent fb packets - %lu \n", DEBUG_NAME,
            msock->inc_packet_list.fbpackets_sent);

    PDEBUG(MSG_FINAL, "%s: ----------------------\n", DEBUG_NAME);
    PDEBUG(MSG_FINAL, "%s: Inc data list stats:\n", DEBUG_NAME);
    PDEBUG(MSG_FINAL, "%s: ----------------------\n", DEBUG_NAME);
    prrtstats_print_list_stats(&msock->inc_packet_list.data_list.stats);
    PDEBUG(MSG_FINAL, "%s: ----------------------\n", DEBUG_NAME);
    PDEBUG(MSG_FINAL, "%s: Inc redundany list stats:\n", DEBUG_NAME);
    PDEBUG(MSG_FINAL, "%s: ----------------------\n", DEBUG_NAME);
    prrtstats_print_list_stats(&msock->inc_packet_list.red_list.stats);
    PDEBUG(MSG_FINAL, "%s: ----------------------\n", DEBUG_NAME);
    PDEBUG(MSG_FINAL, "%s: >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>\n", DEBUG_NAME);
}

static int create_feedback_socket(struct prrt_socket *msock) {
    /*must hold lock here*/
    int err = 0;
    struct timeval totimefb;

    PDEBUG(MSG_INITSEND, "%s: Recreating sender sockets\n", DEBUG_NAME);

    totimefb.tv_sec = POLLTIMEOUTFB;
    totimefb.tv_usec = 0;

    if (msock->sockfb) {
        /*release the socket for feedback from its binding*/
        kernel_sock_shutdown(msock->sockfb, 2);
        sock_release(msock->sockfb);
    }

    /*Normal UDP Sockets are used for the feedback channel*/
    err = sock_create_kern(PF_INET, SOCK_DGRAM, IPPROTO_UDP, &(msock->sockfb));
    if (err < 0)
        goto error;

    /*set a timeout for the socket*/
    err = kernel_setsockopt(msock->sockfb, SOL_SOCKET, SO_RCVTIMEO, (char *) &totimefb,
            sizeof (totimefb));
    if (err < 0)
        goto error;

error:
    return err;
}

static int init_sending(struct prrt_socket *msock) {
    int err = 0;

    msock->out_status = DOWN;
    up(&msock->out_lock);

    /*this is already hold when called*/
    /*down(&msock->out_lock);*/
    if (msock->send_queue == 0) {
        PDEBUG(MSG_INITSEND, "%s: Creating prrt_send workqueue\n", DEBUG_NAME);
        msock->send_queue = create_singlethread_workqueue("prrt_send");
    } else {
        PDEBUG(MSG_INITSEND, "%s: Releasing send thread from lock\n", DEBUG_NAME);
        /*this has to be done to free the send thread out of its loop*/
        up(&msock->out_packet_list.block_list.packet_count);
        PDEBUG(MSG_OUTLIST, "%s: Upped count to release thread, now %u left\n", DEBUG_NAME,
                msock->out_packet_list.block_list.packet_count.count);
    }
    if (msock->writepkt_queue == 0) {
        PDEBUG(MSG_FINAL, "%s: Creating prrt_writepkt workqueue\n", DEBUG_NAME);
        msock->writepkt_queue = create_singlethread_workqueue("prrt_writepkt");
    }
    if (msock->recvfb_queue == 0) {
        PDEBUG(MSG_INITSEND, "%s: Creating prrt_recvfb workqueue\n", DEBUG_NAME);
        msock->recvfb_queue = create_singlethread_workqueue("prrt_recvfb");
    }
    if (msock->sendrepl_queue == 0) {
        PDEBUG(MSG_INITSEND, "%s: Creating prrt_sendrepl workqueue\n", DEBUG_NAME);
        msock->sendrepl_queue = create_singlethread_workqueue("prrt_sendrepl");
    }

    PDEBUG(MSG_INITSEND, "%s: Flushing workqueues\n", DEBUG_NAME);
    /*delete works and flush workqueue*/
    cancel_delayed_work(&msock->out_work);
    cancel_delayed_work(&msock->recvfb_work);
    cancel_delayed_work(&msock->writepkt_work);
    flush_workqueue(msock->send_queue);
    flush_workqueue(msock->recvfb_queue);
    flush_workqueue(msock->writepkt_queue);

    PDEBUG(MSG_INITSEND, "%s: Taking lock\n", DEBUG_NAME);
    if (down_interruptible(&msock->out_lock)) {
        err = -EINTR;
        goto no_lock;
    }
    down(&msock->recvfb_lock);

    msock->out_status = WORKING;
    msock->out_pno = 0;

    err = prrtolist_reinit(&msock->out_packet_list);
    if (err)
        goto out;

    PDEBUG(MSG_INITSEND, "%s: Queuing works\n", DEBUG_NAME);
    queue_delayed_work(msock->send_queue, &msock->out_work, 1);
    queue_delayed_work(msock->writepkt_queue, &msock->writepkt_work, 1);
    queue_delayed_work(msock->recvfb_queue, &msock->recvfb_work, FEEDBACKDELAY);
    queue_delayed_work(msock->sendrepl_queue, &msock->sendrepl_work, REPLACEMENTDELAY);

    err = create_feedback_socket(msock);

out:
    up(&msock->recvfb_lock);
no_lock:
    /*up(&msock->out_lock);*/
    return err;
}

