/*
 * fec.c -- forward error correction based on Vandermonde matrices
 * 980614
 * (C) 1997-98 Luigi Rizzo (luigi@iet.unipi.it)
 *
 * Portions derived from code by Phil Karn (karn@ka9q.ampr.org),
 * Robert Morelos-Zaragoza (robert@spectra.eng.hawaii.edu) and Hari
 * Thirumoorthy (harit@spectra.eng.hawaii.edu), Aug 1995
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:

 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials
 *    provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 */

/*
 * The following parameter defines how many bits are used for
 * field elements. The code supports any value from 2 to 16
 * but fastest operation is achieved with 8 bit elements
 * This is the only parameter you may want to change.
 */
#ifndef PRRT_BLOCK_CODE_H
#define PRRT_BLOCK_CODE_H

#ifndef GF_BITS
#define GF_BITS  8      /* code over GF(2**GF_BITS) - change to suit */
#endif

#include "../defines.h"
#include <linux/types.h>

/*
 * stuff used for testing purposes only
 */

#undef TEST
#ifdef  TEST
#define DEB(x)
#define DDB(x) x
#define FECDEBUG   1       /* minimal debugging */

#include <sys/time.h>
#define DIFF_T(a,b) \
        (1+ 1000000*(a.tv_sec - b.tv_sec) + (a.tv_usec - b.tv_usec) )

#define TICK(t) \
        {struct timeval x ; \
        gettimeofday(&x, NULL) ; \
        t = x.tv_usec + 1000000* (x.tv_sec & 0xff ) ; \
        }
#define TOCK(t) \
        { uint64_t t1 ; TICK(t1) ; \
          if (t1 < t) t = 256000000 + t1 - t ; \
          else t = t1 - t ; \
          if (t == 0) t = 1 ;}

#else
#define DEB(x)
#define DDB(x)
#define TICK(x)
#define TOCK(x)
#endif /* TEST */

/*
 * You should not need to change anything beyond this point.
 * The first part of the file implements linear algebra in GF.
 *
 * gf is the type used to store an element of the Galois Field.
 * Must constain at least GF_BITS bits.
 *
 * Note: unsigned char will work up to GF(256) but int seems to run
 * faster on the Pentium. We use int whenever have to deal with an
 * index, since they are generally faster.
 */
#if (GF_BITS < 2 || GF_BITS >16)
#error "GF_BITS must be 2 .. 16"
#endif
#if (GF_BITS <= 8)
typedef unsigned char gf;
#else
typedef unsigned short gf;
#endif

#define GF_SIZE ((1 << GF_BITS) - 1)    /* powers of \alpha */
#define FEC_MAGIC       0xFECC0DEC

/**
 * Struct holding parameters for a coder.
 */
struct fec_params {
    uint64_t magic;
    int k, n;          /* parameters of the code */
    gf *enc_matrix;
};

/**
 * This struture contains the members for a coder that can en/decode.
 * To change the code you need to free the old coder and allocate a new one.
 */
__PCLASS __NOT_THREAD_SAFE struct prrt_coder {
	struct fec_params params;
	gf gf_exp[2*GF_SIZE];	/* index->poly form conversion table	*/
	int gf_log[GF_SIZE + 1];/* Poly->index form conversion table	*/
	gf inverse[GF_SIZE+1];	/* inverse of field elem.		*/
				/* inv[\alpha**i]=\alpha**(GF_SIZE-i-1)	*/
#if (GF_BITS <= 8)
	gf gf_mul_table[GF_SIZE + 1][GF_SIZE + 1];
#endif
	int count_num;
	uint64_t tot_encoding_time;
	uint64_t myticks[10];       /* vars for timekeeping */
};

/**
 * This functions updates the coder with the new parameters.
 * \param k number of data packets in one block
 * \param n overall number of packets in one block
 * \return 0 on success, -1 if parameters are invalid, -2 if no memory
 */
int prrt_fec_get_coder(struct prrt_coder **cod, int n, int k);

/**
 * Free the coder object.
 */
void prrt_fec_free(struct prrt_coder *cod) ;

/**
 * This function creates an encoded packet from some given packets.
 * We are using a (k,n) systematic block code. This function takes k packets as source (the first
 * k) and creates one of the n-k remaining packets. Index is the number of the packet and therefore
 * lies between k and n (k <= index < n).
 * \param src an array of pointers to the source packets
 * \param dst the pointer to a preallocatet space for the new packet.
 * \param index this is the position of the new packet in the resulting set of encoded packets
 * \param sz the size of the packets that should be encoded
 */
void prrt_fec_encode(struct prrt_coder *cod, gf *src[], gf *dst, int index, int sz) ;

/**
 * This function decodes a set of packets.
 * It takes a set of packets that is given as an array of pointers to the packets. Because this
 * is an (k,n) systematic code there have to be k pointers in the array. If data packet i(< k) is
 * present it should be at position i-1 (e.g. packet 1 at 0, 2 at 1) and index[i-1] should be set
 * to i-1. If data packet i is not present, but replacement packet (k < )j(<= n) is present, its
 * data can be pointed to at pkt[i-1]. Index[i-1] should then be set to j-1.
 * \param pkt these are the pointers to the data of the different packets.
 * \param index this is the position of the packets in the block code.
 * \param sz the size of the data of the packets
 */
int prrt_fec_decode(struct prrt_coder *cod, gf *pkt[], int index[], int sz) ;

/* end of file */
#endif
